"""
CBMPy: default model module
===========================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBCommon.py 660 2022-09-24 14:57:04Z bgoli $)

"""

cbmpy_test_core = """<?xml version="1.0" encoding="utf-8"?>
<!-- Created with PySCeS CBM (0.6.0 [r1115]) on Thu, 20 Sep 2012 12:28:43 by bgoli -->
<sbml xmlns="http://www.sbml.org/sbml/level3/version1/core" xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version1" xmlns:html="http://www.w3.org/1999/xhtml" level="3" version="1" fbc:required="false">
  <model metaid="meta_core_memesa_model">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>
          <html:span style="font-family: Courier New,Courier,monospace;">
            <notes>
              <span xmlns="http://www.w3.org/1999/xhtml">
                <p>
                  <span style="font-family: Courier New,Courier,monospace;"/>
                </p>
                <p>
                  <br/>
                  <span size="small">Model &quot;<strong>core_memesa_model</strong>&quot; (None) generated with <a href="http://pysces.sourceforge.net">PySCeS CBM</a> (0.4.0 [r838]) on Mon, 17 Oct 2011 17:04:00.</span>
              </p>
            </span>
          </notes>
        </html:span>
      </html:p>
      <html:p>
        <html:br/>
        <html:span size="small">Model &quot;<html:strong>core_memesa_model</html:strong>&quot; (SBML_CB_MODEL) generated with <html:a href="http://pysces.sourceforge.net">PySCeS CBM</html:a> (0.6.0 [r1115]) on Thu, 20 Sep 2012 12:28:43.</html:span>
    </html:p>
  </body>
</notes><listOfUnitDefinitions>
  <unitDefinition id="substance" name="substance">
    <listOfUnits>
      <unit kind="mole" exponent="1" scale="0" multiplier="1"/>
    </listOfUnits>
  </unitDefinition>
  <unitDefinition id="area" name="area">
    <listOfUnits>
      <unit kind="metre" exponent="2" scale="0" multiplier="1"/>
    </listOfUnits>
  </unitDefinition>
  <unitDefinition id="volume" name="volume">
    <listOfUnits>
      <unit kind="litre" exponent="1" scale="0" multiplier="1"/>
    </listOfUnits>
  </unitDefinition>
  <unitDefinition id="length" name="length">
    <listOfUnits>
      <unit kind="metre" exponent="1" scale="0" multiplier="1"/>
    </listOfUnits>
  </unitDefinition>
  <unitDefinition id="time" name="time">
    <listOfUnits>
      <unit kind="second" exponent="1" scale="0" multiplier="1"/>
    </listOfUnits>
  </unitDefinition>
  <unitDefinition id="mmol_per_gDW_per_hr" name="mmol_per_gDW_per_hr">
    <listOfUnits>
      <unit kind="mole" exponent="1" scale="-3" multiplier="1"/>
      <unit kind="gram" exponent="-1" scale="0" multiplier="1"/>
      <unit kind="second" exponent="-1" scale="0" multiplier="0.00027777"/>
    </listOfUnits>
  </unitDefinition>
</listOfUnitDefinitions><listOfCompartments>
  <compartment id="Cell" name="Cell" size="1" constant="false"/>
</listOfCompartments><listOfSpecies>
  <species metaid="meta_A" id="A" name="A" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_B" id="B" name="B" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_C" id="C" name="C" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_D" id="D" name="D" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_F1" id="F1" name="F1" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_F2" id="F2" name="F2" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_G" id="G" name="G" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_H" id="H" name="H" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_I1" id="I1" name="I1" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_I2" id="I2" name="I2" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_J1" id="J1" name="J1" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_J2" id="J2" name="J2" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_K1" id="K1" name="K1" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_K2" id="K2" name="K2" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_L" id="L" name="L" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_M" id="M" name="M" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_N" id="N" name="N" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_X0" id="X0" name="X0" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="true" constant="false"/>
  <species metaid="meta_X1" id="X1" name="X1" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="true" constant="false"/>
  <species metaid="meta_X3" id="X3" name="X3" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="true" constant="false"/>
  <species metaid="meta_mE1" id="mE1" name="mE1" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
  <species metaid="meta_mE2" id="mE2" name="mE2" compartment="Cell" initialConcentration="0" hasOnlySubstanceUnits="false" boundaryCondition="false" constant="false"/>
</listOfSpecies><listOfReactions>
  <reaction metaid="meta_R01" id="R01" name="R01" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R01</html:p>
        <html:p>SUBSYSTEM: </html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R01"/>
        <data id="subsystem" type="string"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="X0" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="A" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R10" id="R10" name="R10" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R10</html:p>
        <html:p>SUBSYSTEM: C2</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R10"/>
        <data id="subsystem" type="string" value="C2"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="F1" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="F2" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R03" id="R03" name="R03" reversible="true" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R03</html:p>
        <html:p>SUBSYSTEM: C1</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R03"/>
        <data id="subsystem" type="string" value="C1"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="A" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="C" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R02" id="R02" name="R02" reversible="true" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R02</html:p>
        <html:p>SUBSYSTEM: C1</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R02"/>
        <data id="subsystem" type="string" value="C1"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="A" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="B" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R05" id="R05" name="R05" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R05</html:p>
        <html:p>SUBSYSTEM: </html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R05"/>
        <data id="subsystem" type="string"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="B" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="D" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R04" id="R04" name="R04" reversible="true" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R04</html:p>
        <html:p>SUBSYSTEM: C1</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R04"/>
        <data id="subsystem" type="string" value="C1"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="C" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="B" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R16" id="R16" name="R16" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R16</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R16"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="H" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="J1" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R17" id="R17" name="R17" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R17</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R17"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="J1" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="J2" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R14" id="R14" name="R14" reversible="true" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R14</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R14"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="I1" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="I2" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R15" id="R15" name="R15" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R15</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R15"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="I2" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="L" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R12" id="R12" name="R12" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R12</html:p>
        <html:p>SUBSYSTEM: </html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R12"/>
        <data id="subsystem" type="string"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="G" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="H" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R13" id="R13" name="R13" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R13</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R13"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="H" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="I1" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R07" id="R07" name="R07" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R07</html:p>
        <html:p>SUBSYSTEM: C2</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R07"/>
        <data id="subsystem" type="string" value="C2"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="mE1" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="mE2" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R06" id="R06" name="R06" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R06</html:p>
        <html:p>SUBSYSTEM: C2</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R06"/>
        <data id="subsystem" type="string" value="C2"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="D" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="mE1" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R09" id="R09" name="R09" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R09</html:p>
        <html:p>SUBSYSTEM: C2</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R09"/>
        <data id="subsystem" type="string" value="C2"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="D" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="F1" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R08" id="R08" name="R08" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R08</html:p>
        <html:p>SUBSYSTEM: C2</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R08"/>
        <data id="subsystem" type="string" value="C2"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="mE2" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="G" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R11" id="R11" name="R11" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R11</html:p>
        <html:p>SUBSYSTEM: C2</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R11"/>
        <data id="subsystem" type="string" value="C2"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="F2" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="G" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R18" id="R18" name="R18" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R18</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R18"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="J2" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="L" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R19" id="R19" name="R19" reversible="true" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R19</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R19"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="I1" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="K1" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R26" id="R26" name="R26" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R26</html:p>
        <html:p>SUBSYSTEM: </html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R26"/>
        <data id="subsystem" type="string"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="A" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="N" stoichiometry="0.5" constant="false"/>
      <speciesReference species="X3" stoichiometry="0.5" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R25" id="R25" name="R25" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R25</html:p>
        <html:p>SUBSYSTEM: </html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R25"/>
        <data id="subsystem" type="string"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="N" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="X1" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R24" id="R24" name="R24" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R24</html:p>
        <html:p>SUBSYSTEM: C4</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R24"/>
        <data id="subsystem" type="string" value="C4"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="M" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="N" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R23" id="R23" name="R23" reversible="true" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R23</html:p>
        <html:p>SUBSYSTEM: C4</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R23"/>
        <data id="subsystem" type="string" value="C4"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="M" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="N" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R22" id="R22" name="R22" reversible="false" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R22</html:p>
        <html:p>SUBSYSTEM: </html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R22"/>
        <data id="subsystem" type="string"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="L" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="M" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R21" id="R21" name="R21" reversible="true" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R21</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R21"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="K2" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="I2" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
  <reaction metaid="meta_R20" id="R20" name="R20" reversible="true" fast="false">
    <notes>
      <body xmlns:html="http://www.w3.org/1999/xhtml">
        <html:p>Abbreviation: R20</html:p>
        <html:p>SUBSYSTEM: C3</html:p>
      </body>
    </notes>
    <annotation>
      <listOfKeyValueData xmlns="http://pysces.sourceforge.net/KeyValueData">
        <data id="abbreviation" type="string" value="R20"/>
        <data id="subsystem" type="string" value="C3"/>
      </listOfKeyValueData>
    </annotation>
    <listOfReactants>
      <speciesReference species="K1" stoichiometry="1" constant="false"/>
    </listOfReactants>
    <listOfProducts>
      <speciesReference species="K2" stoichiometry="1" constant="false"/>
    </listOfProducts>
  </reaction>
</listOfReactions><fbc:listOfFluxBounds>
  <fbc:fluxBound fbc:id="c0" fbc:reaction="R16" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c1" fbc:reaction="R16" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c2" fbc:reaction="R03" fbc:operation="greaterEqual" fbc:value="-1000"/>
  <fbc:fluxBound fbc:id="c3" fbc:reaction="R03" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c4" fbc:reaction="R02" fbc:operation="greaterEqual" fbc:value="-1000"/>
  <fbc:fluxBound fbc:id="c5" fbc:reaction="R02" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c6" fbc:reaction="R05" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c7" fbc:reaction="R05" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c8" fbc:reaction="R07" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c9" fbc:reaction="R07" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c10" fbc:reaction="R04" fbc:operation="greaterEqual" fbc:value="-1000"/>
  <fbc:fluxBound fbc:id="c11" fbc:reaction="R04" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c12" fbc:reaction="R01" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c13" fbc:reaction="R01" fbc:operation="lessEqual" fbc:value="1"/>
  <fbc:fluxBound fbc:id="c14" fbc:reaction="R17" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c15" fbc:reaction="R17" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c16" fbc:reaction="R14" fbc:operation="greaterEqual" fbc:value="-1000"/>
  <fbc:fluxBound fbc:id="c17" fbc:reaction="R14" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c18" fbc:reaction="R15" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c19" fbc:reaction="R15" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c20" fbc:reaction="R12" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c21" fbc:reaction="R12" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c22" fbc:reaction="R13" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c23" fbc:reaction="R13" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c24" fbc:reaction="R10" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c25" fbc:reaction="R10" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c26" fbc:reaction="R11" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c27" fbc:reaction="R11" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c28" fbc:reaction="R09" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c29" fbc:reaction="R09" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c30" fbc:reaction="R08" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c31" fbc:reaction="R08" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c32" fbc:reaction="R06" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c33" fbc:reaction="R06" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c34" fbc:reaction="R18" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c35" fbc:reaction="R18" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c36" fbc:reaction="R19" fbc:operation="greaterEqual" fbc:value="-1000"/>
  <fbc:fluxBound fbc:id="c37" fbc:reaction="R19" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c38" fbc:reaction="R26" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c39" fbc:reaction="R26" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c40" fbc:reaction="R25" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c41" fbc:reaction="R25" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c42" fbc:reaction="R24" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c43" fbc:reaction="R24" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c44" fbc:reaction="R23" fbc:operation="greaterEqual" fbc:value="-1000"/>
  <fbc:fluxBound fbc:id="c45" fbc:reaction="R23" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c46" fbc:reaction="R22" fbc:operation="greaterEqual" fbc:value="0"/>
  <fbc:fluxBound fbc:id="c47" fbc:reaction="R22" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c48" fbc:reaction="R21" fbc:operation="greaterEqual" fbc:value="-1000"/>
  <fbc:fluxBound fbc:id="c49" fbc:reaction="R21" fbc:operation="lessEqual" fbc:value="1000"/>
  <fbc:fluxBound fbc:id="c50" fbc:reaction="R20" fbc:operation="greaterEqual" fbc:value="-1000"/>
  <fbc:fluxBound fbc:id="c51" fbc:reaction="R20" fbc:operation="lessEqual" fbc:value="1000"/>
</fbc:listOfFluxBounds><fbc:listOfObjectives fbc:activeObjective="objMaxJ25">
  <fbc:objective fbc:id="objMaxJ25" fbc:type="maximize">
    <fbc:listOfFluxes>
      <fbc:fluxObjective fbc:reaction="R25" fbc:coefficient="1"/>
    </fbc:listOfFluxes>
  </fbc:objective>
</fbc:listOfObjectives>
</model>
</sbml>
"""

cbmpy_test_ecoli = """<?xml version='1.0' encoding='utf-8' standalone='no'?>
<!-- Created by ModelPolisher version 1.7 on 2017-05-09 at 24:05:42 PDT with JSBML version 1.2. -->
<sbml xmlns="http://www.sbml.org/sbml/level3/version1/core" fbc:required="false" groups:required="false" level="3" sboTerm="SBO:0000624" version="1"
xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2" xmlns:groups="http://www.sbml.org/sbml/level3/version1/groups/version1">
  <notes>
    <body xmlns="http://www.w3.org/1999/xhtml" style="font-size: 14px; height: 100%; width: 100%; color: #444444; font-family:Helvetica, sans-serif; margin-top: 0px; margin-left: 0px;">
      <div style="height: 60px; background-color: #09F;">
        <div style="margin-left: auto; margin-right: auto; width: 970px;">
          <h1
          style="           font-family: Lato, Helvetica Neue, Helvetica, Arial, sans-serif;           font-weight: 700;           color:#EFF8FF;           text-shadow: 0px 0px 10px rgba(0, 0, 0, 0.5);            text-decoration-color: -moz-use-text-color;           margin-bottom: 40px;           margin-top: 0px;">
            <div class="dc:title">iJR904 - Escherichia coli str. K-12 substr. MG1655</div>
          </h1>
        </div>
      </div>
    </body>
  </notes>
  <model extentUnits="substance" fbc:strict="true" id="iJR904" metaid="iJR904" name="Escherichia coli str. K-12 substr. MG1655" substanceUnits="substance" timeUnits="time">
    <notes>
      <body xmlns="http://www.w3.org/1999/xhtml" style="margin:0 auto; width: 970px; font-size: 14px; height: 100%; color: #444444; font-family:Helvetica, sans-serif;">
      <h2>Description</h2>
      <div class="dc:description">
        <p>This is a metabolism model of Escherichia coli str. K-12 substr. MG1655 in
        <a href="http://sbml.org" target="_blank" title="Access the definition of the SBML file format.">SBML</a>&#160;format.</p>
      </div>
      <div class="dc:provenance">The content of this model has been carefully created in a manual research effort. This file has been exported from the software
      <a href="http://dx.doi.org/10.1186/1752-0509-7-74" title="Access publication about COBRApy." target="_blank">COBRApy</a>&#160;and further processed with the
      <a href="http://dx.doi.org/10.1093/bioinformatics/btv341" title="Access publication about JSBML." target="_blank">JSBML</a>-based
      <a href="http://github.com/SBRG/ModelPolisher/" target="_blank" title="Access ModelPolisher on Github">ModelPolisher</a>&#160;application.</div>
      <div class="dc:publisher">This file has been produced by the
      <a href="http://systemsbiology.ucsd.edu" title="Website of the Systems Biology Research Group" target="_blank">Systems Biology Research Group</a>&#160;using
      <a href="http://bigg.ucsd.edu" title="Access BiGG Models knowledge-base." target="_blank">BiGG Models knowledge-base</a>&#160;version of Oct 8, 2017, where it is currently hosted and identified
      by:
      <a href="http://identifiers.org/bigg.model/iJR904" title="Access to this model via BiGG Models knowledge-base." target="_blank">iJR904</a>.</div>
      <h2>Terms of use</h2>
      <div class="dc:rightsHolder">Copyright 2017 The Regents of the University of California.</div>
      <div class="dc:license">
        <p>Redistribution and use of any part of this model from BiGG Models knowledge-base, with or without modification, are permitted provided that the following conditions are met:
        <ol>
          <li>Redistributions of this SBML file must retain the above copyright notice, this list of conditions and the following disclaimer.</li>
          <li>Redistributions in a different form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided
          with the distribution.</li>
        </ol>This model is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.</p>
        <p>For specific licensing terms about this particular model and regulations of commercial use, see
        <a href="http://identifiers.org/bigg.model/iJR904" title="Access to this model via BiGG Models knowledge-base." target="_blank">this model in BiGG Models knowledge-base</a>.</p>
      </div>
      <h2>References</h2>When using content from BiGG Models knowledge-base in your research works, please cite
      <dl>
        <dt>King ZA, Lu JS, Draeger A, Miller PC, Federowicz S, Lerman JA, Ebrahim A, Palsson BO, and&#160;Lewis NE. (2015).
        <dd>BiGG Models: A platform for integrating, standardizing, and sharing genome-scale models.
        <i>Nucl Acids Res</i>.
        <a href="https://dx.doi.org/10.1093/nar/gkv1049" target="_blank" title="Access the publication about BiGG Models knowledge-base">doi:10.1093/nar/gkv1049</a></dd></dt>
      </dl></body>
    </notes>
    <annotation>
      <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqmodel="http://biomodels.net/model-qualifiers/" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
        <rdf:Description rdf:about="#iJR904">
          <bqbiol:hasTaxon>
            <rdf:Bag>
              <rdf:li rdf:resource="http://identifiers.org/taxonomy/511145" />
            </rdf:Bag>
          </bqbiol:hasTaxon>
          <bqmodel:is>
            <rdf:Bag>
              <rdf:li rdf:resource="http://identifiers.org/bigg.model/iJR904" />
            </rdf:Bag>
          </bqmodel:is>
          <bqmodel:isDescribedBy>
            <rdf:Bag>
              <rdf:li rdf:resource="http://identifiers.org/pubmed/12952533" />
            </rdf:Bag>
          </bqmodel:isDescribedBy>
        </rdf:Description>
      </rdf:RDF>
    </annotation>
    <fbc:listOfObjectives xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2" fbc:activeObjective="obj">
      <fbc:objective fbc:id="obj" fbc:type="maximize">
        <fbc:listOfFluxObjectives>
          <fbc:fluxObjective fbc:coefficient="1" fbc:reaction="R_BIOMASS_Ecoli" />
        </fbc:listOfFluxObjectives>
      </fbc:objective>
    </fbc:listOfObjectives>
    <fbc:listOfGeneProducts xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
      <fbc:geneProduct fbc:id="G_b3553" fbc:label="b3553" metaid="G_b3553">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3553">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37666" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011604" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12272" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948074" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111614" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0347" fbc:label="b0347" metaid="G_b0347">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0347">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77397" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001196" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20273" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945197" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128332" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0004" fbc:label="b0004" metaid="G_b0004">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0004">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00934" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000012" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11000" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945198" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16127998" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4266" fbc:label="b4266" metaid="G_b4266">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4266">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9P9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013971" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12540" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947109" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132088" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2836" fbc:label="b2836" metaid="G_b2836">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2836">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31119" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009302" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11679" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947315" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130740" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3793" fbc:label="b3793" metaid="G_b3793">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3793">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27835" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012392" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11457" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948293" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131649" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2901" fbc:label="b2901" metaid="G_b2901">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2901">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46829" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009523" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13079" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947378" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130803" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1302" fbc:label="b1302" metaid="G_b1302">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1302">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P50457" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004380" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13187" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945446" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129263" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2662" fbc:label="b2662" metaid="G_b2662">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2662">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22256" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008760" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10361" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948067" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130576" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2663" fbc:label="b2663" metaid="G_b2663">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2663">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25527" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008763" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11330" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948049" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130577" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2222" fbc:label="b2222" metaid="G_b2222">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2222">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76459" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007349" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11670" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946719" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130159" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2221" fbc:label="b2221" metaid="G_b2221">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2221">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76458" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007347" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11669" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947525" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130158" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2224" fbc:label="b2224" metaid="G_b2224">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2224">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76461" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007355" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11672" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946727" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130161" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2223" fbc:label="b2223" metaid="G_b2223">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2223">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76460" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007353" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11671" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946721" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130160" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0351" fbc:label="b0351" metaid="G_b0351">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0351">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77580" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001207" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13625" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945008" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128336" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1993" fbc:label="b1993" metaid="G_b1993">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1993">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AE76" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006616" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13238" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946519" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129934" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2316" fbc:label="b2316" metaid="G_b2316">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2316">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9Q5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007648" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10217" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946796" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130251" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3256" fbc:label="b3256" metaid="G_b3256">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3256">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24182" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010677" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10276" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947761" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131144" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0185" fbc:label="b0185" metaid="G_b0185">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0185">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABD5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000630" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11647" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944895" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128178" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3255" fbc:label="b3255" metaid="G_b3255">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3255">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABD8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010675" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10275" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947758" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131143" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0335" fbc:label="b0335" metaid="G_b0335">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0335">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77495" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001154" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13604" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946891" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128320" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3784" fbc:label="b3784" metaid="G_b3784">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3784">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC78" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012366" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10840" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948789" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131640" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2417" fbc:label="b2417" metaid="G_b2417">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2417">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69783" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007971" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10165" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946880" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130343" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2415" fbc:label="b2415" metaid="G_b2415">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2415">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AA04" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007962" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10788" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946886" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130341" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2416" fbc:label="b2416" metaid="G_b2416">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2416">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08839" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007967" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10789" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946879" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130342" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0679" fbc:label="b0679" metaid="G_b0679">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0679">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09323" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002312" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10635" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945292" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128655" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1101" fbc:label="b1101" metaid="G_b1101">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1101">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69786" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003722" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10787" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945651" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129064" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3959" fbc:label="b3959" metaid="G_b3959">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3959">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6C8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012967" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10064" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948464" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698337" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2818" fbc:label="b2818" metaid="G_b2818">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2818">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6C5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009245" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10063" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947289" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130722" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3670" fbc:label="b3670" metaid="G_b3670">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3670">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ADF8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011997" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10502" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948183" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131540" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3769" fbc:label="b3769" metaid="G_b3769">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3769">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ADG1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012313" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10501" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948279" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131627" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3768" fbc:label="b3768" />
      <fbc:geneProduct fbc:id="G_b0078" fbc:label="b0078" metaid="G_b0078">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0078">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00894" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000294" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10499" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947267" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128071" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3767" fbc:label="b3767" />
      <fbc:geneProduct fbc:id="G_b0077" fbc:label="b0077" metaid="G_b0077">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0077">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00893" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000290" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10500" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948793" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111084" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3671" fbc:label="b3671" metaid="G_b3671">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3671">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08142" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011999" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10494" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948182" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131541" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2296" fbc:label="b2296" metaid="G_b2296">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2296">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6A3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007579" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10027" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946775" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130231" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3115" fbc:label="b3115" metaid="G_b3115">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3115">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11868" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010245" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11172" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947635" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698313" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1849" fbc:label="b1849" metaid="G_b1849">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1849">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33221" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006162" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11809" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946368" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129802" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3794" fbc:label="b3794" metaid="G_b3794">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3794">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27836" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012394" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11458" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948301" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131650" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1819" fbc:label="b1819" metaid="G_b1819">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1819">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69805" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006058" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10569" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946342" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:345452720" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1818" fbc:label="b1818" metaid="G_b1818">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1818">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69801" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006056" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10568" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946332" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129772" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1817" fbc:label="b1817" metaid="G_b1817">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1817">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69797" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006054" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10567" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946334" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129771" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3224" fbc:label="b3224" metaid="G_b3224">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3224">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P41036" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010577" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20145" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947740" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111559" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3225" fbc:label="b3225" metaid="G_b3225">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3225">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6L4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010579" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10637" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947742" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131115" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1091" fbc:label="b1091" metaid="G_b1091">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1091">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6R0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003696" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10277" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946003" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129054" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3957" fbc:label="b3957" metaid="G_b3957">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3957">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23908" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012955" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11286" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948456" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131795" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1519" fbc:label="b1519" metaid="G_b1519">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1519">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76145" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005070" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13812" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946074" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129478" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1276" fbc:label="b1276" metaid="G_b1276">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1276">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25516" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004283" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11325" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946724" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129237" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0118" fbc:label="b0118" metaid="G_b0118">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0118">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P36683" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000411" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12316" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944864" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128111" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3359" fbc:label="b3359" metaid="G_b3359">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3359">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P18335" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010976" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10066" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947864" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131238" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2563" fbc:label="b2563" metaid="G_b2563">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2563">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24224" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008433" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10247" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947037" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130488" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4069" fbc:label="b4069" metaid="G_b4069">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4069">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27550" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013337" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11448" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948572" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131895" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1623" fbc:label="b1623" metaid="G_b1623">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1623">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22333" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005435" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10030" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945851" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129581" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1096" fbc:label="b1096" metaid="G_b1096">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1096">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P28305" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003709" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11493" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946647" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129059" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3360" fbc:label="b3360" metaid="G_b3360">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3360">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00903" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010983" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10682" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947873" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131239" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1812" fbc:label="b1812" metaid="G_b1812">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1812">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05041" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006031" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10683" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946337" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129766" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3665" fbc:label="b3665" metaid="G_b3665">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3665">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31441" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011979" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11692" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948210" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131535" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3654" fbc:label="b3654" metaid="G_b3654">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3654">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGM9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011946" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11194" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948172" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131525" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1241" fbc:label="b1241" metaid="G_b1241">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1241">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9Q7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004164" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10031" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945837" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129202" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0474" fbc:label="b0474" metaid="G_b0474">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0474">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69441" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001645" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10032" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945097" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128458" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0120" fbc:label="b0120" metaid="G_b0120">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0120">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7F6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000416" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10962" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947719" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128113" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3806" fbc:label="b3806" metaid="G_b3806">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3806">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00936" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012432" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10170" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947755" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131658" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2393" fbc:label="b2393" metaid="G_b2393">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2393">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFF2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007893" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11971" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946895" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130325" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2964" fbc:label="b2964" metaid="G_b2964">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2964">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFF4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009728" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10664" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946282" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111521" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2406" fbc:label="b2406" metaid="G_b2406">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2406">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P45562" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007925" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13159" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946868" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130332" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1992" fbc:label="b1992" metaid="G_b1992">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1992">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P36561" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006613" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12150" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946520" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129933" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0469" fbc:label="b0469" metaid="G_b0469">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0469">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69503" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001628" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10051" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945113" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128453" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2750" fbc:label="b2750" metaid="G_b2750">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2750">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6J1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009023" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10185" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947221" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130657" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1131" fbc:label="b1131" metaid="G_b1131">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1131">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB89" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003810" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11314" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945695" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129094" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4177" fbc:label="b4177" metaid="G_b4177">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4177">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7D4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013672" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10790" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948695" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131999" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0677" fbc:label="b0677" metaid="G_b0677">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0677">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AF18" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002301" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10632" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945289" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128653" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3619" fbc:label="b3619" metaid="G_b3619">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3619">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P67910" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011843" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10838" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948134" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131490" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2937" fbc:label="b2937" metaid="G_b2937">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2937">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P60651" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009632" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10960" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947715" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130838" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3958" fbc:label="b3958" metaid="G_b3958">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3958">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11446" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012964" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10065" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948455" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131796" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0159" fbc:label="b0159" metaid="G_b0159">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0159">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AF12" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000545" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11090" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948542" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128152" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4006" fbc:label="b4006" metaid="G_b4006">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4006">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15639" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013097" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10795" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948503" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131836" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0522" fbc:label="b0522" metaid="G_b0522">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0522">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09029" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001794" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10796" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945153" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128506" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0523" fbc:label="b0523" metaid="G_b0523">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0523">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AG18" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001796" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10793" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949031" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128507" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0727" fbc:label="b0727" metaid="G_b0727">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0727">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFG6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002480" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10980" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945307" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128702" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0116" fbc:label="b0116" metaid="G_b0116">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0116">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9P0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000404" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10543" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944854" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128109" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0726" fbc:label="b0726" metaid="G_b0726">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0726">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFG3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002478" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10979" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945303" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128701" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2587" fbc:label="b2587" metaid="G_b2587">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2587">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEX3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008515" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10522" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947069" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130512" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0381" fbc:label="b0381" metaid="G_b0381">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0381">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6J8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001309" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10213" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945313" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128366" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0092" fbc:label="b0092" metaid="G_b0092">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0092">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07862" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000326" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10214" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946324" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128085" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4053" fbc:label="b4053" metaid="G_b4053">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4053">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6B4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013272" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10001" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948564" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131879" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1190" fbc:label="b1190" metaid="G_b1190">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1190">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P29012" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003997" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11408" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945754" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129153" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2551" fbc:label="b2551" metaid="G_b2551">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2551">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A825" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008389" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10408" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947022" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130476" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3457" fbc:label="b3457" metaid="G_b3457">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3457">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEX7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011290" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10538" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947965" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131329" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3455" fbc:label="b3455" metaid="G_b3455">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3455">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9S7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011284" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10537" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947967" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131327" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3456" fbc:label="b3456" metaid="G_b3456">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3456">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22729" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011287" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10541" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947966" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131328" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3454" fbc:label="b3454" metaid="G_b3454">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3454">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22731" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011282" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10536" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947961" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111594" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3460" fbc:label="b3460" metaid="G_b3460">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3460">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AD96" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011299" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10539" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947971" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176358" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0007" fbc:label="b0007" metaid="G_b0007">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0007">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30143" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000020" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11555" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944745" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128001" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4208" fbc:label="b4208" metaid="G_b4208">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4208">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAE0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013764" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12504" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948725" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132030" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0356" fbc:label="b0356" metaid="G_b0356">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0356">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25437" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001221" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG50010" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944988" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128341" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1385" fbc:label="b1385" metaid="G_b1385">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1385">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P80668" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004637" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13179" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945933" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111264" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1300" fbc:label="b1300" metaid="G_b1300">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1300">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23883" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004376" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10036" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947003" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129261" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0516" fbc:label="b0516" metaid="G_b0516">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0516">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77425" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001777" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13623" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945150" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128500" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0512" fbc:label="b0512" metaid="G_b0512">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0512">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77671" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001768" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13619" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945134" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128496" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0511" fbc:label="b0511" metaid="G_b0511">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0511">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P75712" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001764" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13618" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945138" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111144" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3091" fbc:label="b3091" metaid="G_b3091">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3091">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P42604" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010164" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12734" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947603" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130986" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3416" fbc:label="b3416" metaid="G_b3416">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3416">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15977" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011151" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10561" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947923" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131292" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3223" fbc:label="b3223" metaid="G_b3223">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3223">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A761" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010575" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12816" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947745" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131113" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3222" fbc:label="b3222" metaid="G_b3222">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3222">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P45425" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010571" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12815" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947757" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111558" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0774" fbc:label="b0774" metaid="G_b0774">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0774">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P12995" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002635" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10117" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945376" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128742" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3929" fbc:label="b3929" metaid="G_b3929">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3929">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A8R0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012837" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11879" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948419" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131767" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3994" fbc:label="b3994" metaid="G_b3994">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3994">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30136" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013057" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11585" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948492" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131824" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1982" fbc:label="b1982" metaid="G_b1982">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1982">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AE12" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006579" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10039" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946508" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129926" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1263" fbc:label="b1263" metaid="G_b1263">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1263">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00904" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004241" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11027" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945109" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129224" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1264" fbc:label="b1264" metaid="G_b1264">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1264">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00895" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004244" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11028" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945846" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129225" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0776" fbc:label="b0776" metaid="G_b0776">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0776">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P12998" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002646" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10121" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945384" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128744" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0049" fbc:label="b0049" metaid="G_b0049">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0049">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05637" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000167" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10048" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944770" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128043" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0414" fbc:label="b0414" metaid="G_b0414">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0414">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25539" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001441" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11321" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945620" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128399" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0062" fbc:label="b0062" metaid="G_b0062">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0062">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08202" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000221" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10052" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947511" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128056" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1900" fbc:label="b1900" metaid="G_b1900">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1900">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAF3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006328" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10058" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946408" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129850" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1901" fbc:label="b1901" metaid="G_b1901">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1901">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P02924" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006330" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10057" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946409" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129851" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1898" fbc:label="b1898" />
      <fbc:geneProduct fbc:id="G_b1899" fbc:label="b1899" />
      <fbc:geneProduct fbc:id="G_b2841" fbc:label="b2841" metaid="G_b2841">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2841">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AE24" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009319" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10056" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947341" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130745" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2938" fbc:label="b2938" metaid="G_b2938">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2938">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21170" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009634" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10959" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947432" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130839" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4117" fbc:label="b4117" metaid="G_b4117">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4117">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P28629" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013483" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11501" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948638" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:221800782" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1605" fbc:label="b1605" metaid="G_b1605">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1605">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAE5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005362" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13930" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945159" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129563" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3960" fbc:label="b3960" metaid="G_b3960">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3960">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11447" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012969" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11223" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948463" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131798" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3172" fbc:label="b3172" metaid="G_b3172">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3172">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6E4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010428" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10068" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947590" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131063" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2308" fbc:label="b2308" metaid="G_b2308">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2308">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P52094" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007615" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12125" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947235" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130243" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2306" fbc:label="b2306" metaid="G_b2306">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2306">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07109" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007610" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10452" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946789" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130241" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0860" fbc:label="b0860" metaid="G_b0860">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0860">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30860" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002927" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11628" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948981" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128828" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2307" fbc:label="b2307" metaid="G_b2307">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2307">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEU3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007613" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10007" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946790" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130242" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0861" fbc:label="b0861" metaid="G_b0861">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0861">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AE30" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002931" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11627" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949066" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128829" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0864" fbc:label="b0864" metaid="G_b0864">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0864">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAF6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002939" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11624" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945489" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128832" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0862" fbc:label="b0862" metaid="G_b0862">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0862">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AE34" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002933" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11626" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949046" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128830" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2310" fbc:label="b2310" metaid="G_b2310">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2310">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09551" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007621" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10072" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949030" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130245" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3433" fbc:label="b3433" metaid="G_b3433">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3433">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9Q9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011211" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10088" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947939" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131307" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0828" fbc:label="b0828" metaid="G_b0828">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0828">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37595" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002828" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12407" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945456" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128796" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2957" fbc:label="b2957" metaid="G_b2957">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2957">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00805" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009700" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10046" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947454" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130858" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1767" fbc:label="b1767" metaid="G_b1767">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1767">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A962" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005886" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10045" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946278" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129721" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0674" fbc:label="b0674" metaid="G_b0674">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0674">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22106" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002292" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10092" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945281" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128650" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3744" fbc:label="b3744" metaid="G_b3744">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3744">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00963" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012242" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10091" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948258" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131612" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0131" fbc:label="b0131" metaid="G_b0131">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0131">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A790" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000459" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11747" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945686" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128124" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4244" fbc:label="b4244" metaid="G_b4244">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4244">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7F3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013889" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10811" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948763" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132066" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4245" fbc:label="b4245" metaid="G_b4245">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4245">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A786" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013892" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10805" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948767" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132067" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0002" fbc:label="b0002" metaid="G_b0002">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0002">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00561" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000008" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10998" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945803" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16127996" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3940" fbc:label="b3940" metaid="G_b3940">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3940">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00562" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012889" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10590" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948433" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131778" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4024" fbc:label="b4024" metaid="G_b4024">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4024">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08660" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013156" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10550" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948531" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131850" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2574" fbc:label="b2574" metaid="G_b2574">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2574">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10902" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008471" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10631" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947049" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130499" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4139" fbc:label="b4139" metaid="G_b4139">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4139">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC38" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013551" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10095" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948658" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111690" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0928" fbc:label="b0928" metaid="G_b0928">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0928">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00509" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003150" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10096" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945553" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128895" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0652" fbc:label="b0652" metaid="G_b0652">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0652">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAG3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002232" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12663" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945254" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128635" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0653" fbc:label="b0653" metaid="G_b0653">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0653">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AER5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002234" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12662" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947354" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128636" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0654" fbc:label="b0654" metaid="G_b0654">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0654">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AER3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002239" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12661" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945443" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128637" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0655" fbc:label="b0655" metaid="G_b0655">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0655">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37902" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002241" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12700" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946938" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128638" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4077" fbc:label="b4077" metaid="G_b4077">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4077">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21345" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013357" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10405" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948591" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131903" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3528" fbc:label="b3528" metaid="G_b3528">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3528">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A830" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011527" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20044" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948039" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131400" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4138" fbc:label="b4138" metaid="G_b4138">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4138">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABN5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013546" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11225" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948659" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131963" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4123" fbc:label="b4123" metaid="G_b4123">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4123">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABN9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013503" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10006" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948641" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131949" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1747" fbc:label="b1747" metaid="G_b1747">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1747">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AE37" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005822" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13998" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946261" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129701" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2019" fbc:label="b2019" metaid="G_b2019">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2019">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P60757" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006713" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10449" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946549" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129960" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3737" fbc:label="b3737" metaid="G_b3737">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3737">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P68699" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012220" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10102" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948253" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131605" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3731" fbc:label="b3731" metaid="G_b3731">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3731">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6E6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012206" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10100" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948245" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131599" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3738" fbc:label="b3738" metaid="G_b3738">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3738">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB98" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012222" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10099" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948252" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131606" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3736" fbc:label="b3736" metaid="G_b3736">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3736">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABA0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012217" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10103" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948247" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131604" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3739" fbc:label="b3739" metaid="G_b3739">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3739">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABC0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012224" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10106" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948251" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111645" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3732" fbc:label="b3732" metaid="G_b3732">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3732">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABB4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012208" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10101" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948244" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131600" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3733" fbc:label="b3733" metaid="G_b3733">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3733">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABA6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012211" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10104" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948243" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131601" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3734" fbc:label="b3734" metaid="G_b3734">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3734">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABB0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012213" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10098" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948242" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131602" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3735" fbc:label="b3735" metaid="G_b3735">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3735">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABA4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012215" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10105" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948254" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131603" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0312" fbc:label="b0312" metaid="G_b0312">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0312">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17445" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001072" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10110" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947376" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128297" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3551" fbc:label="b3551" metaid="G_b3551">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3551">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P20099" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011597" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10124" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946915" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698325" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0775" fbc:label="b0775" metaid="G_b0775">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0775">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P12996" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002644" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10118" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945370" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128743" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0180" fbc:label="b0180" metaid="G_b0180">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0180">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6Q6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000613" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11284" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944888" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128173" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1095" fbc:label="b1095" metaid="G_b1095">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1095">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAI5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003705" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12606" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946665" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129058" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2323" fbc:label="b2323" metaid="G_b2323">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2323">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A953" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007675" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10274" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946799" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130258" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1288" fbc:label="b1288" metaid="G_b1288">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1288">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEK4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004327" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11528" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945870" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129249" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0954" fbc:label="b0954" metaid="G_b0954">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0954">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6Q3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003229" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10273" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945568" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128921" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1093" fbc:label="b1093" metaid="G_b1093">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1093">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEK2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003700" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11318" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945645" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129056" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4132" fbc:label="b4132" metaid="G_b4132">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4132">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAE8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013530" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10132" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948654" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131958" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1732" fbc:label="b1732" metaid="G_b1732">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1732">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21179" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005780" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10509" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946234" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176140" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3942" fbc:label="b3942" metaid="G_b3942">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3942">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P13029" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012901" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10511" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948431" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131780" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1270" fbc:label="b1270" metaid="G_b1270">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1270">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9H5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004261" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10130" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945839" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129231" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3966" fbc:label="b3966" metaid="G_b3966">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3966">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06129" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012987" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10126" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948468" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131804" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0158" fbc:label="b0158" metaid="G_b0158">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0158">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37028" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000542" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12334" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947574" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128151" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1711" fbc:label="b1711" metaid="G_b1711">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1711">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06609" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005712" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10127" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945877" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129667" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1709" fbc:label="b1709" metaid="G_b1709">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1709">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06611" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005707" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10128" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945751" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129665" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0323" fbc:label="b0323" metaid="G_b0323">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0323">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77624" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001107" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13593" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944984" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128308" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2874" fbc:label="b2874" metaid="G_b2874">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2874">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46807" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009438" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13057" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947358" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130776" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0521" fbc:label="b0521" metaid="G_b0521">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0521">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37306" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001791" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12384" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944972" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128505" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0032" fbc:label="b0032" metaid="G_b0032">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0032">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6F1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000118" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10134" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949025" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128026" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0033" fbc:label="b0033" metaid="G_b0033">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0033">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00968" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000121" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10135" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944775" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128027" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3918" fbc:label="b3918" metaid="G_b3918">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3918">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06282" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012797" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10138" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948410" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131756" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1208" fbc:label="b1208" metaid="G_b1208">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1208">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P62615" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004056" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11294" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945774" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129171" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2131" fbc:label="b2131" metaid="G_b2131">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2131">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33362" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007040" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12012" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946681" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130069" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2128" fbc:label="b2128" metaid="G_b2128">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2128">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33359" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007032" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12009" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949028" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130066" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2129" fbc:label="b2129" metaid="G_b2129">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2129">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33360" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007034" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12010" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946659" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130067" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2130" fbc:label="b2130" metaid="G_b2130">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2130">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33361" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007038" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12011" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946660" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130068" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1801" fbc:label="b1801" metaid="G_b1801">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1801">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABD1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005995" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13508" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947326" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129755" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0314" fbc:label="b0314" metaid="G_b0314">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0314">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABC9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001080" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10112" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945079" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128299" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2600" fbc:label="b2600" metaid="G_b2600">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2600">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07023" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008547" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11039" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947115" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130521" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2599" fbc:label="b2599" metaid="G_b2599">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2599">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9J8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008545" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10707" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947081" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130520" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2329" fbc:label="b2329" metaid="G_b2329">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2329">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P12008" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007693" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10075" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946814" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130264" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4039" fbc:label="b4039" metaid="G_b4039">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4039">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P26602" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013227" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11369" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948545" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111677" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2542" fbc:label="b2542" metaid="G_b2542">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2542">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77650" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008363" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13460" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945427" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130467" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2538" fbc:label="b2538" metaid="G_b2538">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2538">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABR5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008355" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13456" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946998" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130463" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2539" fbc:label="b2539" metaid="G_b2539">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2539">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q47140" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008357" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13457" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946997" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130464" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2540" fbc:label="b2540" metaid="G_b2540">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2540">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABW0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008359" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13458" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947015" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130465" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0617" fbc:label="b0617" metaid="G_b0617">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0617">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69330" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002124" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13543" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945415" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128600" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0615" fbc:label="b0615" metaid="G_b0615">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0615">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P75726" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002119" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13541" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945230" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128598" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0616" fbc:label="b0616" metaid="G_b0616">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0616">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9I1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002122" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13542" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945406" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111153" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0612" fbc:label="b0612" metaid="G_b0612">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0612">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AE74" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002110" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13538" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949070" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128595" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1249" fbc:label="b1249" metaid="G_b1249">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1249">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6H8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004186" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11608" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945821" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129210" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2436" fbc:label="b2436" metaid="G_b2436">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2436">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P36553" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008034" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12189" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946908" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130361" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0038" fbc:label="b0038" metaid="G_b0038">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0038">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31572" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000136" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11559" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948997" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128032" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0036" fbc:label="b0036" metaid="G_b0036">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0036">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31551" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000130" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11557" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948995" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:221800779" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0040" fbc:label="b0040" metaid="G_b0040">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0040">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31553" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000141" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11561" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944765" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128034" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0720" fbc:label="b0720" metaid="G_b0720">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0720">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABH7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002451" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10402" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945323" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128695" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0337" fbc:label="b0337" metaid="G_b0337">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0337">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25524" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001161" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11326" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944996" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128322" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0336" fbc:label="b0336" metaid="G_b0336">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0336">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AA82" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001159" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11327" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944994" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128321" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2780" fbc:label="b2780" metaid="G_b2780">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2780">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7E5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009112" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10810" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946116" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130687" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3425" fbc:label="b3425" metaid="G_b3425">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3425">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6V5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011182" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10395" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947935" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131299" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0340" fbc:label="b0340" metaid="G_b0340">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0340">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00816" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001173" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10175" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948998" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128325" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0341" fbc:label="b0341" metaid="G_b0341">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0341">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17583" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001175" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10177" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946770" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128326" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3708" fbc:label="b3708" metaid="G_b3708">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3708">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A853" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012133" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11005" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948221" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111643" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2414" fbc:label="b2414" metaid="G_b2414">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2414">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABK5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007957" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10192" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946877" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130340" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2421" fbc:label="b2421" metaid="G_b2421">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2421">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16703" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007982" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10193" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946888" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130347" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3008" fbc:label="b3008" metaid="G_b3008">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3008">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06721" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009878" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10583" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946240" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130906" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1622" fbc:label="b1622" metaid="G_b1622">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1622">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23256" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005432" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10564" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945937" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129580" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0733" fbc:label="b0733" metaid="G_b0733">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0733">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABJ9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002499" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10173" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945341" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111166" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0734" fbc:label="b0734" metaid="G_b0734">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0734">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABK2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002501" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10174" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945347" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128709" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0429" fbc:label="b0429" metaid="G_b0429">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0429">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABJ6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001487" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10181" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944918" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128414" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0431" fbc:label="b0431" metaid="G_b0431">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0431">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABI8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001493" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10179" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945615" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128416" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0432" fbc:label="b0432" metaid="G_b0432">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0432">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABJ1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001495" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10178" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945080" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128417" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0430" fbc:label="b0430" metaid="G_b0430">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0430">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABJ3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001491" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10180" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946897" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128415" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2143" fbc:label="b2143" metaid="G_b2143">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2143">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABF6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007086" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10137" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946663" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130081" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2066" fbc:label="b2066" metaid="G_b2066">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2066">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A8F4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006840" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11701" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946597" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111379" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0910" fbc:label="b0910" metaid="G_b0910">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0910">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6I0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003101" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11265" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945535" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128877" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1189" fbc:label="b1189" metaid="G_b1189">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1189">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6J5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003995" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11407" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945752" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129152" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4042" fbc:label="b4042" metaid="G_b4042">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4042">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABN1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013235" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10224" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948543" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131868" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2838" fbc:label="b2838" metaid="G_b2838">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2838">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00861" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009309" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10549" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947313" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130742" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3809" fbc:label="b3809" metaid="G_b3809">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3809">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6K1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012442" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10209" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948364" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111650" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0175" fbc:label="b0175" metaid="G_b0175">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0175">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABG1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000600" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10139" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944876" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111095" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3041" fbc:label="b3041" metaid="G_b3041">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3041">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7J0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009980" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10465" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947526" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130937" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0778" fbc:label="b0778" metaid="G_b0778">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0778">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P13000" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002650" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10120" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945387" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128746" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2065" fbc:label="b2065" metaid="G_b2065">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2065">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P28248" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006835" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11418" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946593" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130005" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3693" fbc:label="b3693" metaid="G_b3693">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3693">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31459" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012076" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11717" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948207" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131561" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3909" fbc:label="b3909" metaid="G_b3909">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3909">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A712" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012764" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11869" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948407" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111665" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3526" fbc:label="b3526" metaid="G_b3526">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3526">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37647" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011519" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12253" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948041" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111606" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1704" fbc:label="b1704" metaid="G_b1704">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1704">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00887" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005686" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10080" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946229" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129660" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0754" fbc:label="b0754" metaid="G_b0754">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0754">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB91" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002560" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10079" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945605" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128722" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2601" fbc:label="b2601" metaid="G_b2601">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2601">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00888" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008549" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10078" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947084" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130522" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3692" fbc:label="b3692" />
      <fbc:geneProduct fbc:id="G_b3648" fbc:label="b3648" metaid="G_b3648">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3648">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P60546" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011930" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10965" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948163" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131519" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3771" fbc:label="b3771" metaid="G_b3771">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3771">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05791" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012318" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10496" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948277" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176404" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1199" fbc:label="b1199" metaid="G_b1199">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1199">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76014" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004026" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13900" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945748" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129162" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1198" fbc:label="b1198" metaid="G_b1198">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1198">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37349" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004024" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12399" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945749" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:226524708" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1200" fbc:label="b1200" metaid="G_b1200">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1200">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76015" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004029" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13901" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945747" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:162135900" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0596" fbc:label="b0596" metaid="G_b0596">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0596">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15047" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002054" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10259" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945284" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128579" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0594" fbc:label="b0594" metaid="G_b0594">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0594">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10378" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002049" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10263" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947426" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128577" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2541" fbc:label="b2541" metaid="G_b2541">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2541">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0CI31" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008361" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13459" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945346" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130466" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0348" fbc:label="b0348" metaid="G_b0348">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0348">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABR9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001198" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20274" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945047" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128333" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0031" fbc:label="b0031" metaid="G_b0031">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0031">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04036" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000112" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10206" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944762" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128025" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2478" fbc:label="b2478" metaid="G_b2478">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2478">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6L2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008163" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10205" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946952" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130403" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0048" fbc:label="b0048" metaid="G_b0048">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0048">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABQ4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000164" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10326" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944790" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128042" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2315" fbc:label="b2315" metaid="G_b2315">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2315">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08192" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007645" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10327" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945451" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130250" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3930" fbc:label="b3930" metaid="G_b3930">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3930">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32166" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012841" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11880" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948418" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131768" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3058" fbc:label="b3058" metaid="G_b3058">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3058">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC16" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010036" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11673" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947544" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111533" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0945" fbc:label="b0945" metaid="G_b0945">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0945">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7E1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003201" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10807" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945556" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128912" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1062" fbc:label="b1062" metaid="G_b1062">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1062">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05020" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003607" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10806" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945787" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129025" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3177" fbc:label="b3177" metaid="G_b3177">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3177">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC13" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010441" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG50011" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947691" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111553" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1693" fbc:label="b1693" metaid="G_b1693">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1693">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05194" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005653" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10076" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946210" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129649" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3389" fbc:label="b3389" metaid="G_b3389">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3389">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07639" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011067" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10074" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947927" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131267" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3012" fbc:label="b3012" metaid="G_b3012">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3012">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46857" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009892" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13015" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947495" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111527" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0207" fbc:label="b0207" metaid="G_b0207">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0207">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30863" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000692" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11648" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944901" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128194" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0421" fbc:label="b0421" metaid="G_b0421">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0421">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22939" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001463" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10508" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945064" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128406" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0029" fbc:label="b0029" metaid="G_b0029">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0029">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P62623" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000103" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11081" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944777" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128023" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2232" fbc:label="b2232" metaid="G_b2232">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2232">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17993" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007374" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11143" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946607" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130167" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0894" fbc:label="b0894" metaid="G_b0894">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0894">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P18775" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003044" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10232" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945508" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111183" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0895" fbc:label="b0895" metaid="G_b0895">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0895">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P18776" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003046" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10233" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945507" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128862" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0896" fbc:label="b0896" metaid="G_b0896">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0896">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P18777" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003048" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10234" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945502" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128863" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1872" fbc:label="b1872" metaid="G_b1872">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1872">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P46923" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006244" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13276" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946389" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111350" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1873" fbc:label="b1873" metaid="G_b1873">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1873">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P52005" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006246" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13277" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946490" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129826" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1865" fbc:label="b1865" metaid="G_b1865">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1865">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFC0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006223" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11138" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946383" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129818" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0099" fbc:label="b0099" metaid="G_b0099">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0099">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08337" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000346" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10626" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944824" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128092" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3575" fbc:label="b3575" metaid="G_b3575">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3575">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37672" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011682" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12279" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948096" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131446" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0425" fbc:label="b0425" metaid="G_b0425">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0425">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9J4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001474" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13271" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945065" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128410" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3774" fbc:label="b3774" metaid="G_b3774">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3774">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05793" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012331" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10495" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948286" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131632" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4381" fbc:label="b4381" metaid="G_b4381">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4381">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6L0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014371" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10221" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948902" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132198" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1098" fbc:label="b1098" metaid="G_b1098">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1098">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A720" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003714" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12302" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945663" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129061" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1238" fbc:label="b1238" metaid="G_b1238">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1238">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23331" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004157" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10994" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945834" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129199" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4382" fbc:label="b4382" metaid="G_b4382">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4382">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07650" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014374" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10219" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948901" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132199" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4384" fbc:label="b4384" metaid="G_b4384">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4384">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABP8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014379" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10222" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945654" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132201" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3640" fbc:label="b3640" metaid="G_b3640">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3640">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06968" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011899" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10251" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948607" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131511" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0173" fbc:label="b0173" metaid="G_b0173">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0173">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P45568" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000592" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12715" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945019" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128166" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0420" fbc:label="b0420" metaid="G_b0420">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0420">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77488" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001461" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13612" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945060" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128405" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3564" fbc:label="b3564" metaid="G_b3564">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3564">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09099" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011637" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11075" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948133" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131435" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3603" fbc:label="b3603" metaid="G_b3603">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3603">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33231" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011777" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11961" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948114" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131474" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2975" fbc:label="b2975" metaid="G_b2975">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2975">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46839" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009763" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12995" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947259" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130875" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2927" fbc:label="b2927" metaid="G_b2927">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2927">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9B6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009607" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10368" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947413" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130828" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1850" fbc:label="b1850" metaid="G_b1850">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1850">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A955" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006164" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10256" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946367" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129803" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1851" fbc:label="b1851" metaid="G_b1851">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1851">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ADF6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006166" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10257" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946362" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129804" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1054" fbc:label="b1054" metaid="G_b1054">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1054">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ACV0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003574" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10464" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946216" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129017" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1855" fbc:label="b1855" metaid="G_b1855">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1855">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24205" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006185" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10614" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945143" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129808" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2378" fbc:label="b2378" metaid="G_b2378">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2378">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ACV2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007848" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12901" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946847" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111426" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2779" fbc:label="b2779" metaid="G_b2779">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2779">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6P9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009110" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10258" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945032" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130686" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0583" fbc:label="b0583" metaid="G_b0583">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0583">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P19925" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002009" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10262" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945194" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:345452717" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2440" fbc:label="b2440" metaid="G_b2440">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2440">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P19636" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008043" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG50007" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946925" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130365" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2441" fbc:label="b2441" metaid="G_b2441">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2441">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEJ6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008046" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG50006" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946924" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111437" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3946" fbc:label="b3946" metaid="G_b3946">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3946">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32669" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012915" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11905" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948439" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131784" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0825" fbc:label="b0825" metaid="G_b0825">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0825">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P78055" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002818" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13471" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945449" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111174" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1805" fbc:label="b1805" metaid="G_b1805">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1805">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69451" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006005" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11530" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946327" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129759" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0221" fbc:label="b0221" metaid="G_b0221">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0221">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q47146" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000743" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13145" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949007" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111100" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3845" fbc:label="b3845" metaid="G_b3845">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3845">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21151" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012562" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10278" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948324" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176430" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3846" fbc:label="b3846" metaid="G_b3846">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3846">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21177" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012564" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10279" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948336" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131692" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1773" fbc:label="b1773" metaid="G_b1773">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1773">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77704" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005906" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13485" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946291" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129727" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2097" fbc:label="b2097" metaid="G_b2097">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2097">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A991" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006941" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14062" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946632" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111385" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2925" fbc:label="b2925" metaid="G_b2925">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2925">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB71" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009600" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10282" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947415" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130826" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4232" fbc:label="b4232" metaid="G_b4232">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4232">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A993" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013842" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10283" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948753" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132054" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2802" fbc:label="b2802" metaid="G_b2802">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2802">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69922" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009189" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10349" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946195" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130709" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2803" fbc:label="b2803" metaid="G_b2803">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2803">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11553" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009191" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10350" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946022" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:674842733" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2738" fbc:label="b2738" metaid="G_b2738">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2738">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46890" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008994" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13106" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947197" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130645" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2800" fbc:label="b2800" metaid="G_b2800">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2800">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB87" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009179" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10348" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947282" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130707" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0475" fbc:label="b0475" metaid="G_b0475">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0475">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23871" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001649" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10431" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947532" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128459" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1476" fbc:label="b1476" metaid="G_b1476">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1476">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEK7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004921" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11229" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946038" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129435" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3892" fbc:label="b3892" metaid="G_b3892">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3892">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEL0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012705" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11856" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948383" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131732" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3893" fbc:label="b3893" metaid="G_b3893">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3893">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAJ5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012708" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11857" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948395" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131733" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4079" fbc:label="b4079" metaid="G_b4079">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4079">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07658" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013363" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10285" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948584" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131905" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1475" fbc:label="b1475" metaid="G_b1475">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1475">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAJ3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004919" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11228" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948794" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129434" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3894" fbc:label="b3894" metaid="G_b3894">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3894">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32176" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012710" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11858" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948394" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131734" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1474" fbc:label="b1474" metaid="G_b1474">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1474">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24183" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004917" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11227" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946035" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129433" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3409" fbc:label="b3409" metaid="G_b3409">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3409">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33650" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011126" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12102" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947919" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131285" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2719" fbc:label="b2719" metaid="G_b2719">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2719">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16433" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008936" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10480" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947191" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130626" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2720" fbc:label="b2720" metaid="G_b2720">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2720">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16432" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008942" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10479" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947048" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130627" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2722" fbc:label="b2722" metaid="G_b2722">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2722">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16430" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008946" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10477" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948994" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130629" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2724" fbc:label="b2724" metaid="G_b2724">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2724">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAK1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008950" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10475" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948002" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130631" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2723" fbc:label="b2723" metaid="G_b2723">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2723">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16429" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008948" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10476" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945327" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130630" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2721" fbc:label="b2721" metaid="G_b2721">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2721">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16431" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008944" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10478" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947396" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130628" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0025" fbc:label="b0025" metaid="G_b0025">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0025">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AG40" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000091" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11079" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949129" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128019" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2492" fbc:label="b2492" metaid="G_b2492">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2492">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77733" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008206" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14220" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949032" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130417" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0904" fbc:label="b0904" metaid="G_b0904">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0904">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC23" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003073" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11258" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945513" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128871" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4154" fbc:label="b4154" metaid="G_b4154">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4154">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00363" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013604" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10330" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948667" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131979" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4152" fbc:label="b4152" metaid="G_b4152">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4152">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A8Q0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013598" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10332" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948680" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131977" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4153" fbc:label="b4153" metaid="G_b4153">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4153">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC47" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013602" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10331" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948666" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131978" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4151" fbc:label="b4151" metaid="G_b4151">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4151">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A8Q3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013595" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10333" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948668" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131976" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2168" fbc:label="b2168" metaid="G_b2168">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2168">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEW9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007174" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10337" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946676" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130106" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2169" fbc:label="b2169" metaid="G_b2169">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2169">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69811" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007180" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10005" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946677" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130107" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2167" fbc:label="b2167" metaid="G_b2167">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2167">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P20966" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007172" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10336" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946672" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130105" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1232" fbc:label="b1232" metaid="G_b1232">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1232">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37051" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004138" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11819" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945827" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129193" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3666" fbc:label="b3666" metaid="G_b3666">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3666">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGC0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011981" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11054" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948201" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131536" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2801" fbc:label="b2801" metaid="G_b2801">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2801">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11551" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009187" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10352" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947487" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130708" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1611" fbc:label="b1611" metaid="G_b1611">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1611">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05042" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005380" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10358" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946147" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129569" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1612" fbc:label="b1612" metaid="G_b1612">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1612">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC33" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005392" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10356" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946826" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129570" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4122" fbc:label="b4122" metaid="G_b4122">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4122">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P14407" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013501" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10357" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948642" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131948" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0621" fbc:label="b0621" metaid="G_b0621">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0621">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABP3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002133" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13545" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945000" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128604" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3730" fbc:label="b3730" metaid="G_b3730">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3730">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ACC7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012201" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11198" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948246" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131598" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1002" fbc:label="b1002" metaid="G_b1002">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1002">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P19926" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003388" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10033" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945773" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128968" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3789" fbc:label="b3789" metaid="G_b3789">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3789">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P61887" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012379" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11454" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948299" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131645" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2039" fbc:label="b2039" metaid="G_b2039">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2039">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37744" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006762" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11978" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945154" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129979" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0154" fbc:label="b0154" metaid="G_b0154">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0154">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23893" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000529" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10432" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946892" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128147" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3608" fbc:label="b3608" metaid="G_b3608">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3608">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6S7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011795" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20091" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948125" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131479" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2241" fbc:label="b2241" metaid="G_b2241">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2241">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9C0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007422" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10391" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946713" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130176" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2242" fbc:label="b2242" metaid="G_b2242">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2242">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P13033" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007424" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10392" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946733" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130177" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3426" fbc:label="b3426" metaid="G_b3426">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3426">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P13035" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011191" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10394" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947934" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131300" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2243" fbc:label="b2243" metaid="G_b2243">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2243">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A996" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007426" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10393" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946735" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130178" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0243" fbc:label="b0243" metaid="G_b0243">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0243">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07004" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000832" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10767" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946680" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128229" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0678" fbc:label="b0678" metaid="G_b0678">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0678">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A759" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002304" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10633" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945290" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128654" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1852" fbc:label="b1852" metaid="G_b1852">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1852">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC53" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006171" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11221" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946370" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129805" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3128" fbc:label="b3128" metaid="G_b3128">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3128">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39829" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010284" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12522" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947641" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131020" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3691" fbc:label="b3691" metaid="G_b3691">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3691">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AA76" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012072" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11715" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948196" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111641" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3127" fbc:label="b3127" metaid="G_b3127">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3127">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AA80" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010279" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12760" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947642" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131019" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0757" fbc:label="b0757" metaid="G_b0757">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0757">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6T3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002568" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10363" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945358" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128725" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2045" fbc:label="b2045" metaid="G_b2045">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2045">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P71242" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006780" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13577" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946569" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129985" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4119" fbc:label="b4119" metaid="G_b4119">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4119">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06720" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013493" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10577" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948636" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131945" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2093" fbc:label="b2093" metaid="G_b2093">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2093">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37188" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006930" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12415" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946610" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130031" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2092" fbc:label="b2092" metaid="G_b2092">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2092">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69831" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006928" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12416" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946603" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2094" fbc:label="b2094" metaid="G_b2094">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2094">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69828" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006932" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12414" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946633" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130032" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1236" fbc:label="b1236" metaid="G_b1236">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1236">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEP3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004150" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11319" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945730" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129197" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3093" fbc:label="b3093" metaid="G_b3093">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3093">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AA78" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010171" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12738" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947601" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:674842734" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2042" fbc:label="b2042" metaid="G_b2042">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2042">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAB6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006771" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13578" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946560" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129982" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2150" fbc:label="b2150" metaid="G_b2150">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2150">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEE5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007108" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10593" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949041" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130088" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2149" fbc:label="b2149" metaid="G_b2149">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2149">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAG8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007103" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10592" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949036" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130087" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2148" fbc:label="b2148" metaid="G_b2148">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2148">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23200" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007101" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10594" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949039" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130086" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2943" fbc:label="b2943" metaid="G_b2943">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2943">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEP1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009654" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12148" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947434" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130844" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1779" fbc:label="b1779" metaid="G_b1779">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1779">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9B2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005920" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10367" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947679" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129733" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1416" fbc:label="b1416" />
      <fbc:geneProduct fbc:id="G_b1417" fbc:label="b1417" />
      <fbc:geneProduct fbc:id="G_b2500" fbc:label="b2500" metaid="G_b2500">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2500">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08179" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008232" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10799" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946973" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130425" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1415" fbc:label="b1415" metaid="G_b1415">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1415">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25553" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004725" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10035" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945672" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129376" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2052" fbc:label="b2052" metaid="G_b2052">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2052">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32055" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006801" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11788" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946563" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129992" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3729" fbc:label="b3729" metaid="G_b3729">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3729">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17169" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012198" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10382" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948241" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131597" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0124" fbc:label="b0124" metaid="G_b0124">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0124">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15877" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000433" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10369" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944830" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128117" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4321" fbc:label="b4321" metaid="G_b4321">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4321">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC94" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014161" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12563" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948848" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132142" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4265" fbc:label="b4265" metaid="G_b4265">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4265">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39344" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013969" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12539" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948798" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132087" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3415" fbc:label="b3415" metaid="G_b3415">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3415">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39835" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011148" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12380" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947924" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176350" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3428" fbc:label="b3428" metaid="G_b3428">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3428">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC86" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011198" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10380" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947931" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131302" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3417" fbc:label="b3417" metaid="G_b3417">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3417">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00490" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011154" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10560" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947922" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176351" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3126" fbc:label="b3126" metaid="G_b3126">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3126">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23522" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010276" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10016" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947630" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131018" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2788" fbc:label="b2788" metaid="G_b2788">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2788">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46915" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009136" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13168" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947261" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130695" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2787" fbc:label="b2787" metaid="G_b2787">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2787">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AES2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009134" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13167" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947258" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130694" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2789" fbc:label="b2789" metaid="G_b2789">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2789">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46916" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009139" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13169" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947265" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130696" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3429" fbc:label="b3429" metaid="G_b3429">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3429">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6U8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011200" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10377" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947932" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131303" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1621" fbc:label="b1621" metaid="G_b1621">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1621">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P19642" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005429" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10563" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946009" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129579" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3430" fbc:label="b3430" metaid="G_b3430">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3430">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6V1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011204" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10379" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947942" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131304" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3870" fbc:label="b3870" metaid="G_b3870">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3870">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9C5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012640" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10383" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948370" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131710" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1297" fbc:label="b1297" metaid="G_b1297">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1297">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P78061" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004365" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13908" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946202" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111244" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0809" fbc:label="b0809" metaid="G_b0809">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0809">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10346" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002764" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10389" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945435" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128777" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0810" fbc:label="b0810" metaid="G_b0810">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0810">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEQ6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002766" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10388" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945621" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128778" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0811" fbc:label="b0811" metaid="G_b0811">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0811">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEQ3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002771" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10386" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944872" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128779" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2091" fbc:label="b2091" metaid="G_b2091">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2091">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9S3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006924" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12417" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946598" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130029" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0242" fbc:label="b0242" metaid="G_b0242">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0242">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7B5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000830" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10768" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946425" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128228" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1492" fbc:label="b1492" metaid="G_b1492">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1492">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P63235" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004973" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12456" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946057" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129451" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2688" fbc:label="b2688" metaid="G_b2688">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2688">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6W9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008845" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10418" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944881" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130600" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1493" fbc:label="b1493" metaid="G_b1493">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1493">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69910" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004976" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11490" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946058" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129452" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3517" fbc:label="b3517" metaid="G_b3517">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3517">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69908" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011490" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG50009" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948027" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131389" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1761" fbc:label="b1761" metaid="G_b1761">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1761">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00370" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005865" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10372" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946802" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129715" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0485" fbc:label="b0485" metaid="G_b0485">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0485">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77454" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001688" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13247" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946187" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128469" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1524" fbc:label="b1524" metaid="G_b1524">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1524">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6W0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005086" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13816" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944973" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129483" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2312" fbc:label="b2312" metaid="G_b2312">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2312">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AG16" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007631" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10794" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946794" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130247" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3967" fbc:label="b3967" metaid="G_b3967">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3967">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22634" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012989" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11204" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948467" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111671" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3213" fbc:label="b3213" metaid="G_b3213">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3213">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09832" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010547" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10404" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947723" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131103" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3212" fbc:label="b3212" metaid="G_b3212">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3212">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09831" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010545" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10403" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947724" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:308209621" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1210" fbc:label="b1210" metaid="G_b1210">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1210">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6X1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004064" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10427" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945777" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129173" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2400" fbc:label="b2400" metaid="G_b2400">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2400">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04805" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007911" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10407" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946906" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130330" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3653" fbc:label="b3653" metaid="G_b3653">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3653">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AER8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011941" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10406" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948166" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131524" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0507" fbc:label="b0507" metaid="G_b0507">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0507">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEP7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001754" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11583" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945394" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128491" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3927" fbc:label="b3927" metaid="G_b3927">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3927">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AER0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012828" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10396" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948422" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131765" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3617" fbc:label="b3617" metaid="G_b3617">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3617">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB77" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011832" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10512" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948138" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131488" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3450" fbc:label="b3450" metaid="G_b3450">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3450">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10907" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011266" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11048" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947953" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111593" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3453" fbc:label="b3453" metaid="G_b3453">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3453">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AG80" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011274" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11047" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947962" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131325" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3452" fbc:label="b3452" metaid="G_b3452">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3452">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10905" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011272" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11046" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947957" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131324" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3451" fbc:label="b3451" metaid="G_b3451">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3451">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10906" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011270" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11049" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947959" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131323" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2240" fbc:label="b2240" metaid="G_b2240">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2240">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08194" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007401" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10401" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946704" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130175" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3945" fbc:label="b3945" metaid="G_b3945">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3945">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9S5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012912" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11904" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948440" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111668" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3124" fbc:label="b3124" metaid="G_b3124">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3124">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23524" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010271" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11175" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947632" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698314" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0514" fbc:label="b0514" metaid="G_b0514">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0514">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77364" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001773" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13621" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945129" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128498" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2904" fbc:label="b2904" metaid="G_b2904">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2904">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6T9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009534" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10371" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947393" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130806" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2905" fbc:label="b2905" metaid="G_b2905">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2905">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27248" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009536" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11442" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947390" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130807" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2903" fbc:label="b2903" metaid="G_b2903">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2903">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33195" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009531" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11810" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947394" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130805" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1033" fbc:label="b1033" metaid="G_b1033">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1033">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P75913" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003503" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13869" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946431" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111205" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2979" fbc:label="b2979" metaid="G_b2979">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2979">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEP9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009773" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12997" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947353" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130879" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2978" fbc:label="b2978" />
      <fbc:geneProduct fbc:id="G_b3926" fbc:label="b3926" metaid="G_b3926">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3926">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6F3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012825" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10398" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948423" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131764" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0212" fbc:label="b0212" metaid="G_b0212">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0212">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC84" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000707" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13330" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944902" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128199" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2053" fbc:label="b2053" metaid="G_b2053">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2053">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC88" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006803" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11787" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946562" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129993" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3052" fbc:label="b3052" metaid="G_b3052">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3052">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76658" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010015" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13416" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947548" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130948" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0200" fbc:label="b0200" metaid="G_b0200">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0200">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P63228" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000675" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11736" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944879" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128193" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0104" fbc:label="b0104" metaid="G_b0104">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0104">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P60560" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000362" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10422" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948986" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128097" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2507" fbc:label="b2507" metaid="G_b2507">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2507">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04079" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008254" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10420" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947334" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130432" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2029" fbc:label="b2029" metaid="G_b2029">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2029">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00350" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006737" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10411" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946554" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129970" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4268" fbc:label="b4268" metaid="G_b4268">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4268">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39208" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013980" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12152" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946066" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132090" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3437" fbc:label="b3437" metaid="G_b3437">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3437">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P46859" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011224" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12629" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947937" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111591" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2239" fbc:label="b2239" metaid="G_b2239">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2239">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09394" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007399" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10399" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946725" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130174" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0477" fbc:label="b0477" metaid="G_b0477">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0477">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEW6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001655" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11102" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946584" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128461" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2988" fbc:label="b2988" metaid="G_b2988">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2988">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AES0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009804" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12882" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947474" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130888" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3500" fbc:label="b3500" metaid="G_b3500">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3500">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06715" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011432" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10412" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948014" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131372" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2947" fbc:label="b2947" metaid="G_b2947">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2947">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04425" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009667" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10419" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947445" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130848" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2153" fbc:label="b2153" metaid="G_b2153">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2153">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6T5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007117" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11375" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949040" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130091" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1277" fbc:label="b1277" metaid="G_b1277">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1277">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7I7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004285" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11331" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945763" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129238" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2883" fbc:label="b2883" metaid="G_b2883">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2883">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76641" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009465" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13066" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947366" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130785" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0238" fbc:label="b0238" metaid="G_b0238">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0238">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9M5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000814" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10414" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944817" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128224" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0125" fbc:label="b0125" metaid="G_b0125">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0125">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9M2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000440" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20098" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946624" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111088" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3092" fbc:label="b3092" metaid="G_b3092">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3092">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A8G3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010167" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12737" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947599" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130987" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4040" fbc:label="b4040" metaid="G_b4040">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4040">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGK1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013229" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11370" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948540" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131866" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0353" fbc:label="b0353" metaid="G_b0353">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0353">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77589" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001213" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13293" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944997" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111117" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0339" fbc:label="b0339" metaid="G_b0339">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0339">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABE9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001171" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10176" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946548" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128324" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0126" fbc:label="b0126" metaid="G_b0126">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0126">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P61517" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000442" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12319" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944832" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128119" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2344" fbc:label="b2344" metaid="G_b2344">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2344">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10384" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007740" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10280" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946820" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698292" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0428" fbc:label="b0428" metaid="G_b0428">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0428">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEA5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001484" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10182" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945073" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128413" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2104" fbc:label="b2104" metaid="G_b2104">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2104">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76423" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006960" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14069" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945142" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130042" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2388" fbc:label="b2388" metaid="G_b2388">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2388">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6V8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007875" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12957" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946858" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130320" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2020" fbc:label="b2020" metaid="G_b2020">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2020">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06988" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006715" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10447" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946531" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129961" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2022" fbc:label="b2022" metaid="G_b2022">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2022">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06987" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006720" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10445" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946552" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111373" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2309" fbc:label="b2309" metaid="G_b2309">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2309">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEU0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007617" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12124" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945309" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130244" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0112" fbc:label="b0112" metaid="G_b0112">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0112">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15993" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000384" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10084" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946018" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128105" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0349" fbc:label="b0349" metaid="G_b0349">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0349">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77044" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001201" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20275" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944954" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:459201369" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3805" fbc:label="b3805" metaid="G_b3805">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3805">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06983" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012427" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10429" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947759" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176416" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2103" fbc:label="b2103" metaid="G_b2103">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2103">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76422" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006958" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14068" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946459" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130041" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2418" fbc:label="b2418" metaid="G_b2418">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2418">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P40191" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007973" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12642" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946881" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130344" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0352" fbc:label="b0352" metaid="G_b0352">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0352">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P51020" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001209" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13292" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945012" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128337" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0142" fbc:label="b0142" metaid="G_b0142">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0142">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P26281" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000491" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11374" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948792" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128135" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0508" fbc:label="b0508" metaid="G_b0508">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0508">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30147" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001756" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11584" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946186" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128492" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0003" fbc:label="b0003" metaid="G_b0003">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0003">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00547" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000010" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10999" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947498" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16127997" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4013" fbc:label="b4013" metaid="G_b4013">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4013">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07623" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013121" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10581" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948513" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131839" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2021" fbc:label="b2021" metaid="G_b2021">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2021">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06986" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006717" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10446" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946551" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129962" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2997" fbc:label="b2997" metaid="G_b2997">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2997">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69741" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009834" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13006" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945902" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130897" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0973" fbc:label="b0973" metaid="G_b0973">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0973">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ACD8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003288" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10469" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945580" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128939" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0974" fbc:label="b0974" metaid="G_b0974">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0974">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAM1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003290" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10470" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945581" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128940" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0972" fbc:label="b0972" metaid="G_b0972">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0972">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69739" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003286" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10468" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945579" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128938" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2994" fbc:label="b2994" metaid="G_b2994">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2994">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ACE0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009828" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11801" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945182" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130894" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1136" fbc:label="b1136" metaid="G_b1136">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1136">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08200" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003823" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10489" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945702" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129099" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0593" fbc:label="b0593" metaid="G_b0593">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0593">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEJ2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002047" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10261" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945511" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128576" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2265" fbc:label="b2265" metaid="G_b2265">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2265">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P38051" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007485" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12362" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946712" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111411" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0595" fbc:label="b0595" metaid="G_b0595">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0595">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ADI4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002052" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10260" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946178" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128578" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4015" fbc:label="b4015" metaid="G_b4015">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4015">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9G6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013128" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10022" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948517" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131841" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4267" fbc:label="b4267" metaid="G_b4267">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4267">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39346" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013975" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12541" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944769" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132089" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2023" fbc:label="b2023" metaid="G_b2023">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2023">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P60595" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006722" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10450" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946544" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129964" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2025" fbc:label="b2025" metaid="G_b2025">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2025">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P60664" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006727" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10448" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946516" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129966" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1262" fbc:label="b1262" metaid="G_b1262">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1262">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00909" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004238" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11026" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945519" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:635341464" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3770" fbc:label="b3770" metaid="G_b3770">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3770">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB80" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012316" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10497" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948278" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176403" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0401" fbc:label="b0401" metaid="G_b0401">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0401">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AD99" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001394" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12168" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945042" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128386" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2508" fbc:label="b2508" metaid="G_b2508">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2508">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ADG7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008257" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10421" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946985" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130433" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3161" fbc:label="b3161" metaid="G_b3161">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3161">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAD2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010387" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10617" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947675" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131053" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2889" fbc:label="b2889" metaid="G_b2889">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2889">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46822" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009484" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13072" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949020" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130791" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0073" fbc:label="b0073" metaid="G_b0073">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0073">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30125" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000265" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11577" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944798" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111082" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0072" fbc:label="b0072" metaid="G_b0072">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0072">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6A6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000261" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11576" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945076" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128066" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0071" fbc:label="b0071" metaid="G_b0071">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0071">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30126" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000259" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11575" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945642" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128065" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0074" fbc:label="b0074" metaid="G_b0074">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0074">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09151" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000267" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11226" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947465" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128068" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0918" fbc:label="b0918" metaid="G_b0918">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0918">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04951" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003122" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10519" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945539" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128885" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1215" fbc:label="b1215" metaid="G_b1215">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1215">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A715" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004077" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10518" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945785" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129178" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3581" fbc:label="b3581" metaid="G_b3581">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3581">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37678" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011698" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12285" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948098" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131452" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4196" fbc:label="b4196" metaid="G_b4196">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4196">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39304" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013729" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12496" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948714" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132018" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0696" fbc:label="b0696" metaid="G_b0696">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0696">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P03961" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002374" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10515" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947508" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128672" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0698" fbc:label="b0698" metaid="G_b0698">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0698">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P03959" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002381" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10513" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946045" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128674" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0697" fbc:label="b0697" metaid="G_b0697">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0697">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P03960" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002376" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10514" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947450" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128673" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3849" fbc:label="b3849" metaid="G_b3849">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3849">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFZ7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012574" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11021" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948333" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176432" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1363" fbc:label="b1363" metaid="G_b1363">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1363">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23849" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004567" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11020" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945932" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129324" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0344" fbc:label="b0344" metaid="G_b0344">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0344">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00722" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001183" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10527" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945006" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128329" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2132" fbc:label="b2132" metaid="G_b2132">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2132">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33363" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007043" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12013" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946682" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130070" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3588" fbc:label="b3588" metaid="G_b3588">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3588">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37685" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011718" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12292" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948104" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111619" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2799" fbc:label="b2799" metaid="G_b2799">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2799">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9S1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009177" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10351" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947273" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:345452723" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0343" fbc:label="b0343" metaid="G_b0343">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0343">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P02920" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001181" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10526" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949083" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128328" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2133" fbc:label="b2133" metaid="G_b2133">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2133">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06149" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007048" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10231" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946653" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130071" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1380" fbc:label="b1380" metaid="G_b1380">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1380">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P52643" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004619" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13186" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946315" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129341" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4054" fbc:label="b4054" metaid="G_b4054">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4054">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04693" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013278" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11040" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948563" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131880" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3458" fbc:label="b3458" metaid="G_b3458">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3458">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04816" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011292" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10540" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947964" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131330" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1651" fbc:label="b1651" metaid="G_b1651">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1651">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC81" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005520" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13421" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946161" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129609" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0182" fbc:label="b0182" metaid="G_b0182">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0182">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10441" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000620" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10546" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944838" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128175" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3825" fbc:label="b3825" metaid="G_b3825">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3825">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07000" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012494" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10739" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948314" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176423" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3622" fbc:label="b3622" metaid="G_b3622">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3622">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27243" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011849" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11424" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948148" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131493" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3626" fbc:label="b3626" metaid="G_b3626">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3626">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27129" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011858" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11353" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948142" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131497" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3631" fbc:label="b3631" metaid="G_b3631">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3631">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25740" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011868" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11339" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948149" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131502" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3621" fbc:label="b3621" metaid="G_b3621">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3621">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24173" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011847" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11189" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948136" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131492" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3620" fbc:label="b3620" metaid="G_b3620">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3620">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37692" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011845" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12210" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948135" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131491" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3627" fbc:label="b3627" metaid="G_b3627">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3627">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27128" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011860" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11352" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948143" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131498" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0186" fbc:label="b0186" metaid="G_b0186">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0186">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P52095" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000633" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13219" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944887" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128179" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4131" fbc:label="b4131" metaid="G_b4131">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4131">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9H3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013526" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10131" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948643" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131957" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2156" fbc:label="b2156" metaid="G_b2156">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2156">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25737" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007131" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11337" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946667" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130094" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3605" fbc:label="b3605" metaid="G_b3605">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3605">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33232" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011784" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11963" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948121" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131476" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3600" fbc:label="b3600" metaid="G_b3600">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3600">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09424" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011767" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10616" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948117" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131471" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4014" fbc:label="b4014" metaid="G_b4014">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4014">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08997" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013125" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10023" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948512" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131840" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2976" fbc:label="b2976" metaid="G_b2976">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2976">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37330" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009767" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20080" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948857" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130876" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4036" fbc:label="b4036" metaid="G_b4036">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4036">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P02943" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013218" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10528" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948548" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131862" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4034" fbc:label="b4034" metaid="G_b4034">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4034">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEX9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013200" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10554" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948538" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131860" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4033" fbc:label="b4033" metaid="G_b4033">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4033">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P02916" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013195" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10555" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948532" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131859" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4032" fbc:label="b4032" metaid="G_b4032">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4032">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P68183" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013193" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10556" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948530" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131858" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4035" fbc:label="b4035" metaid="G_b4035">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4035">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P68187" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013216" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10558" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948537" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131861" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2049" fbc:label="b2049" metaid="G_b2049">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2049">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24174" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006791" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10161" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946580" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129989" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1613" fbc:label="b1613" metaid="G_b1613">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1613">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00946" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005398" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10566" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944840" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176125" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4323" fbc:label="b4323" metaid="G_b4323">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4323">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39160" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014175" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20248" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946795" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132144" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0334" fbc:label="b0334" metaid="G_b0334">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0334">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77243" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001152" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13603" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945931" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128319" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0331" fbc:label="b0331" metaid="G_b0331">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0331">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77541" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001138" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13601" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944990" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128316" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0333" fbc:label="b0333" metaid="G_b0333">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0333">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31660" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001148" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11756" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947528" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128318" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1092" fbc:label="b1092" metaid="G_b1092">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1092">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAI9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003698" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11317" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945766" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129055" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3236" fbc:label="b3236" metaid="G_b3236">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3236">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P61889" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010613" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10576" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947854" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131126" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2210" fbc:label="b2210" metaid="G_b2210">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2210">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33940" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007300" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12069" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946702" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130147" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1479" fbc:label="b1479" metaid="G_b1479">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1479">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P26616" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004931" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10948" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946031" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111281" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2463" fbc:label="b2463" metaid="G_b2463">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2463">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76558" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008111" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14193" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946947" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130388" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2515" fbc:label="b2515" metaid="G_b2515">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2515">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P62620" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008281" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10370" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946991" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130440" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2746" fbc:label="b2746" metaid="G_b2746">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2746">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P62617" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009013" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11816" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945057" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130653" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4120" fbc:label="b4120" metaid="G_b4120">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4120">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P02921" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013495" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10578" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948635" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:345452725" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2747" fbc:label="b2747" metaid="G_b2747">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2747">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q46893" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009017" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13110" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948269" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130654" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2942" fbc:label="b2942" metaid="G_b2942">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2942">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A817" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009650" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10589" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945389" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130843" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0199" fbc:label="b0199" metaid="G_b0199">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0199">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30750" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000671" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11621" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944896" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128192" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0198" fbc:label="b0198" metaid="G_b0198">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0198">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31547" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000668" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11737" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944894" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128191" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0197" fbc:label="b0197" metaid="G_b0197">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0197">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P28635" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000666" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11504" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944893" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128190" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3829" fbc:label="b3829" metaid="G_b3829">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3829">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25665" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012520" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10584" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948323" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131678" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4019" fbc:label="b4019" metaid="G_b4019">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4019">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P13009" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013140" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10587" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948522" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131845" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0963" fbc:label="b0963" metaid="G_b0963">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0963">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A731" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003259" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12307" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945574" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111195" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2533" fbc:label="b2533" metaid="G_b2533">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2533">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ADG4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008337" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10983" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947285" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130458" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0403" fbc:label="b0403" metaid="G_b0403">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0403">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21517" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001402" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10565" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949131" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:345452716" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2919" fbc:label="b2919" metaid="G_b2919">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2919">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P52045" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009581" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12972" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947408" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111512" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2917" fbc:label="b2917" metaid="G_b2917">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2917">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27253" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009577" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11444" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945576" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130818" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3599" fbc:label="b3599" metaid="G_b3599">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3599">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00550" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011761" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10615" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948118" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131470" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4322" fbc:label="b4322" metaid="G_b4322">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4322">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24215" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014173" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11066" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947082" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132143" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3633" fbc:label="b3633" metaid="G_b3633">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3633">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC75" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011873" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10520" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949048" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131504" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0134" fbc:label="b0134" metaid="G_b0134">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0134">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31057" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000467" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11675" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944839" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128127" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0529" fbc:label="b0529" metaid="G_b0529">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0529">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24186" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001818" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10328" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945221" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128513" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3941" fbc:label="b3941" metaid="G_b3941">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3941">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEZ1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012897" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10585" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948432" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131779" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2411" fbc:label="b2411" metaid="G_b2411">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2411">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15042" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007938" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10534" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946885" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130337" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1109" fbc:label="b1109" metaid="G_b1109">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1109">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00393" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003746" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10649" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946792" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129072" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2283" fbc:label="b2283" metaid="G_b2283">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2283">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33602" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007543" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12087" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946762" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698290" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2280" fbc:label="b2280" metaid="G_b2280">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2280">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFE0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007536" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12090" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946756" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130215" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2277" fbc:label="b2277" metaid="G_b2277">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2277">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFE8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007529" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11773" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947731" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130212" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2278" fbc:label="b2278" metaid="G_b2278">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2278">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33607" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007532" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12092" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945540" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130213" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2286" fbc:label="b2286" metaid="G_b2286">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2286">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33599" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007549" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12084" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946759" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698291" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2279" fbc:label="b2279" metaid="G_b2279">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2279">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFE4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007534" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12091" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947580" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130214" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2284" fbc:label="b2284" metaid="G_b2284">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2284">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31979" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007545" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11774" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946753" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130219" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2287" fbc:label="b2287" metaid="G_b2287">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2287">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFC7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007551" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12083" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946738" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130222" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2288" fbc:label="b2288" metaid="G_b2288">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2288">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFC3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007553" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12082" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946764" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176207" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2282" fbc:label="b2282" metaid="G_b2282">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2282">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFD4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007541" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12088" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946761" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130217" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2276" fbc:label="b2276" metaid="G_b2276">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2276">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFF0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007526" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12093" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945136" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698289" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2281" fbc:label="b2281" metaid="G_b2281">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2281">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFD6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007539" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12089" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946757" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130216" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2285" fbc:label="b2285" metaid="G_b2285">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2285">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFD1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007547" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12086" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946746" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130220" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2615" fbc:label="b2615" metaid="G_b2615">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2615">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7B3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008603" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12192" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947092" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130534" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1740" fbc:label="b1740" metaid="G_b1740">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1740">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P18843" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005803" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10663" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946946" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129694" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3962" fbc:label="b3962" metaid="G_b3962">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3962">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27306" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012975" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11428" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948461" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111670" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1602" fbc:label="b1602" metaid="G_b1602">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1602">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB67" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005352" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10745" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946144" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129560" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1603" fbc:label="b1603" metaid="G_b1603">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1603">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07001" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005354" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10744" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946628" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129561" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0931" fbc:label="b0931" metaid="G_b0931">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0931">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P18133" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003167" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10742" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946648" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128898" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1216" fbc:label="b1216" metaid="G_b1216">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1216">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31801" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004082" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11753" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945790" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129179" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1186" fbc:label="b1186" metaid="G_b1186">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1186">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFA7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003984" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11392" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944822" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129149" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0019" fbc:label="b0019" metaid="G_b0019">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0019">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P13738" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000068" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10652" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944758" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128013" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2518" fbc:label="b2518" metaid="G_b2518">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2518">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A763" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008291" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10650" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945611" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130443" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0451" fbc:label="b0451" metaid="G_b0451">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0451">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69681" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001564" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11821" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945084" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128436" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0639" fbc:label="b0639" metaid="G_b0639">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0639">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A752" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002187" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13241" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945248" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128622" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0751" fbc:label="b0751" metaid="G_b0751">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0751">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFK2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002547" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11700" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945350" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128719" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1768" fbc:label="b1768" metaid="G_b1768">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1768">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21369" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005891" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11135" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946276" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111327" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1991" fbc:label="b1991" metaid="G_b1991">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1991">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P36562" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006611" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12151" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946517" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129932" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0109" fbc:label="b0109" metaid="G_b0109">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0109">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30011" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000374" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11546" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948869" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128102" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1223" fbc:label="b1223" metaid="G_b1223">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1223">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10903" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004111" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10642" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945783" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129186" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3367" fbc:label="b3367" metaid="G_b3367">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3367">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC26" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011008" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10654" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/2847757" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111575" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1226" fbc:label="b1226" metaid="G_b1226">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1226">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AF26" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004123" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10641" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945807" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129189" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1227" fbc:label="b1227" metaid="G_b1227">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1227">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11350" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004125" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10640" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945808" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129190" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1225" fbc:label="b1225" metaid="G_b1225">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1225">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11349" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004121" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10639" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945780" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129188" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1224" fbc:label="b1224" metaid="G_b1224">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1224">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09152" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004119" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10638" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945782" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129187" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1469" fbc:label="b1469" metaid="G_b1469">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1469">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37758" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004901" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12153" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945799" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129428" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2262" fbc:label="b2262" metaid="G_b2262">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2262">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABU0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007476" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11368" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946747" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130197" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0480" fbc:label="b0480" metaid="G_b0480">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0480">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07024" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001668" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11060" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947331" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128464" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2781" fbc:label="b2781" metaid="G_b2781">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2781">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEY3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009117" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10572" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947254" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130688" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0160" fbc:label="b0160" metaid="G_b0160">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0160">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15723" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000548" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10225" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947177" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128153" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3365" fbc:label="b3365" metaid="G_b3365">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3365">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08201" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011003" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10653" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947868" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131244" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3366" fbc:label="b3366" metaid="G_b3366">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3366">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9I8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011005" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10655" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947881" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131245" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0902" fbc:label="b0902" metaid="G_b0902">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0902">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9N4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003068" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10028" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945517" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128869" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0903" fbc:label="b0903" metaid="G_b0903">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0903">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09373" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003071" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10701" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945514" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128870" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3114" fbc:label="b3114" metaid="G_b3114">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3114">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P42632" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010242" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12758" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947623" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176316" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4254" fbc:label="b4254" metaid="G_b4254">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4254">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04391" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013927" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10069" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948774" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132076" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0273" fbc:label="b0273" metaid="G_b0273">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0273">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06960" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000939" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10067" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944844" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128258" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3187" fbc:label="b3187" metaid="G_b3187">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3187">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AD57" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010474" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10017" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947364" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131077" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0907" fbc:label="b0907" metaid="G_b0907">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0907">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23721" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003093" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10946" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945527" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128874" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3833" fbc:label="b3833" metaid="G_b3833">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3833">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A887" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012534" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11473" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948926" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176426" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0662" fbc:label="b0662" metaid="G_b0662">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0662">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P75728" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002265" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13658" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945261" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128645" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1281" fbc:label="b1281" metaid="G_b1281">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1281">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08244" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004304" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10809" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947121" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129242" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2907" fbc:label="b2907" metaid="G_b2907">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2907">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25534" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009545" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11324" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947388" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130809" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0350" fbc:label="b0350" metaid="G_b0350">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0350">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77608" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001205" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14274" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944768" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111116" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2311" fbc:label="b2311" metaid="G_b2311">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2311">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AG03" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007628" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11044" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949033" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130246" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3843" fbc:label="b3843" metaid="G_b3843">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3843">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAB4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012557" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11396" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948326" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131689" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3835" fbc:label="b3835" metaid="G_b3835">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3835">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6A0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012539" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11476" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948322" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131684" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0693" fbc:label="b0693" metaid="G_b0693">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0693">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24169" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002362" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10964" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945297" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128669" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2965" fbc:label="b2965" metaid="G_b2965">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2965">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21169" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009730" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10961" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947457" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111522" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3073" fbc:label="b3073" metaid="G_b3073">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3073">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P42588" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010092" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12718" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947120" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698310" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3642" fbc:label="b3642" metaid="G_b3642">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3642">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7E3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011904" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10808" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948157" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131513" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2264" fbc:label="b2264" metaid="G_b2264">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2264">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17109" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007482" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10579" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946720" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130199" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1014" fbc:label="b1014" metaid="G_b1014">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1014">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09546" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003424" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10801" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945600" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128980" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0386" fbc:label="b0386" metaid="G_b0386">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0386">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9L8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001338" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10769" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945034" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128371" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1398" fbc:label="b1398" metaid="G_b1398">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1398">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76085" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004671" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13745" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945963" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129359" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0133" fbc:label="b0133" metaid="G_b0133">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0133">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31663" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000465" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11746" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944958" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128126" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0087" fbc:label="b0087" metaid="G_b0087">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0087">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6W3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000316" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10604" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944814" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128080" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2762" fbc:label="b2762" metaid="G_b2762">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2762">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17854" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009057" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10189" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947230" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130669" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4041" fbc:label="b4041" metaid="G_b4041">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4041">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7A7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013232" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10740" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948541" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111678" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3018" fbc:label="b3018" metaid="G_b3018">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3018">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P26647" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009910" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11377" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947496" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130914" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0115" fbc:label="b0115" metaid="G_b0115">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0115">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06959" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000400" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10025" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944794" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128108" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0114" fbc:label="b0114" metaid="G_b0114">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0114">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFG8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000397" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10024" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944834" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128107" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1638" fbc:label="b1638" metaid="G_b1638">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1638">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFI7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005482" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11487" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946806" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129596" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2564" fbc:label="b2564" metaid="G_b2564">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2564">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A794" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008435" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10693" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947039" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130489" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0052" fbc:label="b0052" metaid="G_b0052">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0052">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P19624" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000178" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10691" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944919" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128046" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1386" fbc:label="b1386" metaid="G_b1386">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1386">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P46883" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004639" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13140" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945939" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:162135920" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2320" fbc:label="b2320" metaid="G_b2320">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2320">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05459" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007663" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10692" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946785" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130255" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1723" fbc:label="b1723" metaid="G_b1723">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1723">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06999" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005748" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10700" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946230" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176138" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3916" fbc:label="b3916" metaid="G_b3916">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3916">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A796" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012789" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10699" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948412" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131754" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3132" fbc:label="b3132" metaid="G_b3132">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3132">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0C8K0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010295" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12763" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947637" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131024" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3137" fbc:label="b3137" metaid="G_b3137">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3137">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AB74" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010307" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12768" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947644" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131029" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3951" fbc:label="b3951" metaid="G_b3951">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3951">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32674" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012934" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11910" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948454" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131789" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3952" fbc:label="b3952" metaid="G_b3952">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3952">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32675" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012937" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11911" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948453" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176447" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3176" fbc:label="b3176" metaid="G_b3176">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3176">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31120" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010439" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11553" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947692" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131066" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2913" fbc:label="b2913" metaid="G_b2913">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2913">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9T0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009561" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10944" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945258" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130814" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4025" fbc:label="b4025" metaid="G_b4025">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4025">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6T1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013163" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10702" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948535" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131851" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2926" fbc:label="b2926" metaid="G_b2926">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2926">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A799" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009605" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10703" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947414" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130827" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3385" fbc:label="b3385" metaid="G_b3385">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3385">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32662" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011058" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11871" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947895" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131263" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4395" fbc:label="b4395" metaid="G_b4395">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4395">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7A2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014416" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12164" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948918" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132212" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0755" fbc:label="b0755" metaid="G_b0755">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0755">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P62707" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002563" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11699" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945068" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128723" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3612" fbc:label="b3612" metaid="G_b3612">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3612">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37689" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011818" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12296" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948130" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131483" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2690" fbc:label="b2690" metaid="G_b2690">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2690">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77475" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008850" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13530" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945776" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130602" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0688" fbc:label="b0688" metaid="G_b0688">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0688">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P36938" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002345" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12144" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945271" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128664" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0418" fbc:label="b0418" metaid="G_b0418">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0418">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P18200" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001454" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10704" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947542" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128403" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1278" fbc:label="b1278" metaid="G_b1278">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1278">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A924" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004294" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10705" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945863" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129239" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1912" fbc:label="b1912" metaid="G_b1912">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1912">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABF8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006367" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10706" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945791" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129859" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0576" fbc:label="b0576" metaid="G_b0576">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0576">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24207" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001976" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10708" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945199" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128559" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3726" fbc:label="b3726" metaid="G_b3726">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3726">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07654" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012184" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10782" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948239" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131594" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3725" fbc:label="b3725" metaid="G_b3725">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3725">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAH0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012179" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10783" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948240" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131593" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3727" fbc:label="b3727" metaid="G_b3727">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3727">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGH8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012186" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10784" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948238" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131595" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3728" fbc:label="b3728" metaid="G_b3728">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3728">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AG82" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012190" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10734" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948237" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131596" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2987" fbc:label="b2987" metaid="G_b2987">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2987">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P43676" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009800" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12883" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947475" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130887" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3493" fbc:label="b3493" metaid="G_b3493">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3493">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFJ7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011407" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12230" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948009" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131365" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3821" fbc:label="b3821" metaid="G_b3821">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3821">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A921" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012479" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10738" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948307" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131671" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2048" fbc:label="b2048" metaid="G_b2048">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2048">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P24175" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006788" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10162" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946574" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129988" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3974" fbc:label="b3974" metaid="G_b3974">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3974">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6I3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013003" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10922" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948479" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131808" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3258" fbc:label="b3258" metaid="G_b3258">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3258">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16256" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010695" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10685" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947752" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111565" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0871" fbc:label="b0871" metaid="G_b0871">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0871">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07003" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002958" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10754" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946132" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128839" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4226" fbc:label="b4226" metaid="G_b4226">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4226">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7A9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013824" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10755" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948748" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132048" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0369" fbc:label="b0369" metaid="G_b0369">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0369">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ACB2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001266" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10428" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945017" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111123" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3956" fbc:label="b3956" metaid="G_b3956">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3956">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00864" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012950" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10756" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948457" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131794" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3403" fbc:label="b3403" metaid="G_b3403">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3403">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22259" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011106" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10688" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945667" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131280" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2920" fbc:label="b2920" metaid="G_b2920">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2920">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P52043" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009583" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12973" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947402" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130821" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4383" fbc:label="b4383" metaid="G_b4383">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4383">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6K6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014377" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10220" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948910" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132200" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3380" fbc:label="b3380" metaid="G_b3380">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3380">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P45549" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011046" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12918" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947890" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131258" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3850" fbc:label="b3850" metaid="G_b3850">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3850">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ACB4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012577" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11485" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948331" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131696" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2536" fbc:label="b2536" metaid="G_b2536">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2536">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q47142" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008348" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13454" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947007" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130461" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1702" fbc:label="b1702" metaid="G_b1702">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1702">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23538" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005678" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10759" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946209" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129658" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4005" fbc:label="b4005" metaid="G_b4005">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4005">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15640" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013095" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10792" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948504" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131835" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2499" fbc:label="b2499" metaid="G_b2499">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2499">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08178" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008230" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10798" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946975" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130424" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2026" fbc:label="b2026" metaid="G_b2026">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2026">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P06989" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006729" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10451" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946515" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129967" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2476" fbc:label="b2476" metaid="G_b2476">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2476">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7D7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008156" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10791" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946957" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130401" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2557" fbc:label="b2557" metaid="G_b2557">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2557">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15254" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008412" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10797" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947032" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176239" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2024" fbc:label="b2024" metaid="G_b2024">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2024">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P10371" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006725" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10444" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946521" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111374" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2677" fbc:label="b2677" metaid="G_b2677">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2677">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P14175" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008814" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10771" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947148" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130591" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2678" fbc:label="b2678" metaid="G_b2678">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2678">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P14176" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008816" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10772" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947145" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130592" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2679" fbc:label="b2679" metaid="G_b2679">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2679">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFM2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008818" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10773" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947165" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130593" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4111" fbc:label="b4111" metaid="G_b4111">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4111">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0C0L7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013462" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11612" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948626" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131937" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1015" fbc:label="b1015" metaid="G_b1015">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1015">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07117" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003433" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10802" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945602" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128981" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1207" fbc:label="b1207" metaid="G_b1207">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1207">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A717" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004054" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10774" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945772" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129170" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0908" fbc:label="b0908" metaid="G_b0908">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0908">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6D3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003095" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10073" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945528" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128875" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4160" fbc:label="b4160" metaid="G_b4160">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4160">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A8K1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013623" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10775" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948673" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131985" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4388" fbc:label="b4388" metaid="G_b4388">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4388">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGB0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014393" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10945" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948913" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132205" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2585" fbc:label="b2585" metaid="G_b2585">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2585">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23830" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008510" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10781" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947059" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111464" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2297" fbc:label="b2297" metaid="G_b2297">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2297">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9M8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007582" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20173" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946778" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130232" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2458" fbc:label="b2458" metaid="G_b2458">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2458">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77218" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008097" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14188" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946940" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130383" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3634" fbc:label="b3634" metaid="G_b3634">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3634">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6I6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011875" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11190" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947386" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131505" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0692" fbc:label="b0692" metaid="G_b0692">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0692">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAF1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002360" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10753" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945422" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128668" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1123" fbc:label="b1123" metaid="G_b1123">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1123">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFK9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003789" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10752" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945682" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129086" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1440" fbc:label="b1440" metaid="G_b1440">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1440">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76108" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004803" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13762" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946005" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129399" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1441" fbc:label="b1441" metaid="G_b1441">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1441">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77795" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004805" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13763" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946007" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129400" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0855" fbc:label="b0855" metaid="G_b0855">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0855">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31134" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002915" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11630" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945476" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111177" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1124" fbc:label="b1124" metaid="G_b1124">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1124">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFK6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003791" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10751" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945691" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129087" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1443" fbc:label="b1443" metaid="G_b1443">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1443">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFR9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004809" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13765" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945903" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129402" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1126" fbc:label="b1126" metaid="G_b1126">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1126">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69874" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003796" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10749" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946323" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129089" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0854" fbc:label="b0854" metaid="G_b0854">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0854">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31133" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002913" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11629" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945480" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128822" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1125" fbc:label="b1125" metaid="G_b1125">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1125">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFK4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003794" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10750" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945692" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:162135921" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1442" fbc:label="b1442" metaid="G_b1442">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1442">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77156" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004807" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13764" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945976" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129401" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0856" fbc:label="b0856" metaid="G_b0856">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0856">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31135" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002917" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11631" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945475" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128824" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0857" fbc:label="b0857" metaid="G_b0857">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0857">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFL1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002919" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11632" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945485" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128825" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2407" fbc:label="b2407" metaid="G_b2407">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2407">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P45563" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007927" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20250" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946878" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130333" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1636" fbc:label="b1636" metaid="G_b1636">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1636">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77150" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005475" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13940" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946162" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129594" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1676" fbc:label="b1676" metaid="G_b1676">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1676">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AD61" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005600" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10804" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946179" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129632" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1854" fbc:label="b1854" metaid="G_b1854">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1854">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21599" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006182" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10803" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946527" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129807" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3831" fbc:label="b3831" metaid="G_b3831">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3831">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P12758" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012528" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11045" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948987" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131680" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0750" fbc:label="b0750" metaid="G_b0750">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0750">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11458" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002541" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10630" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945351" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128718" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1662" fbc:label="b1662" metaid="G_b1662">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1662">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AFU8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005557" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11406" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945848" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129620" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0415" fbc:label="b0415" metaid="G_b0415">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0415">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P61714" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001444" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11322" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946453" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128400" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3752" fbc:label="b3752" metaid="G_b3752">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3752">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9J6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012266" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10818" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948260" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131620" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0063" fbc:label="b0063" metaid="G_b0063">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0063">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08204" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000223" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10053" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946017" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128057" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0061" fbc:label="b0061" metaid="G_b0061">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0061">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08203" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000209" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10055" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945294" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128055" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4198" fbc:label="b4198" metaid="G_b4198">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4198">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39306" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013734" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12498" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948711" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132020" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3583" fbc:label="b3583" metaid="G_b3583">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3583">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37680" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011702" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12287" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948099" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131454" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2687" fbc:label="b2687" metaid="G_b2687">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2687">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P45578" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008842" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12712" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947168" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130599" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3749" fbc:label="b3749" metaid="G_b3749">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3749">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04983" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012259" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10814" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948264" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131617" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4231" fbc:label="b4231" metaid="G_b4231">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4231">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37772" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013840" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12439" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948754" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111710" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4229" fbc:label="b4229" />
      <fbc:geneProduct fbc:id="G_b3751" fbc:label="b3751" metaid="G_b3751">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3751">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P02925" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012264" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10815" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948261" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131619" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4228" fbc:label="b4228" />
      <fbc:geneProduct fbc:id="G_b4230" fbc:label="b4230" metaid="G_b4230">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4230">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39328" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013838" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12520" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948743" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698343" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3750" fbc:label="b3750" metaid="G_b3750">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3750">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGI1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012261" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10816" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948262" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131618" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3748" fbc:label="b3748" metaid="G_b3748">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3748">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04982" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012256" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10817" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948267" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111647" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4227" fbc:label="b4227" metaid="G_b4227">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4227">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39325" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013829" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12517" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948746" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132049" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3903" fbc:label="b3903" metaid="G_b3903">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3903">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32170" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012736" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11867" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948400" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176441" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3904" fbc:label="b3904" metaid="G_b3904">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3904">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32171" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012738" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11868" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948399" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131744" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3907" fbc:label="b3907" metaid="G_b3907">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3907">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27125" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012749" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11313" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948398" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131747" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3902" fbc:label="b3902" metaid="G_b3902">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3902">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P32169" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012731" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11866" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948401" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131742" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2234" fbc:label="b2234" metaid="G_b2234">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2234">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00452" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007383" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10660" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946612" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130169" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2235" fbc:label="b2235" metaid="G_b2235">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2235">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69924" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007386" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10661" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946732" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130170" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2675" fbc:label="b2675" metaid="G_b2675">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2675">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39452" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008806" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20257" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947155" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130589" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2676" fbc:label="b2676" metaid="G_b2676">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2676">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37146" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008808" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12381" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947149" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130590" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4238" fbc:label="b4238" metaid="G_b4238">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4238">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P28903" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013865" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11417" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948755" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132060" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3386" fbc:label="b3386" metaid="G_b3386">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3386">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AG07" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011061" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11960" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947896" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131264" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4301" fbc:label="b4301" metaid="G_b4301">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4301">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39362" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0014097" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12553" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948829" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132122" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2914" fbc:label="b2914" metaid="G_b2914">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2914">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7Z0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009567" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11443" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947407" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130815" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4090" fbc:label="b4090" metaid="G_b4090">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4090">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37351" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013405" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11827" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948602" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131916" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0638" fbc:label="b0638" metaid="G_b0638">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0638">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P52086" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002185" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13240" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945246" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128621" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0222" fbc:label="b0222" metaid="G_b0222">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0222">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P63224" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000751" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13146" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949134" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128208" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1748" fbc:label="b1748" metaid="G_b1748">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1748">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77581" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005824" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13999" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946255" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129702" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2751" fbc:label="b2751" metaid="G_b2751">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2751">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23845" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009026" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10194" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947219" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130658" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2752" fbc:label="b2752" metaid="G_b2752">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2752">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21156" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009029" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10186" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947217" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130659" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2705" fbc:label="b2705" metaid="G_b2705">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2705">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05707" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008896" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10971" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948937" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130612" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2704" fbc:label="b2704" metaid="G_b2704">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2704">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05706" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008894" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10970" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948971" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130611" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2702" fbc:label="b2702" metaid="G_b2702">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2702">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P56579" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008890" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10969" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947575" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176260" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2703" fbc:label="b2703" metaid="G_b2703">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2703">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P56580" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008892" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14373" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948933" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176261" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2472" fbc:label="b2472" metaid="G_b2472">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2472">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AED7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008142" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10208" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948313" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130397" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1764" fbc:label="b1764" metaid="G_b1764">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1764">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16456" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005872" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10943" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946768" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129718" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0586" fbc:label="b0586" metaid="G_b0586">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0586">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11454" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002021" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10264" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945184" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128569" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3607" fbc:label="b3607" metaid="G_b3607">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3607">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9D4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011793" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10187" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948126" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131478" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2366" fbc:label="b2366" metaid="G_b2366">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2366">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00926" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007804" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10249" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946837" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130298" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1814" fbc:label="b1814" metaid="G_b1814">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1814">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16095" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006039" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10930" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946331" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129768" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3111" fbc:label="b3111" />
      <fbc:geneProduct fbc:id="G_b3112" fbc:label="b3112" />
      <fbc:geneProduct fbc:id="G_b2797" fbc:label="b2797" metaid="G_b2797">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2797">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30744" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009172" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11623" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947262" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130704" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2796" fbc:label="b2796" metaid="G_b2796">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2796">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAD6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009169" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12142" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947264" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130703" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3116" fbc:label="b3116" metaid="G_b3116">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3116">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAD8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010247" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10991" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947629" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131009" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3089" fbc:label="b3089" metaid="G_b3089">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3089">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGE4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010159" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12732" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947605" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130984" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1744" fbc:label="b1744" metaid="G_b1744">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1744">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76215" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005815" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13995" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946256" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129698" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1746" fbc:label="b1746" metaid="G_b1746">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1746">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76217" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005820" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13997" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946260" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129700" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3368" fbc:label="b3368" metaid="G_b3368">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3368">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEA8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011010" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10188" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947880" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131246" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3281" fbc:label="b3281" metaid="G_b3281">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3281">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P15770" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010763" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10077" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947776" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131162" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1692" fbc:label="b1692" metaid="G_b1692">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1692">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6D5" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005651" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11234" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946200" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129648" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0388" fbc:label="b0388" metaid="G_b0388">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0388">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6E1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001353" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10082" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945031" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128373" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3390" fbc:label="b3390" metaid="G_b3390">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3390">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6D7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011069" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10081" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/2847759" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111581" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3939" fbc:label="b3939" metaid="G_b3939">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3939">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00935" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012887" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10582" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948434" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131777" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1745" fbc:label="b1745" metaid="G_b1745">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1745">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76216" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005817" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13996" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946259" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129699" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1584" fbc:label="b1584" metaid="G_b1584">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1584">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A951" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005286" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12447" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946117" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129542" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0121" fbc:label="b0121" metaid="G_b0121">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0121">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09158" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000418" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10963" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947726" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128114" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1656" fbc:label="b1656" metaid="G_b1656">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1656">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGD3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005537" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10954" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944953" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129614" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1646" fbc:label="b1646" metaid="G_b1646">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1646">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGD1" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005505" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13419" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945343" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129604" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3908" fbc:label="b3908" metaid="G_b3908">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3908">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00448" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012761" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10953" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948403" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176442" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2661" fbc:label="b2661" metaid="G_b2661">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2661">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P25526" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008758" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11329" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948060" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130575" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2260" fbc:label="b2260" metaid="G_b2260">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2260">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37353" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007471" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12437" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946741" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130195" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2261" fbc:label="b2261" metaid="G_b2261">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2261">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P29208" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007473" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11532" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946734" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130196" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3579" fbc:label="b3579" metaid="G_b3579">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3579">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37676" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011694" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12283" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948091" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131450" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0723" fbc:label="b0723" metaid="G_b0723">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0723">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC41" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002466" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10931" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945402" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128698" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0724" fbc:label="b0724" metaid="G_b0724">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0724">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07014" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002468" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10932" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945300" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128699" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0722" fbc:label="b0722" metaid="G_b0722">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0722">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC44" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002464" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10934" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945322" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128697" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0721" fbc:label="b0721" metaid="G_b0721">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0721">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P69054" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002460" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10933" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945316" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128696" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0729" fbc:label="b0729" metaid="G_b0729">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0729">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGE9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002485" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10982" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945314" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128704" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0728" fbc:label="b0728" metaid="G_b0728">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0728">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A836" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002483" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10981" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945312" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128703" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2429" fbc:label="b2429" metaid="G_b2429">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2429">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77272" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008009" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14163" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946894" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130354" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2764" fbc:label="b2764" metaid="G_b2764">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2764">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P38038" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009063" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10191" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947239" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130671" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2763" fbc:label="b2763" metaid="G_b2763">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2763">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17846" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009061" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10190" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947231" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130670" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2423" fbc:label="b2423" metaid="G_b2423">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2423">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AEB0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007988" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10198" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/2847743" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111430" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2424" fbc:label="b2424" metaid="G_b2424">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2424">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16701" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007990" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10197" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946882" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130349" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2422" fbc:label="b2422" metaid="G_b2422">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2422">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16676" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007986" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10183" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946889" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130348" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2425" fbc:label="b2425" metaid="G_b2425">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2425">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P16700" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007992" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10195" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946883" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130350" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3917" fbc:label="b3917" metaid="G_b3917">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3917">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AG78" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012794" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10929" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948411" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131755" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2413" fbc:label="b2413" metaid="G_b2413">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2413">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6J3" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0007953" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10003" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946875" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130339" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1521" fbc:label="b1521" metaid="G_b1521">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1521">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6L7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0005076" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11065" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945542" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176119" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0008" fbc:label="b0008" metaid="G_b0008">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0008">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A870" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000027" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11556" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944748" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128002" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2464" fbc:label="b2464" metaid="G_b2464">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2464">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A867" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008115" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11797" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947006" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130389" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3061" fbc:label="b3061" metaid="G_b3061">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3061">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P05847" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010050" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11168" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947565" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130957" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3062" fbc:label="b3062" metaid="G_b3062">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3062">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AC35" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010052" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11169" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947568" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130958" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3063" fbc:label="b3063" metaid="G_b3063">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3063">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39414" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010056" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12393" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947576" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130959" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0368" fbc:label="b0368" metaid="G_b0368">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0368">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37610" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001263" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12423" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945021" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128353" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0365" fbc:label="b0365" metaid="G_b0365">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0365">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q47537" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001256" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13300" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945030" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111122" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0367" fbc:label="b0367" metaid="G_b0367">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0367">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q47539" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001261" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13301" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945026" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128352" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0366" fbc:label="b0366" metaid="G_b0366">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0366">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/Q47538" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001259" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13299" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945027" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128351" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3790" fbc:label="b3790" metaid="G_b3790">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3790">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27832" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012382" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11455" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948298" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698332" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3791" fbc:label="b3791" metaid="G_b3791">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3791">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27833" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012384" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11456" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948296" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131647" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2038" fbc:label="b2038" metaid="G_b2038">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2038">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37745" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006759" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11979" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947482" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129978" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2040" fbc:label="b2040" metaid="G_b2040">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2040">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37760" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006764" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12411" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947117" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129980" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2041" fbc:label="b2041" metaid="G_b2041">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2041">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37759" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006766" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12412" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945276" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129981" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3788" fbc:label="b3788" metaid="G_b3788">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3788">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27830" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012376" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11453" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948300" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176411" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0915" fbc:label="b0915" metaid="G_b0915">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0915">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27300" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003115" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11409" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945526" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128882" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2096" fbc:label="b2096" metaid="G_b2096">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2096">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0C8J6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006938" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12419" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946636" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111384" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2095" fbc:label="b2095" metaid="G_b2095">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2095">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0C8J8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006934" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12418" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946641" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130033" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0166" fbc:label="b0166" metaid="G_b0166">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0166">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9D8" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000565" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10207" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944862" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128159" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0068" fbc:label="b0068" metaid="G_b0068">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0068">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31550" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000249" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11574" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946306" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128062" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0066" fbc:label="b0066" metaid="G_b0066">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0066">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31548" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000244" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11572" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944785" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128060" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0067" fbc:label="b0067" metaid="G_b0067">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0067">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31549" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000246" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11573" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944784" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128061" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0870" fbc:label="b0870" metaid="G_b0870">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0870">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P75823" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002954" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13690" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944955" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128838" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3616" fbc:label="b3616" metaid="G_b3616">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3616">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P07913" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011830" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10993" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948139" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131487" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3772" fbc:label="b3772" metaid="G_b3772">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3772">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P04968" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012321" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10493" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948287" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131630" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3117" fbc:label="b3117" metaid="G_b3117">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3117">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGF6" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010249" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10990" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947633" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131010" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3992" fbc:label="b3992" metaid="G_b3992">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3992">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30138" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013053" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11587" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948500" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111672" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0423" fbc:label="b0423" metaid="G_b0423">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0423">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77718" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001470" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13273" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945067" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128408" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3991" fbc:label="b3991" metaid="G_b3991">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3991">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30139" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013050" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11589" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948493" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:33347812" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2530" fbc:label="b2530" metaid="G_b2530">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2530">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A6B7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008324" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12677" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947004" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176235" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4407" fbc:label="b4407" metaid="G_b4407">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4407">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/O32583" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0023907" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG14363" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/2847702" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176448" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3990" fbc:label="b3990" metaid="G_b3990">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3990">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30140" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013048" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11590" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948494" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131820" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2935" fbc:label="b2935" metaid="G_b2935">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2935">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27302" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009625" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11427" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947420" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176286" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2465" fbc:label="b2465" metaid="G_b2465">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2465">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33570" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008117" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12100" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945865" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130390" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0996" fbc:label="b0996" metaid="G_b0996">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0996">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33226" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003370" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11815" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946252" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128962" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0997" fbc:label="b0997" metaid="G_b0997">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0997">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P33225" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003372" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11814" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946267" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128963" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2827" fbc:label="b2827" metaid="G_b2827">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2827">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A884" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0009268" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11002" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949035" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130731" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0417" fbc:label="b0417" metaid="G_b0417">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0417">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGG0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001452" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20227" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947387" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128402" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3993" fbc:label="b3993" metaid="G_b3993">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3993">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P30137" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013055" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11586" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948491" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131823" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3919" fbc:label="b3919" metaid="G_b3919">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3919">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A858" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012799" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11015" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948409" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131757" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0888" fbc:label="b0888" metaid="G_b0888">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0888">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A9P4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003018" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11032" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/949054" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128855" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4239" fbc:label="b4239" metaid="G_b4239">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4239">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P28904" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013870" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11402" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948762" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132061" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1897" fbc:label="b1897" metaid="G_b1897">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1897">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31678" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006320" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11752" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946406" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129849" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1896" fbc:label="b1896" metaid="G_b1896">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1896">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P31677" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006318" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11751" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946405" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129848" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3519" fbc:label="b3519" metaid="G_b3519">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3519">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P62601" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011497" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12245" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948037" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131391" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1197" fbc:label="b1197" metaid="G_b1197">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1197">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P13482" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004021" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11017" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945757" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129160" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4240" fbc:label="b4240" metaid="G_b4240">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4240">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P36672" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013872" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12127" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948761" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132062" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1261" fbc:label="b1261" metaid="G_b1261">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1261">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A879" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004234" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11025" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945768" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129222" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1260" fbc:label="b1260" metaid="G_b1260">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1260">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A877" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0004232" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11024" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946204" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129221" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3709" fbc:label="b3709" metaid="G_b3709">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3709">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P23173" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012135" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11006" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948220" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131577" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0509" fbc:label="b0509" metaid="G_b0509">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0509">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77161" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001759" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13265" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945146" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128493" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3125" fbc:label="b3125" metaid="G_b3125">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3125">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0ABQ2" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010273" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11176" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947631" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:145698315" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1907" fbc:label="b1907" metaid="G_b1907">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1907">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AAD4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006355" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11041" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946412" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129857" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0179" fbc:label="b0179" metaid="G_b0179">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0179">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P21645" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000610" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10316" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944882" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128172" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0085" fbc:label="b0085" metaid="G_b0085">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0085">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P22188" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000311" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10621" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944791" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128078" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3787" fbc:label="b3787" metaid="G_b3787">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3787">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27829" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012374" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11452" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948977" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176410" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3786" fbc:label="b3786" metaid="G_b3786">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3786">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P27828" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012372" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11451" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944789" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176409" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0181" fbc:label="b0181" metaid="G_b0181">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0181">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A722" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000618" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10545" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944849" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128174" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3189" fbc:label="b3189" metaid="G_b3189">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3189">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A749" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0010479" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11358" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/947703" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131079" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0090" fbc:label="b0090" metaid="G_b0090">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0090">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17443" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000322" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10623" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946321" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128083" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0088" fbc:label="b0088" metaid="G_b0088">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0088">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P14900" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000318" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10620" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944818" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128081" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0091" fbc:label="b0091" metaid="G_b0091">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0091">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P17952" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000324" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10619" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946153" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128084" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3972" fbc:label="b3972" metaid="G_b3972">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3972">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P08373" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012999" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11205" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948470" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131806" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0174" fbc:label="b0174" metaid="G_b0174">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0174">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P60472" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000595" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13329" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944874" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128167" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0759" fbc:label="b0759" metaid="G_b0759">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0759">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09147" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002575" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10362" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945354" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176045" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2036" fbc:label="b2036" metaid="G_b2036">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2036">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37747" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006755" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11981" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945235" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129976" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2028" fbc:label="b2028" metaid="G_b2028">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2028">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P76373" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0006734" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13407" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946571" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16129969" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0758" fbc:label="b0758" metaid="G_b0758">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0758">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09148" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0002573" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10366" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945357" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128726" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0505" fbc:label="b0505" metaid="G_b0505">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0505">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P77731" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0001747" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13615" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/945141" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128489" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0086" fbc:label="b0086" metaid="G_b0086">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0086">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P11880" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000313" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10622" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944813" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128079" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0096" fbc:label="b0096" metaid="G_b0096">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0096">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A725" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000336" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10265" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944816" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128089" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b0171" fbc:label="b0171" metaid="G_b0171">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b0171">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A7E9" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0000582" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11539" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/944989" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16128164" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3803" fbc:label="b3803" metaid="G_b3803">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3803">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09127" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012419" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10433" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948446" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131655" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3804" fbc:label="b3804" metaid="G_b3804">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3804">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09126" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0012425" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10430" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948587" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131656" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3997" fbc:label="b3997" metaid="G_b3997">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3997">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P29680" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013066" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11543" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948497" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131827" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2498" fbc:label="b2498" metaid="G_b2498">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2498">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0A8F0" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008224" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11332" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946979" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111448" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b2497" fbc:label="b2497" metaid="G_b2497">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b2497">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGM7" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0008219" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12129" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946978" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16130422" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b1006" fbc:label="b1006" metaid="G_b1006">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b1006">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P75892" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0003401" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG13853" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/946589" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:90111201" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3572" fbc:label="b3572" metaid="G_b3572">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3572">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P09053" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011671" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG10107" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948087" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:49176374" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4197" fbc:label="b4197" metaid="G_b4197">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4197">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P39305" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013732" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12497" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948712" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16132019" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3565" fbc:label="b3565" metaid="G_b3565">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3565">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P00944" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011639" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11074" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948141" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131436" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3566" fbc:label="b3566" metaid="G_b3566">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3566">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37387" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011646" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG20252" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948090" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131437" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3568" fbc:label="b3568" metaid="G_b3568">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3568">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGI4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011650" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12276" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948083" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131439" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b3567" fbc:label="b3567" metaid="G_b3567">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b3567">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P37388" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0011648" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG12275" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948127" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131438" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
      <fbc:geneProduct fbc:id="G_b4031" fbc:label="b4031" metaid="G_b4031">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#G_b4031">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/uniprot/P0AGF4" />
                </rdf:Bag>
              </bqbiol:is>
              <bqbiol:isEncodedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/asap/ABE-0013189" />
                  <rdf:li rdf:resource="http://identifiers.org/ecogene/EG11076" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigene/948529" />
                  <rdf:li rdf:resource="http://identifiers.org/ncbigi/gi:16131857" />
                </rdf:Bag>
              </bqbiol:isEncodedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </fbc:geneProduct>
    </fbc:listOfGeneProducts>
    <groups:listOfGroups xmlns:groups="http://www.sbml.org/sbml/level3/version1/groups/version1">
      <groups:group groups:id="g1" groups:kind="partonomy" groups:name="Transport, Extracellular" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_12PPDt" />
          <groups:member groups:idRef="R_ABUTt2r" />
          <groups:member groups:idRef="R_ACALDt" />
          <groups:member groups:idRef="R_ACGApts" />
          <groups:member groups:idRef="R_ACMANApts" />
          <groups:member groups:idRef="R_ACNAMt2" />
          <groups:member groups:idRef="R_ACt2r" />
          <groups:member groups:idRef="R_ADEt2" />
          <groups:member groups:idRef="R_ADNt2_copy1" />
          <groups:member groups:idRef="R_ADNt2_copy2" />
          <groups:member groups:idRef="R_AKGt2r" />
          <groups:member groups:idRef="R_ALAabc" />
          <groups:member groups:idRef="R_ARBabc" />
          <groups:member groups:idRef="R_ARBt2r" />
          <groups:member groups:idRef="R_ARGabc" />
          <groups:member groups:idRef="R_ASNabc" />
          <groups:member groups:idRef="R_ASNt2r" />
          <groups:member groups:idRef="R_ASPabc" />
          <groups:member groups:idRef="R_ASPt2r" />
          <groups:member groups:idRef="R_ASPt2_2" />
          <groups:member groups:idRef="R_ASPt2_3" />
          <groups:member groups:idRef="R_CADVt" />
          <groups:member groups:idRef="R_CBL1abc" />
          <groups:member groups:idRef="R_CHLt2" />
          <groups:member groups:idRef="R_CITt7" />
          <groups:member groups:idRef="R_CO2t" />
          <groups:member groups:idRef="R_CRNt7" />
          <groups:member groups:idRef="R_CSNt2" />
          <groups:member groups:idRef="R_CYSabc" />
          <groups:member groups:idRef="R_CYTDt2_copy1" />
          <groups:member groups:idRef="R_CYTDt2_copy2" />
          <groups:member groups:idRef="R_DADNt2" />
          <groups:member groups:idRef="R_DALAt2r" />
          <groups:member groups:idRef="R_DAPabc" />
          <groups:member groups:idRef="R_DCYTt2" />
          <groups:member groups:idRef="R_DDGLCNt2r" />
          <groups:member groups:idRef="R_DGSNt2" />
          <groups:member groups:idRef="R_DHAt" />
          <groups:member groups:idRef="R_DINSt2" />
          <groups:member groups:idRef="R_DSERt2" />
          <groups:member groups:idRef="R_DURIt2" />
          <groups:member groups:idRef="R_D_LACt2" />
          <groups:member groups:idRef="R_ETOHt2r" />
          <groups:member groups:idRef="R_FE2abc" />
          <groups:member groups:idRef="R_FORt" />
          <groups:member groups:idRef="R_FRUpts" />
          <groups:member groups:idRef="R_FRUpts2" />
          <groups:member groups:idRef="R_FUCPt6_2" />
          <groups:member groups:idRef="R_FUCt" />
          <groups:member groups:idRef="R_FUMt2_2" />
          <groups:member groups:idRef="R_FUMt2_3" />
          <groups:member groups:idRef="R_G6Pt6_2" />
          <groups:member groups:idRef="R_GALCTNt2r" />
          <groups:member groups:idRef="R_GALTpts" />
          <groups:member groups:idRef="R_GALURt2r" />
          <groups:member groups:idRef="R_GALabc" />
          <groups:member groups:idRef="R_GALt2" />
          <groups:member groups:idRef="R_GAMpts" />
          <groups:member groups:idRef="R_GLCNt2r" />
          <groups:member groups:idRef="R_GLCURt2r" />
          <groups:member groups:idRef="R_GLCpts" />
          <groups:member groups:idRef="R_GLCt2" />
          <groups:member groups:idRef="R_GLNabc" />
          <groups:member groups:idRef="R_GLUabc" />
          <groups:member groups:idRef="R_GLUt2r" />
          <groups:member groups:idRef="R_GLUt4" />
          <groups:member groups:idRef="R_GLYALDt" />
          <groups:member groups:idRef="R_GLYC3Pabc" />
          <groups:member groups:idRef="R_GLYC3Pt6" />
          <groups:member groups:idRef="R_GLYCLTt2r" />
          <groups:member groups:idRef="R_GLYCt" />
          <groups:member groups:idRef="R_GLYt2r" />
          <groups:member groups:idRef="R_GSNt2" />
          <groups:member groups:idRef="R_GUAt" />
          <groups:member groups:idRef="R_H2Ot" />
          <groups:member groups:idRef="R_HDCAt2" />
          <groups:member groups:idRef="R_HISabc" />
          <groups:member groups:idRef="R_HISt2r" />
          <groups:member groups:idRef="R_HYXNt" />
          <groups:member groups:idRef="R_IDONt2r" />
          <groups:member groups:idRef="R_ILEabc" />
          <groups:member groups:idRef="R_ILEt2r" />
          <groups:member groups:idRef="R_INDOLEt2r" />
          <groups:member groups:idRef="R_INSt2_copy1" />
          <groups:member groups:idRef="R_INSt2_copy2" />
          <groups:member groups:idRef="R_Kabc" />
          <groups:member groups:idRef="R_Kt2r" />
          <groups:member groups:idRef="R_LCTSt" />
          <groups:member groups:idRef="R_LEUabc" />
          <groups:member groups:idRef="R_LEUt2r" />
          <groups:member groups:idRef="R_LYSabc" />
          <groups:member groups:idRef="R_LYSt2r" />
          <groups:member groups:idRef="R_L_LACt2r" />
          <groups:member groups:idRef="R_MALTHXabc" />
          <groups:member groups:idRef="R_MALTPTabc" />
          <groups:member groups:idRef="R_MALTTRabc" />
          <groups:member groups:idRef="R_MALTTTRabc" />
          <groups:member groups:idRef="R_MALTabc" />
          <groups:member groups:idRef="R_MALTpts" />
          <groups:member groups:idRef="R_MALt2_2" />
          <groups:member groups:idRef="R_MALt2_3" />
          <groups:member groups:idRef="R_MAN6Pt6_2" />
          <groups:member groups:idRef="R_MANpts" />
          <groups:member groups:idRef="R_MELIBt2" />
          <groups:member groups:idRef="R_METDabc" />
          <groups:member groups:idRef="R_METabc" />
          <groups:member groups:idRef="R_MNLpts" />
          <groups:member groups:idRef="R_NACt" />
          <groups:member groups:idRef="R_NAt3_1" />
          <groups:member groups:idRef="R_NAt3_15" />
          <groups:member groups:idRef="R_NAt3_2" />
          <groups:member groups:idRef="R_NH4t" />
          <groups:member groups:idRef="R_NMNP" />
          <groups:member groups:idRef="R_NMNt7" />
          <groups:member groups:idRef="R_NO2t2r" />
          <groups:member groups:idRef="R_NO3t7" />
          <groups:member groups:idRef="R_O2t" />
          <groups:member groups:idRef="R_OCDCAt2" />
          <groups:member groups:idRef="R_ORNabc" />
          <groups:member groups:idRef="R_PHEt2r" />
          <groups:member groups:idRef="R_PIabc" />
          <groups:member groups:idRef="R_PIt2r" />
          <groups:member groups:idRef="R_PNTOt4" />
          <groups:member groups:idRef="R_PROabc" />
          <groups:member groups:idRef="R_PROt2r" />
          <groups:member groups:idRef="R_PROt4" />
          <groups:member groups:idRef="R_PTRCORNt7" />
          <groups:member groups:idRef="R_PTRCabc" />
          <groups:member groups:idRef="R_PTRCt2" />
          <groups:member groups:idRef="R_PYRt2" />
          <groups:member groups:idRef="R_RIBabc" />
          <groups:member groups:idRef="R_RMNt" />
          <groups:member groups:idRef="R_SBTpts" />
          <groups:member groups:idRef="R_SERt2r" />
          <groups:member groups:idRef="R_SERt4" />
          <groups:member groups:idRef="R_SPMDabc" />
          <groups:member groups:idRef="R_SUCCt2_2" />
          <groups:member groups:idRef="R_SUCCt2_3" />
          <groups:member groups:idRef="R_SUCCt2r" />
          <groups:member groups:idRef="R_SUCFUMt" />
          <groups:member groups:idRef="R_SUCpts" />
          <groups:member groups:idRef="R_SULabc" />
          <groups:member groups:idRef="R_TAURabc" />
          <groups:member groups:idRef="R_THMDt2_copy1" />
          <groups:member groups:idRef="R_THMDt2_copy2" />
          <groups:member groups:idRef="R_THMabc" />
          <groups:member groups:idRef="R_THRabc" />
          <groups:member groups:idRef="R_THRt2r" />
          <groups:member groups:idRef="R_THRt4" />
          <groups:member groups:idRef="R_TREpts" />
          <groups:member groups:idRef="R_TRPt2r" />
          <groups:member groups:idRef="R_TSULabc" />
          <groups:member groups:idRef="R_TTDCAt2" />
          <groups:member groups:idRef="R_TYRt2r" />
          <groups:member groups:idRef="R_URAt2_copy1" />
          <groups:member groups:idRef="R_URAt2_copy2" />
          <groups:member groups:idRef="R_UREAt" />
          <groups:member groups:idRef="R_URIt2_copy1" />
          <groups:member groups:idRef="R_URIt2_copy2" />
          <groups:member groups:idRef="R_VALabc" />
          <groups:member groups:idRef="R_VALt2r" />
          <groups:member groups:idRef="R_XANt" />
          <groups:member groups:idRef="R_XTSNt2" />
          <groups:member groups:idRef="R_XYLabc" />
          <groups:member groups:idRef="R_XYLt2" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g2" groups:kind="partonomy" groups:name="Alternate Carbon Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_2DGLCNRx" />
          <groups:member groups:idRef="R_2DGLCNRy" />
          <groups:member groups:idRef="R_2DGULRx" />
          <groups:member groups:idRef="R_2DGULRy" />
          <groups:member groups:idRef="R_3HCINNMH" />
          <groups:member groups:idRef="R_3HPPPNH" />
          <groups:member groups:idRef="R_5DGLCNR" />
          <groups:member groups:idRef="R_A5PISO" />
          <groups:member groups:idRef="R_AB6PGH" />
          <groups:member groups:idRef="R_ACACCT" />
          <groups:member groups:idRef="R_ACALD" />
          <groups:member groups:idRef="R_ACCOAL" />
          <groups:member groups:idRef="R_ACNML" />
          <groups:member groups:idRef="R_AGDC" />
          <groups:member groups:idRef="R_ALCD19" />
          <groups:member groups:idRef="R_ALDD19xr" />
          <groups:member groups:idRef="R_ALDD2x" />
          <groups:member groups:idRef="R_ALTRH" />
          <groups:member groups:idRef="R_AMALT1" />
          <groups:member groups:idRef="R_AMALT2" />
          <groups:member groups:idRef="R_AMALT3" />
          <groups:member groups:idRef="R_AMALT4" />
          <groups:member groups:idRef="R_ARAI" />
          <groups:member groups:idRef="R_BUTCT" />
          <groups:member groups:idRef="R_CINNDO" />
          <groups:member groups:idRef="R_DDGALK" />
          <groups:member groups:idRef="R_DDGLK" />
          <groups:member groups:idRef="R_DDPGALA" />
          <groups:member groups:idRef="R_DHAPT" />
          <groups:member groups:idRef="R_DHCIND" />
          <groups:member groups:idRef="R_DHCINDO_1" />
          <groups:member groups:idRef="R_DHPPD" />
          <groups:member groups:idRef="R_DKGLCNR1" />
          <groups:member groups:idRef="R_DKGLCNR2x" />
          <groups:member groups:idRef="R_DKGLCNR2y" />
          <groups:member groups:idRef="R_DOGULNR" />
          <groups:member groups:idRef="R_DRPA" />
          <groups:member groups:idRef="R_FAO4" />
          <groups:member groups:idRef="R_FCI" />
          <groups:member groups:idRef="R_FCLK" />
          <groups:member groups:idRef="R_FCLPA" />
          <groups:member groups:idRef="R_FFSD" />
          <groups:member groups:idRef="R_FRUK" />
          <groups:member groups:idRef="R_G3PD2" />
          <groups:member groups:idRef="R_G6PDA" />
          <groups:member groups:idRef="R_GALCTD" />
          <groups:member groups:idRef="R_GALCTND" />
          <groups:member groups:idRef="R_GALKr" />
          <groups:member groups:idRef="R_GALS3" />
          <groups:member groups:idRef="R_GALUi_copy1" />
          <groups:member groups:idRef="R_GALUi_copy2" />
          <groups:member groups:idRef="R_GLCRAL" />
          <groups:member groups:idRef="R_GLCRD" />
          <groups:member groups:idRef="R_GLTPD" />
          <groups:member groups:idRef="R_GLYCDx" />
          <groups:member groups:idRef="R_GLYCTO2" />
          <groups:member groups:idRef="R_GLYCTO3" />
          <groups:member groups:idRef="R_GLYCTO4" />
          <groups:member groups:idRef="R_GLYK" />
          <groups:member groups:idRef="R_GNK" />
          <groups:member groups:idRef="R_GUI1" />
          <groups:member groups:idRef="R_GUI2" />
          <groups:member groups:idRef="R_HKNDDH_1" />
          <groups:member groups:idRef="R_HKNTDH_1" />
          <groups:member groups:idRef="R_HOPNTAL" />
          <groups:member groups:idRef="R_HPPPNDO_1" />
          <groups:member groups:idRef="R_HPYRI" />
          <groups:member groups:idRef="R_HPYRRx" />
          <groups:member groups:idRef="R_HPYRRy" />
          <groups:member groups:idRef="R_IDOND" />
          <groups:member groups:idRef="R_IDOND2" />
          <groups:member groups:idRef="R_KG6PDC" />
          <groups:member groups:idRef="R_LACZ" />
          <groups:member groups:idRef="R_LCADi_copy1" />
          <groups:member groups:idRef="R_LCADi_copy2" />
          <groups:member groups:idRef="R_LCARS" />
          <groups:member groups:idRef="R_M1PD" />
          <groups:member groups:idRef="R_MAN6PI" />
          <groups:member groups:idRef="R_MANAO" />
          <groups:member groups:idRef="R_MCITD" />
          <groups:member groups:idRef="R_MCITL2" />
          <groups:member groups:idRef="R_MCITS" />
          <groups:member groups:idRef="R_MICITDr" />
          <groups:member groups:idRef="R_MLTG1" />
          <groups:member groups:idRef="R_MLTG2" />
          <groups:member groups:idRef="R_MLTG3" />
          <groups:member groups:idRef="R_MLTG4" />
          <groups:member groups:idRef="R_MLTG5" />
          <groups:member groups:idRef="R_MLTP1" />
          <groups:member groups:idRef="R_MLTP2" />
          <groups:member groups:idRef="R_MLTP3" />
          <groups:member groups:idRef="R_MMCD" />
          <groups:member groups:idRef="R_MME" />
          <groups:member groups:idRef="R_MMM2" />
          <groups:member groups:idRef="R_MNNH" />
          <groups:member groups:idRef="R_OBTFL" />
          <groups:member groups:idRef="R_OP4ENH" />
          <groups:member groups:idRef="R_PACCOAL" />
          <groups:member groups:idRef="R_PEAMNO" />
          <groups:member groups:idRef="R_PFK_2" />
          <groups:member groups:idRef="R_PGLYCP" />
          <groups:member groups:idRef="R_PGMT" />
          <groups:member groups:idRef="R_PMANM" />
          <groups:member groups:idRef="R_PPAKr" />
          <groups:member groups:idRef="R_PPCSCT" />
          <groups:member groups:idRef="R_PPM" />
          <groups:member groups:idRef="R_PPM2" />
          <groups:member groups:idRef="R_PPPNDO" />
          <groups:member groups:idRef="R_PTA2" />
          <groups:member groups:idRef="R_RBK" />
          <groups:member groups:idRef="R_RBK_L1" />
          <groups:member groups:idRef="R_RBP4E" />
          <groups:member groups:idRef="R_RMI" />
          <groups:member groups:idRef="R_RMK" />
          <groups:member groups:idRef="R_RMPA" />
          <groups:member groups:idRef="R_SBTPD" />
          <groups:member groups:idRef="R_TAGURr" />
          <groups:member groups:idRef="R_TARTD" />
          <groups:member groups:idRef="R_TAUDO" />
          <groups:member groups:idRef="R_TGBPA" />
          <groups:member groups:idRef="R_TRE6PH" />
          <groups:member groups:idRef="R_TRE6PP" />
          <groups:member groups:idRef="R_TRE6PS" />
          <groups:member groups:idRef="R_TREH" />
          <groups:member groups:idRef="R_TREHe" />
          <groups:member groups:idRef="R_UDPG4E" />
          <groups:member groups:idRef="R_UGLT" />
          <groups:member groups:idRef="R_X5PL3E" />
          <groups:member groups:idRef="R_XYLI1" />
          <groups:member groups:idRef="R_XYLI2" />
          <groups:member groups:idRef="R_XYLK" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g3" groups:kind="partonomy" groups:name="Cofactor and Prosthetic Group Biosynthesis" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_4HTHRS" />
          <groups:member groups:idRef="R_ACBIPGT" />
          <groups:member groups:idRef="R_ACPS1" />
          <groups:member groups:idRef="R_ADCL" />
          <groups:member groups:idRef="R_ADCS" />
          <groups:member groups:idRef="R_ADOCBIK" />
          <groups:member groups:idRef="R_ADOCBLS" />
          <groups:member groups:idRef="R_ALATA_D2" />
          <groups:member groups:idRef="R_ALATA_L2" />
          <groups:member groups:idRef="R_AMAOTr" />
          <groups:member groups:idRef="R_AMMQT8_2" />
          <groups:member groups:idRef="R_AMPMS" />
          <groups:member groups:idRef="R_AOXSr" />
          <groups:member groups:idRef="R_APRAUR" />
          <groups:member groups:idRef="R_ASP1DC" />
          <groups:member groups:idRef="R_ASPO3_1" />
          <groups:member groups:idRef="R_ASPO4_1" />
          <groups:member groups:idRef="R_ASPO5_1" />
          <groups:member groups:idRef="R_ASPO6_1" />
          <groups:member groups:idRef="R_BSORx" />
          <groups:member groups:idRef="R_BSORy" />
          <groups:member groups:idRef="R_BTS2" />
          <groups:member groups:idRef="R_CBIAT_DELETE" />
          <groups:member groups:idRef="R_CBLAT_DELETE" />
          <groups:member groups:idRef="R_CDPMEK" />
          <groups:member groups:idRef="R_CHRPL" />
          <groups:member groups:idRef="R_CPPPGO" />
          <groups:member groups:idRef="R_DB4PS" />
          <groups:member groups:idRef="R_DBTS" />
          <groups:member groups:idRef="R_DHBD" />
          <groups:member groups:idRef="R_DHBSr" />
          <groups:member groups:idRef="R_DHFR" />
          <groups:member groups:idRef="R_DHFS_1" />
          <groups:member groups:idRef="R_DHNAOT" />
          <groups:member groups:idRef="R_DHNPA2r" />
          <groups:member groups:idRef="R_DHPPDA2" />
          <groups:member groups:idRef="R_DHPS2_1" />
          <groups:member groups:idRef="R_DMATT" />
          <groups:member groups:idRef="R_DMPPS" />
          <groups:member groups:idRef="R_DMQMT" />
          <groups:member groups:idRef="R_DNMPPA" />
          <groups:member groups:idRef="R_DNTPPA" />
          <groups:member groups:idRef="R_DPCOAK" />
          <groups:member groups:idRef="R_DPR" />
          <groups:member groups:idRef="R_DXPRIi" />
          <groups:member groups:idRef="R_DXPS" />
          <groups:member groups:idRef="R_DXYLK" />
          <groups:member groups:idRef="R_E4PD" />
          <groups:member groups:idRef="R_ENTCS_1" />
          <groups:member groups:idRef="R_FCLT" />
          <groups:member groups:idRef="R_FMNAT" />
          <groups:member groups:idRef="R_G1SAT" />
          <groups:member groups:idRef="R_GLUCYS" />
          <groups:member groups:idRef="R_GLUTRR" />
          <groups:member groups:idRef="R_GLUTRS" />
          <groups:member groups:idRef="R_GRTT" />
          <groups:member groups:idRef="R_GTHOr" />
          <groups:member groups:idRef="R_GTHS" />
          <groups:member groups:idRef="R_GTPCI_1" />
          <groups:member groups:idRef="R_GTPCII2" />
          <groups:member groups:idRef="R_HBZOPT" />
          <groups:member groups:idRef="R_HEMEOS" />
          <groups:member groups:idRef="R_HETZK" />
          <groups:member groups:idRef="R_HMBS" />
          <groups:member groups:idRef="R_HMPK1" />
          <groups:member groups:idRef="R_HPPK2" />
          <groups:member groups:idRef="R_HYPOE" />
          <groups:member groups:idRef="R_ICHORS" />
          <groups:member groups:idRef="R_ICHORT" />
          <groups:member groups:idRef="R_IPDDI" />
          <groups:member groups:idRef="R_IPDPS" />
          <groups:member groups:idRef="R_MECDPDH" />
          <groups:member groups:idRef="R_MECDPS" />
          <groups:member groups:idRef="R_MEPCT" />
          <groups:member groups:idRef="R_MOHMT" />
          <groups:member groups:idRef="R_NADDP" />
          <groups:member groups:idRef="R_NADDPe" />
          <groups:member groups:idRef="R_NADK" />
          <groups:member groups:idRef="R_NADPPPS" />
          <groups:member groups:idRef="R_NADS1" />
          <groups:member groups:idRef="R_NAMNPP" />
          <groups:member groups:idRef="R_NMNAT" />
          <groups:member groups:idRef="R_NMNDA" />
          <groups:member groups:idRef="R_NMNN" />
          <groups:member groups:idRef="R_NNAM" />
          <groups:member groups:idRef="R_NNATr" />
          <groups:member groups:idRef="R_NNDMBRT" />
          <groups:member groups:idRef="R_NNDPR" />
          <groups:member groups:idRef="R_NPHS" />
          <groups:member groups:idRef="R_OCTDPS" />
          <groups:member groups:idRef="R_OHPBAT" />
          <groups:member groups:idRef="R_OHPHM" />
          <groups:member groups:idRef="R_OMBZLM" />
          <groups:member groups:idRef="R_OMMBLHX" />
          <groups:member groups:idRef="R_OMPHHX" />
          <groups:member groups:idRef="R_OPHBDC" />
          <groups:member groups:idRef="R_OPHHX" />
          <groups:member groups:idRef="R_OXGDC2" />
          <groups:member groups:idRef="R_PANTS" />
          <groups:member groups:idRef="R_PDX5POi" />
          <groups:member groups:idRef="R_PDX5PS" />
          <groups:member groups:idRef="R_PDXPP" />
          <groups:member groups:idRef="R_PERD" />
          <groups:member groups:idRef="R_PMDPHT" />
          <groups:member groups:idRef="R_PMPK" />
          <groups:member groups:idRef="R_PNTK" />
          <groups:member groups:idRef="R_PPBNGS" />
          <groups:member groups:idRef="R_PPCDC" />
          <groups:member groups:idRef="R_PPNCL2" />
          <groups:member groups:idRef="R_PPPGO" />
          <groups:member groups:idRef="R_PTPATi" />
          <groups:member groups:idRef="R_PYAM5PO" />
          <groups:member groups:idRef="R_PYDAMK" />
          <groups:member groups:idRef="R_PYDXK" />
          <groups:member groups:idRef="R_PYDXNK" />
          <groups:member groups:idRef="R_PYDXPP" />
          <groups:member groups:idRef="R_QULNS_1" />
          <groups:member groups:idRef="R_RBFK" />
          <groups:member groups:idRef="R_RBFSa" />
          <groups:member groups:idRef="R_RBFSb" />
          <groups:member groups:idRef="R_RZ5PP" />
          <groups:member groups:idRef="R_SERASr" />
          <groups:member groups:idRef="R_SHCHCS2" />
          <groups:member groups:idRef="R_SHCHD2" />
          <groups:member groups:idRef="R_SHCHF" />
          <groups:member groups:idRef="R_SUCBZL" />
          <groups:member groups:idRef="R_SUCBZS" />
          <groups:member groups:idRef="R_THZPSN" />
          <groups:member groups:idRef="R_TMK" />
          <groups:member groups:idRef="R_TMPK" />
          <groups:member groups:idRef="R_TMPPP" />
          <groups:member groups:idRef="R_UDCPDPS" />
          <groups:member groups:idRef="R_UPP3MT" />
          <groups:member groups:idRef="R_UPP3S" />
          <groups:member groups:idRef="R_UPPDC1" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g4" groups:kind="partonomy" groups:name="Cell Envelope Biosynthesis" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_AACPS1" />
          <groups:member groups:idRef="R_AACPS2" />
          <groups:member groups:idRef="R_AACPS3" />
          <groups:member groups:idRef="R_AACPS4" />
          <groups:member groups:idRef="R_AACPS5" />
          <groups:member groups:idRef="R_AADDGT" />
          <groups:member groups:idRef="R_ACGAMT" />
          <groups:member groups:idRef="R_ACMAMUT" />
          <groups:member groups:idRef="R_AGMHE" />
          <groups:member groups:idRef="R_ALAALAr" />
          <groups:member groups:idRef="R_DAGK_EC" />
          <groups:member groups:idRef="R_ECAP_EC" />
          <groups:member groups:idRef="R_EDTXS1" />
          <groups:member groups:idRef="R_EDTXS2" />
          <groups:member groups:idRef="R_EDTXS3" />
          <groups:member groups:idRef="R_EDTXS4" />
          <groups:member groups:idRef="R_ETHAAL" />
          <groups:member groups:idRef="R_G1PACT" />
          <groups:member groups:idRef="R_G1PTT" />
          <groups:member groups:idRef="R_GALUi_copy1" />
          <groups:member groups:idRef="R_GALUi_copy2" />
          <groups:member groups:idRef="R_GDMANE" />
          <groups:member groups:idRef="R_GF6PTA" />
          <groups:member groups:idRef="R_GLUR" />
          <groups:member groups:idRef="R_GMAND" />
          <groups:member groups:idRef="R_GMHEPAT" />
          <groups:member groups:idRef="R_GMHEPK" />
          <groups:member groups:idRef="R_GMHEPPA" />
          <groups:member groups:idRef="R_GOFUCR" />
          <groups:member groups:idRef="R_GPDDA1" />
          <groups:member groups:idRef="R_GPDDA2" />
          <groups:member groups:idRef="R_GPDDA3" />
          <groups:member groups:idRef="R_GPDDA4" />
          <groups:member groups:idRef="R_GPDDA5" />
          <groups:member groups:idRef="R_KDOCT2" />
          <groups:member groups:idRef="R_KDOPP" />
          <groups:member groups:idRef="R_KDOPS" />
          <groups:member groups:idRef="R_LPADSS" />
          <groups:member groups:idRef="R_LPLIPA1" />
          <groups:member groups:idRef="R_LPLIPA2" />
          <groups:member groups:idRef="R_LPLIPA3" />
          <groups:member groups:idRef="R_LPLIPA4" />
          <groups:member groups:idRef="R_LPLIPA5" />
          <groups:member groups:idRef="R_LPLIPA6" />
          <groups:member groups:idRef="R_LPSSYN_EC" />
          <groups:member groups:idRef="R_MAN1PT2" />
          <groups:member groups:idRef="R_MI1PP" />
          <groups:member groups:idRef="R_MOAT" />
          <groups:member groups:idRef="R_MOAT2" />
          <groups:member groups:idRef="R_PAPA_EC" />
          <groups:member groups:idRef="R_PAPPT3" />
          <groups:member groups:idRef="R_PEPT_EC" />
          <groups:member groups:idRef="R_PGAMT" />
          <groups:member groups:idRef="R_PLIPA1" />
          <groups:member groups:idRef="R_PLIPA2" />
          <groups:member groups:idRef="R_PLIPA3" />
          <groups:member groups:idRef="R_PPTGS" />
          <groups:member groups:idRef="R_S7PI" />
          <groups:member groups:idRef="R_TDPADGAT" />
          <groups:member groups:idRef="R_TDPAGTA" />
          <groups:member groups:idRef="R_TDPDRE" />
          <groups:member groups:idRef="R_TDPDRR" />
          <groups:member groups:idRef="R_TDPGDH" />
          <groups:member groups:idRef="R_TDSK" />
          <groups:member groups:idRef="R_U23GAAT" />
          <groups:member groups:idRef="R_UAAGDS" />
          <groups:member groups:idRef="R_UACMAMO" />
          <groups:member groups:idRef="R_UAG2E" />
          <groups:member groups:idRef="R_UAGAAT" />
          <groups:member groups:idRef="R_UAGCVT" />
          <groups:member groups:idRef="R_UAGDP" />
          <groups:member groups:idRef="R_UAGPT3" />
          <groups:member groups:idRef="R_UAMAGS" />
          <groups:member groups:idRef="R_UAMAS" />
          <groups:member groups:idRef="R_UAPGR" />
          <groups:member groups:idRef="R_UDCPDP" />
          <groups:member groups:idRef="R_UDPGALM" />
          <groups:member groups:idRef="R_UDPGD" />
          <groups:member groups:idRef="R_UGMDDS" />
          <groups:member groups:idRef="R_UHGADA" />
          <groups:member groups:idRef="R_USHD" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g5" groups:kind="partonomy" groups:name="Arginine and Proline Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ABTA" />
          <groups:member groups:idRef="R_ABUTD" />
          <groups:member groups:idRef="R_ACGK" />
          <groups:member groups:idRef="R_ACGS" />
          <groups:member groups:idRef="R_ACODA" />
          <groups:member groups:idRef="R_ACOTA" />
          <groups:member groups:idRef="R_ADMDC" />
          <groups:member groups:idRef="R_AGMT" />
          <groups:member groups:idRef="R_AGPR" />
          <groups:member groups:idRef="R_ARGDC" />
          <groups:member groups:idRef="R_ARGSL" />
          <groups:member groups:idRef="R_ARGSS" />
          <groups:member groups:idRef="R_AST" />
          <groups:member groups:idRef="R_CBPS" />
          <groups:member groups:idRef="R_DKMPPD" />
          <groups:member groups:idRef="R_DKMPPD2" />
          <groups:member groups:idRef="R_G5SADs" />
          <groups:member groups:idRef="R_G5SD" />
          <groups:member groups:idRef="R_GLU5K" />
          <groups:member groups:idRef="R_GSPMDA" />
          <groups:member groups:idRef="R_GSPMDS" />
          <groups:member groups:idRef="R_MDRPD" />
          <groups:member groups:idRef="R_MTAN" />
          <groups:member groups:idRef="R_MTRI" />
          <groups:member groups:idRef="R_MTRK" />
          <groups:member groups:idRef="R_NACODA" />
          <groups:member groups:idRef="R_OCBT" />
          <groups:member groups:idRef="R_ORNDC" />
          <groups:member groups:idRef="R_ORNTA" />
          <groups:member groups:idRef="R_P5CD" />
          <groups:member groups:idRef="R_P5CR" />
          <groups:member groups:idRef="R_PROD2" />
          <groups:member groups:idRef="R_PTRCTA" />
          <groups:member groups:idRef="R_SADH" />
          <groups:member groups:idRef="R_SGDS" />
          <groups:member groups:idRef="R_SGSAD" />
          <groups:member groups:idRef="R_SOTA" />
          <groups:member groups:idRef="R_SPMDAT1" />
          <groups:member groups:idRef="R_SPMDAT2" />
          <groups:member groups:idRef="R_SPMS" />
          <groups:member groups:idRef="R_SSALx" />
          <groups:member groups:idRef="R_SSALy" />
          <groups:member groups:idRef="R_UNK3" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g6" groups:kind="partonomy" groups:name="Membrane Lipid Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ACACT1r" />
          <groups:member groups:idRef="R_ACCOAC" />
          <groups:member groups:idRef="R_ACOATA" />
          <groups:member groups:idRef="R_C120SN" />
          <groups:member groups:idRef="R_C140SN" />
          <groups:member groups:idRef="R_C141SN" />
          <groups:member groups:idRef="R_C160SN" />
          <groups:member groups:idRef="R_C161SN" />
          <groups:member groups:idRef="R_C181SN" />
          <groups:member groups:idRef="R_CDAPPA_EC" />
          <groups:member groups:idRef="R_CLPNS_EC" />
          <groups:member groups:idRef="R_DASYN_EC" />
          <groups:member groups:idRef="R_FAO1" />
          <groups:member groups:idRef="R_FAO2" />
          <groups:member groups:idRef="R_FAO3" />
          <groups:member groups:idRef="R_KAS14" />
          <groups:member groups:idRef="R_KAS15" />
          <groups:member groups:idRef="R_KAS16" />
          <groups:member groups:idRef="R_MACPD" />
          <groups:member groups:idRef="R_MCOATA" />
          <groups:member groups:idRef="R_PASYN_EC" />
          <groups:member groups:idRef="R_PGPP_EC" />
          <groups:member groups:idRef="R_PGSA_EC" />
          <groups:member groups:idRef="R_PSD_EC" />
          <groups:member groups:idRef="R_PSSA_EC" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g7" groups:kind="partonomy" groups:name="Putative Transporters" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ACACt2" />
          <groups:member groups:idRef="R_ALAt2r" />
          <groups:member groups:idRef="R_ALLTNt2r" />
          <groups:member groups:idRef="R_ARGORNt7" />
          <groups:member groups:idRef="R_BUTt2r" />
          <groups:member groups:idRef="R_CHLabc" />
          <groups:member groups:idRef="R_CYNTt2" />
          <groups:member groups:idRef="R_GALCTt2r" />
          <groups:member groups:idRef="R_GLCRt2r" />
          <groups:member groups:idRef="R_GLUABUTt7" />
          <groups:member groups:idRef="R_GLYBabc" />
          <groups:member groups:idRef="R_GLYBt2r" />
          <groups:member groups:idRef="R_GUAt2r" />
          <groups:member groups:idRef="R_HCINNMt2r" />
          <groups:member groups:idRef="R_HPPPNt2r" />
          <groups:member groups:idRef="R_PPPNt2r" />
          <groups:member groups:idRef="R_SUCCabc" />
          <groups:member groups:idRef="R_TARTRt7" />
          <groups:member groups:idRef="R_URAt2_copy1" />
          <groups:member groups:idRef="R_URAt2_copy2" />
          <groups:member groups:idRef="R_XANt2" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g8" groups:kind="partonomy" groups:name="Valine, Leucine, and Isoleucine Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ACHBS" />
          <groups:member groups:idRef="R_ACLS" />
          <groups:member groups:idRef="R_DHAD1" />
          <groups:member groups:idRef="R_DHAD2" />
          <groups:member groups:idRef="R_ILETA" />
          <groups:member groups:idRef="R_IPMD" />
          <groups:member groups:idRef="R_IPPMIa" />
          <groups:member groups:idRef="R_IPPMIb" />
          <groups:member groups:idRef="R_IPPS" />
          <groups:member groups:idRef="R_KARA1" />
          <groups:member groups:idRef="R_KARA2" />
          <groups:member groups:idRef="R_LEUTAi" />
          <groups:member groups:idRef="R_OMCDC" />
          <groups:member groups:idRef="R_THRD_L" />
          <groups:member groups:idRef="R_VALTA" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g9" groups:kind="partonomy" groups:name="Pyruvate Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ACKr" />
          <groups:member groups:idRef="R_ACS" />
          <groups:member groups:idRef="R_ADHEr" />
          <groups:member groups:idRef="R_FHL" />
          <groups:member groups:idRef="R_LDH_D" />
          <groups:member groups:idRef="R_PFL" />
          <groups:member groups:idRef="R_PTAr" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g10" groups:kind="partonomy" groups:name="Unassigned" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ACONMT" />
          <groups:member groups:idRef="R_BETALDHx" />
          <groups:member groups:idRef="R_BETALDHy" />
          <groups:member groups:idRef="R_CAT" />
          <groups:member groups:idRef="R_CYANST" />
          <groups:member groups:idRef="R_HCO3E" />
          <groups:member groups:idRef="R_SELNPS" />
          <groups:member groups:idRef="R_SPODM" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g11" groups:kind="partonomy" groups:name="Citric Acid Cycle" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ACONT" />
          <groups:member groups:idRef="R_AKGDH" />
          <groups:member groups:idRef="R_CITL" />
          <groups:member groups:idRef="R_CS" />
          <groups:member groups:idRef="R_FRD2" />
          <groups:member groups:idRef="R_FRD3" />
          <groups:member groups:idRef="R_FUM" />
          <groups:member groups:idRef="R_ICDHyr" />
          <groups:member groups:idRef="R_MDH" />
          <groups:member groups:idRef="R_MDH2" />
          <groups:member groups:idRef="R_MDH3" />
          <groups:member groups:idRef="R_SUCD1" />
          <groups:member groups:idRef="R_SUCOAS" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g12" groups:kind="partonomy" groups:name="Nucleotide Salvage Pathway" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ADA" />
          <groups:member groups:idRef="R_ADD" />
          <groups:member groups:idRef="R_ADK1" />
          <groups:member groups:idRef="R_ADK3" />
          <groups:member groups:idRef="R_ADK4" />
          <groups:member groups:idRef="R_ADNCYC" />
          <groups:member groups:idRef="R_ADNK1" />
          <groups:member groups:idRef="R_ADPT" />
          <groups:member groups:idRef="R_AMPN" />
          <groups:member groups:idRef="R_AP4AH" />
          <groups:member groups:idRef="R_AP5AH" />
          <groups:member groups:idRef="R_CMPN" />
          <groups:member groups:idRef="R_CSND" />
          <groups:member groups:idRef="R_CYTD" />
          <groups:member groups:idRef="R_CYTDK2" />
          <groups:member groups:idRef="R_CYTK1" />
          <groups:member groups:idRef="R_CYTK2" />
          <groups:member groups:idRef="R_DADA" />
          <groups:member groups:idRef="R_DADK" />
          <groups:member groups:idRef="R_DCTPD" />
          <groups:member groups:idRef="R_DCYTD" />
          <groups:member groups:idRef="R_DGK1" />
          <groups:member groups:idRef="R_DTMPK" />
          <groups:member groups:idRef="R_DURIK1" />
          <groups:member groups:idRef="R_DURIPP" />
          <groups:member groups:idRef="R_DUTPDP" />
          <groups:member groups:idRef="R_GK1" />
          <groups:member groups:idRef="R_GP4GH" />
          <groups:member groups:idRef="R_GSNK" />
          <groups:member groups:idRef="R_GUAD" />
          <groups:member groups:idRef="R_GUAPRT" />
          <groups:member groups:idRef="R_HXPRT" />
          <groups:member groups:idRef="R_INSK" />
          <groups:member groups:idRef="R_NDPK1" />
          <groups:member groups:idRef="R_NDPK2" />
          <groups:member groups:idRef="R_NDPK3" />
          <groups:member groups:idRef="R_NDPK4" />
          <groups:member groups:idRef="R_NDPK5" />
          <groups:member groups:idRef="R_NDPK6" />
          <groups:member groups:idRef="R_NDPK7" />
          <groups:member groups:idRef="R_NDPK8" />
          <groups:member groups:idRef="R_NTD1" />
          <groups:member groups:idRef="R_NTD10" />
          <groups:member groups:idRef="R_NTD11" />
          <groups:member groups:idRef="R_NTD2" />
          <groups:member groups:idRef="R_NTD3" />
          <groups:member groups:idRef="R_NTD4" />
          <groups:member groups:idRef="R_NTD5" />
          <groups:member groups:idRef="R_NTD6" />
          <groups:member groups:idRef="R_NTD7" />
          <groups:member groups:idRef="R_NTD8" />
          <groups:member groups:idRef="R_NTD9" />
          <groups:member groups:idRef="R_NTPP1" />
          <groups:member groups:idRef="R_NTPP2" />
          <groups:member groups:idRef="R_NTPP3" />
          <groups:member groups:idRef="R_NTPP4" />
          <groups:member groups:idRef="R_NTPP5" />
          <groups:member groups:idRef="R_NTPP6" />
          <groups:member groups:idRef="R_NTPP7" />
          <groups:member groups:idRef="R_NTPP8" />
          <groups:member groups:idRef="R_NTPTP1" />
          <groups:member groups:idRef="R_NTPTP2" />
          <groups:member groups:idRef="R_PUNP1" />
          <groups:member groups:idRef="R_PUNP2" />
          <groups:member groups:idRef="R_PUNP3" />
          <groups:member groups:idRef="R_PUNP4" />
          <groups:member groups:idRef="R_PUNP5" />
          <groups:member groups:idRef="R_PUNP6" />
          <groups:member groups:idRef="R_PUNP7" />
          <groups:member groups:idRef="R_PYNP2r" />
          <groups:member groups:idRef="R_RNDR1" />
          <groups:member groups:idRef="R_RNDR2" />
          <groups:member groups:idRef="R_RNDR3" />
          <groups:member groups:idRef="R_RNDR4" />
          <groups:member groups:idRef="R_RNTR1" />
          <groups:member groups:idRef="R_RNTR2" />
          <groups:member groups:idRef="R_RNTR3" />
          <groups:member groups:idRef="R_RNTR4" />
          <groups:member groups:idRef="R_TMDK1" />
          <groups:member groups:idRef="R_TMDPP" />
          <groups:member groups:idRef="R_TMDS" />
          <groups:member groups:idRef="R_UMPK" />
          <groups:member groups:idRef="R_UPPRT" />
          <groups:member groups:idRef="R_URIDK2r" />
          <groups:member groups:idRef="R_URIK2" />
          <groups:member groups:idRef="R_XPPT" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g13" groups:kind="partonomy" groups:name="Cysteine Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ADSK" />
          <groups:member groups:idRef="R_BPNT" />
          <groups:member groups:idRef="R_CYSDS" />
          <groups:member groups:idRef="R_CYSS" />
          <groups:member groups:idRef="R_PAPSR" />
          <groups:member groups:idRef="R_SADT2" />
          <groups:member groups:idRef="R_SERAT" />
          <groups:member groups:idRef="R_SULR" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g14" groups:kind="partonomy" groups:name="Purine and Pyrimidine Biosynthesis" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ADSL1r" />
          <groups:member groups:idRef="R_ADSL2r" />
          <groups:member groups:idRef="R_ADSS" />
          <groups:member groups:idRef="R_AICART" />
          <groups:member groups:idRef="R_AIRC2" />
          <groups:member groups:idRef="R_AIRC3" />
          <groups:member groups:idRef="R_ASPCT" />
          <groups:member groups:idRef="R_CTPS2" />
          <groups:member groups:idRef="R_DHORD2" />
          <groups:member groups:idRef="R_DHORD5" />
          <groups:member groups:idRef="R_DHORTS" />
          <groups:member groups:idRef="R_GARFT" />
          <groups:member groups:idRef="R_GART" />
          <groups:member groups:idRef="R_GLUPRT" />
          <groups:member groups:idRef="R_GMPR" />
          <groups:member groups:idRef="R_GMPS2" />
          <groups:member groups:idRef="R_IMPC" />
          <groups:member groups:idRef="R_IMPD" />
          <groups:member groups:idRef="R_OMPDC" />
          <groups:member groups:idRef="R_ORPT" />
          <groups:member groups:idRef="R_PRAGSr" />
          <groups:member groups:idRef="R_PRAIS" />
          <groups:member groups:idRef="R_PRASCSi" />
          <groups:member groups:idRef="R_PRFGS" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g15" groups:kind="partonomy" groups:name="Methionine Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_AHCi" />
          <groups:member groups:idRef="R_AHCYSNS" />
          <groups:member groups:idRef="R_CYSTL" />
          <groups:member groups:idRef="R_DHPTDCs" />
          <groups:member groups:idRef="R_HSST" />
          <groups:member groups:idRef="R_METAT" />
          <groups:member groups:idRef="R_METS_1" />
          <groups:member groups:idRef="R_RHCCE" />
          <groups:member groups:idRef="R_SHSL1" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g16" groups:kind="partonomy" groups:name="Alanine and Aspartate Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ALAR_copy1" />
          <groups:member groups:idRef="R_ALAR_copy2" />
          <groups:member groups:idRef="R_ALATA_L" />
          <groups:member groups:idRef="R_ASNN" />
          <groups:member groups:idRef="R_ASNS1" />
          <groups:member groups:idRef="R_ASNS2" />
          <groups:member groups:idRef="R_ASPT" />
          <groups:member groups:idRef="R_ASPTA" />
          <groups:member groups:idRef="R_DAAD" />
          <groups:member groups:idRef="R_VPAMTr" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g17" groups:kind="partonomy" groups:name="Nitrogen Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ALLTAHr" />
          <groups:member groups:idRef="R_ALLTN" />
          <groups:member groups:idRef="R_CYNTAH" />
          <groups:member groups:idRef="R_UGLYCH" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g18" groups:kind="partonomy" groups:name="Putative" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_AMANAPEr" />
          <groups:member groups:idRef="R_AMANK" />
          <groups:member groups:idRef="R_CBMKr" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g19" groups:kind="partonomy" groups:name="Tyrosine, Tryptophan, and Phenylalanine Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ANPRT" />
          <groups:member groups:idRef="R_ANS" />
          <groups:member groups:idRef="R_CHORM" />
          <groups:member groups:idRef="R_CHORS" />
          <groups:member groups:idRef="R_DDPA" />
          <groups:member groups:idRef="R_DHQTi" />
          <groups:member groups:idRef="R_DHQS" />
          <groups:member groups:idRef="R_IGPS" />
          <groups:member groups:idRef="R_PHETA1" />
          <groups:member groups:idRef="R_PPND" />
          <groups:member groups:idRef="R_PPNDH" />
          <groups:member groups:idRef="R_PRAIi" />
          <groups:member groups:idRef="R_PSCVT" />
          <groups:member groups:idRef="R_SHK3Dr" />
          <groups:member groups:idRef="R_SHKK" />
          <groups:member groups:idRef="R_TRPAS2" />
          <groups:member groups:idRef="R_TRPS1" />
          <groups:member groups:idRef="R_TRPS2" />
          <groups:member groups:idRef="R_TRPS3" />
          <groups:member groups:idRef="R_TYRTA" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g20" groups:kind="partonomy" groups:name="Threonine and Lysine Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ASAD" />
          <groups:member groups:idRef="R_ASPK" />
          <groups:member groups:idRef="R_DAPDC" />
          <groups:member groups:idRef="R_DAPE" />
          <groups:member groups:idRef="R_DHDPRy" />
          <groups:member groups:idRef="R_DHDPS" />
          <groups:member groups:idRef="R_HSDy" />
          <groups:member groups:idRef="R_HSK" />
          <groups:member groups:idRef="R_LYSDC" />
          <groups:member groups:idRef="R_SDPDS" />
          <groups:member groups:idRef="R_SDPTA" />
          <groups:member groups:idRef="R_THDPS" />
          <groups:member groups:idRef="R_THRA" />
          <groups:member groups:idRef="R_THRS" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g21" groups:kind="partonomy" groups:name="Biomass and maintenance functions" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ATPM" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g22" groups:kind="partonomy" groups:name="Histidine Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ATPPRT" />
          <groups:member groups:idRef="R_HISTD" />
          <groups:member groups:idRef="R_HISTP" />
          <groups:member groups:idRef="R_HSTPT" />
          <groups:member groups:idRef="R_IG3PS" />
          <groups:member groups:idRef="R_IGPDH" />
          <groups:member groups:idRef="R_PRAMPC" />
          <groups:member groups:idRef="R_PRATPP" />
          <groups:member groups:idRef="R_PRMICI" />
          <groups:member groups:idRef="R_PRPPS" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g23" groups:kind="partonomy" groups:name="Oxidative Phosphorylation" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ATPS4r" />
          <groups:member groups:idRef="R_CRNBTCT" />
          <groups:member groups:idRef="R_CRNCBCT" />
          <groups:member groups:idRef="R_CRNCDH" />
          <groups:member groups:idRef="R_CYTBD" />
          <groups:member groups:idRef="R_CYTBO3" />
          <groups:member groups:idRef="R_DMSOR1" />
          <groups:member groups:idRef="R_DMSOR1e" />
          <groups:member groups:idRef="R_DMSOR2" />
          <groups:member groups:idRef="R_DMSOR2e" />
          <groups:member groups:idRef="R_FDH2" />
          <groups:member groups:idRef="R_FDH3" />
          <groups:member groups:idRef="R_G3PD5" />
          <groups:member groups:idRef="R_G3PD6" />
          <groups:member groups:idRef="R_G3PD7" />
          <groups:member groups:idRef="R_GLCDe" />
          <groups:member groups:idRef="R_HYD1" />
          <groups:member groups:idRef="R_HYD2" />
          <groups:member groups:idRef="R_HYD3" />
          <groups:member groups:idRef="R_LDH_D2" />
          <groups:member groups:idRef="R_NADH10" />
          <groups:member groups:idRef="R_NADH5" />
          <groups:member groups:idRef="R_NADH6" />
          <groups:member groups:idRef="R_NADH7" />
          <groups:member groups:idRef="R_NADH8" />
          <groups:member groups:idRef="R_NADH9" />
          <groups:member groups:idRef="R_NADTRHD" />
          <groups:member groups:idRef="R_NO3R1" />
          <groups:member groups:idRef="R_NO3R2" />
          <groups:member groups:idRef="R_NTRIR2x" />
          <groups:member groups:idRef="R_POX" />
          <groups:member groups:idRef="R_SUCD4" />
          <groups:member groups:idRef="R_THD2" />
          <groups:member groups:idRef="R_TMAOR1" />
          <groups:member groups:idRef="R_TMAOR1e" />
          <groups:member groups:idRef="R_TMAOR2" />
          <groups:member groups:idRef="R_TMAOR2e" />
          <groups:member groups:idRef="R_TRDR" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g24" groups:kind="partonomy" groups:name="Pentose Phosphate Pathway" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_EDA" />
          <groups:member groups:idRef="R_EDD" />
          <groups:member groups:idRef="R_G6PDH2r" />
          <groups:member groups:idRef="R_GND" />
          <groups:member groups:idRef="R_PGL" />
          <groups:member groups:idRef="R_RPE" />
          <groups:member groups:idRef="R_RPI" />
          <groups:member groups:idRef="R_TALA" />
          <groups:member groups:idRef="R_TKT1" />
          <groups:member groups:idRef="R_TKT2" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g25" groups:kind="partonomy" groups:name="Glycolysis/Gluconeogenesis" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ENO" />
          <groups:member groups:idRef="R_F6PA" />
          <groups:member groups:idRef="R_FBA" />
          <groups:member groups:idRef="R_FBP" />
          <groups:member groups:idRef="R_G1PP" />
          <groups:member groups:idRef="R_GAPD" />
          <groups:member groups:idRef="R_GLCP" />
          <groups:member groups:idRef="R_GLCS1" />
          <groups:member groups:idRef="R_GLGC" />
          <groups:member groups:idRef="R_HEX1" />
          <groups:member groups:idRef="R_PDH" />
          <groups:member groups:idRef="R_PFK" />
          <groups:member groups:idRef="R_PGI" />
          <groups:member groups:idRef="R_PGK" />
          <groups:member groups:idRef="R_PGM" />
          <groups:member groups:idRef="R_PPS" />
          <groups:member groups:idRef="R_PYK" />
          <groups:member groups:idRef="R_TPI" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g26" groups:kind="partonomy" groups:name="Folate Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_FTHFD" />
          <groups:member groups:idRef="R_GCALDD" />
          <groups:member groups:idRef="R_GLYCL" />
          <groups:member groups:idRef="R_MTHFC_1" />
          <groups:member groups:idRef="R_MTHFD_1" />
          <groups:member groups:idRef="R_MTHFR2_1" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g27" groups:kind="partonomy" groups:name="Glycine and Serine Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_GHMT2r" />
          <groups:member groups:idRef="R_GLYAT" />
          <groups:member groups:idRef="R_PGCD" />
          <groups:member groups:idRef="R_PSERT" />
          <groups:member groups:idRef="R_PSP_L" />
          <groups:member groups:idRef="R_SERD_D" />
          <groups:member groups:idRef="R_SERD_L" />
          <groups:member groups:idRef="R_THRD" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g28" groups:kind="partonomy" groups:name="Glutamate metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_GLNS" />
          <groups:member groups:idRef="R_GLUDC" />
          <groups:member groups:idRef="R_GLUDy" />
          <groups:member groups:idRef="R_GLUN" />
          <groups:member groups:idRef="R_GLUSy" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g29" groups:kind="partonomy" groups:name="Glyoxylate Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_GLXCL" />
          <groups:member groups:idRef="R_GLYCK" />
          <groups:member groups:idRef="R_GLYCLTDx" />
          <groups:member groups:idRef="R_GLYCLTDy" />
          <groups:member groups:idRef="R_TRSARr" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g30" groups:kind="partonomy" groups:name="Methylglyoxal Metabolism" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_GLYOX" />
          <groups:member groups:idRef="R_LGTHL" />
          <groups:member groups:idRef="R_MGSA" />
        </groups:listOfMembers>
      </groups:group>
      <groups:group groups:id="g31" groups:kind="partonomy" groups:name="Anaplerotic reactions" sboTerm="SBO:0000633">
        <groups:listOfMembers>
          <groups:member groups:idRef="R_ICL" />
          <groups:member groups:idRef="R_MALS" />
          <groups:member groups:idRef="R_ME1" />
          <groups:member groups:idRef="R_ME2" />
          <groups:member groups:idRef="R_PPA" />
          <groups:member groups:idRef="R_PPC" />
          <groups:member groups:idRef="R_PPCK" />
        </groups:listOfMembers>
      </groups:group>
    </groups:listOfGroups>
    <listOfUnitDefinitions>
      <unitDefinition id="mmol_per_gDW_per_hr" metaid="mmol_per_gDW_per_hr" name="Millimoles per gram (dry weight) per hour">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#mmol_per_gDW_per_hr">
              <bqbiol:isDescribedBy>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/pubmed/7986045" />
                </rdf:Bag>
              </bqbiol:isDescribedBy>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfUnits>
          <unit exponent="1" kind="mole" metaid="_65fdbe47-3274-4ea9-8b58-4731bc5ab5b0" multiplier="1" scale="-3">
            <annotation>
              <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
                <rdf:Description rdf:about="#_65fdbe47-3274-4ea9-8b58-4731bc5ab5b0">
                  <bqbiol:is>
                    <rdf:Bag>
                      <rdf:li rdf:resource="http://identifiers.org/unit/UO:0000040" />
                    </rdf:Bag>
                  </bqbiol:is>
                </rdf:Description>
              </rdf:RDF>
            </annotation>
          </unit>
          <unit exponent="-1" kind="gram" metaid="cdd1752a-9866-44e7-8d61-cbef21e48098" multiplier="1" scale="0">
            <annotation>
              <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
                <rdf:Description rdf:about="#cdd1752a-9866-44e7-8d61-cbef21e48098">
                  <bqbiol:isVersionOf>
                    <rdf:Bag>
                      <rdf:li rdf:resource="UO:0000021" />
                    </rdf:Bag>
                  </bqbiol:isVersionOf>
                </rdf:Description>
              </rdf:RDF>
            </annotation>
          </unit>
          <unit exponent="-1" kind="second" metaid="cb50b172-6fc5-4f80-ab2e-4a8d7d6b7fe7" multiplier="3600" scale="0">
            <annotation>
              <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
                <rdf:Description rdf:about="#cb50b172-6fc5-4f80-ab2e-4a8d7d6b7fe7">
                  <bqbiol:isVersionOf>
                    <rdf:Bag>
                      <rdf:li rdf:resource="http://identifiers.org/unit/UO:0000032" />
                    </rdf:Bag>
                  </bqbiol:isVersionOf>
                </rdf:Description>
              </rdf:RDF>
            </annotation>
          </unit>
        </listOfUnits>
      </unitDefinition>
      <unitDefinition id="substance" name="Millimoles per gram (dry weight)">
        <listOfUnits>
          <unit exponent="1" kind="mole" metaid="f5267609-913b-4559-b549-3e7ea398d12c" multiplier="1" scale="-3">
            <annotation>
              <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
                <rdf:Description rdf:about="#f5267609-913b-4559-b549-3e7ea398d12c">
                  <bqbiol:is>
                    <rdf:Bag>
                      <rdf:li rdf:resource="http://identifiers.org/unit/UO:0000040" />
                    </rdf:Bag>
                  </bqbiol:is>
                </rdf:Description>
              </rdf:RDF>
            </annotation>
          </unit>
          <unit exponent="-1" kind="gram" metaid="_977485fd-0b37-4d0b-a795-4d1822174b57" multiplier="1" scale="0">
            <annotation>
              <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
                <rdf:Description rdf:about="#_977485fd-0b37-4d0b-a795-4d1822174b57">
                  <bqbiol:isVersionOf>
                    <rdf:Bag>
                      <rdf:li rdf:resource="UO:0000021" />
                    </rdf:Bag>
                  </bqbiol:isVersionOf>
                </rdf:Description>
              </rdf:RDF>
            </annotation>
          </unit>
        </listOfUnits>
      </unitDefinition>
      <unitDefinition id="time" name="Hour">
        <listOfUnits>
          <unit exponent="1" kind="second" metaid="_2faa81fb-8850-4b65-9558-978b345ad44a" multiplier="3600" scale="0">
            <annotation>
              <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
                <rdf:Description rdf:about="#_2faa81fb-8850-4b65-9558-978b345ad44a">
                  <bqbiol:is>
                    <rdf:Bag>
                      <rdf:li rdf:resource="http://identifiers.org/unit/UO:0000032" />
                    </rdf:Bag>
                  </bqbiol:is>
                </rdf:Description>
              </rdf:RDF>
            </annotation>
          </unit>
        </listOfUnits>
      </unitDefinition>
    </listOfUnitDefinitions>
    <listOfCompartments>
      <compartment constant="true" id="c" metaid="_3dd523b2-19ba-4124-98a7-efa139c51f22" name="cytosol" sboTerm="SBO:0000290" units="dimensionless">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#_3dd523b2-19ba-4124-98a7-efa139c51f22">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.compartment/c" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </compartment>
      <compartment constant="true" id="e" metaid="_05b1c435-0af2-4acd-9f6e-0d31d7466c2d" name="extracellular space" sboTerm="SBO:0000290" units="dimensionless">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#_05b1c435-0af2-4acd-9f6e-0d31d7466c2d">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.compartment/e" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </compartment>
    </listOfCompartments>
    <listOfSpecies>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C20H21N7O7" hasOnlySubstanceUnits="false" id="M_10fthf_c" metaid="M_10fthf_c"
      name="10-Formyltetrahydrofolate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_10fthf_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/10fthf" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:10-FORMYL-THF" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11304" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15637" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19108" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19109" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57454" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:698" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00972" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00234" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM237" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00201" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C1836H3448O250" hasOnlySubstanceUnits="false" id="M_12dgr_EC_c" metaid="M_12dgr_EC_c"
      name="1,2-Diacylglycerol  (E.coli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_12dgr_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/12dgr_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90584" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H8O2" hasOnlySubstanceUnits="false" id="M_12ppd__S_c" metaid="M_12ppd__S_c"
      name="(S)-Propane-1,2-diol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_12ppd__S_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/12ppd__S" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROPANE-1-2-DIOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18799" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:440" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45065" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06213" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02917" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1255" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19024" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H8O2" hasOnlySubstanceUnits="false" id="M_12ppd__S_e" metaid="M_12ppd__S_e"
      name="(S)-Propane-1,2-diol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_12ppd__S_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/12ppd__S" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROPANE-1-2-DIOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18799" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:440" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45065" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06213" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02917" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1255" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19024" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C3H4O10P2" hasOnlySubstanceUnits="false" id="M_13dpg_c" metaid="M_13dpg_c"
      name="3-Phospho-D-glyceroyl phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_13dpg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/13dpg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DPG" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11881" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16001" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1658" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57604" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62758" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00236" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM261" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00203" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="2" fbc:chemicalFormula="C5H16N2" hasOnlySubstanceUnits="false" id="M_15dap_c" metaid="M_15dap_c"
      name="1,5-Diaminopentane">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_15dap_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/15dap" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CADAVERINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13928" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18127" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22974" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3288" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58384" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02322" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01672" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM943" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01155" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="2" fbc:chemicalFormula="C5H16N2" hasOnlySubstanceUnits="false" id="M_15dap_e" metaid="M_15dap_e"
      name="1,5-Diaminopentane">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_15dap_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/15dap" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CADAVERINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13928" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18127" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22974" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3288" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58384" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02322" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01672" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM943" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01155" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H6NO2" hasOnlySubstanceUnits="false" id="M_1pyr5c_c" metaid="M_1pyr5c_c"
      name="1-Pyrroline-5-carboxylate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_1pyr5c_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/1pyr5c" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-DELTA1-PYRROLINE_5-CARBOXYLATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11297" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11689" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12409" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15893" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17388" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18727" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19873" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26458" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:371" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01301" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02240" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03912" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04322" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1617" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02431" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02651" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29655" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H7O4" hasOnlySubstanceUnits="false" id="M_23ddhb_c" metaid="M_23ddhb_c"
      name="2,3-Dihydro-2,3-dihydroxybenzoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_23ddhb_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/23ddhb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDRO-DIOH-BENZOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11420" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11423" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19311" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48968" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57576" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58764" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:877" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04171" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114261" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02566" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29666" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H5O4" hasOnlySubstanceUnits="false" id="M_23dhb_c" metaid="M_23dhb_c"
      name="2,3-Dihydroxybenzoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_23dhb_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/23dhb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-3-DIHYDROXYBENZOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18026" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19320" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36654" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41901" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:885" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00196" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM455" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00168" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C17H16N5O10P" hasOnlySubstanceUnits="false" id="M_23dhba_c" metaid="M_23dhba_c"
      name="(2,3-Dihydroxybenzoyl)adenylate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_23dhba_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/23dhba" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-62" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15572" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:170" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18521" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57417" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04030" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1978" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02494" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H5NO4" hasOnlySubstanceUnits="false" id="M_23dhdp_c" metaid="M_23dhdp_c"
      name="2,3-Dihydrodipicolinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_23dhdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/23dhdp" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11421" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48052" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM163139" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H9O4" hasOnlySubstanceUnits="false" id="M_23dhmb_c" metaid="M_23dhmb_c"
      name="(R)-2,3-Dihydroxy-3-methylbutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_23dhmb_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/23dhmb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-13357" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10966" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15684" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18645" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:305" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49072" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12141" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04272" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050453" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114097" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19030" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O4" hasOnlySubstanceUnits="false" id="M_23dhmp_c" metaid="M_23dhmp_c"
      name="(R)-2,3-Dihydroxy-3-methylpentanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_23dhmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/23dhmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:1-KETO-2-METHYLVALERATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18646" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27512" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:306" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49258" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12140" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06007" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050452" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1202" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19046" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H7O7" hasOnlySubstanceUnits="false" id="M_23doguln_c" metaid="M_23doguln_c"
      name="2,3-Dioxo-L-gulonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_23doguln_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/23doguln" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-19692" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-334" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10900" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15622" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18578" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:226" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57441" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60793" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB05971" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06511" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62803" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04575" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060195" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM958" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02784" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C13H15N4O12P" hasOnlySubstanceUnits="false" id="M_25aics_c" metaid="M_25aics_c"
      name="(S)-2-[5-Amino-1-(5-phospho-D-ribosyl)imidazole-4-carboxamido]succinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_25aics_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/25aics" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:P-RIBOSYL-4-SUCCCARB-AMINOIMIDAZOLE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18971" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28208" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:576" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00797" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06274" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04811" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM32478" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02921" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H7O7" hasOnlySubstanceUnits="false" id="M_25dkglcn_c" metaid="M_25dkglcn_c"
      name="2,5-diketo-D-gluconate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_25dkglcn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/25dkglcn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:25-DIDEHYDRO-D-GLUCONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1070" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11449" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18281" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19378" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58428" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02780" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050471" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM936" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01793" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H14N5O8P" hasOnlySubstanceUnits="false" id="M_25drapp_c" metaid="M_25drapp_c"
      name="2,5-Diamino-6-(ribosylamino)-4-(3H)-pyrimidinone 5'-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_25drapp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/25drapp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIAMINO-OH-PHOSPHORIBOSYLAMINO-PYR" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11446" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19368" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29114" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58614" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59545" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59546" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:927" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01304" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM648" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00957" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29703" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H14N2O4" hasOnlySubstanceUnits="false" id="M_26dap_LL_c" metaid="M_26dap_LL_c"
      name="LL-2,6-Diaminoheptanedioate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_26dap_LL_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/26dap_LL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LL-DIAMINOPIMELATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16026" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21428" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21429" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47031" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57609" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6341" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01370" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00666" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170102" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM644" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00504" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H14N2O4" hasOnlySubstanceUnits="false" id="M_26dap__M_c" metaid="M_26dap__M_c"
      name="Meso-2,6-Diaminoheptanedioate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_26dap__M_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/26dap__M" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MESO-DIAMINOPIMELATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10598" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12822" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12823" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12825" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16488" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25203" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25204" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40838" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57791" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00680" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM529" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00516" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H14N2O4" hasOnlySubstanceUnits="false" id="M_26dap__M_e" metaid="M_26dap__M_e"
      name="Meso-2,6-Diaminoheptanedioate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_26dap__M_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/26dap__M" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MESO-DIAMINOPIMELATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10598" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12822" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12823" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12825" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16488" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25203" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25204" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40838" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57791" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00680" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM529" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00516" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O4" hasOnlySubstanceUnits="false" id="M_2ahbut_c" metaid="M_2ahbut_c"
      name="(S)-2-Aceto-2-hydroxybutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2ahbut_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2ahbut" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-ACETO-2-HYDROXY-BUTYRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18730" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27681" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:373" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49256" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06854" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06900" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06006" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050383" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114220" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19045" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H7NO3" hasOnlySubstanceUnits="false" id="M_2aobut_c" metaid="M_2aobut_c"
      name="L-2-Amino-3-oxobutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2aobut_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2aobut" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMINO-OXOBUT" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13048" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16944" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21195" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35229" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40668" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40673" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6156" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:78948" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06454" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03508" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060172" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114087" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02211" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C12H13NO9P" hasOnlySubstanceUnits="false" id="M_2cpr5p_c" metaid="M_2cpr5p_c"
      name="1-(2-Carboxyphenylamino)-1-deoxy-D-ribulose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2cpr5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2cpr5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBOXYPHENYLAMINO-DEOXYRIBULOSE-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11186" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18959" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28625" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29112" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:565" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58613" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01302" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1455" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00956" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C7H10O10P" hasOnlySubstanceUnits="false" id="M_2dda7p_c" metaid="M_2dda7p_c"
      name="2-Dehydro-3-deoxy-D-arabino-heptonate 7-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dda7p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dda7p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-DEOXY-D-ARABINO-HEPTULOSONATE-7-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11544" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11785" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11786" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18150" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19523" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20003" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29477" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58394" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04691" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1219" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02857" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H8O9P" hasOnlySubstanceUnits="false" id="M_2ddg6p_c" metaid="M_2ddg6p_c"
      name="2-Dehydro-3-deoxy-D-gluconate 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2ddg6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2ddg6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-KETO-3-DEOXY-6-P-GLUCONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1051" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11543" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11551" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15925" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19522" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57569" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01376" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04442" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM649" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02711" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O6" hasOnlySubstanceUnits="false" id="M_2ddglcn_c" metaid="M_2ddglcn_c"
      name="2-Dehydro-3-deoxy-D-gluconate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2ddglcn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2ddglcn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-DEHYDRO-3-DEOXY-D-GALACTONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-DEHYDRO-3-DEOXY-D-GLUCONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1056" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1059" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11547" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17028" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17032" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19527" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57989" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57990" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01353" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00204" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01216" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050475" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050486" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1627" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00176" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00894" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O6" hasOnlySubstanceUnits="false" id="M_2ddglcn_e" metaid="M_2ddglcn_e"
      name="2-Dehydro-3-deoxy-D-gluconate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2ddglcn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2ddglcn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-DEHYDRO-3-DEOXY-D-GALACTONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-DEHYDRO-3-DEOXY-D-GLUCONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1056" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1059" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11547" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17028" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17032" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19527" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57989" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57990" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01353" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00204" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01216" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050475" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050486" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1627" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00176" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00894" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H8O9P" hasOnlySubstanceUnits="false" id="M_2dh3dgal6p_c" metaid="M_2dh3dgal6p_c"
      name="2-Dehydro-3-deoxy-D-galactonate 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dh3dgal6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dh3dgal6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEHYDRO-DEOXY-GALACTONATE-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1057" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11548" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17860" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19528" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58298" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01286" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90136" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00945" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O6" hasOnlySubstanceUnits="false" id="M_2dh3dgal_c" metaid="M_2dh3dgal_c"
      name="2-Dehydro-3-deoxy-D-galactonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dh3dgal_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dh3dgal" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-DEHYDRO-3-DEOXY-D-GALACTONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-DEHYDRO-3-DEOXY-D-GLUCONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1056" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1059" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11547" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17028" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17032" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19527" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57989" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57990" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01353" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00204" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01216" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050475" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050486" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1627" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00176" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00894" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_2dhglcn_c" metaid="M_2dhglcn_c"
      name="2-Dehydro-D-gluconate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dhglcn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dhglcn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-377" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11559" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1180" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16808" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19539" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19669" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27469" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58512" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59504" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06473" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM582" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00480" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03889" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_2dhguln_c" metaid="M_2dhguln_c"
      name="2-Dehydro-L-gulonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dhguln_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dhguln" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114283" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15351" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O4" hasOnlySubstanceUnits="false" id="M_2dhp_c" metaid="M_2dhp_c"
      name="2-Dehydropantoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dhp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dhp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-DEHYDROPANTOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1071" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11561" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19545" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00966" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM959" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00712" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C50H70O2" hasOnlySubstanceUnits="false" id="M_2dmmq8_c" metaid="M_2dmmq8_c"
      name="2-Demethylmenaquinone 8">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dmmq8_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dmmq8" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEMETHYLMENAQUINONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48455" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR02030027" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2178" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15352" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C50H72O2" hasOnlySubstanceUnits="false" id="M_2dmmql8_c" metaid="M_2dmmql8_c"
      name="2-Demethylmenaquinol 8">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dmmql8_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dmmql8" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-12115" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61873" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM558" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O7P" hasOnlySubstanceUnits="false" id="M_2dr1p_c" metaid="M_2dr1p_c"
      name="2-Deoxy-D-ribose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dr1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dr1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXY-D-RIBOSE-1-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1081" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11563" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19558" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28542" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57259" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58576" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01351" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00672" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM789" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00509" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd26845" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O7P" hasOnlySubstanceUnits="false" id="M_2dr5p_c" metaid="M_2dr5p_c"
      name="2-Deoxy-D-ribose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2dr5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2dr5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXY-RIBOSE-5P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1082" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16132" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19559" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42055" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:55513" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62877" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01031" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00673" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2179" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00510" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd26846" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H3O4" hasOnlySubstanceUnits="false" id="M_2h3oppan_c" metaid="M_2h3oppan_c"
      name="2-Hydroxy-3-oxopropanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2h3oppan_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2h3oppan" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TARTRONATE-S-ALD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1123" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11583" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131382" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15194" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16992" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19605" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57978" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06781" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06938" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01146" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM475" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00843" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H8O4" hasOnlySubstanceUnits="false" id="M_2ippm_c" metaid="M_2ippm_c"
      name="2-Isopropylmaleate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2ippm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2ippm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-9451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11604" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1179" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17275" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19668" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58085" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02631" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1706" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01710" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H7O3S" hasOnlySubstanceUnits="false" id="M_2kmb_c" metaid="M_2kmb_c"
      name="2-keto-4-methylthiobutyrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2kmb_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2kmb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-479" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12029" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16723" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1902" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22458" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33574" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43720" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:63388" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01553" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB13210" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01180" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060170" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM276" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00869" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H8N3O7P2" hasOnlySubstanceUnits="false" id="M_2mahmp_c" metaid="M_2mahmp_c"
      name="2-Methyl-4-amino-5-hydroxymethylpyrimidine diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2mahmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2mahmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMINO-HYDROXYMETHYL-METHYLPYRIMIDINE-PP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11612" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1194" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11953" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16629" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19684" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29079" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57841" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04752" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1135" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02894" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C7H5O6" hasOnlySubstanceUnits="false" id="M_2mcacn_c" metaid="M_2mcacn_c"
      name="Cis-2-Methylaconitate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2mcacn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2mcacn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-1136" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11084" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16717" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18819" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57872" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06357" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04225" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1792" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02597" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C7H7O7" hasOnlySubstanceUnits="false" id="M_2mcit_c" metaid="M_2mcit_c"
      name="2-Methylcitrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2mcit_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2mcit" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-622" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10860" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11592" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11618" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1203" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15598" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19630" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19695" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19696" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30835" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30836" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:50948" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58853" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00379" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03610" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02225" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050442" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90279" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01501" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd24620" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H11O7P" hasOnlySubstanceUnits="false" id="M_2me4p_c" metaid="M_2me4p_c"
      name="2-C-methyl-D-erythritol 4-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2me4p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2me4p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-C-METHYL-D-ERYTHRITOL-4-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1030" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11483" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17764" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58262" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11434" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1626" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08286" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H10O9P2" hasOnlySubstanceUnits="false" id="M_2mecdp_c" metaid="M_2mecdp_c"
      name="2-C-methyl-D-erythritol 2,4-cyclodiphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2mecdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2mecdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2C-METH-D-ERYTHRITOL-CYCLODIPHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1029" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11481" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11482" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18425" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58483" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11453" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1168" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08301" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H5O3" hasOnlySubstanceUnits="false" id="M_2obut_c" metaid="M_2obut_c"
      name="2-Oxobutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2obut_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2obut" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-OXOBUTANOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11636" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1250" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16763" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19741" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30831" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39748" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00109" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060002" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM159" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00094" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C46H70O2" hasOnlySubstanceUnits="false" id="M_2ohph_c" metaid="M_2ohph_c"
      name="2-Octaprenyl-6-hydroxyphenol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2ohph_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2ohph" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-OCTAPRENYL-6-HYDROXYPHENOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1233" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62730" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05811" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1635" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03445" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C47H72O3" hasOnlySubstanceUnits="false" id="M_2ombzl_c" metaid="M_2ombzl_c"
      name="2-Octaprenyl-6-methoxy-1,4-benzoquinol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2ombzl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2ombzl" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162799" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15359" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C48H74O4" hasOnlySubstanceUnits="false" id="M_2omhmbl_c" metaid="M_2omhmbl_c"
      name="2-Octaprenyl-3-methyl-5-hydroxy-6-methoxy-1,4-benzoquinol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2omhmbl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2omhmbl" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OCTAPRENYL-METHYL-OH-METHOXY-BENZQ" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1231" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19728" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27688" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61705" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2187" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15360" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C48H74O3" hasOnlySubstanceUnits="false" id="M_2ommbl_c" metaid="M_2ommbl_c"
      name="2-Octaprenyl-3-methyl-6-methoxy- 1,4-benzoquinol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2ommbl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2ommbl" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11661" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM35383" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15361" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C47H72O2" hasOnlySubstanceUnits="false" id="M_2omph_c" metaid="M_2omph_c"
      name="2-Octaprenyl-6-methoxyphenol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2omph_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2omph" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-OCTAPRENYL-6-METHOXYPHENOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1235" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05812" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1707" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03446" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C46H70O" hasOnlySubstanceUnits="false" id="M_2oph_c" metaid="M_2oph_c"
      name="2-Octaprenylphenol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2oph_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2oph" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-OCTAPRENYLPHENOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1236" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40407" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05810" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1886" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03444" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C14H22N3O17P3" hasOnlySubstanceUnits="false" id="M_2p4c2me_c" metaid="M_2p4c2me_c"
      name="2-phospho-4-(cytidine 5'-diphospho)-2-C-methyl-D-erythritol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2p4c2me_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2p4c2me" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-PHOSPHO-4-CYTIDINE-5-DIPHOSPHO-2-C-MET" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11649" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11650" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1266" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16840" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57919" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11436" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1641" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08288" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C3H4O7P" hasOnlySubstanceUnits="false" id="M_2pg_c" metaid="M_2pg_c"
      name="D-Glycerate 2-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2pg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2pg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-PG" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1267" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17835" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21028" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24344" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39868" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58289" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:88350" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03391" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62707" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00631" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM275" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00482" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C2H2O6P" hasOnlySubstanceUnits="false" id="M_2pglyc_c" metaid="M_2pglyc_c"
      name="2-Phosphoglycolate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2pglyc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2pglyc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-67" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11652" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1268" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17150" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19763" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19764" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44849" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58033" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00988" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2074" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00727" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C11H10O6" hasOnlySubstanceUnits="false" id="M_2shchc_c" metaid="M_2shchc_c"
      name="2-Succinyl-6-hydroxy-2,4-cyclohexadiene-1-carboxylate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_2shchc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/2shchc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCINYL-OH-CYCLOHEXADIENE-COOH" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1277" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57272" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162721" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28162" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H7O4" hasOnlySubstanceUnits="false" id="M_34hpp_c" metaid="M_34hpp_c"
      name="3-(4-Hydroxyphenyl)pyruvate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_34hpp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/34hpp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:P-HYDROXY-PHENYLPYRUVATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1118" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11725" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11727" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12016" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19599" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20425" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20426" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27683" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:594665" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00707" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06915" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01179" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05350" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM153" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00868" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03174" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H10O5" hasOnlySubstanceUnits="false" id="M_3c2hmp_c" metaid="M_3c2hmp_c"
      name="3-Carboxy-2-hydroxy-4-methylpentanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3c2hmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3c2hmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-D-THREO-HYDROXY-3-CARBOXY-ISOCAPROATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11760" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11843" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15592" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1565" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35114" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35120" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35121" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35122" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43465" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43468" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12156" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04411" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM891" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02693" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H10O5" hasOnlySubstanceUnits="false" id="M_3c3hmp_c" metaid="M_3c3hmp_c"
      name="3-Carboxy-3-hydroxy-4-methylpentanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3c3hmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3c3hmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-CARBOXY-3-HYDROXY-ISOCAPROATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1178" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35128" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02504" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170083" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM985" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01646" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H8O5" hasOnlySubstanceUnits="false" id="M_3c4mop_c" metaid="M_3c4mop_c"
      name="3-Carboxy-4-methyl-2-oxopentanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3c4mop_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3c4mop" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-7100" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11765" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1467" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17214" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19975" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12149" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04236" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1602" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02605" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_3dhguln_c" metaid="M_3dhguln_c"
      name="3-Dehydro-L-gulonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3dhguln_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3dhguln" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-KETO-L-GULONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11777" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1482" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16142" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19992" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57655" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06334" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00618" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM736" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00473" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H8O10P" hasOnlySubstanceUnits="false" id="M_3dhgulnp_c" metaid="M_3dhgulnp_c"
      name="3-keto-L-gulonate-6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3dhgulnp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3dhgulnp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-2343" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49039" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58774" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C14899" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1170" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd10596" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H9O6" hasOnlySubstanceUnits="false" id="M_3dhq_c" metaid="M_3dhq_c"
      name="3-Dehydroquinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3dhq_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3dhq" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEHYDROQUINATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11781" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12122" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1487" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17947" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19997" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32364" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42078" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12710" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00944" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM478" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00699" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H7O5" hasOnlySubstanceUnits="false" id="M_3dhsk_c" metaid="M_3dhsk_c"
      name="3-Dehydroshikimate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3dhsk_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3dhsk" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-DEHYDRO-SHIKIMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11782" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12123" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1488" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16630" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19998" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2052" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30918" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42005" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02637" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM611" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01716" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H7O3" hasOnlySubstanceUnits="false" id="M_3hcinnm_c" metaid="M_3hcinnm_c"
      name="3-hydroxycinnamic acid">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3hcinnm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3hcinnm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-10797" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32357" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47925" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47927" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47928" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62774" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C12621" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1799" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd09252" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H7O3" hasOnlySubstanceUnits="false" id="M_3hcinnm_e" metaid="M_3hcinnm_e"
      name="3-hydroxycinnamic acid">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3hcinnm_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3hcinnm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-10797" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32357" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47925" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47927" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47928" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62774" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C12621" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1799" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd09252" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C25H47N2O9PRS" hasOnlySubstanceUnits="false" id="M_3hmrsACP_c" metaid="M_3hmrsACP_c"
      name="(R)-3-Hydroxytetradecanoyl-[acyl-carrier protein]">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3hmrsACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3hmrsACP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:R-3-hydroxymyristoyl-ACPs" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04688" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM7733" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11484" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H9O3" hasOnlySubstanceUnits="false" id="M_3hpppn_c" metaid="M_3hpppn_c"
      name="3-(3-hydroxy-phenyl)propionate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3hpppn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3hpppn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-HYDROXYPHENYL-PROPIONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57277" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00375" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11457" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1634" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08304" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H9O3" hasOnlySubstanceUnits="false" id="M_3hpppn_e" metaid="M_3hpppn_e"
      name="3-(3-hydroxy-phenyl)propionate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3hpppn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3hpppn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-HYDROXYPHENYL-PROPIONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57277" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00375" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11457" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1634" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08304" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C11H12NO6P" hasOnlySubstanceUnits="false" id="M_3ig3p_c" metaid="M_3ig3p_c"
      name="C'-(3-Indolyl)-glycerol 3-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3ig3p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3ig3p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:INDOLE-3-GLYCEROL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:51793" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58866" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03506" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM866" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02210" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29667" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H7O3" hasOnlySubstanceUnits="false" id="M_3mob_c" metaid="M_3mob_c"
      name="3-Methyl-2-oxobutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3mob_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3mob" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-KETO-ISOVALERATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132177" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1584" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20115" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43714" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00141" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01020274" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM238" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00123" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O3" hasOnlySubstanceUnits="false" id="M_3mop_c" metaid="M_3mop_c"
      name="(S)-3-Methyl-2-oxopentanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3mop_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3mop" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-KETO-3-METHYL-VALERATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10888" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11049" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15614" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18567" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18568" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18755" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:213" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35146" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:401" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00671" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01020275" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM439" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00508" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C47H69O3" hasOnlySubstanceUnits="false" id="M_3ophb_c" metaid="M_3ophb_c"
      name="3-Octaprenyl-4-hydroxybenzoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3ophb_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3ophb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-OCTAPRENYL-4-HYDROXYBENZOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1617" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:50116" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05809" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2195" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03443" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C3H4O7P" hasOnlySubstanceUnits="false" id="M_3pg_c" metaid="M_3pg_c"
      name="3-Phospho-D-glycerate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3pg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3pg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:G3P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11879" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11880" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1657" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17794" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21029" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58272" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00197" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM126" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00169" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C3H2O7P" hasOnlySubstanceUnits="false" id="M_3php_c" metaid="M_3php_c"
      name="3-Phosphohydroxypyruvate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3php_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3php" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-P-HYDROXYPYRUVATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11883" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11884" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1661" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18110" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20191" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30933" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03232" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM541" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02069" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H9O10P" hasOnlySubstanceUnits="false" id="M_3psme_c" metaid="M_3psme_c"
      name="5-O-(1-Carboxyvinyl)-3-phosphoshikimate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_3psme_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/3psme" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-ENOLPYRUVYL-SHIKIMATE-5P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12097" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16257" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20533" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2104" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57701" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01269" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1365" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00932" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H9NO2" hasOnlySubstanceUnits="false" id="M_4abut_c" metaid="M_4abut_c"
      name="4-Aminobutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4abut_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4abut" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-AMINO-BUTYRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16865" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1786" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:193777" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20317" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20318" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40483" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59888" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00112" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00334" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00058" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100039" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM192" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00281" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H9NO2" hasOnlySubstanceUnits="false" id="M_4abut_e" metaid="M_4abut_e"
      name="4-Aminobutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4abut_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4abut" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-AMINO-BUTYRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16865" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1786" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:193777" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20317" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20318" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40483" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59888" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00112" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00334" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00058" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100039" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM192" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00281" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C4H10NO" hasOnlySubstanceUnits="false" id="M_4abutn_c" metaid="M_4abutn_c"
      name="4-Aminobutanal">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4abutn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4abutn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-AMINO-BUTYRALDEHYDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11960" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17769" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1785" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20316" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58264" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01080" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60247" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00555" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM422" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00434" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12219" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H6NO2" hasOnlySubstanceUnits="false" id="M_4abz_c" metaid="M_4abz_c"
      name="4-Aminobenzoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4abz_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4abz" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:P-AMINO-BENZOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:113372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11959" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1783" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17836" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20314" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20315" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30753" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44778" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00568" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02456" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM421" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00443" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H10NO5" hasOnlySubstanceUnits="false" id="M_4adcho_c" metaid="M_4adcho_c"
      name="4-amino-4-deoxychorismate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4adcho_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4adcho" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-AMINO-4-DEOXYCHORISMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11956" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18198" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35180" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35181" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58406" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11355" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1458" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08210" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H9N3O" hasOnlySubstanceUnits="false" id="M_4ahmmp_c" metaid="M_4ahmmp_c"
      name="4-Amino-5-hydroxymethyl-2-methylpyrimidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4ahmmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4ahmmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11957" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16892" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1781" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20312" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43206" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01279" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM874" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00939" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H8N3O4P" hasOnlySubstanceUnits="false" id="M_4ampm_c" metaid="M_4ampm_c"
      name="4-Amino-2-methyl-5-phosphomethylpyrimidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4ampm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4ampm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMINO-HYDROXYMETHYL-METHYL-PYR-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11954" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11958" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1779" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18032" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20309" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44219" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58354" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04556" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM790" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02775" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C14H23N3O14P2" hasOnlySubstanceUnits="false" id="M_4c2me_c" metaid="M_4c2me_c"
      name="4-(cytidine 5'-diphospho)-2-C-methyl-D-erythritol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4c2me_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4c2me" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-CYTIDINE-5-DIPHOSPHO-2-C" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11938" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16578" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1770" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57823" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11435" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1642" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08287" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H7O4" hasOnlySubstanceUnits="false" id="M_4h2opntn_c" metaid="M_4h2opntn_c"
      name="4-Hydroxy-2-oxopentanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4h2opntn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4h2opntn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-14442" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11994" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17655" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1840" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20377" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58222" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03589" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1176" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02258" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H8O2" hasOnlySubstanceUnits="false" id="M_4hba_c" metaid="M_4hba_c"
      name="4-Hydroxy-benzyl alcohol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4hba_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4hba" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-HYDROXY-BENZYL-ALCOHOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:67410" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11724" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C17467" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM107505" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15378" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd17595" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H5O3" hasOnlySubstanceUnits="false" id="M_4hbz_c" metaid="M_4hbz_c"
      name="4-Hydroxybenzoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4hbz_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4hbz" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-hydroxybenzoate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12003" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17879" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1858" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30763" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44949" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00500" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00156" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM164" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00136" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H9NO4" hasOnlySubstanceUnits="false" id="M_4hthr_c" metaid="M_4hthr_c"
      name="4-Hydroxy-L-threonine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4hthr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4hthr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-2189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1853" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20393" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28330" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60904" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06056" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1802" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03608" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H9NOS" hasOnlySubstanceUnits="false" id="M_4mhetz_c" metaid="M_4mhetz_c"
      name="4-Methyl-5-(2-hydroxyethyl)-thiazole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4mhetz_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4mhetz" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THZ" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12021" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12022" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12086" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17957" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2011" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20522" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46320" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB32985" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04294" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM962" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02636" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O3" hasOnlySubstanceUnits="false" id="M_4mop_c" metaid="M_4mop_c"
      name="4-Methyl-2-oxopentanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4mop_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4mop" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2K-4CH3-PENTANOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12020" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17865" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1891" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20438" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41619" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48430" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00233" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM404" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00200" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H8NO4PS" hasOnlySubstanceUnits="false" id="M_4mpetz_c" metaid="M_4mpetz_c"
      name="4-Methyl-5-(2-phosphoethyl)-thiazole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4mpetz_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4mpetz" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THZ-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12023" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17857" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1892" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20439" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30638" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46182" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58296" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04327" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM960" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02654" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H6NO7P" hasOnlySubstanceUnits="false" id="M_4pasp_c" metaid="M_4pasp_c"
      name="4-Phospho-L-aspartate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4pasp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4pasp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-BETA-ASPARTYL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12042" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13062" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15836" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1925" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20471" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20472" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21246" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57535" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12250" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03082" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1177" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01977" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C4H6O8P" hasOnlySubstanceUnits="false" id="M_4per_c" metaid="M_4per_c"
      name="4-Phospho-D-erythronate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4per_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4per" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ERYTHRONATE-4P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1924" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41926" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49003" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49055" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58766" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03393" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1407" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02147" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C9H15NO8P" hasOnlySubstanceUnits="false" id="M_4ppan_c" metaid="M_4ppan_c"
      name="D-4'-Phosphopantothenate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4ppan_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4ppan" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-P-PANTOTHENATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12886" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15905" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18702" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20891" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4082" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01016" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62700" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03492" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM415" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02201" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C12H20N2O9PS" hasOnlySubstanceUnits="false" id="M_4ppcys_c" metaid="M_4ppcys_c"
      name="N-((R)-4-Phosphopantothenoyl)-L-cysteine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4ppcys_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4ppcys" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:R-4-PHOSPHOPANTOTHENOYL-L-CYSTEINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12438" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15769" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21461" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:328" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57507" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59458" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04352" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM483" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02666" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H16N4O6" hasOnlySubstanceUnits="false" id="M_4r5au_c" metaid="M_4r5au_c"
      name="4-(1-D-Ribitylamino)-5-aminouracil">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_4r5au_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/4r5au" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMINO-RIBOSYLAMINO-1H-3H-PYR-DIONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11929" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15934" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1761" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20283" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:52403" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57573" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04732" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM791" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02882" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29668" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C9H11N3O9P" hasOnlySubstanceUnits="false" id="M_5aizc_c" metaid="M_5aizc_c"
      name="5-amino-1-(5-phospho-D-ribosyl)imidazole-4-carboxylate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5aizc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5aizc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHORIBOSYL-CARBOXY-AMINOIMIDAZOLE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12103" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18968" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28413" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41336" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43272" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:575" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58564" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:77657" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06273" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04751" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM507" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02893" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H9NO3" hasOnlySubstanceUnits="false" id="M_5aop_c" metaid="M_5aop_c"
      name="5-Amino-4-oxopentanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5aop_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5aop" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-AMINO-LEVULINATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12109" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132970" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17549" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2034" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20547" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:356416" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01149" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00430" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07567" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100055" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM405" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00338" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H15N4O9P" hasOnlySubstanceUnits="false" id="M_5aprbu_c" metaid="M_5aprbu_c"
      name="5-Amino-6-(5'-phosphoribitylamino)uracil">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5aprbu_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5aprbu" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-1086" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2031" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58421" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03841" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04454" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1178" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02720" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H13N4O9P" hasOnlySubstanceUnits="false" id="M_5apru_c" metaid="M_5apru_c"
      name="5-Amino-6-(5'-phosphoribosylamino)uracil">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5apru_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5apru" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-602" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12108" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18337" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2032" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20546" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58453" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59551" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01268" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1532" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00931" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C9H11N3O9P" hasOnlySubstanceUnits="false" id="M_5caiz_c" metaid="M_5caiz_c"
      name="5-phosphoribosyl-5-carboxyaminoimidazole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5caiz_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5caiz" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-181" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48000" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58730" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12268" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C15667" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1409" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11310" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd25523" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_5dglcn_c" metaid="M_5dglcn_c"
      name="5-Dehydro-D-gluconate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5dglcn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5dglcn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-DEHYDROGLUCONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12120" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12121" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17426" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2051" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20564" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58143" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01062" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM963" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00781" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H6O7" hasOnlySubstanceUnits="false" id="M_5dh4dglc_c" metaid="M_5dh4dglc_c"
      name="5-Dehydro-4-deoxy-D-glucarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5dh4dglc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5dh4dglc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-KETO-4-DEOXY-D-GLUCARATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12117" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16369" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2048" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20561" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35453" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35454" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42815" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42819" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43704" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00679" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM854" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00515" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O7PS" hasOnlySubstanceUnits="false" id="M_5mdr1p_c" metaid="M_5mdr1p_c"
      name="5-Methylthio-5-deoxy-D-ribose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5mdr1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5mdr1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-444" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12767" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12769" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12770" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20614" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20616" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2102" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27859" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30309" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30310" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30311" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58533" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04188" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM407" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02574" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29669" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O7PS" hasOnlySubstanceUnits="false" id="M_5mdru1p_c" metaid="M_5mdru1p_c"
      name="5-Methylthio-5-deoxy-D-ribulose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5mdru1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5mdru1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-1063" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12768" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20615" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2100" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28096" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44141" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58548" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01299" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62647" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04582" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM522" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02791" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C11H15N5O3S" hasOnlySubstanceUnits="false" id="M_5mta_c" metaid="M_5mta_c"
      name="5-Methylthioadenosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5mta_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5mta" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-METHYLTHIOADENOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12055" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12064" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12771" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14605" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17509" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18175" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1966" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20491" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20494" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44181" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01173" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00170" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM150" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00147" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C20H24N7O6" hasOnlySubstanceUnits="false" id="M_5mthf_c" metaid="M_5mthf_c"
      name="5-Methyltetrahydrofolate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5mthf_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5mthf" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-METHYL-THF" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12146" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:136009" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15641" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18608" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20612" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2097" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01396" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00440" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D09353" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM318" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00345" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O4S" hasOnlySubstanceUnits="false" id="M_5mtr_c" metaid="M_5mtr_c"
      name="5-Methylthio-D-ribose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5mtr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5mtr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-560" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12148" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16895" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2101" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22007" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:78440" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03089" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1592" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01981" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C14H17N2O7P" hasOnlySubstanceUnits="false" id="M_5prdmbz_c" metaid="M_5prdmbz_c"
      name="N1-(5-Phospho-alpha-D-ribosyl)-5,6-dimethylbenzimidazole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_5prdmbz_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/5prdmbz" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALPHA-RIBAZOLE-5-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12622" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16837" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21797" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30599" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45218" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45255" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45256" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57918" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7355" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04778" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1224" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02904" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H9N5O2" hasOnlySubstanceUnits="false" id="M_6hmhpt_c" metaid="M_6hmhpt_c"
      name="6-hydroxymethyl dihydropterin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_6hmhpt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/6hmhpt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMINO-OH-HYDROXYMETHYL-DIHYDROPTERIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1003" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11511" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17083" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44841" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01300" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM937" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00954" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29670" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C7H8N5O8P2" hasOnlySubstanceUnits="false" id="M_6hmhptpp_c" metaid="M_6hmhptpp_c"
      name="6-hydroxymethyl-dihydropterin pyrophosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_6hmhptpp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/6hmhptpp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROPTERIN-CH2OH-PP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1011" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11512" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11517" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15998" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19465" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57602" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:72950" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:73083" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04807" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1160" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02920" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H10O10P" hasOnlySubstanceUnits="false" id="M_6pgc_c" metaid="M_6pgc_c"
      name="6-Phospho-D-gluconate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_6pgc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/6pgc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-2961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12232" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16863" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2231" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40282" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48928" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58759" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01316" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62800" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00345" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM325" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00284" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H9O9P" hasOnlySubstanceUnits="false" id="M_6pgl_c" metaid="M_6pgl_c"
      name="6-phospho-D-glucono-1,5-lactone">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_6pgl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/6pgl" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-6-P-GLUCONO-DELTA-LACTONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12233" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12958" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16938" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20989" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4160" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57955" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62628" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01236" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM429" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00911" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H17NO3" hasOnlySubstanceUnits="false" id="M_8aonn_c" metaid="M_8aonn_c"
      name="8-Amino-7-oxononanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_8aonn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/8aonn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:8-AMINO-7-OXONONANOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12266" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15830" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20808" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57532" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB37790" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01092" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060168" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100058" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1141" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00800" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C11H21N2O7PRS" hasOnlySubstanceUnits="false" id="M_ACP_c" metaid="M_ACP_c"
      name="Acyl carrier protein">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ACP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14405" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18359" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2458" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00229" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM925" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11493" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="2" fbc:chemicalFormula="C9H23N3O" hasOnlySubstanceUnits="false" id="M_N1aspmd_c" metaid="M_N1aspmd_c"
      name="N1-Acetylspermidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_N1aspmd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/N1aspmd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-568" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12625" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17927" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21798" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58324" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7356" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01276" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00612" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM501" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00470" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C2H6NO" hasOnlySubstanceUnits="false" id="M_aacald_c" metaid="M_aacald_c"
      name="Aminoacetaldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_aacald_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/aacald" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-1772" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13766" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17628" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22486" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2647" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42833" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58213" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06735" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2212" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd04122" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C25H36N7O18P3S" hasOnlySubstanceUnits="false" id="M_aacoa_c" metaid="M_aacoa_c"
      name="Acetoacetyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_aacoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/aacoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETOACETYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11756" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13706" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15345" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22173" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2392" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41333" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57286" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01484" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11665" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00332" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050030" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM133" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00279" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C13H23N2O8PRS" hasOnlySubstanceUnits="false" id="M_acACP_c" metaid="M_acACP_c"
      name="Acetyl-ACP">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acACP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYL-ACP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13713" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17093" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2409" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03939" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1269" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11494" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C2H3O2" hasOnlySubstanceUnits="false" id="M_ac_c" metaid="M_ac_c" name="Acetate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ac_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ac" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACET" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13704" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15366" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22165" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22169" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2387" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30089" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40480" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40486" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00042" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00033" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00010" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010002" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM26" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00029" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C2H3O2" hasOnlySubstanceUnits="false" id="M_ac_e" metaid="M_ac_e" name="Acetate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ac_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ac" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACET" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13704" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15366" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22165" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22169" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2387" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30089" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40480" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40486" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00042" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00033" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00010" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010002" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM26" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00029" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H5O3" hasOnlySubstanceUnits="false" id="M_acac_c" metaid="M_acac_c"
      name="Acetoacetate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acac_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acac" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-KETOBUTYRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13705" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15344" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22172" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2391" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:38871" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40507" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00060" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00164" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01030978" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060003" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM154" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00142" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H5O3" hasOnlySubstanceUnits="false" id="M_acac_e" metaid="M_acac_e"
      name="Acetoacetate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acac_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acac" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-KETOBUTYRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13705" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15344" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22172" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2391" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:38871" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40507" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00060" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00164" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01030978" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060003" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM154" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00142" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H4O" hasOnlySubstanceUnits="false" id="M_acald_c" metaid="M_acald_c"
      name="Acetaldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acald_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acald" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETALD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13703" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15343" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22158" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2383" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40533" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00990" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00084" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM75" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00071" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H4O" hasOnlySubstanceUnits="false" id="M_acald_e" metaid="M_acald_e"
      name="Acetaldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acald_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acald" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETALD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13703" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15343" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22158" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2383" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40533" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00990" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00084" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM75" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00071" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C23H34N7O17P3S" hasOnlySubstanceUnits="false" id="M_accoa_c" metaid="M_accoa_c"
      name="Acetyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_accoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/accoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13712" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15351" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2408" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40470" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57288" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01206" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00024" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050029" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050281" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM21" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00022" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C7H9NO8P" hasOnlySubstanceUnits="false" id="M_acg5p_c" metaid="M_acg5p_c"
      name="N-Acetyl-L-glutamyl 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acg5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acg5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-ACETYL-GLUTAMYL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12441" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12576" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16878" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57936" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7151" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06456" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04133" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1384" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02552" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H10NO4" hasOnlySubstanceUnits="false" id="M_acg5sa_c" metaid="M_acg5sa_c"
      name="N-Acetyl-L-glutamate 5-semialdehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acg5sa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acg5sa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-469" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11493" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12461" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12577" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21551" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29123" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7152" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06488" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01250" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1062" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00918" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C8H14NO9P" hasOnlySubstanceUnits="false" id="M_acgam1p_c" metaid="M_acgam1p_c"
      name="N-Acetyl-D-glucosamine 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acgam1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acgam1p" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7125" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04256" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91871" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02611" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C8H14NO9P" hasOnlySubstanceUnits="false" id="M_acgam6p_c" metaid="M_acgam6p_c"
      name="N-Acetyl-D-glucosamine 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acgam6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acgam6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-16168" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-ACETYL-D-GLUCOSAMINE-6-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12564" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15784" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21521" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39577" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49353" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49359" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57513" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7127" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01062" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00357" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM63556" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00293" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H15NO6" hasOnlySubstanceUnits="false" id="M_acgam_e" metaid="M_acgam_e"
      name="N-Acetyl-D-glucosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acgam_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acgam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-acetyl-D-glucosamine" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12455" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12563" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17411" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21517" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:506227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58134" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7123" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00215" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00533" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB14288" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62641" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00140" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM143" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00122" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27608" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H9NO5" hasOnlySubstanceUnits="false" id="M_acglu_c" metaid="M_acglu_c"
      name="N-Acetyl-L-glutamate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acglu_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acglu" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYL-GLU" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12575" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17533" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21549" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21552" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44335" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44337" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64040" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7150" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:87274" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00624" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM730" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00477" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H15NO6" hasOnlySubstanceUnits="false" id="M_acmana_c" metaid="M_acmana_c"
      name="N-Acetyl-D-mannosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acmana_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acmana" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-acetyl-D-mannosamine" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12459" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12573" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17122" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21538" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58019" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:63153" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7141" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62724" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00645" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2403" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00492" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H15NO6" hasOnlySubstanceUnits="false" id="M_acmana_e" metaid="M_acmana_e"
      name="N-Acetyl-D-mannosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acmana_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acmana" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-acetyl-D-mannosamine" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12459" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12573" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17122" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21538" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58019" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:63153" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7141" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62724" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00645" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2403" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00492" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C8H14NO9P" hasOnlySubstanceUnits="false" id="M_acmanap_c" metaid="M_acmanap_c"
      name="N-Acetyl-D-mannosamine 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acmanap_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acmanap" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-ACETYL-D-MANNOSAMINE-6P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21539" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28273" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58557" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62168" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7142" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01121" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04257" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM4708" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02612" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C11H18NO9" hasOnlySubstanceUnits="false" id="M_acnam_c" metaid="M_acnam_c"
      name="N-Acetylneuraminate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acnam_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acnam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-ACETYLNEURAMINATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12471" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12579" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17012" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21617" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21620" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35418" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7214" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00230" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00800" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00270" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM227" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00232" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27569" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C11H18NO9" hasOnlySubstanceUnits="false" id="M_acnam_e" metaid="M_acnam_e"
      name="N-Acetylneuraminate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acnam_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acnam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-ACETYLNEURAMINATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12471" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12579" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17012" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21617" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21620" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35418" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7214" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00230" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00800" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00270" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM227" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00232" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27569" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H3O6" hasOnlySubstanceUnits="false" id="M_acon_T_c" metaid="M_acon_T_c"
      name="Trans-Aconitate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acon_T_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acon_T" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-225" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10719" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12869" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12878" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15708" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22210" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22211" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27069" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27070" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46108" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00958" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02341" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1388" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01563" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H6O6" hasOnlySubstanceUnits="false" id="M_aconm_c" metaid="M_aconm_c"
      name="E-3-carboxy-2-pentenedioate 6-methyl ester">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_aconm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/aconm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MONOMETHYL-ESTER-OF-TRANS-ACONITATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10951" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15663" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:277" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57470" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11514" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07010698" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM3742" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08353" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H14N2O3" hasOnlySubstanceUnits="false" id="M_acorn_c" metaid="M_acorn_c"
      name="N2-Acetyl-L-ornithine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acorn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acorn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-ALPHA-ACETYLORNITHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12634" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16543" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21814" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40771" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57805" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7368" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03357" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06489" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00437" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM817" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00342" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H9NO4" hasOnlySubstanceUnits="false" id="M_acser_c" metaid="M_acser_c"
      name="O-Acetyl-L-serine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_acser_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/acser" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYLSERINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12685" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12710" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12724" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17981" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21938" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44568" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7668" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03011" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00979" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM418" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00722" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C15H25N2O9PRS" hasOnlySubstanceUnits="false" id="M_actACP_c" metaid="M_actACP_c"
      name="Acetoacetyl-ACP">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_actACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/actACP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Acetoacetyl-ACP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Acetoacetyl-ACPs" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2393" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05744" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1223" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11488" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C2H3O5P" hasOnlySubstanceUnits="false" id="M_actp_c" metaid="M_actp_c"
      name="Acetyl phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_actp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/actp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13711" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15350" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22191" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46262" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01494" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00227" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM280" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00196" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H5N5" hasOnlySubstanceUnits="false" id="M_ade_c" metaid="M_ade_c" name="Adenine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ade_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ade" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENINE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENINE-RING" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16708" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22236" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2470" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40579" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00034" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00147" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00034" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM168" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00128" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22238" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H5N5" hasOnlySubstanceUnits="false" id="M_ade_e" metaid="M_ade_e" name="Adenine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ade_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ade" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENINE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENINE-RING" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16708" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22236" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2470" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40579" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00034" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00147" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00034" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM168" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00128" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22238" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H13N5O4" hasOnlySubstanceUnits="false" id="M_adn_c" metaid="M_adn_c" name="Adenosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13734" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16335" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2472" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40558" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40825" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40906" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40911" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00050" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04401" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04402" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04421" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00212" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00045" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM212" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00182" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H13N5O4" hasOnlySubstanceUnits="false" id="M_adn_e" metaid="M_adn_e" name="Adenosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13734" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16335" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2472" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40558" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40825" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40906" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40911" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00050" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04401" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04402" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04421" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00212" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00045" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM212" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00182" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C58H84CoN16O11" hasOnlySubstanceUnits="false" id="M_adocbi_c" metaid="M_adocbi_c"
      name="Adenosyl cobinamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adocbi_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adocbi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSYLCOBINAMIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06903" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06964" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06508" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90790" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03918" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C58H83CoN16O14P" hasOnlySubstanceUnits="false" id="M_adocbip_c" metaid="M_adocbip_c"
      name="Adenosyl cobinamide phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adocbip_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adocbip" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSYLCOBINAMIDE-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2481" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58502" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12184" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06509" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM856" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03919" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C72H100CoN18O17P" hasOnlySubstanceUnits="false" id="M_adocbl_c" metaid="M_adocbl_c"
      name="Adenosylcobalamin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adocbl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adocbl" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSYLCOBALAMIN" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02086" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00194" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00042" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90703" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00166" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C10H12N5O10P2" hasOnlySubstanceUnits="false" id="M_adp_c" metaid="M_adp_c"
      name="ADP C10H12N5O10P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-1651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13222" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16761" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2342" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40553" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:456216" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:87518" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01341" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00008" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G11113" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM7" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00008" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H23N5O15P2" hasOnlySubstanceUnits="false" id="M_adpglc_c" metaid="M_adpglc_c"
      name="ADPglucose C16H23N5O15P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adpglc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adpglc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADP-D-GLUCOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13230" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15751" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2349" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40615" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57498" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06557" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12186" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62621" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00498" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G11109" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM561" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00387" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C17H25N5O16P2" hasOnlySubstanceUnits="false" id="M_adphep_DD_c" metaid="M_adphep_DD_c"
      name="ADP-D-glycero-D-manno-heptose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adphep_DD_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adphep_DD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADP-D-GLYCERO-D-MANNO-HEPTOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13223" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16693" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20842" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2343" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57863" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59966" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59967" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06397" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1648" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03830" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29671" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C17H25N5O16P2" hasOnlySubstanceUnits="false" id="M_adphep_LD_c" metaid="M_adphep_LD_c"
      name="ADP-L-glycero-D-manno-heptose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_adphep_LD_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/adphep_LD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADP-L-GLYCERO-D-MANNO-HEPTOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13226" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15915" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20844" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2347" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57564" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61506" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61530" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06398" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM857" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03831" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C68H95CoN21O21P2" hasOnlySubstanceUnits="false" id="M_agdpcbi_c" metaid="M_agdpcbi_c"
      name="Adenosine-GDP-cobinamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_agdpcbi_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/agdpcbi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSYLCOBINAMIDE-GDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2479" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60487" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12185" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06510" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1105" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03920" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="2" fbc:chemicalFormula="C5H16N4" hasOnlySubstanceUnits="false" id="M_agm_c" metaid="M_agm_c" name="Agmatine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_agm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/agm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AGMATHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13747" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18576" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2514" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40556" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58145" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01432" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60259" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00179" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM328" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00152" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C1243H2524N50O350P50" hasOnlySubstanceUnits="false" id="M_agpc_EC_c" metaid="M_agpc_EC_c"
      name="Acyl-glycerophosphocholine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_agpc_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/agpc_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91473" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15646" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C1093H2224O350N50P50" hasOnlySubstanceUnits="false" id="M_agpe_EC_c" metaid="M_agpe_EC_c"
      name="Acyl-glycerophosphoethanolamine  (E.coli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_agpe_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/agpe_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM4432" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-50" fbc:chemicalFormula="C1143H2224O450P50" hasOnlySubstanceUnits="false" id="M_agpg_EC_c" metaid="M_agpg_EC_c"
      name="Acyl-glycerophosphoglycerol  (E.coli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_agpg_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/agpg_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM4433" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C14H20N6O5S" hasOnlySubstanceUnits="false" id="M_ahcys_c" metaid="M_ahcys_c"
      name="S-Adenosyl-L-homocysteine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ahcys_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ahcys" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSYL-HOMO-CYS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12741" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12759" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12761" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16680" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22034" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45495" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57856" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:67009" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8945" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00939" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00021" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM19" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00019" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C9H12N5O13P3" hasOnlySubstanceUnits="false" id="M_ahdt_c" metaid="M_ahdt_c"
      name="2-Amino-4-hydroxy-6-(erythro-1,2,3-trihydroxypropyl)dihydropteridine triphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ahdt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ahdt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDRONEOPTERIN-P3" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11509" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12201" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19455" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20684" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28069" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58462" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00980" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01144" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12181" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04895" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM397" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02978" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H13N4O8P" hasOnlySubstanceUnits="false" id="M_aicar_c" metaid="M_aicar_c"
      name="5-Amino-1-(5-Phospho-D-ribosyl)imidazole-4-carboxamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_aicar_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/aicar" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AICAR" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12102" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18406" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18966" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40739" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:573" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58475" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01517" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04677" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM365" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02851" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C8H12N3O7P" hasOnlySubstanceUnits="false" id="M_air_c" metaid="M_air_c"
      name="5-amino-1-(5-phospho-D-ribosyl)imidazole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_air_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/air" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12101" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18969" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2655" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28843" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58592" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62575" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162266" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H4O5" hasOnlySubstanceUnits="false" id="M_akg_c" metaid="M_akg_c"
      name="2-Oxoglutarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_akg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/akg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-KETOGLUTARATE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-16852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11638" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1253" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16810" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19749" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30915" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30916" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40661" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62781" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00026" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM20" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00024" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H4O5" hasOnlySubstanceUnits="false" id="M_akg_e" metaid="M_akg_e"
      name="2-Oxoglutarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_akg_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/akg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-KETOGLUTARATE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-16852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11638" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1253" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16810" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19749" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30915" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30916" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40661" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62781" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00026" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM20" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00024" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO2" hasOnlySubstanceUnits="false" id="M_ala_B_c" metaid="M_ala_B_c"
      name="Beta-Alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ala_B_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ala_B" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:B-ALANINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10343" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12389" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16958" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22821" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57966" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:63070" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00056" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00099" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07561" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM144" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00085" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO2" hasOnlySubstanceUnits="false" id="M_ala__D_c" metaid="M_ala__D_c"
      name="D-Alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ala__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ala__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-ALANINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10840" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12899" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15570" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20893" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32435" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32436" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41756" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41798" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41848" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41877" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57416" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01310" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00133" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM156" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00117" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO2" hasOnlySubstanceUnits="false" id="M_ala__D_e" metaid="M_ala__D_e"
      name="D-Alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ala__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ala__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-ALANINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10840" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12899" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15570" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20893" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32435" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32436" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41756" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41798" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41848" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41877" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57416" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01310" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00133" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM156" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00117" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO2" hasOnlySubstanceUnits="false" id="M_ala__L_c" metaid="M_ala__L_c"
      name="L-Alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ala__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ala__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ALPHA-ALANINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13069" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132498" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16449" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16977" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21216" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22277" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2539" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32432" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32439" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32440" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40734" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40735" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6171" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:66916" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76050" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00161" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62251" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00041" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01401" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00012" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM32" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00035" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01003" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO2" hasOnlySubstanceUnits="false" id="M_ala__L_e" metaid="M_ala__L_e"
      name="L-Alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ala__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ala__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ALPHA-ALANINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13069" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132498" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16449" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16977" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21216" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22277" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2539" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32432" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32439" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32440" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40734" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40735" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6171" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:66916" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76050" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00161" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62251" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00041" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01401" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00012" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM32" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00035" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01003" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12N2O3" hasOnlySubstanceUnits="false" id="M_alaala_c" metaid="M_alaala_c"
      name="D-Alanyl-D-alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_alaala_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/alaala" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-ALA-D-ALA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12900" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13749" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16576" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20894" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4088" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57822" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03459" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB28680" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00993" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM669" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00731" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H7O4" hasOnlySubstanceUnits="false" id="M_alac__S_c" metaid="M_alac__S_c"
      name="(S)-2-Acetolactate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_alac__S_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/alac__S" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-ACETO-LACTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11033" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18409" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18731" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:374" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58476" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06855" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06010" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050460" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114079" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19047" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H6N4O3" hasOnlySubstanceUnits="false" id="M_alltn_c" metaid="M_alltn_c"
      name="Allantoin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_alltn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/alltn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALLANTOIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13761" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15676" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2594" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:74345" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00462" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01551" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00121" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM612" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01092" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H6N4O3" hasOnlySubstanceUnits="false" id="M_alltn_e" metaid="M_alltn_e"
      name="Allantoin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_alltn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/alltn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALLANTOIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13761" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15676" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2594" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:74345" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00462" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01551" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00121" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM612" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01092" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H7N4O4" hasOnlySubstanceUnits="false" id="M_alltt_c" metaid="M_alltt_c"
      name="Allantoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_alltt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/alltt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALLANTOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13760" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17536" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22352" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22353" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2593" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30837" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01209" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00499" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM584" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00388" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O7" hasOnlySubstanceUnits="false" id="M_altrn_c" metaid="M_altrn_c"
      name="D-Altronate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_altrn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/altrn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-ALTRONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12908" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17360" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20904" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46644" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58707" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00817" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2011" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00608" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C15H23N6O5S" hasOnlySubstanceUnits="false" id="M_amet_c" metaid="M_amet_c"
      name="S-Adenosyl-L-methionine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_amet_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/amet" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:S-ADENOSYLMETHIONINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10786" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10833" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12742" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12757" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12760" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15414" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22036" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33440" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33442" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45607" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:527887" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59789" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:67040" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8946" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01185" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62709" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00019" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM16" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00017" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="2" fbc:chemicalFormula="C14H24N6O3S" hasOnlySubstanceUnits="false" id="M_ametam_c" metaid="M_ametam_c"
      name="S-Adenosylmethioninamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ametam_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ametam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:S-ADENOSYLMETHIONINAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10906" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15625" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22035" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57443" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8947" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00988" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01137" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM321" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00837" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C15H19N5O6S" hasOnlySubstanceUnits="false" id="M_amob_c" metaid="M_amob_c"
      name="S-Adenosyl-4-methylthio-2-oxobutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_amob_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/amob" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:S-ADENOSYL-4-METHYLTHIO-2-OXOBUTANOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12758" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16490" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22033" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8944" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04425" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM3090" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02701" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H12N5O7P" hasOnlySubstanceUnits="false" id="M_amp_c" metaid="M_amp_c"
      name="AMP C10H12N5O7P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_amp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/amp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMP-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12056" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13736" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13740" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16027" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22245" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2356" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40510" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40721" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40726" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40786" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40826" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:456215" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47222" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00020" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02769" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM14" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00018" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22272" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H12N5O7P" hasOnlySubstanceUnits="false" id="M_amp_e" metaid="M_amp_e"
      name="AMP C10H12N5O7P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_amp_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/amp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMP-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12056" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13736" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13740" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16027" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22245" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2356" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40510" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40721" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40726" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40786" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40826" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:456215" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47222" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00020" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02769" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM14" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00018" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22272" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H6NO2" hasOnlySubstanceUnits="false" id="M_anth_c" metaid="M_anth_c"
      name="Anthranilate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_anth_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/anth" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ANTHRANILATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16567" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22575" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22577" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22578" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2757" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30754" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40980" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01123" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00108" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM188" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00093" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C20H24N10O19P4" hasOnlySubstanceUnits="false" id="M_ap4a_c" metaid="M_ap4a_c"
      name="P1,P4-Bis(5'-adenosyl) tetraphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ap4a_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ap4a" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSYL-P4" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12726" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12729" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21998" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58141" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7875" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06502" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01260" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1089" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00924" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C20H24N10O22P5" hasOnlySubstanceUnits="false" id="M_ap5a_c" metaid="M_ap5a_c"
      name="P1,P5-Bis(5'-adenosyl) pentaphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ap5a_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ap5a" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-1137" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28898" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3120" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62041" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04058" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2412" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02511" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-50" fbc:chemicalFormula="C2829H5272O550P50" hasOnlySubstanceUnits="false" id="M_apg_EC_c" metaid="M_apg_EC_c"
      name="Acyl phosphatidylglycerol (E.coli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_apg_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/apg_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM5599" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="RHO" hasOnlySubstanceUnits="false" id="M_apoACP_c" metaid="M_apoACP_c"
      name="Apoprotein [acyl carrier protein]">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_apoACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/apoACP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2214" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12370" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29672" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H12N5O10PS" hasOnlySubstanceUnits="false" id="M_aps_c" metaid="M_aps_c"
      name="Adenosine 5'-phosphosulfate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_aps_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/aps" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:APS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12059" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13741" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17709" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2486" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40562" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58243" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00224" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM287" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00193" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O8P" hasOnlySubstanceUnits="false" id="M_ara5p_c" metaid="M_ara5p_c"
      name="D-Arabinose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ara5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ara5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ARABINOSE-5P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12916" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16241" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20915" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4104" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57693" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:79058" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11734" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01112" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1184" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00817" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O5" hasOnlySubstanceUnits="false" id="M_arab__L_c" metaid="M_arab__L_c"
      name="L-Arabinose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_arab__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/arab__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15699" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ARABINOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6182" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11476" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91126" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19109" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27351" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O5" hasOnlySubstanceUnits="false" id="M_arab__L_e" metaid="M_arab__L_e"
      name="L-Arabinose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_arab__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/arab__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15699" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ARABINOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6182" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11476" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91126" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19109" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27351" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C12H15O10P" hasOnlySubstanceUnits="false" id="M_arbt6p_c" metaid="M_arbt6p_c"
      name="Arbutin 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_arbt6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/arbt6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-1162" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2807" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60929" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06187" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2806" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03697" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C6H15N4O2" hasOnlySubstanceUnits="false" id="M_arg__L_c" metaid="M_arg__L_c"
      name="L-Arginine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_arg__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/arg__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ARG" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13077" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:133495" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16467" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22616" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2643" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29016" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32681" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32682" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32683" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32695" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32696" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32697" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42927" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6185" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00517" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62762" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00062" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02385" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02982" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM70" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00051" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19021" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="C6H15N4O2" hasOnlySubstanceUnits="false" id="M_arg__L_e" metaid="M_arg__L_e"
      name="L-Arginine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_arg__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/arg__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ARG" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13077" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:133495" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16467" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22616" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2643" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29016" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32681" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32682" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32683" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32695" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32696" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32697" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42927" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6185" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00517" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62762" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00062" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02385" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02982" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM70" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00051" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19021" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H17N4O6" hasOnlySubstanceUnits="false" id="M_argsuc_c" metaid="M_argsuc_c"
      name="N(omega)-(L-Arginino)succinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_argsuc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/argsuc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ARGININO-SUCCINATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10960" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15682" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21475" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57472" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7098" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03406" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM550" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02152" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29673" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H8N2O3" hasOnlySubstanceUnits="false" id="M_asn__L_c" metaid="M_asn__L_c"
      name="L-Asparagine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_asn__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/asn__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13083" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17196" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22653" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32650" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32660" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32661" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40902" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58048" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6191" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00168" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00152" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16438" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM147" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00132" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15142" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H8N2O3" hasOnlySubstanceUnits="false" id="M_asn__L_e" metaid="M_asn__L_e"
      name="L-Asparagine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_asn__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/asn__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13083" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17196" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22653" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32650" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32660" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32661" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40902" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58048" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6191" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00168" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00152" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16438" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM147" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00132" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15142" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H6NO4" hasOnlySubstanceUnits="false" id="M_asp__L_c" metaid="M_asp__L_c"
      name="L-Aspartate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_asp__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/asp__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ASPARTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13085" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22659" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22660" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29991" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29992" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29993" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29995" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35391" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40853" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40900" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40913" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40942" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6193" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00191" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62186" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62501" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00049" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16433" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00013" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM42" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00041" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19181" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H6NO4" hasOnlySubstanceUnits="false" id="M_asp__L_e" metaid="M_asp__L_e"
      name="L-Aspartate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_asp__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/asp__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ASPARTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13085" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22659" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22660" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29991" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29992" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29993" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29995" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35391" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40853" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40900" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40913" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40942" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6193" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00191" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62186" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62501" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00049" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16433" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00013" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM42" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00041" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19181" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H7NO3" hasOnlySubstanceUnits="false" id="M_aspsa_c" metaid="M_aspsa_c"
      name="L-Aspartate 4-semialdehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_aspsa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/aspsa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ASPARTATE-SEMIALDEHYDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13086" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18051" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21245" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40847" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:537519" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6194" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12249" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00441" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM361" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00346" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H12N5O13P3" hasOnlySubstanceUnits="false" id="M_atp_c" metaid="M_atp_c"
      name="ATP C10H12N5O13P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_atp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/atp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ATP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-1634" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10789" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13236" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22249" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2359" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:237958" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30616" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40938" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57299" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00002" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D08646" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM3" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00002" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C28H46N8O17P3S" hasOnlySubstanceUnits="false" id="M_bbtcoa_c" metaid="M_bbtcoa_c"
      name="Gamma-butyrobetainyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_bbtcoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/bbtcoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GAMMA-BUTYROBETAINYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61513" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61517" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C20749" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050321" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM5762" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15412" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C5H12NO" hasOnlySubstanceUnits="false" id="M_betald_c" metaid="M_betald_c"
      name="Betaine aldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_betald_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/betald" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BETAINE_ALDEHYDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13896" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15710" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22859" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3074" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41256" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01252" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00576" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM457" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00447" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C25H38N7O17P3S" hasOnlySubstanceUnits="false" id="M_btcoa_c" metaid="M_btcoa_c"
      name="Butanoyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_btcoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/btcoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BUTYRYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13926" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15517" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22953" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22973" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57371" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01088" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00136" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050292" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM233" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00120" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H15N2O3S" hasOnlySubstanceUnits="false" id="M_btn_c" metaid="M_btn_c" name="Biotin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_btn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/btn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BIOTIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13905" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15956" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22882" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22884" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3108" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41236" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57586" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00030" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00120" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00029" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM304" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00104" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H15N2O4S" hasOnlySubstanceUnits="false" id="M_btnso_c" metaid="M_btnso_c"
      name="D-biotin d-sulfoxide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_btnso_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/btnso" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-722" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62046" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62193" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C20386" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM3932" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15414" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H7O2" hasOnlySubstanceUnits="false" id="M_but_c" metaid="M_but_c"
      name="Butyrate (n-C4:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_but_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/but" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162281" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H7O2" hasOnlySubstanceUnits="false" id="M_but_e" metaid="M_but_e"
      name="Butyrate (n-C4:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_but_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/but" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162281" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H11N5O6P" hasOnlySubstanceUnits="false" id="M_camp_c" metaid="M_camp_c"
      name="CAMP C10H11N5O6P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_camp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/camp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CAMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11673" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1325" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17489" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19827" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41588" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58165" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00058" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00575" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM243" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00446" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H6N2O5" hasOnlySubstanceUnits="false" id="M_cbasp_c" metaid="M_cbasp_c"
      name="N-Carbamoyl-L-aspartate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cbasp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cbasp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBAMYUL-L-ASPARTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12496" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12593" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15859" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21687" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21688" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32814" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44316" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7257" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00438" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM465" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00343" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C48H72CoN11O8" hasOnlySubstanceUnits="false" id="M_cbi_c" metaid="M_cbi_c"
      name="Cobinamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cbi_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cbi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:COBINAMIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48529" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1549" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C62H88CoN13O14P" hasOnlySubstanceUnits="false" id="M_cbl1_c" metaid="M_cbl1_c"
      name="Cob(I)alamin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cbl1_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cbl1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:COB-I-ALAMIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14004" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15982" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23328" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3784" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60488" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03429" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06470" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00853" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91519" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C62H88CoN13O14P" hasOnlySubstanceUnits="false" id="M_cbl1_e" metaid="M_cbl1_e"
      name="Cob(I)alamin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cbl1_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cbl1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:COB-I-ALAMIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14004" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15982" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23328" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3784" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60488" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03429" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06470" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00853" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91519" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="CH2NO5P" hasOnlySubstanceUnits="false" id="M_cbp_c" metaid="M_cbp_c"
      name="Carbamoyl phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cbp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cbp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBAMOYL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13942" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17672" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23005" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3389" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41567" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58228" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01096" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00169" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM138" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00146" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C9H12N3O11P2" hasOnlySubstanceUnits="false" id="M_cdp_c" metaid="M_cdp_c"
      name="CDP C9H12N3O11P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13254" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17239" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23519" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3260" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58069" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00112" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM220" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00096" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-100" fbc:chemicalFormula="C2286H3998N150O750P100" hasOnlySubstanceUnits="false" id="M_cdpdag_EC_c"
      metaid="M_cdpdag_EC_c" name="CDPdiacylglycerol (E coli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cdpdag_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cdpdag_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM45899" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd16066" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C11H19N4O11P2" hasOnlySubstanceUnits="false" id="M_cdpea_c" metaid="M_cdpea_c"
      name="CDPethanolamine C11H19N4O11P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cdpea_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cdpea" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CDP-ETHANOLAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13257" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13270" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16732" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20869" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3270" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57876" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01564" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00570" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM449" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00444" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H11O4" hasOnlySubstanceUnits="false" id="M_cechddd_c" metaid="M_cechddd_c"
      name="Cis-3-(3-carboxyethyl)-3,5-cyclohexadiene-1,2-diol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cechddd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cechddd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBOXYETHYL-3-5-CYCLOHEXADIENE-1-2-DIOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48690" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48691" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60088" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1636" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H9O4" hasOnlySubstanceUnits="false" id="M_cenchddd_c" metaid="M_cenchddd_c"
      name="Cis-3-(3-carboxyethenyl)-3,5-cyclohexadiene-1,2-diol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cenchddd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cenchddd" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49071" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60108" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61462" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C12622" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2161" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd09253" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C5H14NO" hasOnlySubstanceUnits="false" id="M_chol_c" metaid="M_chol_c"
      name="Choline C5H14NO">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_chol_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/chol" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CHOLINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13985" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23212" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3665" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41524" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:72322" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00097" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00114" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07690" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D10498" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00098" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="C5H14NO" hasOnlySubstanceUnits="false" id="M_chol_e" metaid="M_chol_e"
      name="Choline C5H14NO">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_chol_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/chol" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CHOLINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13985" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23212" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3665" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41524" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:72322" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00097" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00114" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07690" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D10498" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00098" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H8O6" hasOnlySubstanceUnits="false" id="M_chor_c" metaid="M_chor_c"
      name="Chorismate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_chor_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/chor" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CHORISMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13993" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17333" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23225" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3677" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12199" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00251" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM337" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00216" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H7O2" hasOnlySubstanceUnits="false" id="M_cinnm_c" metaid="M_cinnm_c"
      name="Trans-Cinnamate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cinnm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cinnm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-674" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10720" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10955" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12871" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12879" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15669" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23248" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23250" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27072" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27073" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27386" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35697" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3710" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45845" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00567" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00930" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00423" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C10438" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM438" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00333" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd07326" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H5O7" hasOnlySubstanceUnits="false" id="M_cit_c" metaid="M_cit_c" name="Citrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cit_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cit" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CIT" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132362" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:133748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16947" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23322" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30769" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35802" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35804" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35808" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35809" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35810" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3727" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41523" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42563" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76049" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:79399" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00094" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00158" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C13660" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00037" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM131" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00137" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H5O7" hasOnlySubstanceUnits="false" id="M_cit_e" metaid="M_cit_e" name="Citrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cit_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cit" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CIT" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132362" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:133748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16947" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23322" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30769" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35802" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35804" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35808" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35809" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35810" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3727" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41523" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42563" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76049" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:79399" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00094" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00158" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C13660" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00037" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM131" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00137" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H13N3O3" hasOnlySubstanceUnits="false" id="M_citr__L_c" metaid="M_citr__L_c"
      name="L-Citrulline">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_citr__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/citr__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-CITRULLINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13092" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16349" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18211" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21257" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3730" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41489" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6203" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:66922" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00904" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00327" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07706" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM211" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00274" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C17H24N3O15P" hasOnlySubstanceUnits="false" id="M_ckdo_c" metaid="M_ckdo_c"
      name="CMP-3-deoxy-D-manno-octulosonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ckdo_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ckdo" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CMP-KDO" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13275" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18014" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20874" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3277" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59825" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:85987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:86284" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04121" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM535" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02546" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-100" fbc:chemicalFormula="C3822H7096O850P100" hasOnlySubstanceUnits="false" id="M_clpn_EC_c"
      metaid="M_clpn_EC_c" name="Cardiolipin (Ecoli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_clpn_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/clpn_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91138" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15650" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H12N3O8P" hasOnlySubstanceUnits="false" id="M_cmp_c" metaid="M_cmp_c"
      name="CMP C9H12N3O8P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13274" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17361" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23520" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3275" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41312" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41666" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41691" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47362" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48799" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58120" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60377" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00055" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10621" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM31" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00046" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="CO2" hasOnlySubstanceUnits="false" id="M_co2_c" metaid="M_co2_c" name="CO2 CO2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_co2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/co2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBON-DIOXIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13282" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13283" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16526" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23011" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3283" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48829" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01967" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00011" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00004" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM13" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00011" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="CO2" hasOnlySubstanceUnits="false" id="M_co2_e" metaid="M_co2_e" name="CO2 CO2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_co2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/co2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBON-DIOXIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13282" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13283" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16526" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23011" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3283" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48829" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01967" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00011" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00004" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM13" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00011" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C21H32N7O16P3S" hasOnlySubstanceUnits="false" id="M_coa_c" metaid="M_coa_c"
      name="Coenzyme A">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_coa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/coa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CO-A" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:COA-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13294" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13295" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13298" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15346" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23355" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3771" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41597" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41631" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57287" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:741566" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01423" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00010" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM12" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00010" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22528" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C36H40N4O8" hasOnlySubstanceUnits="false" id="M_cpppg3_c" metaid="M_cpppg3_c"
      name="Coproporphyrinogen III">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cpppg3_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cpppg3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:COPROPORPHYRINOGEN_III" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14020" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15439" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23386" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3880" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41560" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57309" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01261" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03263" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM410" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02083" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H15NO3" hasOnlySubstanceUnits="false" id="M_crn_c" metaid="M_crn_c" name="L-Carnitine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_crn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/crn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARNITINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11817" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13091" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13947" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16347" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20047" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21256" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23038" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29085" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3424" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39547" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6202" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00062" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01467" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62496" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00318" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00487" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02176" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM173" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00266" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19003" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H15NO3" hasOnlySubstanceUnits="false" id="M_crn_e" metaid="M_crn_e" name="L-Carnitine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_crn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/crn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARNITINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11817" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13091" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13947" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16347" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20047" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21256" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23038" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29085" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3424" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39547" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6202" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00062" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01467" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62496" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00318" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00487" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02176" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM173" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00266" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19003" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C28H46N8O18P3S" hasOnlySubstanceUnits="false" id="M_crncoa_c" metaid="M_crncoa_c"
      name="L-Carnitinyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_crncoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/crncoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-CARNITINYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41482" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60932" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C20750" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050150" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1451" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H5N3O" hasOnlySubstanceUnits="false" id="M_csn_c" metaid="M_csn_c" name="Cytosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_csn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/csn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYTOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16040" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23531" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4072" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41732" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00630" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00380" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM761" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00307" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H5N3O" hasOnlySubstanceUnits="false" id="M_csn_e" metaid="M_csn_e" name="Cytosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_csn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/csn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYTOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16040" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23531" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4072" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41732" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00630" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00380" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM761" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00307" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H13NO2" hasOnlySubstanceUnits="false" id="M_ctbt_c" metaid="M_ctbt_c"
      name="Crotonobetaine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ctbt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ctbt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CROTONO-BETAINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11946" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1774" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20299" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48867" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04114" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1607" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02543" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C28H44N8O17P3S" hasOnlySubstanceUnits="false" id="M_ctbtcoa_c" metaid="M_ctbtcoa_c"
      name="Crotonobetainyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ctbtcoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ctbtcoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CROTONOBETAINYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60933" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61123" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C20748" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050143" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050306" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2363" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15441" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C9H12N3O14P3" hasOnlySubstanceUnits="false" id="M_ctp_c" metaid="M_ctp_c"
      name="CTP C9H12N3O14P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ctp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ctp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CTP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13286" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17677" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23522" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37563" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41675" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58231" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00063" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM63" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00052" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="CHN" hasOnlySubstanceUnits="false" id="M_cyan_c" metaid="M_cyan_c"
      name="Hydrogen cyanide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cyan_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cyan" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-13584" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HCN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13362" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14038" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17514" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36856" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3969" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41780" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5786" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60292" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00177" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01326" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C18673" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM254" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00150" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19012" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="CNO" hasOnlySubstanceUnits="false" id="M_cynt_c" metaid="M_cynt_c" name="Cyanate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cynt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cynt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-69" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14037" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23419" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29195" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29202" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3968" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02078" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01417" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C19067" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1191" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01015" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="CNO" hasOnlySubstanceUnits="false" id="M_cynt_e" metaid="M_cynt_e" name="Cyanate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cynt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cynt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-69" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14037" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23419" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29195" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29202" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3968" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02078" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01417" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C19067" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1191" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01015" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO2S" hasOnlySubstanceUnits="false" id="M_cys__L_c" metaid="M_cys__L_c"
      name="L-Cysteine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cys__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cys__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14061" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15356" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17561" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21261" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23508" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32442" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32443" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32445" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32457" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32458" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41700" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41768" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41781" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6207" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00574" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00097" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00736" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00026" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM55" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00084" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00547" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO2S" hasOnlySubstanceUnits="false" id="M_cys__L_e" metaid="M_cys__L_e"
      name="L-Cysteine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cys__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cys__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14061" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15356" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17561" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21261" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23508" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32442" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32443" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32445" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32457" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32458" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41700" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41768" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41781" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6207" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00574" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00097" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00736" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00026" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM55" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00084" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00547" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H14N2O4S" hasOnlySubstanceUnits="false" id="M_cyst__L_c" metaid="M_cyst__L_c"
      name="L-Cystathionine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cyst__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cyst__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-CYSTATHIONINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13093" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17482" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21259" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58161" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6205" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00099" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03614" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02291" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM319" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19019" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H13N3O5" hasOnlySubstanceUnits="false" id="M_cytd_c" metaid="M_cytd_c" name="Cytidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cytd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cytd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYTIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14063" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17562" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23515" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41649" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41686" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41704" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41709" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00089" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00475" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07769" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM338" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00367" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H13N3O5" hasOnlySubstanceUnits="false" id="M_cytd_e" metaid="M_cytd_e" name="Cytidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_cytd_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/cytd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYTIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14063" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17562" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23515" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41649" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41686" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41704" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41709" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00089" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00475" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07769" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM338" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00367" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H13N5O3" hasOnlySubstanceUnits="false" id="M_dad_2_c" metaid="M_dad_2_c"
      name="Deoxyadenosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dad_2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dad_2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYADENOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14112" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17256" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39863" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40535" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40560" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40565" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4405" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00101" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB05778" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11722" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00559" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM625" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00438" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H13N5O3" hasOnlySubstanceUnits="false" id="M_dad_2_e" metaid="M_dad_2_e"
      name="Deoxyadenosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dad_2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dad_2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYADENOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14112" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17256" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39863" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40535" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40560" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40565" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4405" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00101" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB05778" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11722" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00559" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM625" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00438" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C10H12N5O9P2" hasOnlySubstanceUnits="false" id="M_dadp_c" metaid="M_dadp_c"
      name="DADP C10H12N5O9P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dadp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dadp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DADP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10489" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14067" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16174" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41890" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57667" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01508" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00206" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM374" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00177" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H12N5O6P" hasOnlySubstanceUnits="false" id="M_damp_c" metaid="M_damp_c"
      name="DAMP C10H12N5O6P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_damp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/damp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DAMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10490" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14068" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17713" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19236" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40419" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40499" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40505" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40544" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40607" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40614" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41815" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41864" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41870" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58245" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00360" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM432" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00294" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C9H21N2O2" hasOnlySubstanceUnits="false" id="M_dann_c" metaid="M_dann_c"
      name="7,8-Diaminononanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dann_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dann" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIAMINONONANOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17830" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20765" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58500" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01037" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100062" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1140" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00764" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H12N5O12P3" hasOnlySubstanceUnits="false" id="M_datp_c" metaid="M_datp_c"
      name="DATP C10H12N5O12P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_datp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/datp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DATP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10491" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14069" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19238" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42290" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:495505" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61404" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00131" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM286" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00115" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H7O6P" hasOnlySubstanceUnits="false" id="M_db4p_c" metaid="M_db4p_c"
      name="3,4-dihydroxy-2-butanone 4-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_db4p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/db4p" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:50606" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C15556" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2323" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11225" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C14H14N5O11P" hasOnlySubstanceUnits="false" id="M_dcamp_c" metaid="M_dcamp_c"
      name="N6-(1,2-Dicarboxyethyl)-AMP">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dcamp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dcamp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENYLOSUCC" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12656" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12657" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15919" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21857" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22262" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39869" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57567" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7405" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB59653" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03794" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM565" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02375" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C9H12N3O10P2" hasOnlySubstanceUnits="false" id="M_dcdp_c" metaid="M_dcdp_c"
      name="DCDP C9H12N3O10P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dcdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dcdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DCDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10492" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19241" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49966" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58593" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01245" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00705" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM411" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00533" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H12N3O7P" hasOnlySubstanceUnits="false" id="M_dcmp_c" metaid="M_dcmp_c"
      name="DCMP C9H12N3O7P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dcmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dcmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DCMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10493" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14070" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14071" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14115" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15918" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41838" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41875" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41881" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57566" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01202" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00239" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM266" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00206" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C9H12N3O13P3" hasOnlySubstanceUnits="false" id="M_dctp_c" metaid="M_dctp_c"
      name="DCTP C9H12N3O13P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dctp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dctp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DCTP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10494" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14072" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16311" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19243" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57724" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61481" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00998" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00458" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM360" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00356" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H13N3O4" hasOnlySubstanceUnits="false" id="M_dcyt_c" metaid="M_dcyt_c"
      name="Deoxycytidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dcyt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dcyt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYCYTIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15698" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19240" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41417" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41430" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41434" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4407" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00014" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00881" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM704" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00654" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H13N3O4" hasOnlySubstanceUnits="false" id="M_dcyt_e" metaid="M_dcyt_e"
      name="Deoxycytidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dcyt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dcyt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYCYTIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15698" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19240" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41417" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41430" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41434" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4407" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00014" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00881" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM704" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00654" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C23H43N2O8PRS" hasOnlySubstanceUnits="false" id="M_ddcaACP_c" metaid="M_ddcaACP_c"
      name="Dodecanoyl-ACP (n-C12:0ACP)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ddcaACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ddcaACP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89851" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C10H12N5O10P2" hasOnlySubstanceUnits="false" id="M_dgdp_c" metaid="M_dgdp_c"
      name="DGDP C10H12N5O10P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dgdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dgdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DGDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10495" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19245" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28862" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41949" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58595" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00960" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00361" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM436" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00295" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H12N5O7P" hasOnlySubstanceUnits="false" id="M_dgmp_c" metaid="M_dgmp_c"
      name="DGMP C10H12N5O7P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dgmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dgmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DGMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10496" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14074" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19246" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41902" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41939" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41944" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42676" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42726" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45049" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47449" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57673" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00362" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM546" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00296" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H13N5O4" hasOnlySubstanceUnits="false" id="M_dgsn_c" metaid="M_dgsn_c"
      name="Deoxyguanosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dgsn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dgsn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYGUANOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14116" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17172" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23624" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42667" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42867" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42874" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4412" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00085" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00330" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM647" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00277" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H13N5O4" hasOnlySubstanceUnits="false" id="M_dgsn_e" metaid="M_dgsn_e"
      name="Deoxyguanosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dgsn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dgsn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYGUANOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14116" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17172" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23624" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42667" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42867" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42874" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4412" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00085" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00330" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM647" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00277" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H12N5O13P3" hasOnlySubstanceUnits="false" id="M_dgtp_c" metaid="M_dgtp_c"
      name="DGTP C10H12N5O13P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dgtp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dgtp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DGTP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10497" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14076" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16497" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57794" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61429" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01440" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00286" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM344" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00241" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H6O3" hasOnlySubstanceUnits="false" id="M_dha_c" metaid="M_dha_c"
      name="Dihydroxyacetone">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dha_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dha" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROXYACETONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16016" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39809" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5453" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01882" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00184" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07841" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM460" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00157" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H6O3" hasOnlySubstanceUnits="false" id="M_dha_e" metaid="M_dha_e"
      name="Dihydroxyacetone">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dha_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dha" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROXYACETONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16016" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39809" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5453" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01882" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00184" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07841" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM460" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00157" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C3H5O6P" hasOnlySubstanceUnits="false" id="M_dhap_c" metaid="M_dhap_c"
      name="Dihydroxyacetone phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhap_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhap" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROXY-ACETONE-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14341" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14342" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16108" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24355" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39571" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5454" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57642" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01473" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11735" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00111" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM77" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00095" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H7O4" hasOnlySubstanceUnits="false" id="M_dhcinnm_c" metaid="M_dhcinnm_c"
      name="2,3-dihydroxicinnamic acid">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhcinnm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhcinnm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-10796" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32356" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48681" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58642" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C12623" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1130" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd09254" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29675" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C19H19N7O6" hasOnlySubstanceUnits="false" id="M_dhf_c" metaid="M_dhf_c"
      name="7,8-Dihydrofolate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhf_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhf" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROFOLATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12245" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14150" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15633" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20768" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42000" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4564" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57451" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01056" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00415" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM281" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00330" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C11H7O4" hasOnlySubstanceUnits="false" id="M_dhna_c" metaid="M_dhna_c"
      name="1,4-Dihydroxy-2-naphthoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhna_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhna" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROXYNAPHTHOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11173" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18933" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:539" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03657" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM552" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02295" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H13N5O4" hasOnlySubstanceUnits="false" id="M_dhnpt_c" metaid="M_dhnpt_c"
      name="Dihydroneopterin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhnpt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhnpt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-11770" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDRO-NEO-PTERIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:71175" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02275" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04874" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C21008" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM993" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02961" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd23083" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H5N2O4" hasOnlySubstanceUnits="false" id="M_dhor__S_c" metaid="M_dhor__S_c"
      name="(S)-Dihydroorotate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhor__S_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhor__S" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DI-H-OROTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11063" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17025" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18777" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18778" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30864" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:417" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42132" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03349" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00337" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM252" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00282" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H12N5O7P" hasOnlySubstanceUnits="false" id="M_dhpmp_c" metaid="M_dhpmp_c"
      name="Dihydroneopterin monophosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhpmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhpmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDRONEOPTERIN-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23756" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4577" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48954" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58762" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06824" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05925" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1645" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03521" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29676" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H9O4" hasOnlySubstanceUnits="false" id="M_dhpppn_c" metaid="M_dhpppn_c"
      name="3-(2,3-Dihydroxyphenyl)propanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhpppn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhpppn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-3-DIHYDROXYPHENYL-PROPIONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11713" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11718" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1419" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18136" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19918" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19919" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46951" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04044" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM826" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02501" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C14H13N6O3" hasOnlySubstanceUnits="false" id="M_dhpt_c" metaid="M_dhpt_c"
      name="Dihydropteroate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhpt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhpt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:7-8-DIHYDROPTEROATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14160" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17839" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4581" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01412" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00921" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM965" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00683" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28673" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28675" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H8O4" hasOnlySubstanceUnits="false" id="M_dhptd_c" metaid="M_dhptd_c"
      name="4,5-dihydroxy-2,3-pentanedione">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dhptd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dhptd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROXYPENTANEDIONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29484" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11838" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2449" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08636" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08638" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H12N4O4" hasOnlySubstanceUnits="false" id="M_din_c" metaid="M_din_c"
      name="Deoxyinosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_din_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/din" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYINOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19248" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23629" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28997" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43293" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43436" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43437" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4413" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00071" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05512" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM935" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03279" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H12N4O4" hasOnlySubstanceUnits="false" id="M_din_e" metaid="M_din_e"
      name="Deoxyinosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_din_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/din" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYINOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19248" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23629" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28997" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43293" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43436" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43437" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4413" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00071" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05512" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM935" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03279" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H9O6PS" hasOnlySubstanceUnits="false" id="M_dkmpp_c" metaid="M_dkmpp_c"
      name="2,3-diketo-5-methylthio-1-phosphopentane">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dkmpp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dkmpp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-HYDROXY-3-KETO-5-METHYLTHIO-1-PHOSPHOP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-8999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:50605" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58829" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59505" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB59620" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60266" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62750" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C15650" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C15651" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM409" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11295" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11296" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H10N2" hasOnlySubstanceUnits="false" id="M_dmbzid_c" metaid="M_dmbzid_c"
      name="5,6-Dimethylbenzimidazole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dmbzid_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dmbzid" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIMETHYLBENZIMIDAZOLE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14172" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15890" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20516" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4620" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03701" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03114" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1531" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01997" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C13H18N4O6" hasOnlySubstanceUnits="false" id="M_dmlz_c" metaid="M_dmlz_c"
      name="6,7-Dimethyl-8-(1-D-ribityl)lumazine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dmlz_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dmlz" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIMETHYL-D-RIBITYL-LUMAZINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12185" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17601" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20682" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2154" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58201" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03826" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04332" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1313" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02656" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C5H9O7P2" hasOnlySubstanceUnits="false" id="M_dmpp_c" metaid="M_dmpp_c"
      name="Dimethylallyl diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dmpp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dmpp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-4211" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12280" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14169" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14883" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16057" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18108" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23803" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26245" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:341937" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42074" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4616" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57623" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8394" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01120" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02101" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00235" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR01010001" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM132" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00202" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H6S" hasOnlySubstanceUnits="false" id="M_dms_c" metaid="M_dms_c"
      name="Dimethyl sulfide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dms_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dms" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-7670" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14168" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14175" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17437" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23800" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44169" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4611" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02303" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00580" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM444" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00450" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H6S" hasOnlySubstanceUnits="false" id="M_dms_e" metaid="M_dms_e"
      name="Dimethyl sulfide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dms_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dms" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-7670" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14168" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14175" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17437" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23800" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44169" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4611" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02303" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00580" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM444" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00450" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H6OS" hasOnlySubstanceUnits="false" id="M_dmso_c" metaid="M_dmso_c"
      name="Dimethyl sulfoxide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dmso_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dmso" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DMSO" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23801" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28262" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42138" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4612" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02151" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11143" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D01043" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM745" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08021" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H6OS" hasOnlySubstanceUnits="false" id="M_dmso_e" metaid="M_dmso_e"
      name="Dimethyl sulfoxide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dmso_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dmso" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DMSO" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23801" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28262" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42138" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4612" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02151" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11143" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D01043" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM745" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08021" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C21H24N6O15P2" hasOnlySubstanceUnits="false" id="M_dnad_c" metaid="M_dnad_c"
      name="Deamino-NAD+">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dnad_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dnad" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEAMIDO-NAD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14103" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14104" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14105" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18304" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23567" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58437" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00857" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM309" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00638" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C21H33N7O13P2S" hasOnlySubstanceUnits="false" id="M_dpcoa_c" metaid="M_dpcoa_c"
      name="Dephospho-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dpcoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dpcoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEPHOSPHO-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11793" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14124" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15468" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23642" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41614" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4436" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57328" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01373" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00882" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050315" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM481" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00655" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-7" fbc:chemicalFormula="C42H41N4O16" hasOnlySubstanceUnits="false" id="M_dscl_c" metaid="M_dscl_c"
      name="Dihydrosirohydrochlorin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dscl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dscl" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROSIROHYDROCHLORIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14870" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:50602" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58827" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8370" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02463" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM672" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01620" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H17N2O3" hasOnlySubstanceUnits="false" id="M_dtbt_c" metaid="M_dtbt_c"
      name="Dethiobiotin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtbt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtbt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DETHIOBIOTIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14132" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16691" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23649" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36990" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42279" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42280" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4457" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57861" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03581" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01909" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1020" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01311" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C18H27N3O15P2" hasOnlySubstanceUnits="false" id="M_dtdp4aaddg_c" metaid="M_dtdp4aaddg_c"
      name="DTDP-4-acetamido-4,6-dideoxy-D-galactose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtdp4aaddg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtdp4aaddg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TDP-FUC4NAC" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61562" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61661" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68493" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68518" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12222" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C20415" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM3015" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15454" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C16H26N3O14P2" hasOnlySubstanceUnits="false" id="M_dtdp4addg_c" metaid="M_dtdp4addg_c"
      name="DTDP-4-amino-4,6-dideoxy-D-galactose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtdp4addg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtdp4addg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TDP-D-FUCOSAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10507" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14078" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23539" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57596" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68492" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68516" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04346" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114295" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02661" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd24124" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H22N2O15P2" hasOnlySubstanceUnits="false" id="M_dtdp4d6dg_c" metaid="M_dtdp4d6dg_c"
      name="DTDP-4-dehydro-6-deoxy-D-glucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtdp4d6dg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtdp4d6dg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDP-DEOH-DEOXY-GLUCOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10510" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14081" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16128" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23542" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57649" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62829" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60081" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11907" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM451" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00521" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H22N2O15P2" hasOnlySubstanceUnits="false" id="M_dtdp4d6dm_c" metaid="M_dtdp4d6dm_c"
      name="DTDP-4-dehydro-6-deoxy-L-mannose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtdp4d6dm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtdp4d6dm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDP-DEOH-DEOXY-MANNOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10511" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14082" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15744" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23543" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45868" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57494" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62830" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00688" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2371" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00522" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08300" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C10H13N2O11P2" hasOnlySubstanceUnits="false" id="M_dtdp_c" metaid="M_dtdp_c"
      name="DTDP C10H13N2O11P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10500" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14077" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18075" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26998" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46061" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58369" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01274" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00363" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM152" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00297" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H24N2O16P2" hasOnlySubstanceUnits="false" id="M_dtdpglu_c" metaid="M_dtdpglu_c"
      name="DTDPglucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtdpglu_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtdpglu" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDP-D-GLUCOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10526" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10528" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14089" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14091" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15700" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23556" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57477" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62805" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00842" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1376" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00626" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H24N2O15P2" hasOnlySubstanceUnits="false" id="M_dtdprmn_c" metaid="M_dtdprmn_c"
      name="DTDP-L-rhamnose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtdprmn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtdprmn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDP-RHAMNOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10518" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14083" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15774" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23547" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35452" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46114" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57510" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61119" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03319" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM899" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02113" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H13N2O8P" hasOnlySubstanceUnits="false" id="M_dtmp_c" metaid="M_dtmp_c"
      name="DTMP C10H13N2O8P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dtmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dtmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10529" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14092" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15246" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17013" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45759" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45772" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45926" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46013" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46036" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46960" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47711" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:63528" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:63549" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00364" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM257" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00298" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H13N2O14P3" hasOnlySubstanceUnits="false" id="M_dttp_c" metaid="M_dttp_c"
      name="DTTP C10H13N2O14P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dttp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dttp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TTP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14093" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18077" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27000" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37568" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46175" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58370" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00459" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM394" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00357" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C9H11N2O11P2" hasOnlySubstanceUnits="false" id="M_dudp_c" metaid="M_dudp_c"
      name="DUDP C9H11N2O11P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dudp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dudp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DUDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10531" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19262" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28850" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60471" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01000" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01346" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM572" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00978" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H11N2O8P" hasOnlySubstanceUnits="false" id="M_dump_c" metaid="M_dump_c"
      name="DUMP C9H11N2O8P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dump_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dump" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DUMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10532" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17622" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19263" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:246422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42245" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46286" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46288" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47722" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01409" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00365" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM234" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00299" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H12N2O5" hasOnlySubstanceUnits="false" id="M_duri_c" metaid="M_duri_c"
      name="Deoxyuridine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_duri_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/duri" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYURIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11572" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14123" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16450" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19261" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23640" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29113" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42178" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4434" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46165" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46289" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46293" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00012" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00526" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM492" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00412" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H12N2O5" hasOnlySubstanceUnits="false" id="M_duri_e" metaid="M_duri_e"
      name="Deoxyuridine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_duri_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/duri" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYURIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11572" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14123" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16450" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19261" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23640" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29113" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42178" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4434" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46165" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46289" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46293" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00012" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00526" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM492" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00412" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C9H11N2O14P3" hasOnlySubstanceUnits="false" id="M_dutp_c" metaid="M_dutp_c"
      name="DUTP C9H11N2O14P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dutp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dutp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DUTP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10533" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17625" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19264" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58212" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61555" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00460" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM452" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00358" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O7P" hasOnlySubstanceUnits="false" id="M_dxyl5p_c" metaid="M_dxyl5p_c"
      name="1-deoxy-D-xylulose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dxyl5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dxyl5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYXYLULOSE-5P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11254" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16493" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57792" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:622" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01213" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11437" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM622" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08289" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O4" hasOnlySubstanceUnits="false" id="M_dxyl_c" metaid="M_dxyl_c"
      name="1-deoxy-D-xylulose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_dxyl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/dxyl" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-1093" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19038" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:621" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01292" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06257" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2454" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03738" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H7O7P" hasOnlySubstanceUnits="false" id="M_e4p_c" metaid="M_e4p_c"
      name="D-Erythrose 4-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_e4p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/e4p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ERYTHROSE-4P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12921" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16897" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20927" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4114" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42349" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48153" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01321" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00279" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM258" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00236" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C24H36N3O15" hasOnlySubstanceUnits="false" id="M_eca_EC_c" metaid="M_eca_EC_c"
      name="Enterobacterial common antigen polysaccharide (Ecoli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_eca_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/eca_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM11492" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15651" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H9N2O6P" hasOnlySubstanceUnits="false" id="M_eig3p_c" metaid="M_eig3p_c"
      name="D-erythro-1-(Imidazol-4-yl)glycerol 3-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_eig3p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/eig3p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-ERYTHRO-IMIDAZOLE-GLYCEROL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12888" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17805" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20923" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4270" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58278" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12208" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04666" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1422" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02843" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C30H27N3O15" hasOnlySubstanceUnits="false" id="M_enter_c" metaid="M_enter_c"
      name="Enterochelin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_enter_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/enter" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ENTEROBACTIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23923" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28855" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:38150" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4799" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:77805" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05821" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM883" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03453" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C2H8NO" hasOnlySubstanceUnits="false" id="M_etha_c" metaid="M_etha_c"
      name="Ethanolamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_etha_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/etha" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ETHANOL-AMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14223" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16000" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23979" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:272066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42323" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4880" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57603" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00149" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62693" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00189" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D05074" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM218" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00162" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H6O" hasOnlySubstanceUnits="false" id="M_etoh_c" metaid="M_etoh_c" name="Ethanol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_etoh_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/etoh" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ETOH" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14222" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16236" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30878" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30880" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42377" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44594" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4879" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:52092" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00108" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00469" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00068" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02798" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04855" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D06542" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM303" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00363" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H6O" hasOnlySubstanceUnits="false" id="M_etoh_e" metaid="M_etoh_e" name="Ethanol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_etoh_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/etoh" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ETOH" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14222" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16236" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30878" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30880" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42377" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44594" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4879" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:52092" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00108" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00469" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00068" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02798" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04855" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D06542" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM303" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00363" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_f1p_c" metaid="M_f1p_c"
      name="D-Fructose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_f1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/f1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-fructofuranose-1-phosphate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20930" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37515" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5174" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58674" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01094" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM145568" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00802" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_f6p_c" metaid="M_f6p_c"
      name="D-Fructose 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_f6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/f6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FRUCTOSE-6P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10375" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12352" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16084" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22768" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42378" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57634" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03971" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05345" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89621" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19035" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C27H31N9O15P2" hasOnlySubstanceUnits="false" id="M_fad_c" metaid="M_fad_c"
      name="Flavin adenine dinucleotide oxidized">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fad_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fad" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FAD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13315" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16238" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42388" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4956" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57692" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01248" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00016" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00005" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM33" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00015" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C27H33N9O15P2" hasOnlySubstanceUnits="false" id="M_fadh2_c" metaid="M_fadh2_c"
      name="Flavin adenine dinucleotide reduced">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fadh2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fadh2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FADH2" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13316" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17877" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4957" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58307" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01352" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM38" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00982" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O8P" hasOnlySubstanceUnits="false" id="M_fc1p_c" metaid="M_fc1p_c"
      name="L-Fuculose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fc1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fc1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FUCULOSE-1P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13104" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13105" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16647" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21296" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6220" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01099" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1749" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00806" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O5" hasOnlySubstanceUnits="false" id="M_fcl__L_c" metaid="M_fcl__L_c"
      name="L-fuculose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fcl__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fcl__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-FUCULOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13103" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17617" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21295" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58208" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6219" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60267" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01721" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1748" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01186" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C6H10O12P2" hasOnlySubstanceUnits="false" id="M_fdp_c" metaid="M_fdp_c"
      name="D-Fructose 1,6-bisphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fdp" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37736" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49299" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00354" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM417" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00290" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="2" fbc:chemicalFormula="Fe" hasOnlySubstanceUnits="false" id="M_fe2_c" metaid="M_fe2_c" name="Fe2+ mitochondria">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fe2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fe2" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21129" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24876" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29033" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:34754" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49599" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00692" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C14818" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM111" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd10515" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="2" fbc:chemicalFormula="Fe" hasOnlySubstanceUnits="false" id="M_fe2_e" metaid="M_fe2_e" name="Fe2+ mitochondria">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fe2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fe2" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21129" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24876" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29033" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:34754" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49599" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00692" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C14818" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM111" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd10515" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C8H13N2O9P" hasOnlySubstanceUnits="false" id="M_fgam_c" metaid="M_fgam_c"
      name="N2-Formyl-N1-(5-phospho-D-ribosyl)glycinamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fgam_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fgam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-P-RIBOSYL-N-FORMYLGLYCINEAMIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12635" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18272" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1982" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20498" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42411" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58426" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01308" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06476" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04376" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM453" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02678" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C17H19N4O9P" hasOnlySubstanceUnits="false" id="M_fmn_c" metaid="M_fmn_c"
      name="FMN C17H19N4O9P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fmn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fmn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FMN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13317" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17621" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21127" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42587" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4960" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58210" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00061" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM119" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00050" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="CH1O2" hasOnlySubstanceUnits="false" id="M_for_c" metaid="M_for_c" name="Formate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_for_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/for" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBOXYL-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-9845" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD1G-1532" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD1G-1533" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD1G-1534" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD1G-1535" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FORMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14276" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15740" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24081" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24082" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30751" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5145" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00142" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00058" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM39" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00047" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22511" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="CH1O2" hasOnlySubstanceUnits="false" id="M_for_e" metaid="M_for_e" name="Formate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_for_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/for" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBOXYL-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-9845" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD1G-1532" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD1G-1533" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD1G-1534" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD1G-1535" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FORMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14276" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15740" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24081" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24082" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30751" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5145" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00142" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00058" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM39" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00047" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22511" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C8H15N3O8P" hasOnlySubstanceUnits="false" id="M_fpram_c" metaid="M_fpram_c"
      name="2-(Formamido)-N1-(5-phospho-D-ribosyl)acetamidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fpram_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fpram" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11474" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18413" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58478" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:971" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06211" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM568" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H13N4O9P" hasOnlySubstanceUnits="false" id="M_fprica_c" metaid="M_fprica_c"
      name="5-Formamido-1-(5-phospho-D-ribosyl)imidazole-4-carboxamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fprica_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fprica" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHORIBOSYL-FORMAMIDO-CARBOXAMIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18381" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18967" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:574" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58467" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62708" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04734" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM456" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02884" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C15H25O7P2" hasOnlySubstanceUnits="false" id="M_frdp_c" metaid="M_frdp_c"
      name="Farnesyl diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_frdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/frdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FARNESYL-PP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10700" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11488" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11491" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12854" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12874" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14231" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:175763" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19789" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24016" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42496" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:50277" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00961" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04201" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00448" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR0103010002" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM34" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00350" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6" hasOnlySubstanceUnits="false" id="M_fru_c" metaid="M_fru_c" name="D-Fructose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fru_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fru" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15382" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Fructopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FRU" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Fructofuranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12923" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15824" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20929" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24104" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24110" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28757" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37714" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37721" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4118" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4119" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47424" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5172" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62538" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00095" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01496" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05003" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C10906" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00114" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM175" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00082" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19015" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27040" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6" hasOnlySubstanceUnits="false" id="M_fru_e" metaid="M_fru_e" name="D-Fructose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fru_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fru" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15382" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Fructopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FRU" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Fructofuranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12923" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15824" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20929" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24104" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24110" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28757" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37714" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37721" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4118" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4119" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47424" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5172" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62538" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00095" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01496" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05003" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C10906" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00114" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM175" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00082" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19015" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27040" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_fruur_c" metaid="M_fruur_c"
      name="D-Fructuronate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fruur_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fruur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FRUCTURONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12927" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16849" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20936" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20937" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24112" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24113" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47950" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59863" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59881" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00905" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1660" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00672" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27042" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O8P" hasOnlySubstanceUnits="false" id="M_fuc1p__L_c" metaid="M_fuc1p__L_c"
      name="L-Fucose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fuc1p__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fuc1p__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-488" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12212" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12387" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21294" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57268" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6218" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02985" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1727" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01912" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O8P" hasOnlySubstanceUnits="false" id="M_fuc1p__L_e" metaid="M_fuc1p__L_e"
      name="L-Fucose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fuc1p__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fuc1p__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-488" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12212" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12387" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21294" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57268" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6218" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02985" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1727" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01912" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O5" hasOnlySubstanceUnits="false" id="M_fuc__L_c" metaid="M_fuc__L_c"
      name="L-Fucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fuc__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fuc__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15619" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-fucoses" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48204" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62489" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM40586" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O5" hasOnlySubstanceUnits="false" id="M_fuc__L_e" metaid="M_fuc__L_e"
      name="L-Fucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fuc__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fuc__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15619" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-fucoses" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48204" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62489" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM40586" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H2O4" hasOnlySubstanceUnits="false" id="M_fum_c" metaid="M_fum_c" name="Fumarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fum_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fum" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FUM" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18012" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22956" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22957" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22958" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24122" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24124" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36180" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37154" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37155" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42511" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5190" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00134" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00122" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02308" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM93" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00106" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H2O4" hasOnlySubstanceUnits="false" id="M_fum_e" metaid="M_fum_e" name="Fumarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_fum_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/fum" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FUM" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18012" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22956" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22957" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22958" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24122" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24124" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36180" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37154" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37155" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42511" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5190" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00134" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00122" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02308" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM93" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00106" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_g1p_c" metaid="M_g1p_c"
      name="D-Glucose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-glucose-1-phosphates" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLC-1-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10246" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12320" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12967" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12970" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16077" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21001" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21004" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29042" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4169" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42623" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57629" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58601" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01306" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01586" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00103" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89588" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00089" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28817" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C3H5O6P" hasOnlySubstanceUnits="false" id="M_g3p_c" metaid="M_g3p_c"
      name="Glyceraldehyde 3-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g3p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g3p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GAP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12983" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12984" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14333" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17138" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:181" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18324" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21026" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29052" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5446" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58027" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59776" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01112" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00118" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00661" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM74" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00102" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19005" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H20NO6P" hasOnlySubstanceUnits="false" id="M_g3pc_c" metaid="M_g3pc_c"
      name="Sn-Glycero-3-phosphocholine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g3pc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g3pc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-1-GLYCERO-PHOSPHORYLCHOLINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10646" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12847" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14343" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16870" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26697" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41458" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:55397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76433" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00670" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07349" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM367" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00507" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H14NO6P" hasOnlySubstanceUnits="false" id="M_g3pe_c" metaid="M_g3pe_c"
      name="Sn-Glycero-3-phosphoethanolamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g3pe_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g3pe" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-1-GLYCEROPHOSPHORYLETHANOL-AMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10647" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12842" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16929" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26699" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57952" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB59660" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01233" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM368" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00908" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H14O8P" hasOnlySubstanceUnits="false" id="M_g3pg_c" metaid="M_g3pg_c"
      name="Glycerophosphoglycerol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g3pg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g3pg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCEROPHOSPHOGLYCEROL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5457" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61933" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03274" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM598" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02090" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H18O11P" hasOnlySubstanceUnits="false" id="M_g3pi_c" metaid="M_g3pi_c"
      name="Sn-Glycero-3-phospho-1-inositol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g3pi_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g3pi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-541" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10645" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11200" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26695" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58444" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64715" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11649" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01225" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1517" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00902" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H13NO8P" hasOnlySubstanceUnits="false" id="M_g3ps_c" metaid="M_g3ps_c"
      name="Glycerophosphoserine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g3ps_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g3ps" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-2030" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61931" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62013" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64765" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64945" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM97117" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15468" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_g6p_c" metaid="M_g6p_c"
      name="D-Glucose 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-glucopyranose-6-phosphate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14314" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4170" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61548" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01401" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01549" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06793" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00092" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM160" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00079" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd26836" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_g6p_e" metaid="M_g6p_e"
      name="D-Glucose 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_g6p_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/g6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-glucopyranose-6-phosphate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14314" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4170" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61548" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01401" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01549" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06793" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00092" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM160" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00079" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd26836" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_gal1p_c" metaid="M_gal1p_c"
      name="Alpha-D-Galactose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gal1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gal1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GALACTOSE-1P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10232" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12305" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12306" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17973" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22374" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42878" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58336" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00645" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03359" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00446" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM336" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00348" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19025" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6" hasOnlySubstanceUnits="false" id="M_gal_c" metaid="M_gal_c" name="D-Galactose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gal_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gal" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALPHA-D-GALACTOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-galactopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10231" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12936" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22373" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28061" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42741" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00143" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB05762" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00124" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00984" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04291" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM390" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00108" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00724" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6" hasOnlySubstanceUnits="false" id="M_gal_e" metaid="M_gal_e" name="D-Galactose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gal_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gal" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALPHA-D-GALACTOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-galactopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10231" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12936" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22373" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28061" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42741" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00143" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB05762" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00124" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00984" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04291" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM390" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00108" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00724" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H8O8" hasOnlySubstanceUnits="false" id="M_galct__D_c" metaid="M_galct__D_c"
      name="D-Galactarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_galct__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/galct__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-GALACTARATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12929" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16537" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17444" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20944" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24135" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24136" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24137" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33799" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35390" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4130" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48871" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5250" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00639" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03435" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00879" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170107" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1080" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00652" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H8O8" hasOnlySubstanceUnits="false" id="M_galct__D_e" metaid="M_galct__D_e"
      name="D-Galactarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_galct__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/galct__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-GALACTARATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12929" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16537" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17444" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20944" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24135" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24136" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24137" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33799" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35390" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4130" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48871" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5250" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00639" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03435" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00879" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170107" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1080" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00652" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O7" hasOnlySubstanceUnits="false" id="M_galctn__D_c" metaid="M_galctn__D_c"
      name="D-Galactonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_galctn__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/galctn__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-GALACTONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12931" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16534" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24148" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24149" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4132" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00565" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00880" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1734" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00653" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O7" hasOnlySubstanceUnits="false" id="M_galctn__D_e" metaid="M_galctn__D_e"
      name="D-Galactonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_galctn__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/galctn__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-GALACTONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12931" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16534" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24148" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24149" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4132" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00565" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00880" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1734" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00653" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H13O9P" hasOnlySubstanceUnits="false" id="M_galt1p_c" metaid="M_galt1p_c"
      name="Galactitol 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_galt1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/galt1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GALACTITOL-1-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14287" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24140" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28663" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5252" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60083" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06311" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1667" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03752" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H14O6" hasOnlySubstanceUnits="false" id="M_galt_e" metaid="M_galt_e" name="Galactitol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_galt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/galt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GALACTITOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14286" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16813" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5251" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:53575" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00107" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01697" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1233" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01171" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_galur_c" metaid="M_galur_c"
      name="D-Galacturonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_galur_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/galur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Galactopyranuronate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20976" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4153" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:75525" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00333" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM48464" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00280" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_galur_e" metaid="M_galur_e"
      name="D-Galacturonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_galur_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/galur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Galactopyranuronate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20976" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4153" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:75525" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00333" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM48464" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00280" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H13NO8P" hasOnlySubstanceUnits="false" id="M_gam1p_c" metaid="M_gam1p_c"
      name="D-Glucosamine 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gam1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gam1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCOSAMINE-1P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12316" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20994" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27625" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4163" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42868" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58516" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06156" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM969" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03671" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H13NO8P" hasOnlySubstanceUnits="false" id="M_gam6p_c" metaid="M_gam6p_c"
      name="D-Glucosamine 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gam6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gam6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-GLUCOSAMINE-6-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12962" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58725" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00352" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM370" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00288" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="C6H14NO5" hasOnlySubstanceUnits="false" id="M_gam_e" metaid="M_gam_e"
      name="D-Glucosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gam_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCOSAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17315" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4162" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47977" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5417" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58723" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00329" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04334" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM533" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00276" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01247" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27103" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H14N2O8P" hasOnlySubstanceUnits="false" id="M_gar_c" metaid="M_gar_c"
      name="N1-(5-Phospho-D-ribosyl)glycinamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gar_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gar" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-PHOSPHO-RIBOSYL-GLYCINEAMIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12623" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18349" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1983" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20499" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58457" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03838" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM463" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02394" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H15NO2" hasOnlySubstanceUnits="false" id="M_gbbtn_c" metaid="M_gbbtn_c"
      name="Gamma-butyrobetaine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gbbtn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gbbtn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GAMMA-BUTYROBETAINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12047" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20484" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01161" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06831" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01181" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM626" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00870" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C7H15NO2" hasOnlySubstanceUnits="false" id="M_gbbtn_e" metaid="M_gbbtn_e"
      name="Gamma-butyrobetaine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gbbtn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gbbtn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GAMMA-BUTYROBETAINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12047" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20484" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01161" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06831" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01181" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM626" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00870" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H4O2" hasOnlySubstanceUnits="false" id="M_gcald_c" metaid="M_gcald_c"
      name="Glycolaldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gcald_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gcald" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCOLALDEHYDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131198" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14347" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17071" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24386" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5474" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02165" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03344" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00266" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM349" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00229" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C10H12N5O11P2" hasOnlySubstanceUnits="false" id="M_gdp_c" metaid="M_gdp_c"
      name="GDP C10H12N5O11P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13327" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14379" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17552" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24448" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42738" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5212" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:65180" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00035" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10620" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM30" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00031" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H21N5O15P2" hasOnlySubstanceUnits="false" id="M_gdpddman_c" metaid="M_gdpddman_c"
      name="GDP-4-dehydro-6-deoxy-D-mannose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gdpddman_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gdpddman" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GDP-4-DEHYDRO-6-DEOXY-D-MANNOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13329" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16955" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21153" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5214" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57964" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01222" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM516" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00900" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H23N5O15P2" hasOnlySubstanceUnits="false" id="M_gdpfuc_c" metaid="M_gdpfuc_c"
      name="GDP-L-fucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gdpfuc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gdpfuc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-13118" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANOSINE_DIPHOSPHATE_FUCOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13332" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13335" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17009" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21162" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5221" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57273" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57984" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01095" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00325" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10615" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM193" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00272" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd23744" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27127" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H23N5O16P2" hasOnlySubstanceUnits="false" id="M_gdpmann_c" metaid="M_gdpmann_c"
      name="GDP-D-mannose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gdpmann_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gdpmann" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GDP-MANNOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13328" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13333" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15820" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21159" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42729" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5225" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57527" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:84880" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:84881" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62759" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00096" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10614" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM82" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00083" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C16H21N5O15P2" hasOnlySubstanceUnits="false" id="M_gdpofuc_c" metaid="M_gdpofuc_c"
      name="GDP-4-oxo-L-fucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gdpofuc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gdpofuc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-2128" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21155" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5216" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM54801" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15472" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6" hasOnlySubstanceUnits="false" id="M_glc__D_c" metaid="M_glc__D_c"
      name="D-Glucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glc__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glc__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Glucopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12965" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17634" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4167" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00122" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06564" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00031" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00009" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM41" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00027" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd26821" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6" hasOnlySubstanceUnits="false" id="M_glc__D_e" metaid="M_glc__D_e"
      name="D-Glucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glc__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glc__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Glucopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12965" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17634" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4167" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00122" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06564" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00031" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00009" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM41" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00027" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd26821" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O7" hasOnlySubstanceUnits="false" id="M_glcn_c" metaid="M_glcn_c"
      name="D-Gluconate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glcn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glcn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12955" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18391" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20983" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20985" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24265" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24266" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33198" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4157" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42715" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00625" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03373" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00257" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM341" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00222" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O7" hasOnlySubstanceUnits="false" id="M_glcn_e" metaid="M_glcn_e"
      name="D-Gluconate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glcn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glcn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12955" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18391" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20983" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20985" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24265" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24266" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33198" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4157" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42715" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00625" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03373" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00257" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM341" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00222" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H8O8" hasOnlySubstanceUnits="false" id="M_glcr_c" metaid="M_glcr_c"
      name="D-Glucarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glcr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glcr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-GLUCARATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12953" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20982" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30612" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33801" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4155" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42731" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB29881" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00818" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170108" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM744" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00609" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H8O8" hasOnlySubstanceUnits="false" id="M_glcr_e" metaid="M_glcr_e"
      name="D-Glucarate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glcr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glcr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-GLUCARATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12953" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20982" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30612" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33801" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4155" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42731" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB29881" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00818" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170108" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM744" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00609" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_glcur_c" metaid="M_glcur_c"
      name="D-Glucuronate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glcur_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glcur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Glucopyranuronate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12975" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21013" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24297" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24298" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4178" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47952" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58720" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00191" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM241" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00164" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_glcur_e" metaid="M_glcur_e"
      name="D-Glucuronate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glcur_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glcur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Glucopyranuronate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12975" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21013" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24297" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24298" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4178" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47952" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58720" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00191" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM241" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00164" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10N2O3" hasOnlySubstanceUnits="false" id="M_gln__L_c" metaid="M_gln__L_c"
      name="L-Glutamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gln__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gln__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13110" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24316" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28300" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32665" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32666" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32678" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32679" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42812" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42814" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42899" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5432" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58359" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6227" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00641" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00064" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00303" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00015" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM37" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00053" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00253" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10N2O3" hasOnlySubstanceUnits="false" id="M_gln__L_e" metaid="M_gln__L_e"
      name="L-Glutamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gln__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gln__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13110" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24316" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28300" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32665" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32666" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32678" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32679" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42812" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42814" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42899" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5432" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58359" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6227" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00641" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00064" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00303" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00015" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM37" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00053" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00253" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H9NO3" hasOnlySubstanceUnits="false" id="M_glu1sa_c" metaid="M_glu1sa_c"
      name="L-Glutamate 1-semialdehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glu1sa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glu1sa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTAMATE-1-SEMIALDEHYDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11022" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15757" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18756" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24312" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:404" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42823" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42827" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57501" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03741" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1453" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02345" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H8NO7P" hasOnlySubstanceUnits="false" id="M_glu5p_c" metaid="M_glu5p_c"
      name="L-Glutamate 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glu5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glu5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-GLUTAMATE-5-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13108" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13113" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17798" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21312" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58274" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6230" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03287" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1280" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02097" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H9NO3" hasOnlySubstanceUnits="false" id="M_glu5sa_c" metaid="M_glu5sa_c"
      name="L-Glutamate 5-semialdehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glu5sa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glu5sa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-GLUTAMATE_GAMMA-SEMIALDEHYDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13109" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17232" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21302" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6225" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02104" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01165" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM245" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00858" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H8NO4" hasOnlySubstanceUnits="false" id="M_glu__D_c" metaid="M_glu__D_c"
      name="D-Glutamate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glu__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glu__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-GLT" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12979" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15966" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21022" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21023" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29989" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4183" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03339" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00217" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM331" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00186" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H8NO4" hasOnlySubstanceUnits="false" id="M_glu__L_c" metaid="M_glu__L_c"
      name="L-Glutamate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glu__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glu__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Glutamates" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16015" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21301" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21304" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24314" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29985" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29988" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42825" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6224" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76051" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00148" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60475" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00025" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00302" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00007" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04341" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89557" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00023" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19002" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27177" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H8NO4" hasOnlySubstanceUnits="false" id="M_glu__L_e" metaid="M_glu__L_e"
      name="L-Glutamate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glu__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glu__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Glutamates" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16015" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21301" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21304" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24314" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29985" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29988" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42825" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6224" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76051" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00148" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60475" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00025" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00302" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00007" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04341" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89557" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00023" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19002" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27177" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C8H13N2O5S" hasOnlySubstanceUnits="false" id="M_glucys_c" metaid="M_glucys_c"
      name="Gamma-L-Glutamyl-L-cysteine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glucys_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glucys" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-GAMMA-GLUTAMYLCYSTEINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10570" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12400" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12404" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17515" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17971" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24185" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24194" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39975" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58173" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00669" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM412" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00506" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H7NO3R" hasOnlySubstanceUnits="false" id="M_glutrna_c" metaid="M_glutrna_c"
      name="L-Glutamyl-tRNA(Glu)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glutrna_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glutrna" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13114" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29157" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6232" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02987" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89752" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12227" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C2H1O3" hasOnlySubstanceUnits="false" id="M_glx_c" metaid="M_glx_c" name="Glyoxylate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glx_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYOX" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14368" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16891" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24420" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24421" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35977" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36655" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42767" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5509" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00119" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00048" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM69" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00040" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H5NO2" hasOnlySubstanceUnits="false" id="M_gly_c" metaid="M_gly_c" name="Glycine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gly_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gly" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLY" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10792" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132194" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14344" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15428" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24368" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32507" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32508" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42964" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57305" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00123" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00037" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00011" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM29" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00033" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H5NO2" hasOnlySubstanceUnits="false" id="M_gly_e" metaid="M_gly_e" name="Glycine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gly_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gly" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLY" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10792" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132194" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14344" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15428" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24368" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32507" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32508" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42964" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57305" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00123" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00037" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00011" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM29" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00033" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H6O3" hasOnlySubstanceUnits="false" id="M_glyald_c" metaid="M_glyald_c"
      name="D-Glyceraldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyald_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyald" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCERALD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12982" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17378" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21025" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39973" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4186" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00577" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM435" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00448" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H6O3" hasOnlySubstanceUnits="false" id="M_glyald_e" metaid="M_glyald_e"
      name="D-Glyceraldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyald_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyald" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCERALD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12982" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17378" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21025" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39973" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4186" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00577" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM435" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00448" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H11NO2" hasOnlySubstanceUnits="false" id="M_glyb_c" metaid="M_glyb_c"
      name="Glycine betaine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyb_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BETAINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12531" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13895" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15264" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17750" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22858" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27128" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3073" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41134" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41139" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00043" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00719" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07523" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM289" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00540" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H11NO2" hasOnlySubstanceUnits="false" id="M_glyb_e" metaid="M_glyb_e"
      name="Glycine betaine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyb_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BETAINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12531" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13895" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15264" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17750" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22858" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27128" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3073" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41134" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41139" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00043" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00719" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07523" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM289" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00540" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C3H7O6P" hasOnlySubstanceUnits="false" id="M_glyc3p_c" metaid="M_glyc3p_c"
      name="Glycerol 3-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyc3p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyc3p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCEROL-3P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10648" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12843" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12848" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26705" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42793" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57597" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00093" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM66" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00080" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C3H7O6P" hasOnlySubstanceUnits="false" id="M_glyc3p_e" metaid="M_glyc3p_e"
      name="Glycerol 3-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyc3p_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyc3p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCEROL-3P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10648" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12843" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12848" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26705" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42793" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57597" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00093" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM66" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00080" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H5O4" hasOnlySubstanceUnits="false" id="M_glyc__R_c" metaid="M_glyc__R_c"
      name="(R)-Glycerate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyc__R_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyc__R" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCERATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12985" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16659" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21027" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21030" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24348" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24349" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33508" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33871" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4187" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41990" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00139" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00258" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM189" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00223" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H8O3" hasOnlySubstanceUnits="false" id="M_glyc_c" metaid="M_glyc_c" name="Glycerol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCEROL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14334" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17754" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24351" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42998" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5448" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00131" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00116" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00028" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89612" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00100" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H8O3" hasOnlySubstanceUnits="false" id="M_glyc_e" metaid="M_glyc_e" name="Glycerol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyc_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCEROL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14334" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17754" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24351" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42998" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5448" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00131" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00116" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00028" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89612" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00100" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C2H3O3" hasOnlySubstanceUnits="false" id="M_glyclt_c" metaid="M_glyclt_c"
      name="Glycolate C2H3O3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyclt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyclt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCOLLATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14348" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17497" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24388" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24390" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29805" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42865" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5475" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00115" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03035" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00160" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM222" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00139" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12347" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C2H3O3" hasOnlySubstanceUnits="false" id="M_glyclt_e" metaid="M_glyclt_e"
      name="Glycolate C2H3O3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glyclt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glyclt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCOLLATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14348" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17497" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24388" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24390" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29805" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42865" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5475" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00115" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03035" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00160" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM222" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00139" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12347" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H10O5" hasOnlySubstanceUnits="false" id="M_glycogen_c" metaid="M_glycogen_c"
      name="Glycogen C6H10O5">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_glycogen_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/glycogen" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Glycogens" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24379" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24384" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5466" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00757" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00182" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G11603" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM55375" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00155" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27186" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C7H12O13P2" hasOnlySubstanceUnits="false" id="M_gmhep17bp_c" metaid="M_gmhep17bp_c"
      name="D-Glycero-D-manno-heptose 1,7-bisphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gmhep17bp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gmhep17bp" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4188" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59957" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM48518" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29677" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H13O10P" hasOnlySubstanceUnits="false" id="M_gmhep1p_c" metaid="M_gmhep1p_c"
      name="D-Glycero-D-manno-heptose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gmhep1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gmhep1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-BETA-D-HEPTOSE-1-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21031" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28137" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61593" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C07838" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1661" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd04920" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H13O10P" hasOnlySubstanceUnits="false" id="M_gmhep7p_c" metaid="M_gmhep7p_c"
      name="D-Glycero-D-manno-heptose 7-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gmhep7p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gmhep7p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-12543" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-ALPHABETA-D-HEPTOSE-7-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21032" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28723" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4190" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59955" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60202" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60204" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C07836" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C19882" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM17114" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd04918" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd21122" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd26805" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H12N5O8P" hasOnlySubstanceUnits="false" id="M_gmp_c" metaid="M_gmp_c"
      name="GMP C10H12N5O8P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13341" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14381" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17345" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24449" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24450" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29058" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40119" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42615" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42647" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42831" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42887" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42892" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42979" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47450" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5228" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58115" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00144" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM113" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00126" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C20H24N10O21P4" hasOnlySubstanceUnits="false" id="M_gp4g_c" metaid="M_gp4g_c"
      name="P1,P4-Bis(5'-guanosyl) tetraphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gp4g_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gp4g" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-609" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12727" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12730" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15883" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57553" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7876" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01261" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1582" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00925" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C10H17O7P2" hasOnlySubstanceUnits="false" id="M_grdp_c" metaid="M_grdp_c"
      name="Geranyl diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_grdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/grdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GERANYL-PP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14299" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17211" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24223" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42877" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5332" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58057" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00341" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR0102010001" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM100" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00283" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03476" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H13N5O5" hasOnlySubstanceUnits="false" id="M_gsn_c" metaid="M_gsn_c" name="Guanosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gsn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gsn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14375" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16750" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24444" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42840" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42847" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42853" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:471737" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5564" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00133" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00387" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM401" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00311" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H13N5O5" hasOnlySubstanceUnits="false" id="M_gsn_e" metaid="M_gsn_e" name="Guanosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gsn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gsn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14375" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16750" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24444" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42840" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42847" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42853" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:471737" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5564" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00133" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00387" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM401" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00311" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C20H30N6O12S2" hasOnlySubstanceUnits="false" id="M_gthox_c" metaid="M_gthox_c"
      name="Oxidized glutathione">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gthox_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gthox" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OXIDIZED-GLUTATHIONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14328" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14720" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17858" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24336" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42832" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58297" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7840" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03337" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00127" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00031" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM151" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00111" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H16N3O6S" hasOnlySubstanceUnits="false" id="M_gthrd_c" metaid="M_gthrd_c"
      name="Reduced glutathione">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gthrd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gthrd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTATHIONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12402" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14327" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16856" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24334" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42873" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43049" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5437" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57925" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00125" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01463" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06960" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62697" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00051" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00014" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM57" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00042" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H12N5O14P3" hasOnlySubstanceUnits="false" id="M_gtp_c" metaid="M_gtp_c"
      name="GTP C10H12N5O14P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gtp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gtp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GTP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13342" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15996" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37565" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42934" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57600" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00044" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM51" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00038" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="2" fbc:chemicalFormula="C17H36N6O5S" hasOnlySubstanceUnits="false" id="M_gtspmd_c" metaid="M_gtspmd_c"
      name="Glutathionylspermidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gtspmd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gtspmd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTATHIONYLSPERMIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12624" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14329" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16613" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24341" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46146" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5438" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57835" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05730" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1329" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03411" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H5N5O" hasOnlySubstanceUnits="false" id="M_gua_c" metaid="M_gua_c" name="Guanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gua_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gua" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14371" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24443" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42948" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5563" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00132" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00242" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM259" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00207" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H5N5O" hasOnlySubstanceUnits="false" id="M_gua_e" metaid="M_gua_e" name="Guanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_gua_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/gua" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14371" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16235" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24443" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42948" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5563" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00132" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00242" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM259" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00207" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="H2" hasOnlySubstanceUnits="false" id="M_h2_c" metaid="M_h2_c" name="Hydrogen">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_h2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/h2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYDROGEN-MOLECULE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13350" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18276" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25363" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29294" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29298" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29299" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5785" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01362" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00282" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM195" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11640" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C5H9O8P2" hasOnlySubstanceUnits="false" id="M_h2mb4p_c" metaid="M_h2mb4p_c"
      name="1-hydroxy-2-methyl-2-(E)-butenyl 4-diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_h2mb4p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/h2mb4p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYDROXY-METHYL-BUTENYL-DIP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10952" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:128753" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15664" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:632" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11811" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR01010009" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM606" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd08615" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29213" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="H2O2" hasOnlySubstanceUnits="false" id="M_h2o2_c" metaid="M_h2o2_c"
      name="Hydrogen peroxide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_h2o2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/h2o2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYDROGEN-PEROXIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13355" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16240" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24637" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25940" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44782" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44785" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44812" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5586" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03125" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06491" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00027" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C17732" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C18566" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00008" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM22" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00025" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="H2O" hasOnlySubstanceUnits="false" id="M_h2o_c" metaid="M_h2o_c" name="H2O H2O">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_h2o_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/h2o" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15815" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYDROXYL-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OXONIUM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:WATER" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13352" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13365" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13419" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15377" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27313" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29356" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29373" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29374" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29375" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29412" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30490" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33813" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41979" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41981" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42043" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42857" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43228" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44292" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44641" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44701" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44819" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5585" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5594" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02111" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00001" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01328" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C18714" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00001" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D03703" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D06322" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00001" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15275" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27222" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="H2O" hasOnlySubstanceUnits="false" id="M_h2o_e" metaid="M_h2o_e" name="H2O H2O">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_h2o_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/h2o" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15815" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYDROXYL-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OXONIUM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:WATER" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13352" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13365" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13419" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15377" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27313" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29356" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29373" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29374" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29375" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29412" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30490" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33806" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33813" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41979" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41981" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42043" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42857" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43228" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44292" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44641" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44701" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44819" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5585" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5594" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02111" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00001" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01328" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C18714" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00001" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D03703" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D06322" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00001" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15275" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27222" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="H2S" hasOnlySubstanceUnits="false" id="M_h2s_c" metaid="M_h2s_c" name="Hydrogen sulfide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_h2s_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/h2s" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-7046" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-846" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13356" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14414" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15138" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16136" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24639" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29919" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30488" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30489" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43058" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45489" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5787" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03276" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00283" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89582" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00239" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd24697" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="H" hasOnlySubstanceUnits="false" id="M_h_c" metaid="M_h_c" name="H+">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_h_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/h" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROTON" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10744" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13357" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15378" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24636" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29233" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5584" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB59597" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00080" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00067" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="H" hasOnlySubstanceUnits="false" id="M_h_e" metaid="M_h_e" name="H+">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_h_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/h" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROTON" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10744" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13357" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15378" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24636" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29233" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5584" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB59597" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00080" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00067" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="CHO3" hasOnlySubstanceUnits="false" id="M_hco3_c" metaid="M_hco3_c" name="Bicarbonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hco3_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hco3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CO3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:H2CO3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HCO3" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13351" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13363" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17544" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22863" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23017" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23744" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28976" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29201" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:3401" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41605" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:41609" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5589" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00595" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03538" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB31453" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00288" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01353" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C12893" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM60" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00242" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H9NO2S" hasOnlySubstanceUnits="false" id="M_hcys__L_c" metaid="M_hcys__L_c"
      name="L-Homocysteine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hcys__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hcys__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HOMO-CYS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13122" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14408" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17230" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17588" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21329" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43117" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5751" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58065" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58199" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6245" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:63072" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:66952" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00165" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00742" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00155" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05330" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM123" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00135" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19034" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C16H31O2" hasOnlySubstanceUnits="false" id="M_hdca_c" metaid="M_hdca_c"
      name="Hexadecanoate (n-C16:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hdca_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hdca" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PALMITATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14730" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15756" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:231736" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:233028" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24540" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24541" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24542" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29889" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44952" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7896" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00220" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60083" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00249" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D05341" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010001" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010046" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM108" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00214" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C16H31O2" hasOnlySubstanceUnits="false" id="M_hdca_e" metaid="M_hdca_e"
      name="Hexadecanoate (n-C16:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hdca_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hdca" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PALMITATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14730" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15756" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:231736" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:233028" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24540" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24541" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24542" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29889" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44952" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7896" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00220" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60083" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00249" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D05341" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010001" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010046" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM108" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00214" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C16H29O2" hasOnlySubstanceUnits="false" id="M_hdcea_c" metaid="M_hdcea_c"
      name="Hexadecenoate (n-C16:1)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hdcea_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hdcea" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM950" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C27H49N2O8PRS" hasOnlySubstanceUnits="false" id="M_hdeACP_c" metaid="M_hdeACP_c"
      name="Cis-hexadec-9-enoyl-[acyl-carrier protein] (n-C16:1)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hdeACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hdeACP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89949" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C49H56FeN4O5" hasOnlySubstanceUnits="false" id="M_hemeO_c" metaid="M_hemeO_c"
      name="Heme O C49H56FeN4O5">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hemeO_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hemeO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-17063" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HEME_O" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24480" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36301" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61719" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01162" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C15672" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1278" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11313" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H9N3O2" hasOnlySubstanceUnits="false" id="M_his__L_c" metaid="M_his__L_c"
      name="L-Histidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_his__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/his__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HIS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13117" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15971" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21324" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24598" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27570" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32510" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32511" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32512" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32513" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32529" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32531" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32532" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43048" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43114" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43118" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43190" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43239" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57595" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6240" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00177" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03412" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00135" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00768" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00032" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM134" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00119" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00572" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H9N3O2" hasOnlySubstanceUnits="false" id="M_his__L_e" metaid="M_his__L_e"
      name="L-Histidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_his__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/his__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HIS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13117" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15971" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21324" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24598" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27570" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32510" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32511" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32512" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32513" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32529" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32531" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32532" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43048" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43114" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43118" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43190" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43239" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57595" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6240" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00177" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03412" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00135" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00768" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00032" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM134" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00119" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00572" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11N3O4P" hasOnlySubstanceUnits="false" id="M_hisp_c" metaid="M_hisp_c"
      name="L-Histidinol phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hisp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hisp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-HISTIDINOL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13119" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16996" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21327" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6242" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01100" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1482" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00807" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C6H12N3O" hasOnlySubstanceUnits="false" id="M_histd_c" metaid="M_histd_c"
      name="L-Histidinol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_histd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/histd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HISTIDINOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13118" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16255" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21326" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57699" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6241" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03431" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00860" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1281" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00641" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H8O6" hasOnlySubstanceUnits="false" id="M_hkndd_c" metaid="M_hkndd_c"
      name="2-Hydroxy-6-oxonona-2,4-diene-1,9-dioate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hkndd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hkndd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-157" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1140" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11589" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19615" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19621" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61449" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61470" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:66887" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:66936" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04479" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1693" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02731" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H6O6" hasOnlySubstanceUnits="false" id="M_hkntd_c" metaid="M_hkntd_c"
      name="2-hydroxy-6-ketononatrienedioate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hkntd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hkntd" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61450" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61467" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C12624" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2040" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd09255" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-8" fbc:chemicalFormula="C40H38N4O17" hasOnlySubstanceUnits="false" id="M_hmbil_c" metaid="M_hmbil_c"
      name="Hydroxymethylbilane">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hmbil_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hmbil" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYDROXYMETHYLBILANE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14423" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16645" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24716" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57845" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5809" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01137" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01024" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM547" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00755" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H6O3" hasOnlySubstanceUnits="false" id="M_hmfurn_c" metaid="M_hmfurn_c"
      name="4-hydroxy-5-methyl-3(2H)-furanone">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hmfurn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hmfurn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-10204" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:74456" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB31859" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM6359" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15486" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H9NO3" hasOnlySubstanceUnits="false" id="M_hom__L_c" metaid="M_hom__L_c"
      name="L-Homoserine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hom__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hom__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HOMO-SER" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13123" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15699" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21330" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43131" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57476" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6246" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00719" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00263" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM353" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00227" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H3O4" hasOnlySubstanceUnits="false" id="M_hpyr_c" metaid="M_hpyr_c"
      name="Hydroxypyruvate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hpyr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hpyr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OH-PYR" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11837" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14425" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17180" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20082" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20083" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5813" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01352" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00168" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM392" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00145" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H6O2" hasOnlySubstanceUnits="false" id="M_hqn_c" metaid="M_hqn_c" name="Hydroquinone">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hqn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hqn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYDROQUINONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14416" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17594" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24645" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5793" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02434" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00530" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00073" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM376" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00415" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H4N4O" hasOnlySubstanceUnits="false" id="M_hxan_c" metaid="M_hxan_c"
      name="Hypoxanthine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hxan_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hxan" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYPOXANTHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17368" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5841" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00157" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00262" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM213" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00226" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H4N4O" hasOnlySubstanceUnits="false" id="M_hxan_e" metaid="M_hxan_e"
      name="Hypoxanthine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_hxan_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/hxan" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYPOXANTHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17368" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5841" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00157" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00262" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM213" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00226" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H3NO4" hasOnlySubstanceUnits="false" id="M_iasp_c" metaid="M_iasp_c"
      name="Iminoaspartate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_iasp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/iasp" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24784" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:50616" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5878" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58831" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:77875" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01131" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05840" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170110" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM599" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03470" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H8O6" hasOnlySubstanceUnits="false" id="M_ichor_c" metaid="M_ichor_c"
      name="Isochorismate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ichor_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ichor" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ISOCHORISMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14464" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17582" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24883" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29780" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5997" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00885" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM799" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00658" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C6H5O7" hasOnlySubstanceUnits="false" id="M_icit_c" metaid="M_icit_c" name="Isocitrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_icit_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/icit" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14465" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24884" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24886" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30887" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36453" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36454" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5998" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00193" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00311" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89661" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00260" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O7" hasOnlySubstanceUnits="false" id="M_idon__L_c" metaid="M_idon__L_c"
      name="L-Idonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_idon__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/idon__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-IDONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17796" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21335" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21336" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57659" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58494" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6250" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00770" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1565" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00573" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O7" hasOnlySubstanceUnits="false" id="M_idon__L_e" metaid="M_idon__L_e"
      name="L-Idonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_idon__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/idon__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-IDONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17796" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21335" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21336" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57659" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58494" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6250" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00770" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1565" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00573" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C10H11N4O11P2" hasOnlySubstanceUnits="false" id="M_idp_c" metaid="M_idp_c"
      name="IDP C10H11N4O11P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_idp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/idp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:IDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13371" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17808" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19270" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43252" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58280" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5848" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03335" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00104" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM495" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00090" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H13NO2" hasOnlySubstanceUnits="false" id="M_ile__L_c" metaid="M_ile__L_c"
      name="L-Isoleucine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ile__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ile__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ILE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13127" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17191" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21344" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24898" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32604" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32605" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32612" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32613" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43290" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43342" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43366" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58045" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6255" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00172" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00407" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00065" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100047" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM231" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00322" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H13NO2" hasOnlySubstanceUnits="false" id="M_ile__L_e" metaid="M_ile__L_e"
      name="L-Isoleucine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ile__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ile__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ILE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13127" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17191" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21344" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24898" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32604" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32605" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32612" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32613" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43290" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43342" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43366" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58045" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6255" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00172" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00407" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00065" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100047" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM231" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00322" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H7N2O5P" hasOnlySubstanceUnits="false" id="M_imacp_c" metaid="M_imacp_c"
      name="3-(Imidazol-4-yl)-2-oxopropyl phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_imacp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/imacp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:IMIDAZOLE-ACETOL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11736" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1437" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16426" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57766" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12236" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01267" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1456" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00930" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H11N4O8P" hasOnlySubstanceUnits="false" id="M_imp_c" metaid="M_imp_c"
      name="IMP C10H11N4O8P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_imp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/imp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:IMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12057" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12063" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13373" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14457" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17202" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19271" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43418" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43475" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43524" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43528" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43563" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43611" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47501" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5849" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00130" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM125" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00114" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H7N" hasOnlySubstanceUnits="false" id="M_indole_c" metaid="M_indole_c" name="Indole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_indole_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/indole" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:INDOLE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14444" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16881" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24794" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35579" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35581" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43537" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5900" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00738" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00463" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM377" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00359" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H7N" hasOnlySubstanceUnits="false" id="M_indole_e" metaid="M_indole_e" name="Indole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_indole_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/indole" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:INDOLE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14444" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16881" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24794" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35579" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35581" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43537" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5900" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00738" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00463" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM377" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00359" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6" hasOnlySubstanceUnits="false" id="M_inost_c" metaid="M_inost_c"
      name="Myo-Inositol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_inost_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/inost" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-8052" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MYO-INOSITOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10601" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12826" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12831" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17268" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19183" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24848" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4200" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43559" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:52772" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00211" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02256" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB34220" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00137" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C19891" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D08079" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM127" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00121" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd21131" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H12N4O5" hasOnlySubstanceUnits="false" id="M_ins_c" metaid="M_ins_c" name="Inosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ins_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ins" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:INOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17596" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5927" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00195" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00294" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00054" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM334" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00246" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H12N4O5" hasOnlySubstanceUnits="false" id="M_ins_e" metaid="M_ins_e" name="Inosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ins_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ins" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:INOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17596" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5927" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00195" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00294" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00054" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM334" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00246" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C5H9O7P2" hasOnlySubstanceUnits="false" id="M_ipdp_c" metaid="M_ipdp_c"
      name="Isopentenyl diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ipdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ipdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DELTA3-ISOPENTENYL-PP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:128769" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14473" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16584" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24907" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6037" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01347" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04196" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00129" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR01010008" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM83" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00113" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H11N4O14P3" hasOnlySubstanceUnits="false" id="M_itp_c" metaid="M_itp_c"
      name="ITP C10H11N4O14P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_itp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/itp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ITP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13374" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16039" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19272" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43508" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57614" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61402" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00081" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM423" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00068" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="K" hasOnlySubstanceUnits="false" id="M_k_c" metaid="M_k_c" name="Potassium">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_k_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/k" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KCL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26219" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29103" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49685" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8345" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00586" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00238" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C12568" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D01168" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D08403" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM95" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00205" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="K" hasOnlySubstanceUnits="false" id="M_k_e" metaid="M_k_e" name="Potassium">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_k_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/k" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KCL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26219" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29103" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49685" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8345" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00586" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00238" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C12568" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D01168" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D08403" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM95" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00205" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-6" fbc:chemicalFormula="C96H170N2O38P2" hasOnlySubstanceUnits="false" id="M_kdo2lipid4L_c"
      metaid="M_kdo2lipid4L_c" name="KDO(2)-lipid IV(A) with laurate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_kdo2lipid4L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/kdo2lipid4L" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25015" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27422" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61524" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61554" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6393" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06251" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMSL02000002" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1870" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03736" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29678" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-6" fbc:chemicalFormula="C84H148N2O37P2" hasOnlySubstanceUnits="false" id="M_kdo2lipid4_c"
      metaid="M_kdo2lipid4_c" name="KDO(2)-lipid IV(A)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_kdo2lipid4_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/kdo2lipid4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDO2-LIPID-IVA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23657" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28526" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4477" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60365" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06025" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G11160" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMSL02000003" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM824" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03586" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-6" fbc:chemicalFormula="C100H176N2O38P2" hasOnlySubstanceUnits="false" id="M_kdo2lipid4p_c"
      metaid="M_kdo2lipid4p_c" name="KDO(2)-lipid IV(A) with palmitoleoyl">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_kdo2lipid4p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/kdo2lipid4p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-2265" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM58983" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15491" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C8H12O11P" hasOnlySubstanceUnits="false" id="M_kdo8p_c" metaid="M_kdo8p_c"
      name="3-Deoxy-D-manno-octulosonate 8-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_kdo8p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/kdo8p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDO-8P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11552" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11788" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1494" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18069" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20006" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42115" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58368" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:85985" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04478" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM146080" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02730" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C8H13O8" hasOnlySubstanceUnits="false" id="M_kdo_c" metaid="M_kdo_c"
      name="3-Deoxy-D-manno-2-octulosonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_kdo_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/kdo" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDO" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43577" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:85986" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM36403" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C76H137N2O30P2" hasOnlySubstanceUnits="false" id="M_kdolipid4_c" metaid="M_kdolipid4_c"
      name="KDO-lipid IV(A)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_kdolipid4_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/kdolipid4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDO-LIPID-IVA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1496" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20009" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27439" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60364" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06024" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G11161" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1206" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03585" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H5O3" hasOnlySubstanceUnits="false" id="M_lac__D_c" metaid="M_lac__D_c"
      name="D-Lactate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lac__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lac__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-LACTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11001" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16004" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18684" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:341" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42105" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42111" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43701" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00171" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01311" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00256" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM285" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00221" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H5O3" hasOnlySubstanceUnits="false" id="M_lac__D_e" metaid="M_lac__D_e"
      name="D-Lactate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lac__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lac__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-LACTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11001" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16004" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18684" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:341" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42105" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42111" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43701" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00171" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01311" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00256" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM285" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00221" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H5O3" hasOnlySubstanceUnits="false" id="M_lac__L_c" metaid="M_lac__L_c"
      name="L-Lactate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lac__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lac__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-LACTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11065" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12411" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18783" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:422" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00190" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03328" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62492" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00186" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM179" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00159" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H5O3" hasOnlySubstanceUnits="false" id="M_lac__L_e" metaid="M_lac__L_e"
      name="L-Lactate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lac__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lac__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-LACTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11065" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12411" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18783" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:422" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00190" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03328" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62492" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00186" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM179" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00159" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H6O2" hasOnlySubstanceUnits="false" id="M_lald__L_c" metaid="M_lald__L_c"
      name="L-Lactaldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lald__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lald__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LACTALD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11015" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11064" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13130" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18041" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18419" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18782" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24994" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:421" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6349" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03052" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00424" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05999" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2387" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00334" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03571" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27393" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_lcts_c" metaid="M_lcts_c"
      name="Lactose C12H22O11">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lcts_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lcts" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15971" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10296" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10380" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10428" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14497" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17716" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22760" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25005" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27755" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27968" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28577" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35463" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36218" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43665" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:613009" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB41627" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00243" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01970" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00046" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10504" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM362" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00208" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01354" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03809" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_lcts_e" metaid="M_lcts_e"
      name="Lactose C12H22O11">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lcts_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lcts" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15971" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10296" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10380" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10428" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14497" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17716" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22760" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25005" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27755" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27968" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28577" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35463" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36218" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43665" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:613009" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB41627" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00243" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01970" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00046" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10504" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM362" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00208" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01354" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03809" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H13NO2" hasOnlySubstanceUnits="false" id="M_leu__L_c" metaid="M_leu__L_c"
      name="L-Leucine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_leu__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/leu__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LEU" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10866" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13131" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15603" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21348" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25017" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32619" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32620" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32627" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32628" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43646" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43695" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43814" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6260" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00687" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62203" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00123" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16439" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00030" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100048" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM140" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00107" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15143" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H13NO2" hasOnlySubstanceUnits="false" id="M_leu__L_e" metaid="M_leu__L_e"
      name="L-Leucine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_leu__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/leu__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LEU" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10866" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13131" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15603" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21348" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25017" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32619" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32620" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32627" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32628" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43646" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43695" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43814" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6260" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00687" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62203" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00123" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16439" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00030" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100048" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM140" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00107" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15143" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C13H20N3O8S" hasOnlySubstanceUnits="false" id="M_lgt__S_c" metaid="M_lgt__S_c"
      name="(R)-S-Lactoylglutathione">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lgt__S_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lgt__S" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:S-LACTOYL-GLUTATHIONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11014" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15694" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18678" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:355" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57474" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01066" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62624" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03451" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1253" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02182" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-6" fbc:chemicalFormula="C110H196N2O39P2" hasOnlySubstanceUnits="false" id="M_lipa_c" metaid="M_lipa_c"
      name="KDO(2)-lipid (A)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lipa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lipa" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23656" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27963" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4476" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58540" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06026" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMSL02000001" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM946" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03587" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-6" fbc:chemicalFormula="C114H202N2O39P2" hasOnlySubstanceUnits="false" id="M_lipa_cold_c" metaid="M_lipa_cold_c"
      name="Cold adapted KDO(2)-lipid (A)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lipa_cold_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lipa_cold" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDO2-LIPID-IVA-COLD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61522" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61556" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM6016" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15493" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C68H126N2O23P2" hasOnlySubstanceUnits="false" id="M_lipidA_c" metaid="M_lipidA_c"
      name="2,3,2'3'-Tetrakis(beta-hydroxymyristoyl)-D-glucosaminyl-1,6-beta-D-glucosamine 1,4'-bisphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lipidA_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lipidA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LIPID-IV-A" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19292" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19294" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28165" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29056" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58603" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:863" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04919" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMSL01040001" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1754" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02993" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C68H127N2O20P" hasOnlySubstanceUnits="false" id="M_lipidAds_c" metaid="M_lipidAds_c"
      name="Lipid A Disaccharide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lipidAds_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lipidAds" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BISOHMYR-GLC" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11414" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11415" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18380" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19293" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58466" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:862" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04932" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1569" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03002" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C34H64NO12P" hasOnlySubstanceUnits="false" id="M_lipidX_c" metaid="M_lipidX_c"
      name="2,3-Bis(3-hydroxytetradecanoyl)-beta-D-glucosaminyl 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lipidX_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lipidX" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BISOHMYR-GLUCOSAMINYL-1P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11413" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16942" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19303" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57957" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:870" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04824" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMSL01020001" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1839" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02930" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29679" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-7" fbc:chemicalFormula="C171H297N4O88P2" hasOnlySubstanceUnits="false" id="M_lps_EC_c" metaid="M_lps_EC_c"
      name="Lipopolysaccharide (Ecoli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lps_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lps_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91640" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15652" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C6H15N2O2" hasOnlySubstanceUnits="false" id="M_lys__L_c" metaid="M_lys__L_c"
      name="L-Lysine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lys__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lys__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LYS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13135" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:133538" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18019" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21351" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32551" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32552" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32563" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32564" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32565" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43950" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6264" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62809" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00047" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16440" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02304" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM78" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00039" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19182" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="C6H15N2O2" hasOnlySubstanceUnits="false" id="M_lys__L_e" metaid="M_lys__L_e"
      name="L-Lysine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_lys__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/lys__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LYS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13135" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:133538" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18019" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21351" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32551" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32552" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32563" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32564" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32565" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43950" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6264" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62809" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00047" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16440" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02304" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM78" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00039" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19182" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C14H22N2O10PRS" hasOnlySubstanceUnits="false" id="M_malACP_c" metaid="M_malACP_c"
      name="Malonyl-[acyl-carrier protein]">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malACP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALONYL-ACP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17330" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6662" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01209" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM184" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11492" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H4O5" hasOnlySubstanceUnits="false" id="M_mal__L_c" metaid="M_mal__L_c"
      name="L-Malate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mal__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mal__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MAL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13140" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15589" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18784" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18785" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30797" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:423" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00156" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00149" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM98" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00130" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H4O5" hasOnlySubstanceUnits="false" id="M_mal__L_e" metaid="M_mal__L_e"
      name="L-Malate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mal__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mal__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MAL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13140" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15589" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18784" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18785" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30797" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:423" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00156" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00149" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM98" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00130" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C24H33N7O19P3S" hasOnlySubstanceUnits="false" id="M_malcoa_c" metaid="M_malcoa_c"
      name="Malonyl CoA C24H33N7O19P3S">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malcoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malcoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALONYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14565" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15531" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25135" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43979" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57384" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6661" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01175" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00083" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050031" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050345" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM40" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00070" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12270" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C12H21O14P" hasOnlySubstanceUnits="false" id="M_malt6p_c" metaid="M_malt6p_c"
      name="Maltose 6'-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malt6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malt6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-1244" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15982" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:111006" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14569" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15703" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25142" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57478" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6669" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6670" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:91177" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02995" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10519" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2394" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01919" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_malt_c" metaid="M_malt_c"
      name="Maltose C12H22O11">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALPHA-MALTOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10300" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14568" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17306" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18167" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22463" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25144" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43893" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6668" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00208" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00897" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00044" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00275" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM165" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00179" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00665" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_malt_e" metaid="M_malt_e"
      name="Maltose C12H22O11">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALPHA-MALTOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10300" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14568" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17306" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18167" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22463" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25144" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43893" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6668" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00208" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00897" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00044" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00275" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM165" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00179" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00665" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C42H72O36" hasOnlySubstanceUnits="false" id="M_malthp_c" metaid="M_malthp_c"
      name="Maltoheptaose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malthp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malthp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-1133" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62010" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB13000" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00689" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM61025" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15494" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C36H62O31" hasOnlySubstanceUnits="false" id="M_malthx_c" metaid="M_malthx_c"
      name="Maltohexaose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malthx_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malthx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOHEXAOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25141" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27445" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61953" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6667" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12253" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01936" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00755" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1317" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01329" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C36H62O31" hasOnlySubstanceUnits="false" id="M_malthx_e" metaid="M_malthx_e"
      name="Maltohexaose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malthx_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malthx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOHEXAOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25141" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27445" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61953" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6667" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12253" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01936" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00755" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1317" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01329" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C30H52O26" hasOnlySubstanceUnits="false" id="M_maltpt_c" metaid="M_maltpt_c"
      name="Maltopentaose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_maltpt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/maltpt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOPENTAOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61952" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62006" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12254" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00343" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1538" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15495" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C30H52O26" hasOnlySubstanceUnits="false" id="M_maltpt_e" metaid="M_maltpt_e"
      name="Maltopentaose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_maltpt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/maltpt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOPENTAOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61952" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62006" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12254" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00343" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1538" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15495" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C18H32O16" hasOnlySubstanceUnits="false" id="M_malttr_c" metaid="M_malttr_c"
      name="Maltotriose C18H32O16">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malttr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malttr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOTRIOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25146" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27931" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43937" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61993" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6672" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01262" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01835" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM468" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01262" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C18H32O16" hasOnlySubstanceUnits="false" id="M_malttr_e" metaid="M_malttr_e"
      name="Maltotriose C18H32O16">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_malttr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/malttr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOTRIOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25146" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27931" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43937" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61993" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6672" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01262" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01835" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM468" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01262" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C24H42O21" hasOnlySubstanceUnits="false" id="M_maltttr_c" metaid="M_maltttr_c"
      name="Maltotetraose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_maltttr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/maltttr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-13205" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-2595" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOTETRAOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25145" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44299" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61988" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62974" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6671" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01296" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02013" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02052" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00459" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM5663" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01376" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01399" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C24H42O21" hasOnlySubstanceUnits="false" id="M_maltttr_e" metaid="M_maltttr_e"
      name="Maltotetraose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_maltttr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/maltttr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-13205" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-2595" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTOTETRAOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25145" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44299" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61988" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62974" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6671" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01296" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02013" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02052" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00459" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM5663" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01376" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01399" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_man1p_c" metaid="M_man1p_c"
      name="D-Mannose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_man1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/man1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MANNOSE-1P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10261" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12327" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18205" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21058" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22404" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35374" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4210" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43854" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58409" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01436" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06330" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00636" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM721" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00485" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_man6p_c" metaid="M_man6p_c"
      name="D-Mannose 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_man6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/man6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15979" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15980" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Mannose-6-phosphate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13000" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17369" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4211" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48042" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58735" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00275" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM427" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00235" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_man6p_e" metaid="M_man6p_e"
      name="D-Mannose 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_man6p_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/man6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15979" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15980" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Mannose-6-phosphate" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13000" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17369" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4211" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48042" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58735" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00275" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM427" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00235" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6" hasOnlySubstanceUnits="false" id="M_man_e" metaid="M_man_e" name="D-Mannose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_man_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/man" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-mannopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14575" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21057" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33930" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37684" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4208" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00169" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00159" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM182" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00138" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27437" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O7" hasOnlySubstanceUnits="false" id="M_mana_c" metaid="M_mana_c"
      name="D-Mannonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mana_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mana" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-MANNONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12998" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17767" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21052" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33076" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4206" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49545" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58654" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00514" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1045" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00403" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_melib_c" metaid="M_melib_c"
      name="Melibiose C12H22O11">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_melib_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/melib" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MELIBIOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25182" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60170" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61827" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6733" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00048" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05402" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00835" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G01275" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1434" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03198" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_melib_e" metaid="M_melib_e"
      name="Melibiose C12H22O11">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_melib_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/melib" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MELIBIOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25182" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28053" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60170" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61827" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6733" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00048" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05402" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00835" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G01275" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1434" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03198" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10NO2S" hasOnlySubstanceUnits="false" id="M_met__D_c" metaid="M_met__D_c"
      name="D-Methionine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_met__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/met__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-218" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13005" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16867" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21065" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32637" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32638" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4215" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44071" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57932" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00855" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1275" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00637" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10NO2S" hasOnlySubstanceUnits="false" id="M_met__D_e" metaid="M_met__D_e"
      name="D-Methionine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_met__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/met__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-218" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13005" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16867" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21065" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32637" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32638" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4215" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44071" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57932" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00855" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1275" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00637" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H11NO2S" hasOnlySubstanceUnits="false" id="M_met__L_c" metaid="M_met__L_c"
      name="L-Methionine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_met__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/met__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MET" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13141" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14590" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16643" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21360" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25229" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32631" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32632" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32644" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32646" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43990" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57844" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6271" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64558" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6829" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00696" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB33951" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00073" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01733" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00019" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04983" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04984" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM61" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00060" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd30746" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H11NO2S" hasOnlySubstanceUnits="false" id="M_met__L_e" metaid="M_met__L_e"
      name="L-Methionine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_met__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/met__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MET" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13141" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14590" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16643" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21360" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25229" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32631" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32632" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32644" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32646" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43990" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57844" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6271" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64558" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6829" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00696" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB33951" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00073" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01733" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00019" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04983" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04984" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM61" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00060" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd30746" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C20H20N7O6" hasOnlySubstanceUnits="false" id="M_methf_c" metaid="M_methf_c"
      name="5,10-Methenyltetrahydrofolate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_methf_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/methf" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-10-METHENYL-THF" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12068" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12069" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15638" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57455" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62623" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00445" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM511" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00347" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_mi1p__D_c" metaid="M_mi1p__D_c"
      name="1D-myo-Inositol 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mi1p__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mi1p__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-MYO-INOSITOL-1-MONOPHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11220" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12828" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12895" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18297" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19207" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43432" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58433" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:818" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01177" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM646" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00867" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C7H7O7" hasOnlySubstanceUnits="false" id="M_micit_c" metaid="M_micit_c"
      name="Methylisocitrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_micit_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/micit" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-618" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10869" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15607" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18550" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57429" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04593" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01050444" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1694" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02799" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C20H21N7O6" hasOnlySubstanceUnits="false" id="M_mlthf_c" metaid="M_mlthf_c"
      name="5,10-Methylenetetrahydrofolate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mlthf_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mlthf" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHYLENE-THF" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10925" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15636" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18602" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1989" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00143" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM183" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00125" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C25H35N7O19P3S" hasOnlySubstanceUnits="false" id="M_mmcoa__R_c" metaid="M_mmcoa__R_c"
      name="(R)-Methylmalonyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mmcoa__R_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mmcoa__R" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHYL-MALONYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10976" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15465" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18654" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:313" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57326" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02255" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01213" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050152" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM608" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00891" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29702" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C25H35N7O19P3S" hasOnlySubstanceUnits="false" id="M_mmcoa__S_c" metaid="M_mmcoa__S_c"
      name="(S)-Methylmalonyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mmcoa__S_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mmcoa__S" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-METHYL-MALONYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11038" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11068" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15466" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18742" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:384" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43874" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57327" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02310" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00683" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050164" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89955" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00519" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29701" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H13O9P" hasOnlySubstanceUnits="false" id="M_mnl1p_c" metaid="M_mnl1p_c"
      name="D-Mannitol 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mnl1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mnl1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MANNITOL-1P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12997" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16298" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21051" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4205" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61381" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01530" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00644" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90126" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00491" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27436" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H14O6" hasOnlySubstanceUnits="false" id="M_mnl_e" metaid="M_mnl_e" name="D-Mannitol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mnl_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mnl" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MANNITOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12996" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:130180" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14574" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16899" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25163" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29864" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6686" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00765" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01363" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00392" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00062" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM615" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00314" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C51H74O2" hasOnlySubstanceUnits="false" id="M_mql8_c" metaid="M_mql8_c"
      name="Menaquinol 8">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mql8_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mql8" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:REDUCED-MENAQUINONE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61684" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM223" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15499" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C51H72O2" hasOnlySubstanceUnits="false" id="M_mqn8_c" metaid="M_mqn8_c"
      name="Menaquinone 8">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mqn8_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mqn8" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-9728" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44027" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM509" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15500" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H4O2" hasOnlySubstanceUnits="false" id="M_mthgxl_c" metaid="M_mthgxl_c"
      name="Methylglyoxal">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_mthgxl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/mthgxl" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-10807" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHYL-GLYOXAL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11643" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14599" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17158" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25303" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6875" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01167" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00546" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM310" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00428" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22830" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C25H47N2O8PRS" hasOnlySubstanceUnits="false" id="M_myrsACP_c" metaid="M_myrsACP_c"
      name="Myristoyl-ACP (n-C14:0ACP)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_myrsACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/myrsACP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89779" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="2" fbc:chemicalFormula="C9H23N3O" hasOnlySubstanceUnits="false" id="M_n8aspmd_c" metaid="M_n8aspmd_c"
      name="N8-Acetylspermidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_n8aspmd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/n8aspmd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-3462" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21900" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27911" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58535" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7420" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02189" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60288" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01029" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1679" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00758" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="Na" hasOnlySubstanceUnits="false" id="M_na1_c" metaid="M_na1_c" name="Sodium">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_na1_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/na1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-2422" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NACL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NaF" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26717" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29101" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49766" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9175" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00588" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01330" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C14042" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04874" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM27" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00971" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="Na" hasOnlySubstanceUnits="false" id="M_na1_e" metaid="M_na1_e" name="Sodium">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_na1_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/na1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-2422" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NACL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NaF" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26717" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29101" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49766" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9175" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00588" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01330" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C14042" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D04874" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM27" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00971" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H4NO2" hasOnlySubstanceUnits="false" id="M_nac_c" metaid="M_nac_c" name="Nicotinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nac_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nac" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NIACINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14650" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15940" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25538" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32544" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7559" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01488" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00253" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00049" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM274" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00218" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H4NO2" hasOnlySubstanceUnits="false" id="M_nac_e" metaid="M_nac_e" name="Nicotinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nac_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nac" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NIACINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14650" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15940" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25538" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32544" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44319" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7559" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01488" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00253" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00049" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM274" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00218" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C21H26N7O14P2" hasOnlySubstanceUnits="false" id="M_nad_c" metaid="M_nad_c"
      name="Nicotinamide adenine dinucleotide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nad_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nad" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NAD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13393" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13394" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21901" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29867" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44214" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44215" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44281" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57540" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7422" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00003" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00002" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM8" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00003" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C21H26N7O14P2" hasOnlySubstanceUnits="false" id="M_nad_e" metaid="M_nad_e"
      name="Nicotinamide adenine dinucleotide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nad_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nad" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NAD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13393" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13394" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21901" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29867" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44214" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44215" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44281" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57540" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7422" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00003" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00002" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM8" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00003" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C21H27N7O14P2" hasOnlySubstanceUnits="false" id="M_nadh_c" metaid="M_nadh_c"
      name="Nicotinamide adenine dinucleotide - reduced">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nadh_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nadh" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NADH" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13395" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13396" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16908" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21902" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44216" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57945" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7423" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01487" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00004" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM10" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00004" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C21H25N7O17P3" hasOnlySubstanceUnits="false" id="M_nadp_c" metaid="M_nadp_c"
      name="Nicotinamide adenine dinucleotide phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nadp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nadp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NADP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18009" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21903" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25523" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29868" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44405" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44409" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58349" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7424" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00006" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM5" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00006" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C21H26N7O17P3" hasOnlySubstanceUnits="false" id="M_nadph_c" metaid="M_nadph_c"
      name="Nicotinamide adenine dinucleotide phosphate - reduced">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nadph_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nadph" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NADPH" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13399" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13400" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16474" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21904" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44286" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57783" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7425" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00221" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00799" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06341" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00005" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM6" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00005" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H6N2O" hasOnlySubstanceUnits="false" id="M_ncam_c" metaid="M_ncam_c"
      name="Nicotinamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ncam_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ncam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NIACINAMIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14645" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17154" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25521" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44258" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7556" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01406" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00153" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00036" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM216" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00133" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="H4N" hasOnlySubstanceUnits="false" id="M_nh4_c" metaid="M_nh4_c" name="Ammonium">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nh4_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nh4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMMONIA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMMONIUM" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13405" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13406" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:135980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13771" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16134" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22533" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22534" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28938" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29337" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44269" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44404" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49783" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7434" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7435" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00051" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB41827" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00014" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01342" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02915" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02916" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM15" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00013" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19013" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="H4N" hasOnlySubstanceUnits="false" id="M_nh4_e" metaid="M_nh4_e" name="Ammonium">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nh4_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nh4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMMONIA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMMONIUM" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13405" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13406" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:135980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13771" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16134" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22533" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22534" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28938" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29337" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29340" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44269" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44404" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49783" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7434" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7435" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00051" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB41827" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00014" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01342" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02915" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02916" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM15" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00013" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19013" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C11H12NO9P" hasOnlySubstanceUnits="false" id="M_nicrnt_c" metaid="M_nicrnt_c"
      name="Nicotinate D-ribonucleotide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nicrnt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nicrnt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NICOTINATE_NUCLEOTIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14652" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15763" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25532" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57502" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7561" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB59646" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01185" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM194" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00873" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C11H14N2O8P" hasOnlySubstanceUnits="false" id="M_nmn_c" metaid="M_nmn_c"
      name="NMN C11H14N2O8P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nmn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nmn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NICOTINAMIDE_NUCLEOTIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10433" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13409" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14646" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14647" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14648" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14649" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16171" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18201" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22850" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25522" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7557" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB59645" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00455" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM355" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00355" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C11H14N2O8P" hasOnlySubstanceUnits="false" id="M_nmn_e" metaid="M_nmn_e"
      name="NMN C11H14N2O8P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_nmn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/nmn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NICOTINAMIDE_NUCLEOTIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10433" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13409" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14646" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14647" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14648" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14649" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16171" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18201" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22850" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25522" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7557" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB59645" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00455" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM355" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00355" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="NO2" hasOnlySubstanceUnits="false" id="M_no2_c" metaid="M_no2_c" name="Nitrite">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_no2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/no2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NITRITE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14658" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16301" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25567" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44396" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7585" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02786" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00088" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM107" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00075" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="NO2" hasOnlySubstanceUnits="false" id="M_no2_e" metaid="M_no2_e" name="Nitrite">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_no2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/no2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NITRITE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14658" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16301" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25567" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44396" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7585" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02786" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00088" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM107" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00075" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="NO3" hasOnlySubstanceUnits="false" id="M_no3_c" metaid="M_no3_c" name="Nitrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_no3_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/no3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15028" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NITRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14654" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17632" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25545" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44487" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:71263" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7580" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01853" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02878" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00244" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02313" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM207" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00209" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="NO3" hasOnlySubstanceUnits="false" id="M_no3_e" metaid="M_no3_e" name="Nitrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_no3_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/no3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15028" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NITRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14654" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17632" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25545" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44487" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:71263" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7580" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01853" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02878" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00244" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02313" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM207" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00209" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="O2" hasOnlySubstanceUnits="false" id="M_o2_c" metaid="M_o2_c" name="O2 O2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_o2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/o2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OXYGEN-MOLECULE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10745" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13416" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15379" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23833" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25366" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26689" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27140" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29097" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29793" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30491" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44742" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7860" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01377" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00007" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00003" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM4" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00007" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="O2" hasOnlySubstanceUnits="false" id="M_o2_e" metaid="M_o2_e" name="O2 O2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_o2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/o2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OXYGEN-MOLECULE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10745" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13416" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15379" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23833" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25366" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26689" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27140" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29097" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29793" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30491" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44742" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7860" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01377" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00007" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00003" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM4" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00007" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="O2" hasOnlySubstanceUnits="false" id="M_o2s_c" metaid="M_o2s_c" name="Superoxide anion">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_o2s_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/o2s" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUPER-OXIDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15143" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18421" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25935" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26839" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30492" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7710" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02168" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00704" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM330" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H2O5" hasOnlySubstanceUnits="false" id="M_oaa_c" metaid="M_oaa_c" name="Oxaloacetate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_oaa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/oaa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ENOL-OXALOACETATE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OXALACETIC_ACID" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12820" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14703" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16452" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24958" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24959" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25731" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25734" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29049" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30744" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7812" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00223" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00036" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03981" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170061" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170120" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM46" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00032" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02469" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C18H35O2" hasOnlySubstanceUnits="false" id="M_ocdca_c" metaid="M_ocdca_c"
      name="Octadecanoate (n-C18:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ocdca_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ocdca" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:STEARIC_ACID" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:231588" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25629" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25631" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28842" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45710" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00827" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01530" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00119" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010018" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010047" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM236" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01080" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C18H35O2" hasOnlySubstanceUnits="false" id="M_ocdca_e" metaid="M_ocdca_e"
      name="Octadecanoate (n-C18:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ocdca_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ocdca" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:STEARIC_ACID" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:231588" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25629" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25631" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28842" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45710" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00827" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01530" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00119" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010018" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01010047" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM236" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01080" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C18H33O2" hasOnlySubstanceUnits="false" id="M_ocdcea_c" metaid="M_ocdcea_c"
      name="Octadecenoate (n-C18:1)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ocdcea_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ocdcea" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OLEATE-CPD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:104361" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14684" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16196" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25663" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25664" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30823" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44741" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7741" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00712" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D02315" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01030002" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01030763" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01030810" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM306" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00536" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C40H65O7P2" hasOnlySubstanceUnits="false" id="M_octdp_c" metaid="M_octdp_c"
      name="All-trans-Octaprenyl diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_octdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/octdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OCTAPRENYL-DIPHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10194" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12781" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16275" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22346" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44585" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:53045" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57711" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01094" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04146" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR01070226" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM811" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02557" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd30761" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C29H53N2O8PRS" hasOnlySubstanceUnits="false" id="M_octeACP_c" metaid="M_octeACP_c"
      name="Cis-octadec-11-enoyl-[acyl-carrier protein] (n-C18:1)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_octeACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/octeACP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89950" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C4H4O8P" hasOnlySubstanceUnits="false" id="M_ohpb_c" metaid="M_ohpb_c"
      name="2-Oxo-3-hydroxy-4-phosphobutanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ohpb_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ohpb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3OH-4P-OH-ALPHA-KETOBUTYRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1240" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19731" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27951" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58538" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06801" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06054" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1514" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03606" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H5O3" hasOnlySubstanceUnits="false" id="M_op4en_c" metaid="M_op4en_c"
      name="2-Oxopent-4-enoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_op4en_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/op4en" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-14447" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-6761" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OXOPENTENOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1113" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48643" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60886" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:67152" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:67181" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00596" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM9829" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00461" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27712" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C5H13N2O2" hasOnlySubstanceUnits="false" id="M_orn_c" metaid="M_orn_c" name="Ornithine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_orn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/orn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ORNITHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11448" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13148" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15729" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18257" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32964" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44667" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46911" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46912" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46913" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46914" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6280" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7784" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00214" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00818" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04984" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62654" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62718" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00077" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01602" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D08302" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100057" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89689" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00064" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19016" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="C5H13N2O2" hasOnlySubstanceUnits="false" id="M_orn_e" metaid="M_orn_e" name="Ornithine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_orn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/orn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-ORNITHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11448" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13148" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15729" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18257" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32964" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44667" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46911" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46912" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46913" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46914" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6280" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7784" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00214" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00818" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB04984" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62654" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62718" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00077" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01602" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D08302" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100057" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89689" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00064" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19016" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C10H10N2O11P" hasOnlySubstanceUnits="false" id="M_orot5p_c" metaid="M_orot5p_c"
      name="Orotidine 5'-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_orot5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/orot5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OROTIDINE-5-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14699" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15842" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25723" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57538" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7788" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00218" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01103" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM519" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00810" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H3N2O4" hasOnlySubstanceUnits="false" id="M_orot_c" metaid="M_orot_c"
      name="Orotate C5H3N2O4">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_orot_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/orot" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OROTATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14698" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16742" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25719" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25720" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30839" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44781" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7787" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00226" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00295" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00055" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM235" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00247" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-100" fbc:chemicalFormula="C1836H3398O400P50" hasOnlySubstanceUnits="false" id="M_pa_EC_c" metaid="M_pa_EC_c"
      name="Phosphatidate  (E.coli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pa_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pa_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11149" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14795" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16337" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18879" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26023" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29089" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57739" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58608" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8122" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00416" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMGP10010000" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM96054" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11422" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C8H7O2" hasOnlySubstanceUnits="false" id="M_pac_c" metaid="M_pac_c"
      name="Phenylacetic acid">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pac_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pac" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHENYLACETATE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Phenyl-Acetates" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14779" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18401" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25975" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25977" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30745" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44686" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8082" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8085" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00209" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01326" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB40733" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00548" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C07086" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C15583" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM497" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00430" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19069" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H8O" hasOnlySubstanceUnits="false" id="M_pacald_c" metaid="M_pacald_c"
      name="Phenylacetaldehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pacald_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pacald" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHENYLACETALDEHYDE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14778" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16424" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43163" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8084" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06236" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00601" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM473" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00464" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C27H51N2O8PRS" hasOnlySubstanceUnits="false" id="M_palmACP_c" metaid="M_palmACP_c"
      name="Palmitoyl-ACP (n-C16:0ACP)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_palmACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/palmACP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2026" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C11H21N2O7PS" hasOnlySubstanceUnits="false" id="M_pan4p_c" metaid="M_pan4p_c"
      name="Pantetheine 4'-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pan4p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pan4p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PANTETHEINE-P" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHOPANTOTHEINE-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11916" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14735" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14736" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14738" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14825" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16858" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25844" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29890" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4222" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45012" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47942" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61723" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7914" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01416" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01134" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM373" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00834" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27792" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H11O4" hasOnlySubstanceUnits="false" id="M_pant__R_c" metaid="M_pant__R_c"
      name="(R)-Pantoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pant__R_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pant__R" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-PANTOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11006" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14737" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18695" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18696" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18697" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:350" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44662" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00522" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM593" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00408" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H11N5O10P2" hasOnlySubstanceUnits="false" id="M_pap_c" metaid="M_pap_c"
      name="Adenosine 3',5'-bisphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pap_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pap" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-5-ADP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13735" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17985" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22240" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2475" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58343" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00054" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM45" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00045" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C10H11N5O13P2S" hasOnlySubstanceUnits="false" id="M_paps_c" metaid="M_paps_c"
      name="3'-Phosphoadenylyl sulfate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_paps_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/paps" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PAPS" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11679" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11680" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1353" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19857" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58339" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62646" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00053" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM49" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00044" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C2086H4048N50O400P50" hasOnlySubstanceUnits="false" id="M_pc_EC_c" metaid="M_pc_EC_c"
      name="Phosphatidylcholine (E.coli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pc_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pc_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM93806" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C8H10NO6P" hasOnlySubstanceUnits="false" id="M_pdx5p_c" metaid="M_pdx5p_c"
      name="Pyridoxine 5'-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pdx5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pdx5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRIDOXINE-5P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26430" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28803" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45202" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58589" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8672" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01319" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00627" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM454" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00478" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C1936H3748N50O400P50" hasOnlySubstanceUnits="false" id="M_pe_EC_c" metaid="M_pe_EC_c"
      name="Phosphatidylethanolamine (ecoli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pe_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pe_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91271" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15653" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C8H12N" hasOnlySubstanceUnits="false" id="M_peamn_c" metaid="M_peamn_c"
      name="Phenethylamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_peamn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/peamn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHENYLETHYLAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14782" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:225237" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25965" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45001" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:50048" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8063" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12275" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60276" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05332" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM660" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03161" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C3H2O6P" hasOnlySubstanceUnits="false" id="M_pep_c" metaid="M_pep_c"
      name="Phosphoenolpyruvate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pep_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pep" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHO-ENOL-PYRUVATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14812" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18021" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26054" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26055" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44894" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44897" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58702" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8147" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00263" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00074" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM73" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00061" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C40H62N8O21" hasOnlySubstanceUnits="false" id="M_peptido_EC_c" metaid="M_peptido_EC_c"
      name="Peptidoglycan subunit of Escherichia coli">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_peptido_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/peptido_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM4770" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15654" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-50" fbc:chemicalFormula="C1986H3748O500P50" hasOnlySubstanceUnits="false" id="M_pg_EC_c" metaid="M_pg_EC_c"
      name="Phosphatidylglycerol (Ecoli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pg_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pg_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90562" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15655" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-150" fbc:chemicalFormula="C1986H3698O650P100" hasOnlySubstanceUnits="false" id="M_pgp_EC_c" metaid="M_pgp_EC_c"
      name="Phosphatidylglycerophosphate (Ecoli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pgp_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pgp_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-1-PHOSPHATIDYL-GLYCEROL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11681" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19930" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37393" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60110" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03892" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMGP05010000" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM871" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11454" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29689" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C29H38N7O17P3S" hasOnlySubstanceUnits="false" id="M_phaccoa_c" metaid="M_phaccoa_c"
      name="Phenylacetyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_phaccoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/phaccoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-207" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14780" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15537" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57390" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8086" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06503" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00582" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM502" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00452" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H11NO2" hasOnlySubstanceUnits="false" id="M_phe__L_c" metaid="M_phe__L_c"
      name="L-Phenylalanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_phe__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/phe__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13151" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17295" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25984" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28044" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32486" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32487" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32504" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32505" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44885" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45079" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6282" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76052" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8089" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00159" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00612" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00079" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02057" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00021" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM97" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00066" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01400" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H11NO2" hasOnlySubstanceUnits="false" id="M_phe__L_e" metaid="M_phe__L_e"
      name="L-Phenylalanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_phe__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/phe__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13151" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17295" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21370" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25984" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28044" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32486" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32487" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32504" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32505" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44885" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45079" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6282" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:76052" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8089" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00159" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00612" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00079" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02057" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00021" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM97" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00066" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01400" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C34H30FeN4O4" hasOnlySubstanceUnits="false" id="M_pheme_c" metaid="M_pheme_c"
      name="Protoheme C34H30FeN4O4">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pheme_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pheme" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROTOHEME" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14957" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17627" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26355" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5651" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60344" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00032" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM249" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H8NO6P" hasOnlySubstanceUnits="false" id="M_phom_c" metaid="M_phom_c"
      name="O-Phospho-L-homoserine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_phom_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/phom" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:O-PHOSPHO-L-HOMOSERINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12693" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12717" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21965" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57590" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7691" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62692" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01102" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1334" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00809" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H7O3" hasOnlySubstanceUnits="false" id="M_phpyr_c" metaid="M_phpyr_c"
      name="Phenylpyruvate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_phpyr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/phpyr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ENOL-PHENYLPYRUVATE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHENYL-PYRUVATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10548" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16815" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23912" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23914" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:23915" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26009" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32815" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45005" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8105" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12225" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00166" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02763" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM210" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00143" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01784" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H8NO7P" hasOnlySubstanceUnits="false" id="M_phthr_c" metaid="M_phthr_c"
      name="O-Phospho-4-hydroxy-L-threonine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_phthr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/phthr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4-PHOSPHONOOXY-THREONINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11945" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18336" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21963" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58452" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59459" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7690" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06802" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06055" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM759" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03607" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd22119" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="HO4P" hasOnlySubstanceUnits="false" id="M_pi_c" metaid="M_pi_c" name="Phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pi_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-16459" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-9010" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHATE-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Pi" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14791" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26078" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29137" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35780" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39739" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39745" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43470" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43474" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7793" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00973" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01429" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02105" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02142" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB05947" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00009" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C13558" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D05467" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM9" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00009" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27787" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="HO4P" hasOnlySubstanceUnits="false" id="M_pi_e" metaid="M_pi_e" name="Phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pi_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-16459" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-9010" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHATE-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Pi" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14791" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26078" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29137" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35780" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39739" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39745" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43470" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43474" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7793" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00973" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01429" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02105" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02142" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB05947" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00009" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C13558" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D05467" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM9" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00009" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27787" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C28H41N7O19P3S" hasOnlySubstanceUnits="false" id="M_pmcoa_c" metaid="M_pmcoa_c"
      name="Pimeloyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pmcoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pmcoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-558" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12210" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14838" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15504" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2176" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26131" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57360" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01063" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050363" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM951" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00782" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H16NO5" hasOnlySubstanceUnits="false" id="M_pnto__R_c" metaid="M_pnto__R_c"
      name="(R)-Pantothenate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pnto__R_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pnto__R" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PANTOTHENATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11008" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14739" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16454" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18700" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18701" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29032" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44679" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46905" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7916" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00210" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62717" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00864" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07413" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM364" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00644" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H16NO5" hasOnlySubstanceUnits="false" id="M_pnto__R_e" metaid="M_pnto__R_e"
      name="(R)-Pantothenate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pnto__R_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pnto__R" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PANTOTHENATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11008" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14739" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16454" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18700" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18701" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:25846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29032" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44679" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46905" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7916" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00210" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62717" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00864" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D07413" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM364" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00644" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H5O2" hasOnlySubstanceUnits="false" id="M_ppa_c" metaid="M_ppa_c"
      name="Propionate (n-C3:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ppa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ppa" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162260" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C3H5O5P" hasOnlySubstanceUnits="false" id="M_ppap_c" metaid="M_ppap_c"
      name="Propanoyl phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ppap_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ppap" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROPIONYL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58933" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8478" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02876" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1499" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01844" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H13N2O4" hasOnlySubstanceUnits="false" id="M_ppbng_c" metaid="M_ppbng_c"
      name="Porphobilinogen">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ppbng_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ppbng" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PORPHOBILINOGEN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14867" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17381" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26212" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44832" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8335" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00245" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00931" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM554" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00689" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C24H36N7O17P3S" hasOnlySubstanceUnits="false" id="M_ppcoa_c" metaid="M_ppcoa_c"
      name="Propanoyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ppcoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ppcoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROPIONYL-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14904" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14907" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15539" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26295" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57392" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8479" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01275" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00100" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050364" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM86" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00086" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12196" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H8O6" hasOnlySubstanceUnits="false" id="M_pphn_c" metaid="M_pphn_c"
      name="Prephenate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pphn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pphn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PREPHENATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14884" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16666" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26256" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26257" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29934" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45028" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8399" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:84387" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12283" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00254" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM503" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00219" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="HO7P2" hasOnlySubstanceUnits="false" id="M_ppi_c" metaid="M_ppi_c" name="Diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ppi_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ppi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PPI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYROPHOSPHATE-GROUP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13420" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18361" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29888" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33017" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33018" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33019" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42009" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45067" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45208" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45212" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8683" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00250" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00013" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM11" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00012" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27828" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C34H32N4O4" hasOnlySubstanceUnits="false" id="M_ppp9_c" metaid="M_ppp9_c"
      name="Protoporphyrin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ppp9_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ppp9" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROTOPORPHYRIN_IX" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14959" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14960" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15430" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26358" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36159" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57306" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8592" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00241" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02191" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM346" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01476" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C34H38N4O4" hasOnlySubstanceUnits="false" id="M_pppg9_c" metaid="M_pppg9_c"
      name="Protoporphyrinogen IX">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pppg9_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pppg9" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROTOPORPHYRINOGEN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14962" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15435" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26359" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26360" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57307" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8593" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01097" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01079" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM351" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00791" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="HO10P3" hasOnlySubstanceUnits="false" id="M_pppi_c" metaid="M_pppi_c"
      name="Inorganic triphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pppi_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pppi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:P3I" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18036" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29203" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39942" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39949" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48313" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48314" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48315" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48316" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5926" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9744" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03379" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12282" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00536" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM332" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00421" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11672" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12039" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H9O2" hasOnlySubstanceUnits="false" id="M_pppn_c" metaid="M_pppn_c"
      name="Phenylpropanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pppn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pppn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-PHENYLPROPIONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20186" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20187" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26005" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28631" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43112" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:51057" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8103" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00764" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05629" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1403" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03343" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H9O2" hasOnlySubstanceUnits="false" id="M_pppn_e" metaid="M_pppn_e"
      name="Phenylpropanoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pppn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pppn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-PHENYLPROPIONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20186" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20187" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26005" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28631" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43112" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:51057" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8103" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00764" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05629" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1403" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03343" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C5H11NO7P" hasOnlySubstanceUnits="false" id="M_pram_c" metaid="M_pram_c"
      name="5-Phospho-beta-D-ribosylamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pram_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pram" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-P-BETA-D-RIBOSYL-AMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12161" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20626" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2123" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37737" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42834" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58681" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01128" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02347" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62576" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03090" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90003" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01982" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C12H13NO9P" hasOnlySubstanceUnits="false" id="M_pran_c" metaid="M_pran_c"
      name="N-(5-Phospho-D-ribosyl)anthranilate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pran_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pran" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-5-PHOSPHORIBOSYL-ANTHRANILATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12431" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12432" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18277" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21472" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21488" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7091" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04302" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1489" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02642" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C15H19N5O14P2" hasOnlySubstanceUnits="false" id="M_prbamp_c" metaid="M_prbamp_c"
      name="1-(5-Phosphoribosyl)-AMP">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_prbamp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/prbamp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHORIBOSYL-AMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11195" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18374" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20628" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37522" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59457" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8166" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12276" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02741" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1515" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01777" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-6" fbc:chemicalFormula="C15H19N5O20P4" hasOnlySubstanceUnits="false" id="M_prbatp_c" metaid="M_prbatp_c"
      name="1-(5-Phosphoribosyl)-ATP">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_prbatp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/prbatp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHORIBOSYL-ATP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11196" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18263" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18970" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58424" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:73183" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:73200" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8167" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03665" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02739" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1351" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01775" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C15H21N5O15P2" hasOnlySubstanceUnits="false" id="M_prfp_c" metaid="M_prfp_c"
      name="1-(5-Phosphoribosyl)-5-[(5-phosphoribosylamino)methylideneamino]imidazole-4-carboxamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_prfp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/prfp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHORIBOSYL-FORMIMINO-AICAR-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11194" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18302" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2020" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20528" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58435" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:75912" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12277" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04896" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1397" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02979" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C15H21N5O15P2" hasOnlySubstanceUnits="false" id="M_prlp_c" metaid="M_prlp_c"
      name="5-[(5-phospho-1-deoxyribulos-1-ylamino)methylideneamino]-1-(5-phosphoribosyl)imidazole-4-carboxamide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_prlp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/prlp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHORIBULOSYL-FORMIMINO-AICAR-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12100" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21471" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27735" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:39698" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58525" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7090" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04916" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1408" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02991" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H9NO2" hasOnlySubstanceUnits="false" id="M_pro__L_c" metaid="M_pro__L_c"
      name="L-Proline">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pro__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pro__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PRO" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13154" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17203" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:184637" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21373" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26271" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32862" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32864" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32871" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32872" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42067" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45040" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45100" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45159" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58054" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60039" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6286" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00162" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00148" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16435" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00035" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00129" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15140" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H9NO2" hasOnlySubstanceUnits="false" id="M_pro__L_e" metaid="M_pro__L_e"
      name="L-Proline">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pro__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pro__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PRO" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13154" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17203" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:184637" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21373" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26271" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32862" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32864" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32871" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32872" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42067" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45040" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45100" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45159" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58054" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60039" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6286" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00162" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00148" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16435" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00035" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00129" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15140" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C5H8O14P3" hasOnlySubstanceUnits="false" id="M_prpp_c" metaid="M_prpp_c"
      name="5-Phospho-alpha-D-ribose 1-diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_prpp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/prpp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PRPP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12159" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12160" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17111" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20625" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:2121" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58017" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62643" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00119" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00103" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-50" fbc:chemicalFormula="C1986H3698N50O500P50" hasOnlySubstanceUnits="false" id="M_ps_EC_c" metaid="M_ps_EC_c"
      name="Phosphatidylserine (Ecoli)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ps_EC_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ps_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM91692" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15657" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C3H6NO6P" hasOnlySubstanceUnits="false" id="M_pser__L_c" metaid="M_pser__L_c"
      name="O-Phospho-L-serine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pser__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pser__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-P-SERINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12718" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21966" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57524" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7692" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00272" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01005" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM379" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00738" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="2" fbc:chemicalFormula="C4H14N2" hasOnlySubstanceUnits="false" id="M_ptrc_c" metaid="M_ptrc_c" name="Putrescine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ptrc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ptrc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PUTRESCINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17148" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26405" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:326268" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45092" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8650" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01414" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60243" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00134" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM118" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00118" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12215" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="2" fbc:chemicalFormula="C4H14N2" hasOnlySubstanceUnits="false" id="M_ptrc_e" metaid="M_ptrc_e" name="Putrescine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ptrc_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ptrc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PUTRESCINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17148" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26405" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:326268" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45092" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8650" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01414" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB60243" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00134" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM118" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00118" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd12215" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C8H12N2O5P" hasOnlySubstanceUnits="false" id="M_pyam5p_c" metaid="M_pyam5p_c"
      name="Pyridoxamine 5'-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pyam5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pyam5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRIDOXAMINE-5P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14979" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14980" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18335" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45037" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8670" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01555" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00647" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM366" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00493" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C8H13N2O2" hasOnlySubstanceUnits="false" id="M_pydam_c" metaid="M_pydam_c"
      name="Pyridoxamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pydam_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pydam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRIDOXAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131533" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16410" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26426" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45228" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57761" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8669" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01431" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62696" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00534" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM548" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00419" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C8H8NO6P" hasOnlySubstanceUnits="false" id="M_pydx5p_c" metaid="M_pydx5p_c"
      name="Pyridoxal 5'-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pydx5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pydx5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRIDOXAL_PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14977" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18405" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:234571" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26424" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:358848" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45145" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:597326" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8668" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01491" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00018" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM161" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00016" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H9NO3" hasOnlySubstanceUnits="false" id="M_pydx_c" metaid="M_pydx_c" name="Pyridoxal">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pydx_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pydx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRIDOXAL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131530" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14976" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17310" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26423" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45112" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8667" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01545" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00250" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM311" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00215" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C8H11NO3" hasOnlySubstanceUnits="false" id="M_pydxn_c" metaid="M_pydxn_c"
      name="Pyridoxine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pydxn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pydxn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRIDOXINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14981" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16709" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26429" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8671" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00239" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02075" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00314" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D08454" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM419" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00263" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H3O3" hasOnlySubstanceUnits="false" id="M_pyr_c" metaid="M_pyr_c" name="Pyruvate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pyr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pyr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRUVATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15361" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26462" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26466" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32816" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45253" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:86354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8685" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00243" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62676" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00022" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060077" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM23" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00020" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H3O3" hasOnlySubstanceUnits="false" id="M_pyr_e" metaid="M_pyr_e" name="Pyruvate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_pyr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/pyr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRUVATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14987" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15361" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26462" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26466" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32816" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45253" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:86354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8685" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00243" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62676" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00022" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01060077" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM23" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00020" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C49H74O4" hasOnlySubstanceUnits="false" id="M_q8_c" metaid="M_q8_c" name="Ubiquinone-8">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_q8_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/q8" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UBIQUINONE-8" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61683" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C17569" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR02010005" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM232" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15560" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C49H76O4" hasOnlySubstanceUnits="false" id="M_q8h2_c" metaid="M_q8h2_c"
      name="Ubiquinol-8">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_q8h2_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/q8h2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-9956" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61682" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01060" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM191" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15561" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29608" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H3NO4" hasOnlySubstanceUnits="false" id="M_quln_c" metaid="M_quln_c"
      name="Quinolinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_quln_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/quln" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:QUINOLINATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132942" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14975" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16675" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26417" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26418" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29959" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44529" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46828" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46832" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46833" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8663" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00232" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03722" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM555" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02333" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O8P" hasOnlySubstanceUnits="false" id="M_r1p_c" metaid="M_r1p_c"
      name="Alpha-D-Ribose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_r1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/r1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBOSE-1P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10269" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12329" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12330" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16300" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22412" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45429" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45482" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57720" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01489" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00620" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM295" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00475" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O8P" hasOnlySubstanceUnits="false" id="M_r5p_c" metaid="M_r5p_c"
      name="Alpha-D-Ribose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_r5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/r5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15318" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10270" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12331" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22413" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03736" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM15900" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19028" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O5" hasOnlySubstanceUnits="false" id="M_rbl__L_c" metaid="M_rbl__L_c"
      name="L-Ribulose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rbl__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rbl__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-RIBULOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13163" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16880" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21382" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6295" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00508" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM685" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00397" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C14H18N2O4" hasOnlySubstanceUnits="false" id="M_rdmbzi_c" metaid="M_rdmbzi_c"
      name="N1-(alpha-D-ribosyl)-5,6-dimethylbenzimidazole">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rdmbzi_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rdmbzi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALPHA-RIBAZOLE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10329" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11112" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05775" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1539" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03423" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H17NO6S" hasOnlySubstanceUnits="false" id="M_rhcys_c" metaid="M_rhcys_c"
      name="S-Ribosyl-L-homocysteine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rhcys_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rhcys" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-564" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12750" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17575" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22071" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45383" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58195" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8968" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:90364" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03539" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1777" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02227" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O5" hasOnlySubstanceUnits="false" id="M_rib__D_c" metaid="M_rib__D_c"
      name="D-Ribose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rib__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rib__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Ribofuranose" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Ribopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13011" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16988" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21078" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4233" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47006" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47013" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00283" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00121" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C21057" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM242" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00105" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O5" hasOnlySubstanceUnits="false" id="M_rib__D_e" metaid="M_rib__D_e"
      name="D-Ribose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rib__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rib__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Ribofuranose" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Ribopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13011" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16988" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21078" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4233" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46999" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47006" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47013" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00283" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00121" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C21057" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM242" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00105" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C17H20N4O6" hasOnlySubstanceUnits="false" id="M_ribflv_c" metaid="M_ribflv_c"
      name="Riboflavin C17H20N4O6">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ribflv_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ribflv" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBOFLAVIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15044" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17015" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27299" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45214" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:529204" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:8843" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00244" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00255" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00050" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM270" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00220" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O8P" hasOnlySubstanceUnits="false" id="M_rml1p_c" metaid="M_rml1p_c"
      name="L-Rhamnulose 1-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rml1p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rml1p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RHAMNULOSE-1P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13162" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17892" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21381" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58313" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6294" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01131" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1675" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00831" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O5" hasOnlySubstanceUnits="false" id="M_rml_c" metaid="M_rml_c" name="L-Rhamnulose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rml_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rml" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-16566" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13161" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17897" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21380" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58316" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6293" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB10207" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00861" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1750" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00642" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O5" hasOnlySubstanceUnits="false" id="M_rmn_c" metaid="M_rmn_c" name="L-Rhamnose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rmn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rmn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15405" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-rhamnopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-rhamnose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13160" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16055" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21378" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26546" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57622" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62345" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62346" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6292" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00849" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00507" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2967" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00396" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27379" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H12O5" hasOnlySubstanceUnits="false" id="M_rmn_e" metaid="M_rmn_e" name="L-Rhamnose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_rmn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/rmn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15405" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-rhamnopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-rhamnose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13160" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16055" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21378" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26546" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57622" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62345" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62346" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6292" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00849" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00507" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2967" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00396" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27379" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O8P" hasOnlySubstanceUnits="false" id="M_ru5p__D_c" metaid="M_ru5p__D_c"
      name="D-Ribulose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ru5p__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ru5p__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBULOSE-5P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13018" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13040" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17363" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21088" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26572" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37455" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:40192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4243" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58121" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00618" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02033" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02694" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00199" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM145" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00171" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O8P" hasOnlySubstanceUnits="false" id="M_ru5p__L_c" metaid="M_ru5p__L_c"
      name="L-Ribulose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ru5p__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ru5p__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-RIBULOSE-5-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13164" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17666" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21383" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58226" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6296" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01101" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM901" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00808" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H13O10P" hasOnlySubstanceUnits="false" id="M_s7p_c" metaid="M_s7p_c"
      name="Sedoheptulose 7-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_s7p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/s7p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-SEDOHEPTULOSE-7-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15073" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15074" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15721" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26621" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57483" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9083" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01068" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62754" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05382" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM271" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00238" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H13O9P" hasOnlySubstanceUnits="false" id="M_sbt6p_c" metaid="M_sbt6p_c"
      name="D-Sorbitol 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sbt6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sbt6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-SORBITOL-6-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13021" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17044" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17116" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21093" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45426" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60084" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9202" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB05831" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01096" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM393" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00804" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C6H14O6" hasOnlySubstanceUnits="false" id="M_sbt__D_e" metaid="M_sbt__D_e"
      name="D-Sorbitol">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sbt__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sbt__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SORBITOL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12954" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13020" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15093" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17220" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17924" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21091" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26724" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26726" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30911" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33795" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33796" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4246" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45559" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:87978" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9201" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00247" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00794" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00096" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM469" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00588" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C32H39N7O20P3S" hasOnlySubstanceUnits="false" id="M_sbzcoa_c" metaid="M_sbzcoa_c"
      name="O-Succinylbenzoyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sbzcoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sbzcoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-6972" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12700" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1279" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12836" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15509" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19779" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57364" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03160" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1132" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02021" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-7" fbc:chemicalFormula="C42H39N4O16" hasOnlySubstanceUnits="false" id="M_scl_c" metaid="M_scl_c"
      name="Sirohydrochlorin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_scl_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/scl" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SIROHYDROCHLORIN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15090" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18023" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26691" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58351" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9167" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05778" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM863" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03426" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="HSe" hasOnlySubstanceUnits="false" id="M_seln_c" metaid="M_seln_c" name="Selenide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_seln_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/seln" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-678" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SE-2" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15076" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16503" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24638" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29317" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30485" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30486" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47675" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9089" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB11110" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01528" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM92652" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01078" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28148" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="H2O3PSe" hasOnlySubstanceUnits="false" id="M_selnp_c" metaid="M_selnp_c"
      name="Selenophosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_selnp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/selnp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SEPO3" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15078" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16144" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29269" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58618" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64331" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9101" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03840" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06407" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05172" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1339" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03078" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO3" hasOnlySubstanceUnits="false" id="M_ser__D_c" metaid="M_ser__D_c"
      name="D-Serine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ser__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ser__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-SERINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13019" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:143888" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16523" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21090" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32840" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42262" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4245" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03406" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00740" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM694" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00550" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO3" hasOnlySubstanceUnits="false" id="M_ser__D_e" metaid="M_ser__D_e"
      name="D-Serine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ser__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ser__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-SERINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13019" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:143888" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16523" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21090" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32840" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32841" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42262" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4245" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03406" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00740" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM694" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00550" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO3" hasOnlySubstanceUnits="false" id="M_ser__L_c" metaid="M_ser__L_c"
      name="L-Serine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ser__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ser__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SER" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Serines" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13167" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15081" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17115" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17822" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21387" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26648" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32836" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32837" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32845" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33384" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35243" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45440" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45590" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45597" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45677" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6301" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9116" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00187" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00589" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62263" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00065" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00716" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00016" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100045" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM53" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00054" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28171" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd30743" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H7NO3" hasOnlySubstanceUnits="false" id="M_ser__L_e" metaid="M_ser__L_e"
      name="L-Serine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ser__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ser__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SER" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Serines" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13167" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15081" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17115" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17822" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21387" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26648" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32836" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32837" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32845" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32846" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33384" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35243" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45440" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45451" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45590" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45597" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45677" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6301" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9116" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00187" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00589" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62263" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00065" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00716" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00016" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100045" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM53" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00054" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28171" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd30743" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C13H19N6O9P" hasOnlySubstanceUnits="false" id="M_seramp_c" metaid="M_seramp_c"
      name="L-seryl-AMP">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_seramp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/seramp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SERYL-AMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:286" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61231" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61645" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05820" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM114390" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03452" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15564" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-8" fbc:chemicalFormula="C42H36FeN4O16" hasOnlySubstanceUnits="false" id="M_sheme_c" metaid="M_sheme_c"
      name="Siroheme C42H36FeN4O16">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sheme_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sheme" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SIROHEME" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26690" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28599" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:60052" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9166" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00748" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM82173" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00557" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C7H8O8P" hasOnlySubstanceUnits="false" id="M_skm5p_c" metaid="M_skm5p_c"
      name="Shikimate 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_skm5p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/skm5p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SHIKIMATE-5P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11886" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:145989" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15084" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17052" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20195" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9134" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03175" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1265" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02030" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C7H9O5" hasOnlySubstanceUnits="false" id="M_skm_c" metaid="M_skm_c" name="Shikimate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_skm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/skm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SHIKIMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15083" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16119" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26662" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26663" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26664" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:36208" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45740" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9133" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03070" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00493" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM602" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00383" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C11H16N2O7" hasOnlySubstanceUnits="false" id="M_sl26da_c" metaid="M_sl26da_c"
      name="N-Succinyl-LL-2,6-diaminoheptanedioate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sl26da_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sl26da" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-SUCCINYLLL-2-6-DIAMINOPIMELATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12617" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12618" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17279" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21791" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21792" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58087" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7342" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12267" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04421" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1488" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02698" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C11H12NO8" hasOnlySubstanceUnits="false" id="M_sl2a6o_c" metaid="M_sl2a6o_c"
      name="N-Succinyl-2-L-amino-6-oxoheptanedioate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sl2a6o_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sl2a6o" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-SUCCINYL-2-AMINO-6-KETOPIMELATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10967" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12616" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15685" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21789" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21790" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35266" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7340" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12266" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04462" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1563" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02724" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="O3S" hasOnlySubstanceUnits="false" id="M_so3_c" metaid="M_so3_c" name="Sulfite">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_so3_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/so3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:H2SO3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HSO3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SO3" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17137" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17359" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26837" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29214" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33543" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45548" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48854" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5598" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9344" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00240" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01033" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03008" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB34829" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00094" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C11481" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM105630" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00081" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="O4S" hasOnlySubstanceUnits="false" id="M_so4_c" metaid="M_so4_c" name="Sulfate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_so4_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/so4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HSO4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SULFATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15135" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26836" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29199" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45687" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45693" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45696" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9335" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01448" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00059" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D05963" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM58" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00048" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="O4S" hasOnlySubstanceUnits="false" id="M_so4_e" metaid="M_so4_e" name="Sulfate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_so4_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/so4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HSO4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SULFATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15135" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26836" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29199" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45687" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45693" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45696" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9335" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01448" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00059" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D05963" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM58" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00048" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="3" fbc:chemicalFormula="C7H22N3" hasOnlySubstanceUnits="false" id="M_spmd_c" metaid="M_spmd_c" name="Spermidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_spmd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/spmd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SPERMIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15097" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16610" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26732" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45647" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57834" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9218" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01257" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00315" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM124" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00264" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="3" fbc:chemicalFormula="C7H22N3" hasOnlySubstanceUnits="false" id="M_spmd_e" metaid="M_spmd_e" name="Spermidine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_spmd_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/spmd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SPERMIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15097" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16610" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26732" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26733" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45647" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57834" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9218" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01257" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00315" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM124" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00264" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C16H21N4O10P2S" hasOnlySubstanceUnits="false" id="M_ssaltpp_c" metaid="M_ssaltpp_c"
      name="Succinate semialdehyde-thiamin diphosphate anion">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ssaltpp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ssaltpp" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM92663" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03450" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C12H21O14P" hasOnlySubstanceUnits="false" id="M_suc6p_c" metaid="M_suc6p_c"
      name="Sucrose 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_suc6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/suc6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15716" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCROSE-6P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:131603" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15129" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15130" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26813" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57723" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:91002" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9315" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02591" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16688" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10508" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1004" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01693" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19187" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C10H17N4O5" hasOnlySubstanceUnits="false" id="M_sucarg_c" metaid="M_sucarg_c"
      name="N2-Succinyl-L-arginine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sucarg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sucarg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-421" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12637" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17705" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21819" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58241" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7372" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03296" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1756" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02103" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C11H8O5" hasOnlySubstanceUnits="false" id="M_sucbz_c" metaid="M_sucbz_c"
      name="O-Succinylbenzoate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sucbz_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sucbz" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:O-SUCCINYLBENZOATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:1278" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12835" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18325" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19778" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37026" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44787" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:44788" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02730" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1212" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01772" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H4O4" hasOnlySubstanceUnits="false" id="M_succ_c" metaid="M_succ_c" name="Succinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_succ_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/succ" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUC" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132287" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15741" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26803" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26807" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30031" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30779" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45639" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:90372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9304" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00254" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00042" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170043" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM25" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00036" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H4O4" hasOnlySubstanceUnits="false" id="M_succ_e" metaid="M_succ_e" name="Succinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_succ_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/succ" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUC" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:132287" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15741" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22943" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26803" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26807" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30031" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30779" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45639" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:90372" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9304" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00254" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00042" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01170043" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM25" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00036" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-5" fbc:chemicalFormula="C25H35N7O19P3S" hasOnlySubstanceUnits="false" id="M_succoa_c" metaid="M_succoa_c"
      name="Succinyl-CoA">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_succoa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/succoa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUC-COA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10746" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15127" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15380" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26811" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45541" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57292" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9310" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01022" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00091" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA07050370" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM92" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00078" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C9H10NO7" hasOnlySubstanceUnits="false" id="M_sucglu_c" metaid="M_sucglu_c"
      name="N2-Succinyl-L-glutamate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sucglu_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sucglu" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N2-SUCCINYLGLUTAMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21821" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48957" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58763" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7373" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05931" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1757" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03525" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H11NO6" hasOnlySubstanceUnits="false" id="M_sucgsa_c" metaid="M_sucgsa_c"
      name="N2-Succinyl-L-glutamate 5-semialdehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sucgsa_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sucgsa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-822" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21820" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27657" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58520" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7374" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05932" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1763" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03526" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C8H12NO6" hasOnlySubstanceUnits="false" id="M_suchms_c" metaid="M_suchms_c"
      name="O-Succinyl-L-homoserine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_suchms_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/suchms" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:O-SUCCINYL-L-HOMOSERINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12699" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12723" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16160" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21975" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57661" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7704" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01118" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM820" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00822" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C9H15N2O5" hasOnlySubstanceUnits="false" id="M_sucorn_c" metaid="M_sucorn_c"
      name="N2-Succinyl-L-ornithine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sucorn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sucorn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N2-SUCCINYLORNITHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21822" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27574" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45896" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58514" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7375" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01199" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03415" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1152" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02160" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_sucr_e" metaid="M_sucr_e"
      name="Sucrose C12H22O11">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sucr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sucr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCROSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15128" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17992" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26812" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45795" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9314" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00258" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00089" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00025" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00370" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM167" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00076" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C4H5O3" hasOnlySubstanceUnits="false" id="M_sucsal_c" metaid="M_sucsal_c"
      name="Succinic semialdehyde">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_sucsal_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/sucsal" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCC-S-ALD" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16265" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26804" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26805" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57706" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9305" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01259" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00232" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA06000118" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM172" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00199" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P" hasOnlySubstanceUnits="false" id="M_tag6p__D_c" metaid="M_tag6p__D_c"
      name="D-Tagatose 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tag6p__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tag6p__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TAGATOSE-6-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13025" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17837" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21097" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4251" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58695" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06873" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01097" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM795" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00805" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C6H10O12P2" hasOnlySubstanceUnits="false" id="M_tagdp__D_c" metaid="M_tagdp__D_c"
      name="D-Tagatose 1,6-biphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tagdp__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tagdp__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TAGATOSE-1-6-DIPHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13024" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16743" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21096" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4250" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58694" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06872" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03785" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1324" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02371" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29681" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C6H9O7" hasOnlySubstanceUnits="false" id="M_tagur_c" metaid="M_tagur_c"
      name="D-Tagaturonate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tagur_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tagur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-TAGATURONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13026" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17886" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21098" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21099" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4252" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58493" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:59451" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00558" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA05000654" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1423" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00437" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H4O6" hasOnlySubstanceUnits="false" id="M_tartr__L_c" metaid="M_tartr__L_c"
      name="L-tartrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tartr__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tartr__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TARTRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11018" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15193" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15671" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18710" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18711" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26849" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30924" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:358" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45866" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00956" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00898" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00103" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM816" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00666" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="C4H4O6" hasOnlySubstanceUnits="false" id="M_tartr__L_e" metaid="M_tartr__L_e"
      name="L-tartrate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tartr__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tartr__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TARTRATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10961" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11018" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15193" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15671" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18710" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18711" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26849" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30924" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:35398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:358" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45866" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00956" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00898" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00103" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM816" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00666" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H7NO3S" hasOnlySubstanceUnits="false" id="M_taur_c" metaid="M_taur_c"
      name="Taurine C2H7NO3S">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_taur_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/taur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TAURINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15195" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15891" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32970" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45877" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:507393" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9406" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00251" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00245" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00047" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM282" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00210" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C2H7NO3S" hasOnlySubstanceUnits="false" id="M_taur_e" metaid="M_taur_e"
      name="Taurine C2H7NO3S">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_taur_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/taur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TAURINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15195" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15891" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32970" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45877" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:507393" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9406" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00251" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00245" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00047" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM282" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00210" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="CNS" hasOnlySubstanceUnits="false" id="M_tcynt_c" metaid="M_tcynt_c" name="Thiocyanate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tcynt_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tcynt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-19512" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HSCN" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18022" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24926" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:24928" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26954" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26956" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29200" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45576" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9550" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01453" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01755" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM762" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01211" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C25H45N2O8PRS" hasOnlySubstanceUnits="false" id="M_tdeACP_c" metaid="M_tdeACP_c"
      name="Cis-tetradec-7-enoyl-[acyl-carrier protein] (n-C14:1)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tdeACP_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tdeACP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162426" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C7H7NO4" hasOnlySubstanceUnits="false" id="M_thdp_c" metaid="M_thdp_c"
      name="2,3,4,5-Tetrahydrodipicolinate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DELTA1-PIPERIDEINE-2-6-DICARBOXYLATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11408" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13042" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16845" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18055" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32976" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6152" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:864" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12289" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03972" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM480" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02465" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29199" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C19H21N7O6" hasOnlySubstanceUnits="false" id="M_thf_c" metaid="M_thf_c"
      name="5,6,7,8-Tetrahydrofolate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thf_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thf" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THF" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10931" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12075" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15220" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15635" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18606" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:20506" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46069" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57453" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68437" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9482" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62806" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00101" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM79" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00087" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C12H17N4OS" hasOnlySubstanceUnits="false" id="M_thm_c" metaid="M_thm_c" name="Thiamin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thm_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THIAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18385" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46393" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9530" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00235" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00378" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM322" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00305" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="C12H17N4OS" hasOnlySubstanceUnits="false" id="M_thm_e" metaid="M_thm_e" name="Thiamin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thm_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thm" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THIAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18385" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26941" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46393" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9530" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00235" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00378" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM322" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00305" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C12H16N4O4PS" hasOnlySubstanceUnits="false" id="M_thmmp_c" metaid="M_thmmp_c"
      name="Thiamin monophosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thmmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thmmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THIAMINE-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15230" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15231" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26945" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37574" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37575" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46189" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9533" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02666" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01081" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM662" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00793" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C12H16N4O7P2S" hasOnlySubstanceUnits="false" id="M_thmpp_c" metaid="M_thmpp_c"
      name="Thiamine diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thmpp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thmpp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THIAMINE-PYROPHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15229" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45930" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45931" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49939" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58937" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9532" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62636" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00068" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM256" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00056" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H9NO3" hasOnlySubstanceUnits="false" id="M_thr__L_c" metaid="M_thr__L_c"
      name="L-Threonine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thr__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thr__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THR" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13175" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16857" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21403" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32820" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32822" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32832" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32833" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42083" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45843" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45983" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57926" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6308" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00167" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00188" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00041" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM142" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00161" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H9NO3" hasOnlySubstanceUnits="false" id="M_thr__L_e" metaid="M_thr__L_e"
      name="L-Threonine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thr__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thr__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THR" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13175" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16857" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21403" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26986" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32820" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32822" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32832" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32833" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42083" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45843" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45983" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57926" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6308" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00167" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00188" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00041" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM142" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00161" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H6N2O2" hasOnlySubstanceUnits="false" id="M_thym_c" metaid="M_thym_c"
      name="Thymine C5H6N2O2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thym_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thym" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THYMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15247" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17821" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27004" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46017" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9580" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00262" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00178" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM386" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00151" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H14N2O5" hasOnlySubstanceUnits="false" id="M_thymd_c" metaid="M_thymd_c"
      name="Thymidine C10H14N2O5">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thymd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thymd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THYMIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19273" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45782" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45834" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45917" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45918" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:53527" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9579" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00273" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00214" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM420" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00184" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H14N2O5" hasOnlySubstanceUnits="false" id="M_thymd_e" metaid="M_thymd_e"
      name="Thymidine C10H14N2O5">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_thymd_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/thymd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THYMIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15244" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17748" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:19273" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45782" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45834" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45917" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45918" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:53527" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9579" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00273" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00214" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM420" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00184" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="1" fbc:chemicalFormula="C3H10N" hasOnlySubstanceUnits="false" id="M_tma_c" metaid="M_tma_c"
      name="Trimethylamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tma_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tma" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRIMETHYLAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15261" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27127" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58389" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9732" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00906" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00565" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM352" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00441" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="1" fbc:chemicalFormula="C3H10N" hasOnlySubstanceUnits="false" id="M_tma_e" metaid="M_tma_e"
      name="Trimethylamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tma_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tma" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRIMETHYLAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15261" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18139" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27127" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58389" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9732" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00906" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00565" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM352" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00441" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H9NO" hasOnlySubstanceUnits="false" id="M_tmao_c" metaid="M_tmao_c"
      name="Trimethylamine N-oxide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tmao_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tmao" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRIMETHYLAMINE-N-O" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15262" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15263" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15724" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9733" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00925" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01104" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM437" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00811" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C3H9NO" hasOnlySubstanceUnits="false" id="M_tmao_e" metaid="M_tmao_e"
      name="Trimethylamine N-oxide">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tmao_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tmao" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRIMETHYLAMINE-N-O" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15262" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15263" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15724" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9733" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00925" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01104" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM437" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00811" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="X" hasOnlySubstanceUnits="false" id="M_trdox_c" metaid="M_trdox_c"
      name="Oxidized thioredoxin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_trdox_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/trdox" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Ox-Thioredoxin" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:14726" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15240" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18191" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:7845" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00343" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM148" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11420" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd27735" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29682" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="XH2" hasOnlySubstanceUnits="false" id="M_trdrd_c" metaid="M_trdrd_c"
      name="Reduced thioredoxin">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_trdrd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/trdrd" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:Red-Thioredoxin" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15033" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00342" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM96993" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11421" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28060" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C12H21O14P" hasOnlySubstanceUnits="false" id="M_tre6p_c" metaid="M_tre6p_c"
      name="Alpha,alpha'-Trehalose 6-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tre6p_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tre6p" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TREHALOSE-6P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10201" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15252" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18283" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22364" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58429" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01124" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00689" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G09795" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM448" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00523" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_tre_c" metaid="M_tre_c" name="Trehalose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tre_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tre" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15990" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TREHALOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10202" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12281" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12287" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15251" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16551" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22365" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27082" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46211" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00975" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01083" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00293" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM198" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00794" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C12H22O11" hasOnlySubstanceUnits="false" id="M_tre_e" metaid="M_tre_e" name="Trehalose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tre_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tre" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-15990" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TREHALOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10202" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12281" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:12287" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15251" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16551" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22365" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27082" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46211" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00975" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01083" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G00293" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM198" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00794" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="R" hasOnlySubstanceUnits="false" id="M_trnaglu_c" metaid="M_trnaglu_c" name="TRNA (Glu)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_trnaglu_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/trnaglu" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLT-tRNAs" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10680" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15175" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29175" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01641" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM90886" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd11912" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C11H12N2O2" hasOnlySubstanceUnits="false" id="M_trp__L_c" metaid="M_trp__L_c"
      name="L-Tryptophan">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_trp__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/trp__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13178" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16828" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:184633" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27163" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27897" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32702" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32704" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32727" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32728" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45988" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46086" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46225" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57912" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6310" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64554" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9769" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00929" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB30396" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00078" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00806" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00020" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM94" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00065" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19007" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C11H12N2O2" hasOnlySubstanceUnits="false" id="M_trp__L_e" metaid="M_trp__L_e"
      name="L-Tryptophan">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_trp__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/trp__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13178" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16828" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:184633" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21407" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27163" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27897" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32702" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32704" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32727" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32728" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45988" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46086" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46225" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57912" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6310" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:64554" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9769" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00929" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB30396" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00078" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00806" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00020" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM94" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00065" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd19007" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="O3S2" hasOnlySubstanceUnits="false" id="M_tsul_c" metaid="M_tsul_c" name="Thiosulfate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tsul_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tsul" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:S2O3" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26977" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29279" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33539" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33540" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33541" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33542" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45922" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5587" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9569" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00257" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00320" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05529" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM323" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00268" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-2" fbc:chemicalFormula="O3S2" hasOnlySubstanceUnits="false" id="M_tsul_e" metaid="M_tsul_e" name="Thiosulfate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tsul_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tsul" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:S2O3" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15242" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16094" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:26977" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:29279" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33539" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33540" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33541" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:33542" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:45922" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:5587" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9569" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00257" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00320" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05529" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM323" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00268" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C14H27O2" hasOnlySubstanceUnits="false" id="M_ttdca_c" metaid="M_ttdca_c"
      name="Tetradecanoate (n-C14:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ttdca_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ttdca" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162239" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="-1" fbc:chemicalFormula="C14H27O2" hasOnlySubstanceUnits="false" id="M_ttdca_e" metaid="M_ttdca_e"
      name="Tetradecanoate (n-C14:0)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ttdca_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ttdca" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162239" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C14H25O2" hasOnlySubstanceUnits="false" id="M_ttdcea_c" metaid="M_ttdcea_c"
      name="Tetradecenoate (n-C14:1)">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ttdcea_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ttdcea" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM641" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H11NO3" hasOnlySubstanceUnits="false" id="M_tyr__L_c" metaid="M_tyr__L_c"
      name="L-Tyrosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tyr__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tyr__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TYR" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13181" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15277" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17895" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18186" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21411" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27176" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32760" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32761" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32784" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32785" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32786" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46070" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46161" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58315" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6313" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9800" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00158" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00647" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00082" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01536" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00022" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM76" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00069" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd30745" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H11NO3" hasOnlySubstanceUnits="false" id="M_tyr__L_e" metaid="M_tyr__L_e"
      name="L-Tyrosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_tyr__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/tyr__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TYR" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13181" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15277" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17895" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18186" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21411" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27176" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32760" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32761" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32784" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32785" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32786" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46070" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46161" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58315" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6313" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9800" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00158" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00647" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00082" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01536" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00022" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM76" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00069" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd30745" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C43H75N3O20P2" hasOnlySubstanceUnits="false" id="M_u23ga_c" metaid="M_u23ga_c"
      name="UDP-2,3-bis(3-hydroxytetradecanoyl)glucosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_u23ga_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/u23ga" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OH-MYRISTOYL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:78847" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:82845" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04652" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1199" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02835" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C31H51N3O19P2" hasOnlySubstanceUnits="false" id="M_u3aga_c" metaid="M_u3aga_c"
      name="UDP-3-O-(3-hydroxytetradecanoyl)-N-acetylglucosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_u3aga_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/u3aga" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-OHMYR-ACETYLGLUCOSAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61494" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61537" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04738" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2157" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02886" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C29H50N3O18P2" hasOnlySubstanceUnits="false" id="M_u3hga_c" metaid="M_u3hga_c"
      name="UDP-3-O-(3-hydroxytetradecanoyl)-D-glucosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_u3hga_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/u3hga" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-OHMYR-GLUCOSAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22095" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27392" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58510" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:71573" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:71619" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9807" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06022" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2036" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03584" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C95H152N8O28P2" hasOnlySubstanceUnits="false" id="M_uaagmda_c" metaid="M_uaagmda_c"
      name="Undecaprenyl-diphospho-N-acetylmuramoyl-(N-acetylglucosamine)-L-ala-D-glu-meso-2,6-diaminopimeloyl-D-ala-D-ala">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uaagmda_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uaagmda" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1242" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03495" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C20H26N3O19P2" hasOnlySubstanceUnits="false" id="M_uaccg_c" metaid="M_uaccg_c"
      name="UDP-N-acetyl-3-O-(1-carboxyvinyl)-D-glucosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uaccg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uaccg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-ACETYL-CARBOXYVINYL-GLUCOSAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13466" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13467" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16435" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22110" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57771" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68483" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68507" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9818" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04631" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1688" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02820" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C17H25N3O17P2" hasOnlySubstanceUnits="false" id="M_uacgam_c" metaid="M_uacgam_c"
      name="UDP-N-acetyl-D-glucosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uacgam_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uacgam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-N-ACETYL-D-GLUCOSAMINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13473" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13475" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13476" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16264" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22115" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46243" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46287" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57705" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9823" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62760" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00043" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10610" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMSL01010002" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM47" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00037" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C17H25N3O17P2" hasOnlySubstanceUnits="false" id="M_uacmam_c" metaid="M_uacmam_c"
      name="UDP-N-acetyl-D-mannosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uacmam_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uacmam" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-MANNAC" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13457" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13474" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16287" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22116" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57715" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68623" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:68678" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9824" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB13112" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01170" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G11112" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM932" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00861" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C17H22N3O18P2" hasOnlySubstanceUnits="false" id="M_uacmamu_c" metaid="M_uacmamu_c"
      name="UDP-N-acetyl-D-mannosaminouronate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uacmamu_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uacmamu" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-MANNACA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22117" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28581" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58578" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:70731" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:70738" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9825" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C06240" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM97052" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03732" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C87H139N7O23P2" hasOnlySubstanceUnits="false" id="M_uagmda_c" metaid="M_uagmda_c"
      name="Undecaprenyl-diphospho-N-acetylmuramoyl-L-alanyl-D-glutamyl-meso-2,6-diaminopimeloyl-D-alanyl-D-alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uagmda_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uagmda" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:C5" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61387" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61543" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C05897" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10556" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2264" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd03494" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C23H33N4O20P2" hasOnlySubstanceUnits="false" id="M_uama_c" metaid="M_uama_c"
      name="UDP-N-acetylmuramoyl-L-alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uama_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uama" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD0-1456" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13459" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13478" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16932" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22123" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57953" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:83898" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:84726" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9830" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01212" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1953" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00890" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd29609" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C28H39N5O23P2" hasOnlySubstanceUnits="false" id="M_uamag_c" metaid="M_uamag_c"
      name="UDP-N-acetylmuramoyl-L-alanyl-D-glutamate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uamag_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uamag" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-AA-GLUTAMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13461" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13482" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16970" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22125" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46143" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57968" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:83900" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9832" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00692" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM955" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00525" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C20H28N3O19P2" hasOnlySubstanceUnits="false" id="M_uamr_c" metaid="M_uamr_c"
      name="UDP-N-acetylmuramate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uamr_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uamr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-N-ACETYLMURAMATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13458" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13477" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17882" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22120" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58309" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:70757" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:70765" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9827" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01050" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM785" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00773" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C55H89O7P2" hasOnlySubstanceUnits="false" id="M_udcpdp_c" metaid="M_udcpdp_c"
      name="Undecaprenyl diphosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_udcpdp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/udcpdp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-9649" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15284" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17047" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27192" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:53042" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57995" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9863" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01469" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR03030004" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM5043" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd25775" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C55H89O4P" hasOnlySubstanceUnits="false" id="M_udcpp_c" metaid="M_udcpp_c"
      name="Undecaprenyl phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_udcpp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/udcpp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UNDECAPRENYL-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15286" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16141" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57654" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9864" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00348" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMPR03020006" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM251" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00286" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C9H11N2O12P2" hasOnlySubstanceUnits="false" id="M_udp_c" metaid="M_udp_c"
      name="UDP C9H11N2O12P2">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_udp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/udp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13445" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17659" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27230" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46402" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58223" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9802" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00015" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10619" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM17" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00014" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C15H22N2O17P2" hasOnlySubstanceUnits="false" id="M_udpg_c" metaid="M_udpg_c"
      name="UDPglucose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_udpg_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/udpg" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-12575" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13498" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13505" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22103" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27234" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28532" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42885" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46228" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46229" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:52249" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58367" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58885" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9845" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9895" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00029" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10608" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM52" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00026" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd28293" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C15H22N2O17P2" hasOnlySubstanceUnits="false" id="M_udpgal_c" metaid="M_udpgal_c"
      name="UDPgalactose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_udpgal_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/udpgal" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-14553" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13487" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13495" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18307" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22100" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:42751" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58439" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:66914" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:67119" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9811" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62573" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00052" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10609" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM89795" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00043" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd30770" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C15H22N2O17P2" hasOnlySubstanceUnits="false" id="M_udpgalfur_c" metaid="M_udpgalfur_c"
      name="UDP-D-galacto-1,4-furanose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_udpgalfur_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/udpgalfur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-D-GALACTO-14-FURANOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:66915" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:67118" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03733" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2158" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02338" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C15H19N2O18P2" hasOnlySubstanceUnits="false" id="M_udpglcur_c" metaid="M_udpglcur_c"
      name="UDP-D-glucuronate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_udpglcur_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/udpglcur" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-GLUCURONATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13489" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13506" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17200" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22104" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46309" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58052" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9846" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00167" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.glycan/G10612" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM87" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00144" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C35H51N7O26P2" hasOnlySubstanceUnits="false" id="M_ugmd_c" metaid="M_ugmd_c"
      name="UDP-N-acetylmuramoyl-L-alanyl-D-gamma-glutamyl-meso-2,6-diaminopimelate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ugmd_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ugmd" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13463" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13481" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22124" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:28639" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:30905" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:58582" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9831" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM162592" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C41H61N9O28P2" hasOnlySubstanceUnits="false" id="M_ugmda_c" metaid="M_ugmda_c"
      name="UDP-N-acetylmuramoyl-L-alanyl-D-glutamyl-meso-2,6-diaminopimeloyl-D-alanyl-D-alanine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ugmda_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ugmda" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:C1" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13462" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18199" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:22126" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61386" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9833" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C04882" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1954" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02932" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02968" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C9H11N2O9P" hasOnlySubstanceUnits="false" id="M_ump_c" metaid="M_ump_c"
      name="UMP C9H11N2O9P">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ump_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ump" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UMP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13508" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13509" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16695" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27231" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46362" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46382" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46385" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:47721" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57865" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9849" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00105" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM80" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00091" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C63H103NO12P2" hasOnlySubstanceUnits="false" id="M_unaga_c" metaid="M_unaga_c"
      name="Undecaprenyl diphospho N-acetyl-glucosamine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_unaga_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/unaga" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYL-D-GLUCOSAMINYLDIPHOSPHO-UNDECAPRE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62958" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62959" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01289" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1333" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00946" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C71H113N2O18P2" hasOnlySubstanceUnits="false" id="M_unagamu_c" metaid="M_unagamu_c"
      name="Undecaprenyl-diphospho-N-acetylglucosamine-N-acetylmannosaminuronate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_unagamu_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/unagamu" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:C55-PP-GLCNAC-MANNACA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61495" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:61561" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:62960" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB12306" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C20497" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM2504" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15581" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-3" fbc:chemicalFormula="C79H126N3O22P2" hasOnlySubstanceUnits="false" id="M_unagamuf_c" metaid="M_unagamuf_c"
      name="Undecaprenyl-diphospho N-acetylglucosamine-N-acetylmannosaminuronate-N-acetamido-4,6-dideoxy-D-galactose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_unagamuf_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/unagamuf" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:C55-PP-GLCNAC-MANNACA-FUC4NAC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM6160" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15582" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-8" fbc:chemicalFormula="C40H36N4O16" hasOnlySubstanceUnits="false" id="M_uppg3_c" metaid="M_uppg3_c"
      name="Uroporphyrinogen III">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uppg3_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uppg3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UROPORPHYRINOGEN-III" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15300" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15437" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27257" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57308" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9905" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01086" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01051" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM414" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00774" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H4N2O2" hasOnlySubstanceUnits="false" id="M_ura_c" metaid="M_ura_c" name="Uracil">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ura_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ura" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:URACIL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15288" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17568" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27210" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43254" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46375" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9882" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00300" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00106" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00027" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D09776" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM158" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00092" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C4H4N2O2" hasOnlySubstanceUnits="false" id="M_ura_e" metaid="M_ura_e" name="Uracil">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_ura_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/ura" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:URACIL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15288" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17568" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27210" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:43254" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46375" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9882" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00300" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00106" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00027" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D09776" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM158" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00092" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-1" fbc:chemicalFormula="C3H5N2O4" hasOnlySubstanceUnits="false" id="M_urdglyc_c" metaid="M_urdglyc_c"
      name="(-)-Ureidoglycolate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_urdglyc_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/urdglyc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPD-1091" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10783" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:11076" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:121" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15412" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18499" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18804" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27224" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:49050" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57296" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9891" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01005" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00603" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C02766" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM490" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00465" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01786" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="CH4N2O" hasOnlySubstanceUnits="false" id="M_urea_c" metaid="M_urea_c" name="Urea CH4N2O">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_urea_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/urea" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UREA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:134711" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15292" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16199" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27218" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46379" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48376" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9888" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00294" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00086" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00023" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D01749" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D10496" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM117" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00073" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="CH4N2O" hasOnlySubstanceUnits="false" id="M_urea_e" metaid="M_urea_e" name="Urea CH4N2O">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_urea_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/urea" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UREA" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:134711" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15292" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16199" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27218" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32285" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46379" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48376" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9888" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00294" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00086" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00023" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D01749" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D10496" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM117" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00073" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H12N2O6" hasOnlySubstanceUnits="false" id="M_uri_c" metaid="M_uri_c" name="Uridine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uri_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uri" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:URIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15296" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16704" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46386" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46391" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46463" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9893" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00296" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00299" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM288" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00249" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C9H12N2O6" hasOnlySubstanceUnits="false" id="M_uri_e" metaid="M_uri_e" name="Uridine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_uri_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/uri" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:URIDINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15296" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16704" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27227" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46386" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46391" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46460" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46463" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9893" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00296" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00299" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM288" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00249" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-4" fbc:chemicalFormula="C9H11N2O15P3" hasOnlySubstanceUnits="false" id="M_utp_c" metaid="M_utp_c"
      name="UTP C9H11N2O15P3">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_utp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/utp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UTP" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13510" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15713" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27233" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:37567" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46397" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46398" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57481" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:9850" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00075" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM121" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00062" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H11NO2" hasOnlySubstanceUnits="false" id="M_val__L_c" metaid="M_val__L_c"
      name="L-Valine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_val__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/val__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:VAL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13186" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16414" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21417" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27266" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32859" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32860" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46282" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46376" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46418" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46484" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:87977" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00883" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB34366" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00183" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16436" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00039" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100046" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM199" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00156" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15141" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H11NO2" hasOnlySubstanceUnits="false" id="M_val__L_e" metaid="M_val__L_e"
      name="L-Valine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_val__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/val__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:VAL" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13186" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16414" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21417" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27266" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32851" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32852" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32859" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:32860" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46282" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46376" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46418" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46484" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57762" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6321" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:87977" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00883" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB34366" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00183" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C16436" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D00039" />
                  <rdf:li rdf:resource="http://identifiers.org/lipidmaps/LMFA01100046" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM199" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00156" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd15141" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H4N4O2" hasOnlySubstanceUnits="false" id="M_xan_c" metaid="M_xan_c" name="Xanthine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xan_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xan" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XANTHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10059" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15318" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17712" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27317" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46377" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48517" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00292" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00385" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM174" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00309" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H4N4O2" hasOnlySubstanceUnits="false" id="M_xan_e" metaid="M_xan_e" name="Xanthine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xan_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xan" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XANTHINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10059" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15318" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17712" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27317" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46377" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:48517" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00292" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00385" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM174" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00309" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C10H11N4O9P" hasOnlySubstanceUnits="false" id="M_xmp_c" metaid="M_xmp_c"
      name="Xanthosine 5'-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xmp_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xmp" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XANTHOSINE-5-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10067" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10938" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15324" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15652" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27328" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57464" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB62755" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00655" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM298" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00497" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H12N4O6" hasOnlySubstanceUnits="false" id="M_xtsn_c" metaid="M_xtsn_c"
      name="Xanthosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xtsn_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xtsn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XANTHOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15323" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27327" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00299" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01762" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM687" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01217" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C10H12N4O6" hasOnlySubstanceUnits="false" id="M_xtsn_e" metaid="M_xtsn_e"
      name="Xanthosine">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xtsn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xtsn" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XANTHOSINE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:10066" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:15323" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:18107" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27327" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00299" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C01762" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM687" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd01217" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O8P" hasOnlySubstanceUnits="false" id="M_xu5p__D_c" metaid="M_xu5p__D_c"
      name="D-Xylulose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xu5p__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xu5p__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XYLULOSE-5-PHOSPHATE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13036" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16332" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21121" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27354" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4269" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57737" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00868" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB06212" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00231" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM186" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00198" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="-2" fbc:chemicalFormula="C5H9O8P" hasOnlySubstanceUnits="false" id="M_xu5p__L_c" metaid="M_xu5p__L_c"
      name="L-Xylulose 5-phosphate">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xu5p__L_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xu5p__L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-XYLULOSE-5-P" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13191" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:16593" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21427" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:57829" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6328" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C03291" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM1238" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd02100" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O5" hasOnlySubstanceUnits="false" id="M_xyl__D_c" metaid="M_xyl__D_c"
      name="D-Xylose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xyl__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xyl__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Xylopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:53455" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:65327" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00098" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03763" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00181" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D06346" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM348" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00154" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="e" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O5" hasOnlySubstanceUnits="false" id="M_xyl__D_e" metaid="M_xyl__D_e"
      name="D-Xylose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xyl__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xyl__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-Xylopyranose" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:53455" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:65327" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00098" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB03763" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00181" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.drug/D06346" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM348" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00154" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
      <species boundaryCondition="false" compartment="c" constant="false" fbc:charge="0" fbc:chemicalFormula="C5H10O5" hasOnlySubstanceUnits="false" id="M_xylu__D_c" metaid="M_xylu__D_c"
      name="D-Xylulose">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#M_xylu__D_c">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.metabolite/xylu__D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-XYLULOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-XYLULOSE" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13035" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:13190" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17140" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:17399" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21120" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:21425" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:27353" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:4268" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:46514" />
                  <rdf:li rdf:resource="http://identifiers.org/chebi/CHEBI:6326" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00654" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB00751" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB01644" />
                  <rdf:li rdf:resource="http://identifiers.org/hmdb/HMDB02245" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00310" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.compound/C00312" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.chemical/MNXM597" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00259" />
                  <rdf:li rdf:resource="http://identifiers.org/seed.compound/cpd00261" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
      </species>
    </listOfSpecies>
    <listOfParameters>
      <parameter constant="true" id="cobra_default_lb" name="cobra default - lb" sboTerm="SBO:0000626" units="mmol_per_gDW_per_hr" value="-999999" />
      <parameter constant="true" id="cobra_default_ub" name="cobra default - ub" sboTerm="SBO:0000626" units="mmol_per_gDW_per_hr" value="999999" />
      <parameter constant="true" id="cobra_0_bound" name="cobra 0 - bound" sboTerm="SBO:0000625" units="mmol_per_gDW_per_hr" value="0" />
      <parameter constant="true" id="R_ATPM_upper_bound" name="R ATPM upper - bound" sboTerm="SBO:0000625" units="mmol_per_gDW_per_hr" value="7.6" />
      <parameter constant="true" id="R_ATPM_lower_bound" name="R ATPM lower - bound" sboTerm="SBO:0000625" units="mmol_per_gDW_per_hr" value="7.6" />
      <parameter constant="true" id="R_EX_glc__D_e_lower_bound" name="R EX glc-D e lower - bound" sboTerm="SBO:0000625" units="mmol_per_gDW_per_hr" value="-10" />
      <parameter constant="true" id="R_EX_o2_e_lower_bound" name="R EX o2 e lower - bound" sboTerm="SBO:0000625" units="mmol_per_gDW_per_hr" value="-20" />
    </listOfParameters>
    <listOfReactions>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_12PPDt" metaid="R_12PPDt" name="S-Propane-1,2-diol facilitated transport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_12PPDt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/12PPDt" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR94684" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_12ppd__S_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_12ppd__S_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_2DGLCNRx" metaid="R_2DGLCNRx" name="2-dehydro-D-gluconate reductase (NADH)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_2DGLCNRx">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/2DGLCNRx" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR94783" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27910" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27911" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27912" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27913" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3553" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dhglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_2DGLCNRy" metaid="R_2DGLCNRy" name="2-dehydro-D-gluconate reductase (NADPH)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_2DGLCNRy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/2DGLCNRy" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:1.1.1.215-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.215" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01739" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR94784" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16653" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16654" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16655" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16656" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3553" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dhglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_2DGULRx" metaid="R_2DGULRx" name="2-dehydro-L-gulonate reductase (NADH)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_2DGULRx">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/2DGULRx" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR94787" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3553" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dhguln_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_idon__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_2DGULRy" metaid="R_2DGULRy" name="2-dehydro-L-gulonate reductase (NADPH)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_2DGULRy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/2DGULRy" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR94788" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3553" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dhguln_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_idon__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_3HCINNMH" metaid="R_3HCINNMH" name="3-hydroxycinnamate hydroxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_3HCINNMH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/3HCINNMH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-10040" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.14.13.127" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06787" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR94893" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27846" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27847" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27848" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27849" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0347" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3hcinnm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhcinnm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_3HPPPNH" metaid="R_3HPPPNH" name="3-(3-hydroxy-phenyl)propionate hydroxylase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_3HPPPNH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/3HPPPNH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MHPHYDROXY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.14.13.127" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06786" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR94903" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24785" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24786" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24787" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24788" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0347" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3hpppn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhpppn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_4HTHRS" metaid="R_4HTHRS" name="4-Hydroxy-L-threonine synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_4HTHRS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/4HTHRS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14125" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.3.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05086" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95024" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30659" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30660" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30661" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30662" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0004" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_phthr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4hthr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_5DGLCNR" metaid="R_5DGLCNR" name="5-dehydro-D-gluconate reductase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_5DGLCNR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/5DGLCNR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14367" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.69" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01740" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95065" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23936" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23937" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23938" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23939" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4266" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5dglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_A5PISO" metaid="R_A5PISO" name="Arabinose-5-phosphate isomerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_A5PISO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/A5PISO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DARAB5PISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01530" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95132" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23104" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23105" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23106" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23107" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ru5p__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ara5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AACPS1" metaid="R_AACPS1" name="Acyl-[acyl-carrier-protein] synthetase (n-C14:0)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AACPS1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AACPS1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95138" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2836" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ttdca_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_myrsACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AACPS2" metaid="R_AACPS2" name="Acyl-[acyl-carrier-protein] synthetase  (n-C14:1)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AACPS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AACPS2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95141" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2836" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ttdcea_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tdeACP_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AACPS3" metaid="R_AACPS3" name="Acyl-[acyl-carrier-protein] synthetase  (n-C16:0)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AACPS3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AACPS3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95142" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2836" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hdca_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_palmACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AACPS4" metaid="R_AACPS4" name="Acyl-[acyl-carrier-protein] synthetase  (n-C16:1)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AACPS4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AACPS4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95143" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2836" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hdcea_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdeACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AACPS5" metaid="R_AACPS5" name="Acyl-[acyl-carrier-protein] synthetase  (n-C18:1)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AACPS5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AACPS5" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95144" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2836" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ocdcea_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_octeACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AADDGT" metaid="R_AADDGT" name="DTDP-N-4-acetamido-4,6-dideoxy-D-galactose transferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AADDGT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AADDGT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FUC4NACTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.325" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95151" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3793" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtdp4aaddg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_unagamu_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_unagamuf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AB6PGH" metaid="R_AB6PGH" name="Arbutin 6-phosphate glucohydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AB6PGH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AB6PGH" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.86" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95165" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2901" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arbt6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hqn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ABTA" metaid="R_ABTA" name="4-aminobutyrate transaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ABTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ABTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GABATRANSAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01648" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95186" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23352" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23353" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23354" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23355" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2662" />
            <fbc:geneProductRef fbc:geneProduct="G_b1302" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4abut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sucsal_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ABUTD" metaid="R_ABUTD" name="Aminobutyraldehyde dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ABUTD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ABUTD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMINOBUTDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02549" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95191" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19105" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19106" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19107" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19108" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4abutn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4abut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ABUTt2r" metaid="R_ABUTt2r"
      name="4 aminobutyrate reversible transport in via proton symport" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ABUTt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ABUTt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-57" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR94994" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28915" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28916" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28917" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28918" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2663" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4abut_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4abut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACACCT" metaid="R_ACACCT" name="Acetyl-CoA:acetoacetyl-CoA transferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACACCT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACACCT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETOACETYL-COA-TRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.8.3.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.8.3.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01359" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95193" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27806" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27807" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27808" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27809" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2222" />
            <fbc:geneProductRef fbc:geneProduct="G_b2221" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aacoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACACT1r" metaid="R_ACACT1r" name="Acetyl-CoA C-acetyltransferase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACACT1r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACACT1r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYL-COA-ACETYLTRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00238" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95194" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21036" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21037" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21038" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21039" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2224" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aacoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACACt2" metaid="R_ACACt2" name="Acetoacetate transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACACt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACACt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95207" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2223" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acac_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACALD" metaid="R_ACALD" name="Acetaldehyde dehydrogenase (acetylating)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACALD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACALD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETALD-DEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00228" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95210" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23288" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23289" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23290" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23291" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0351" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACALDt" metaid="R_ACALDt" name="Acetaldehyde reversible transport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACALDt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACALDt" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95212" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acald_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acald_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACBIPGT" metaid="R_ACBIPGT" name="Adenosyl cobinamide phosphate guanyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACBIPGT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACBIPGT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:COBINPGUANYLYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.62" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05222" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95216" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22712" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22713" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22714" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22715" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1993" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adocbip_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_agdpcbi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACCOAC" metaid="R_ACCOAC" name="Acetyl-CoA carboxylase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACCOAC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACCOAC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYL-COA-CARBOXYLTRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.4.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00742" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95219" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11308" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11309" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11310" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11311" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3256" />
            <fbc:geneProductRef fbc:geneProduct="G_b0185" />
            <fbc:geneProductRef fbc:geneProduct="G_b3255" />
            <fbc:geneProductRef fbc:geneProduct="G_b2316" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hco3_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malcoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACCOAL" metaid="R_ACCOAL" name="Acetate-CoA ligase (ADP-forming)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACCOAL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACCOAL" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.2.1.13" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95221" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0335" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppcoa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACGAMT" metaid="R_ACGAMT"
      name="UDP-N-acetylglucosamine:undecaprenylphosphate N-acetylglucosamine -1-phosphate transferase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACGAMT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACGAMT" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95251" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3784" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uacgam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udcpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ump_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_unaga_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACGApts" metaid="R_ACGApts" name="N-Acetyl-D-glucosamine transport via PEP:Pyr PTS"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACGApts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACGApts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95253" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2416" />
              <fbc:geneProductRef fbc:geneProduct="G_b2415" />
              <fbc:geneProductRef fbc:geneProduct="G_b2417" />
              <fbc:geneProductRef fbc:geneProduct="G_b1101" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2416" />
              <fbc:geneProductRef fbc:geneProduct="G_b2415" />
              <fbc:geneProductRef fbc:geneProduct="G_b0679" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acgam_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acgam6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACGK" metaid="R_ACGK" name="Acetylglutamate kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACGK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACGK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYLGLUTKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02649" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95256" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14629" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14630" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14631" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14632" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3959" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acglu_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acg5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACGS" metaid="R_ACGS" name="N-acetylglutamate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACGS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACGS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-ACETYLTRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00259" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95259" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24292" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24293" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24294" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24295" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2818" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acglu_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACHBS" metaid="R_ACHBS" name="2-aceto-2-hydroxybutanoate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACHBS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACHBS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETOOHBUTSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.2.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R08648" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95260" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27654" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27655" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27656" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27657" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0077" />
              <fbc:geneProductRef fbc:geneProduct="G_b0078" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3670" />
              <fbc:geneProductRef fbc:geneProduct="G_b3671" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3769" />
              <fbc:geneProductRef fbc:geneProduct="G_b3768" />
              <fbc:geneProductRef fbc:geneProduct="G_b3767" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2obut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ahbut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACKr" metaid="R_ACKr" name="Acetate kinase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACKr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACKr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETATEKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.2.15" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00315" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95269" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11352" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11353" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11354" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11355" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3115" />
            <fbc:geneProductRef fbc:geneProduct="G_b2296" />
            <fbc:geneProductRef fbc:geneProduct="G_b1849" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_actp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACLS" metaid="R_ACLS" name="Acetolactate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACLS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACLS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETOLACTSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.2.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00226" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95271" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25249" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25250" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25251" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25252" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0077" />
              <fbc:geneProductRef fbc:geneProduct="G_b0078" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3670" />
              <fbc:geneProductRef fbc:geneProduct="G_b3671" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3769" />
              <fbc:geneProductRef fbc:geneProduct="G_b3768" />
              <fbc:geneProductRef fbc:geneProduct="G_b3767" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_alac__S_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACMAMUT" metaid="R_ACMAMUT" name="UDP-N-acetyl-D-mannosaminuronic acid transferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACMAMUT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACMAMUT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPMANACATRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.180" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04477" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95274" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28366" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28367" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28368" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28369" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3794" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uacmamu_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_unaga_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_unagamu_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACMANApts" metaid="R_ACMANApts" name="N-acetyl-D-mannosamine transport via PTS"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACMANApts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACMANApts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95275" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1817" />
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b1818" />
            <fbc:geneProductRef fbc:geneProduct="G_b1819" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acmana_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acmanap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACNAMt2" metaid="R_ACNAMt2" name="N-acetylneuraminate proton symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACNAMt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACNAMt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-25" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95291" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28987" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28988" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28989" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28990" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3224" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acnam_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acnam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACNML" metaid="R_ACNML" name="N-Acetylneuraminate lyase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACNML">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACNML" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACNEULY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01811" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95296" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23296" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23297" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23298" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23299" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3225" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acnam_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acmana_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACOATA" metaid="R_ACOATA" name="Acetyl-CoA ACP transacylase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACOATA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACOATA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACP-S-ACETYLTRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.179" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.180" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.38" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.41" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.85" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.86" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01624" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95376" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1091" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACODA" metaid="R_ACODA" name="Acetylornithine deacetylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACODA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACODA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYLORNDEACET-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.14" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00669" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95377" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15941" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15942" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15943" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15944" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3957" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acorn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_orn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACONMT" metaid="R_ACONMT" name="Trans-aconitate methyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACONMT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACONMT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-2441" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.1.144" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05763" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95383" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14969" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14970" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14971" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14972" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1519" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acon_T_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amet_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aconm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ahcys_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACONT" metaid="R_ACONT" name="Aconitate hydratase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACONT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACONT" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01324" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95384" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10336" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10337" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10338" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10339" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1276" />
            <fbc:geneProductRef fbc:geneProduct="G_b0118" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cit_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_icit_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACOTA" metaid="R_ACOTA" name="Acetylornithine transaminase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACOTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACOTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETYLORNTRANSAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.11" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02283" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95388" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18049" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18050" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18051" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18052" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3359" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acorn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acg5sa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACPS1" metaid="R_ACPS1" name="Acyl-carrier protein synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACPS1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACPS1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.8.7" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95403" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2563" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_apoACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pap_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ACS" metaid="R_ACS" name="Acetyl-CoA synthetase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETATE--COA-LIGASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.2.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00235" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95413" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23176" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23177" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23178" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23179" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4069" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ACt2r" metaid="R_ACt2r" name="Acetate reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ACt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ACt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-571" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95429" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ac_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADA" metaid="R_ADA" name="Adenosine deaminase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENODEAMIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01560" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95432" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24408" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24409" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24410" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24411" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1623" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ins_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADCL" metaid="R_ADCL" name="4-aminobenzoate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADCL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADCL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADCLY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.38" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05553" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95436" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16201" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16202" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16203" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16204" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1096" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4adcho_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4abz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADCS" metaid="R_ADCS" name="4-amino-4-deoxychorismate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADCS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADCS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PABASYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.85" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01716" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95440" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11672" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11673" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11674" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11675" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3360" />
            <fbc:geneProductRef fbc:geneProduct="G_b1812" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_chor_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4adcho_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADD" metaid="R_ADD" name="Adenine deaminase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENINE-DEAMINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01244" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95443" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23688" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23689" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23690" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23691" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3665" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ade_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hxan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADEt2" metaid="R_ADEt2" name="Adenine transport in via proton symport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADEt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADEt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-198" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-447" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95446" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28719" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28720" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28721" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28722" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3654" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ade_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ade_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADHEr" metaid="R_ADHEr" name="Acetaldehyde dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADHEr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADHEr" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95226" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1241" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_etoh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADK1" metaid="R_ADK1" name="Adenylate kinase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENYL-KIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00127" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95450" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12973" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12974" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12975" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12976" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0474" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADK3" metaid="R_ADK3" name="Adentylate kinase (GTP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADK3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADK3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14074" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.10" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95452" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29863" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29864" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29865" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29866" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0474" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADK4" metaid="R_ADK4" name="Adentylate kinase (ITP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADK4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADK4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95453" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29867" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29868" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29869" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29870" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0474" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_itp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_idp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADMDC" metaid="R_ADMDC" name="Adenosylmethionine decarboxylase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADMDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADMDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SAMDECARB-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.50" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00178" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95455" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15981" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15982" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15983" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15984" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0120" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amet_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ametam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADNCYC" metaid="R_ADNCYC" name="Adenylate cyclase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADNCYC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADNCYC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENYLATECYC-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.6.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00089" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95444" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15389" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15390" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15391" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15392" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3806" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_camp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADNK1" metaid="R_ADNK1" name="Adenosine kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADNK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADNK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSINE-KINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.74" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00185" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95456" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20824" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20825" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20826" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20827" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0474" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADNt2_copy1" metaid="R_ADNt2_copy1" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADNt2_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108A" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-203" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95461" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29987" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29988" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29989" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29990" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or>
            <fbc:geneProductRef fbc:geneProduct="G_b2964" />
            <fbc:geneProductRef fbc:geneProduct="G_b2393" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADNt2_copy2" metaid="R_ADNt2_copy2" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADNt2_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108A" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-203" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95461" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29987" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29988" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29989" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29990" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2406" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADOCBIK" metaid="R_ADOCBIK" name="Adenosyl cobinamide kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADOCBIK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADOCBIK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:COBINAMIDEKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.156" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05221" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95465" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1993" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adocbi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adocbip_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADOCBLS" metaid="R_ADOCBLS" name="Adenosylcobalamin 5'-phosphate synthase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADOCBLS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADOCBLS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:COBALAMINSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.8.26" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05223" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95469" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1992" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_agdpcbi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rdmbzi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adocbl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADPT" metaid="R_ADPT" name="Adenine phosphoribosyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADPT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADPT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENPRIBOSYLTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.7" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00190" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95482" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16609" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16610" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16611" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16612" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0469" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ade_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADSK" metaid="R_ADSK" name="Adenylyl-sulfate kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADSK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADSK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENYLYLSULFKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.25" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00509" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95159" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24152" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24153" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24154" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24155" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2750" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_aps_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_paps_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADSL1r" metaid="R_ADSL1r" name="Adenylsuccinate lyase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADSL1r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADSL1r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01083" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95493" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16853" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16854" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16855" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16856" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1131" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dcamp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ADSL2r" metaid="R_ADSL2r" name="Adenylosuccinate lyase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADSL2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADSL2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AICARSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95494" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1131" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_25aics_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aicar_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ADSS" metaid="R_ADSS" name="Adenylosuccinate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ADSS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ADSS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENYLOSUCCINATE-SYNTHASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.4.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01135" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95495" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15753" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15754" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15755" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15756" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4177" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_imp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dcamp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AGDC" metaid="R_AGDC" name="N-acetylglucosamine-6-phosphate deacetylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AGDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AGDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NAG6PDEACET-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.25" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02059" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95528" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22936" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22937" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22938" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22939" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0677" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acgam6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gam6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AGMHE" metaid="R_AGMHE" name="ADP-D-glycero-D-manno-heptose epimerase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AGMHE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AGMHE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5.1.3.20-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.3.20" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05176" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95551" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17577" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17578" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17579" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17580" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3619" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adphep_DD_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adphep_LD_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AGMT" metaid="R_AGMT" name="Agmatinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AGMT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AGMT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AGMATIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.3.11" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01157" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95554" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13929" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13930" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13931" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13932" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2937" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_agm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ptrc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_urea_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_AGPR" metaid="R_AGPR" name="N-acetyl-g-glutamyl-phosphate reductase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AGPR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AGPR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:N-ACETYLGLUTPREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.38" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03443" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95530" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21588" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21589" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21590" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21591" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3958" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acg5sa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acg5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_AHCi" metaid="R_AHCi" name="Adenosylhomocysteinase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AHCi">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AHCi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSYLHOMOCYSTEINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.3.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00192" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95492" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21708" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21709" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21710" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21711" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ahcys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hcys__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AHCYSNS" metaid="R_AHCYSNS" name="S-adenosylhomocysteine nucleosidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AHCYSNS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AHCYSNS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENOSYLHOMOCYSTEINE-NUCLEOSIDASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.2.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00194" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95625" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17805" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17806" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17807" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17808" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0159" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ahcys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ade_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_rhcys_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_AICART" metaid="R_AICART"
      name="Phosphoribosylaminoimidazolecarboxamide formyltransferase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AICART">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AICART" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04560" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95639" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22192" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22193" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22194" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22195" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4006" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_10fthf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_aicar_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fprica_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AIRC2" metaid="R_AIRC2" name="Phosphoribosylaminoimidazole carboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AIRC2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AIRC2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.4.18" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95644" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19317" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19318" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19319" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19320" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0522" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_air_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hco3_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5caiz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_AIRC3" metaid="R_AIRC3" name="Phosphoribosylaminoimidazole carboxylase (mutase rxn)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AIRC3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AIRC3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-743" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.99.18" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R07405" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95645" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13193" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13194" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13195" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13196" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0523" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5aizc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5caiz_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AKGDH" metaid="R_AKGDH" name="2-Oxogluterate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AKGDH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AKGDH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2OXOGLUTARATEDEH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.52" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.8.1.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.61" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R08549" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95655" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27786" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27787" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27788" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27789" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0116" />
            <fbc:geneProductRef fbc:geneProduct="G_b0727" />
            <fbc:geneProductRef fbc:geneProduct="G_b0726" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succoa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_AKGt2r" metaid="R_AKGt2r" name="2 oxoglutarate reversible transport via symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AKGt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AKGt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-23" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95661" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29011" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29012" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29013" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29014" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2587" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ALAALAr" metaid="R_ALAALAr" name="D-alanine-D-alanine ligase (reversible)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALAALAr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALAALAr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DALADALALIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01150" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95678" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11224" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11225" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11226" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11227" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0381" />
            <fbc:geneProductRef fbc:geneProduct="G_b0092" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__D_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_alaala_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ALAR_copy1" metaid="R_ALAR_copy1" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALAR_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALARACECAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00401" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95693" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20249" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20250" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20251" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20252" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4053" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ala__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALAR_copy2" metaid="R_ALAR_copy2" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALAR_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALARACECAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00401" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95693" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20249" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20250" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20251" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20252" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1190" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ala__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALATA_D2" metaid="R_ALATA_D2" name="D-alanine transaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALATA_D2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALATA_D2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5240" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01147" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95697" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28562" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28563" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28564" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28565" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2551" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pydx5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyam5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ALATA_L" metaid="R_ALATA_L" name="L-alanine transaminase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALATA_L">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALATA_L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALANINE-AMINOTRANSFERASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-13698" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00258" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95698" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19453" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19454" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19455" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19456" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALATA_L2" metaid="R_ALATA_L2" name="Alanine transaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALATA_L2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALATA_L2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95699" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30023" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30024" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30025" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30026" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2551" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pydx5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyam5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALAabc" metaid="R_ALAabc" name="L-alanine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALAabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALAabc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95703" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3460" />
            <fbc:geneProductRef fbc:geneProduct="G_b3457" />
            <fbc:geneProductRef fbc:geneProduct="G_b3455" />
            <fbc:geneProductRef fbc:geneProduct="G_b3454" />
            <fbc:geneProductRef fbc:geneProduct="G_b3456" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ala__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ALAt2r" metaid="R_ALAt2r" name="L alanine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALAt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALAt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5202" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95704" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29443" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29444" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29445" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29446" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0007" />
            <fbc:geneProductRef fbc:geneProduct="G_b4208" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ala__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ALCD19" metaid="R_ALCD19" name="Alcohol dehydrogenase (glycerol)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALCD19">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALCD19" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.21" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.72" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01036" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95709" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30027" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30028" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30029" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30030" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0356" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALDD19xr" metaid="R_ALDD19xr" name="Aldehyde dehydrogenase (phenylacetaldehyde, NAD)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALDD19xr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALDD19xr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHENDEHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.39" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02536" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95743" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21392" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21393" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21394" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21395" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1385" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pacald_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pac_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALDD2x" metaid="R_ALDD2x" name="Aldehyde dehydrogenase (acetaldehyde, NAD)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALDD2x">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALDD2x" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN66-3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00710" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95749" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25294" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25295" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25296" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25297" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1300" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALLTAHr" metaid="R_ALLTAHr" name="Allantoate amidinohydrolase  reversible"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALLTAHr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALLTAHr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALLANTOICASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.3.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02422" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95767" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11016" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11017" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11018" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11019" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0516" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_alltt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_urdglyc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_urea_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALLTN" metaid="R_ALLTN" name="Allantoinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALLTN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALLTN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.2.5" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95769" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30819" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30820" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30821" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30822" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0512" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_alltn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_alltt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ALLTNt2r" metaid="R_ALLTNt2r" name="Allantoin transport in via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALLTNt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALLTNt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95770" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0511" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_alltn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_alltn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ALTRH" metaid="R_ALTRH" name="Altronate hydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ALTRH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ALTRH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALTRODEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01540" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95791" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15957" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15958" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15959" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15960" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3091" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_altrn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ddglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMALT1" metaid="R_AMALT1" name="Amylomaltase (maltotriose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMALT1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMALT1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMYLOMALT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.25" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95800" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3416" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malttr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_maltttr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMALT2" metaid="R_AMALT2" name="Amylomaltase (maltotetraose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMALT2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMALT2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14260" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.25" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95801" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3416" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltttr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_maltpt_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMALT3" metaid="R_AMALT3" name="Amylomaltase (maltopentaose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMALT3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMALT3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14261" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.25" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95802" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3416" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltpt_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malthx_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMALT4" metaid="R_AMALT4" name="Amylomaltase (maltohexaose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMALT4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMALT4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTODEG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.25" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95803" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3416" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malthx_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malthp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMANAPEr" metaid="R_AMANAPEr" name="N-acetylmannosamine 6-phosphate epimerase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMANAPEr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMANAPEr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NANE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.3.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02087" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95804" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25257" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25258" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25259" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25260" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3223" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acmanap_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acgam6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMANK" metaid="R_AMANK" name="N-acetyl-D-mannosamine kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMANK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMANK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NANK-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.60" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02705" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95805" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25253" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25254" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25255" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25256" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3222" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acmana_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acmanap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_AMAOTr" metaid="R_AMAOTr" name="Adenosylmethionine-8-amino-7-oxononanoate transaminase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMAOTr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMAOTr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DAPASYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.62" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03231" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95806" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16861" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16862" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16863" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16864" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0774" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_8aonn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amet_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amob_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dann_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMMQT8_2" metaid="R_AMMQT8_2"
      name="S-adenosylmethione:2-demethylmenaquinone methyltransferase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMMQT8_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMMQT8_2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95823" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3929" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amet_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ahcys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mqn8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMPMS" metaid="R_AMPMS" name="4-amino-2-methyl-5-phosphomethylpyrimidine synthetase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMPMS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMPMS" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95825" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3994" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_air_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4ampm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="4" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AMPN" metaid="R_AMPN" name="AMP nucleosidase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AMPN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AMPN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.2.4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95827" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1982" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ade_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ANPRT" metaid="R_ANPRT" name="Anthranilate phosphoribosyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ANPRT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ANPRT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PRTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.18" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01073" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95842" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11768" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11769" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11770" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11771" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1263" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_anth_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pran_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ANS" metaid="R_ANS" name="Anthranilate synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ANS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ANS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ANTHRANSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.27" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00986" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95843" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21732" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21733" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21734" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21735" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1264" />
            <fbc:geneProductRef fbc:geneProduct="G_b1263" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_chor_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_anth_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_AOXSr" metaid="R_AOXSr" name="8-amino-7-oxononanoate synthase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AOXSr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AOXSr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:7KAPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.47" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03210" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95854" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20712" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20713" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20714" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20715" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0776" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pmcoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_8aonn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AP4AH" metaid="R_AP4AH" name="Ap4A hydrolase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AP4AH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AP4AH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.6.1.41-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.41" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00125" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95856" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24252" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24253" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24254" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24255" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0049" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ap4a_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AP5AH" metaid="R_AP5AH" name="Ap5A hydrolase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AP5AH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AP5AH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-10966" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.61" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R09880" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95859" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30527" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30528" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30529" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30530" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0049" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ap5a_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_APRAUR" metaid="R_APRAUR" name="5-amino-6-(5-phosphoribosylamino)uracil reductase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_APRAUR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/APRAUR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBOFLAVINSYNREDUC-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.193" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03458" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95898" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17845" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17846" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17847" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17848" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0414" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5apru_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5aprbu_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ARAI" metaid="R_ARAI" name="L-arabinose isomerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ARAI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ARAI" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95922" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0062" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arab__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_rbl__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ARBabc" metaid="R_ARBabc" name="L-arabinose transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ARBabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ARBabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-2-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.17" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95935" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1898" />
            <fbc:geneProductRef fbc:geneProduct="G_b1899" />
            <fbc:geneProductRef fbc:geneProduct="G_b1900" />
            <fbc:geneProductRef fbc:geneProduct="G_b1901" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arab__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_arab__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ARBt2r" metaid="R_ARBt2r" name="L-arabinose transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ARBt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ARBt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-10" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95936" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2841" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arab__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_arab__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ARGDC" metaid="R_ARGDC" name="Arginine decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ARGDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ARGDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ARGDECARBOX-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00566" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95940" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17641" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17642" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17643" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17644" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2938" />
            <fbc:geneProductRef fbc:geneProduct="G_b4117" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arg__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_agm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ARGORNt7" metaid="R_ARGORNt7" name="Arginine/ornithine antiporter" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ARGORNt7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ARGORNt7" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95946" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34991" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34992" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34993" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34994" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1605" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arg__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_arg__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_orn_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ARGSL" metaid="R_ARGSL" name="Argininosuccinate lyase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ARGSL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ARGSL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ARGSUCCINLYA-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01086" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95948" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24020" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24021" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24022" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24023" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3960" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_argsuc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_arg__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ARGSS" metaid="R_ARGSS" name="Argininosuccinate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ARGSS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ARGSS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ARGSUCCINSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.4.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01954" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95949" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10932" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10933" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10934" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10935" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3172" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_citr__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_argsuc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ARGabc" metaid="R_ARGabc" name="L-arginine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ARGabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ARGabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-4-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95952" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29879#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29880#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29881#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29882#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2308" />
              <fbc:geneProductRef fbc:geneProduct="G_b2310" />
              <fbc:geneProductRef fbc:geneProduct="G_b2307" />
              <fbc:geneProductRef fbc:geneProduct="G_b2306" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0864" />
              <fbc:geneProductRef fbc:geneProduct="G_b0860" />
              <fbc:geneProductRef fbc:geneProduct="G_b0861" />
              <fbc:geneProductRef fbc:geneProduct="G_b0862" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arg__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_arg__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ASAD" metaid="R_ASAD" name="Aspartate-semialdehyde dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASAD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASAD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASPARTATE-SEMIALDEHYDE-DEHYDROGENASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.11" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02291" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96039" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24284" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24285" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24286" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24287" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3433" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_aspsa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4pasp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASNN" metaid="R_ASNN" name="L-asparaginase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASNN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASNN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASPARAGHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.38" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.5.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00485" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96056" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21016" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21017" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21018" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21019" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2957" />
            <fbc:geneProductRef fbc:geneProduct="G_b1767" />
            <fbc:geneProductRef fbc:geneProduct="G_b0828" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asn__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASNS1" metaid="R_ASNS1" name="Asparagine synthase (glutamine-hydrolysing)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASNS1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASNS1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASNSYNB-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00578" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96059" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12228" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12229" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12230" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12231" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0674" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asn__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASNS2" metaid="R_ASNS2" name="Asparagine synthetase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASNS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASNS2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASNSYNA-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00483" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96060" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11372" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11373" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11374" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11375" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3744" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nh4_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asn__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASNabc" metaid="R_ASNabc" name="L-asparagine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASNabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASNabc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96065" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asn__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asn__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ASNt2r" metaid="R_ASNt2r" name="L asparagine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASNt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASNt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-262" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96066" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asn__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asn__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASP1DC" metaid="R_ASP1DC" name="Aspartate 1-decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASP1DC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASP1DC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASPDECARBOX-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.11" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.15" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00489" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96077" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19497" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19498" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19499" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19500" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0131" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ala_B_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPCT" metaid="R_ASPCT" name="Aspartate carbamoyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPCT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPCT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASPCARBTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.3.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01397" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96080" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20013" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20014" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20015" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20016" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4244" />
            <fbc:geneProductRef fbc:geneProduct="G_b4245" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cbp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cbasp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ASPK" metaid="R_ASPK" name="Aspartate kinase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASPARTATEKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.2.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00480" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96085" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23776" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23777" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23778" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23779" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4024" />
            <fbc:geneProductRef fbc:geneProduct="G_b0002" />
            <fbc:geneProductRef fbc:geneProduct="G_b3940" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4pasp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPO3_1" metaid="R_ASPO3_1" name="L-aspartate oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPO3_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPO3_1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96091" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28683" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28684" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28685" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28686" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2574" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_iasp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPO4_1" metaid="R_ASPO4_1" name="L-aspartate oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPO4_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPO4_1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96092" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29191" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29192" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29193" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29194" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2574" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_iasp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPO5_1" metaid="R_ASPO5_1" name="L-aspartate oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPO5_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPO5_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.5.99.-" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96093" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30043" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30044" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30045" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30046" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2574" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fum_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_iasp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPO6_1" metaid="R_ASPO6_1" name="L-aspartate oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPO6_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPO6_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.3.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00481" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96094" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25876" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25877" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25878" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25879" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2574" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_iasp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPT" metaid="R_ASPT" name="L-aspartase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASPARTASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00490" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96099" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16601" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16602" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16603" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16604" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4139" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ASPTA" metaid="R_ASPTA" name="Aspartate transaminase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ASPAMINOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-13697" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00355" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96079" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21824" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21825" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21826" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21827" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0928" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_oaa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPabc" metaid="R_ASPabc" name="L-aspartate transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-222" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96103" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29039#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29040#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29041#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29042#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0655" />
            <fbc:geneProductRef fbc:geneProduct="G_b0653" />
            <fbc:geneProductRef fbc:geneProduct="G_b0654" />
            <fbc:geneProductRef fbc:geneProduct="G_b0652" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPt2r" metaid="R_ASPt2r" name="L aspartate reversible transport via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96106" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4077" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPt2_2" metaid="R_ASPt2_2" name="Aspartate transport via proton symport (2 H)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPt2_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPt2_2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-122A" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96104" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29287" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29288" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29289" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29290" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3528" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ASPt2_3" metaid="R_ASPt2_3" name="L-asparate transport via proton symport (3 H)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ASPt2_3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ASPt2_3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96105" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4123" />
            <fbc:geneProductRef fbc:geneProduct="G_b4138" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="3" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_AST" metaid="R_AST" name="Arginine succinyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_AST">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/AST" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ARGININE-N-SUCCINYLTRANSFERASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.109" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00832" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96115" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15185" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15186" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15187" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15188" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1747" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arg__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sucarg_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="R_ATPM_lower_bound" fbc:upperFluxBound="R_ATPM_upper_bound" id="R_ATPM" metaid="R_ATPM" name="ATP maintenance requirement" reversible="false"
      sboTerm="SBO:0000630">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ATPM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ATPM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ATPASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.15" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.10" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.11" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.12" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.14" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.15" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.16" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.17" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.18" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.19" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.20" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.22" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.23" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.24" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.25" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.26" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.27" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.28" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.29" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.30" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.31" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.32" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.33" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.34" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.35" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.36" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.37" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.38" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.39" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.40" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.41" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.42" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.43" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.44" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.46" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.47" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.48" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.49" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.50" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.51" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.52" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.53" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.54" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.6" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.7" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.9" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.10" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.11" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.12" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.13" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.7" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.4.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00086" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96131" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13065" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13066" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13067" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13068" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ATPPRT" metaid="R_ATPPRT" name="ATP phosphoribosyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ATPPRT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ATPPRT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ATPPHOSPHORIBOSYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.17" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01071" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96132" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18473" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18474" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18475" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18476" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2019" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_prbatp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ATPS4r" metaid="R_ATPS4r" name="ATP synthase (four protons for one ATP)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ATPS4r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ATPS4r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ATPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.14" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96136" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3736" />
            <fbc:geneProductRef fbc:geneProduct="G_b3737" />
            <fbc:geneProductRef fbc:geneProduct="G_b3738" />
            <fbc:geneProductRef fbc:geneProduct="G_b3739" />
            <fbc:geneProductRef fbc:geneProduct="G_b3731" />
            <fbc:geneProductRef fbc:geneProduct="G_b3732" />
            <fbc:geneProductRef fbc:geneProduct="G_b3735" />
            <fbc:geneProductRef fbc:geneProduct="G_b3734" />
            <fbc:geneProductRef fbc:geneProduct="G_b3733" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="4" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_BETALDHx" metaid="R_BETALDHx" name="Betaine-aldehyde dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BETALDHx">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/BETALDHx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BADH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02565" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96237" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15305" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15306" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15307" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15308" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0312" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_betald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_BETALDHy" metaid="R_BETALDHy" name="Betaine-aldehyde dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BETALDHy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/BETALDHy" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02566" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96238" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30067" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30068" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30069" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30070" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0312" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_betald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_BPNT" metaid="R_BPNT" name="3',5'-bisphosphate nucleotidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BPNT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/BPNT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:325-BISPHOSPHATE-NUCLEOTIDASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.7" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.97" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00188" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96321" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10040" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10041" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10042" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10043" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pap_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_BSORx" metaid="R_BSORx" name="Biotin sulfoxide reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BSORx">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/BSORx" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96325" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30531" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30532" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30533" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30534" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3551" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_btnso_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_btn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_BSORy" metaid="R_BSORy" name="Biotin sulfoxide reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BSORy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/BSORy" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.-.-.-" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10127" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96326" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30535" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30536" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30537" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30538" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3551" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_btnso_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_btn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_BTS2" metaid="R_BTS2" name="Biotin synthase (ala-L producing)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BTS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/BTS2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96341" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0775" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtbt_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ala__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_btn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_BUTCT" metaid="R_BUTCT" name="Acetyl-CoA:butyrate-CoA transferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BUTCT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/BUTCT" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.8.3.8" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96217" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2222" />
            <fbc:geneProductRef fbc:geneProduct="G_b2221" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_but_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_btcoa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_BUTt2r" metaid="R_BUTt2r" name="Butyrate transport via proton symport, reversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BUTt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/BUTt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96353" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2223" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_but_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_but_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_BIOMASS_Ecoli" metaid="R_BIOMASS_Ecoli" name="BiomassEcoli" reversible="false"
      sboTerm="SBO:0000629">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_BIOMASS_Ecoli">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/Ecoli" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5mthf_c" stoichiometry="0.05" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="5E-5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_c" stoichiometry="0.488" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amp_c" stoichiometry="0.001" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arg__L_c" stoichiometry="0.281" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asn__L_c" stoichiometry="0.229" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="0.229" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="45.7318" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_clpn_EC_c" stoichiometry="0.000129" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="6E-6" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ctp_c" stoichiometry="0.126" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_c" stoichiometry="0.087" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_datp_c" stoichiometry="0.0247" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dctp_c" stoichiometry="0.0254" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgtp_c" stoichiometry="0.0254" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dttp_c" stoichiometry="0.0247" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fad_c" stoichiometry="1E-5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="0.25" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="0.25" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gly_c" stoichiometry="0.582" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glycogen_c" stoichiometry="0.154" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="0.203" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="45.5608" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_his__L_c" stoichiometry="0.09" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ile__L_c" stoichiometry="0.276" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_leu__L_c" stoichiometry="0.428" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lps_EC_c" stoichiometry="0.0084" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lys__L_c" stoichiometry="0.326" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_met__L_c" stoichiometry="0.146" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="0.00215" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="5E-5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="0.00013" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="0.0004" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pe_EC_c" stoichiometry="0.001935" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_peptido_EC_c" stoichiometry="0.0276" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pg_EC_c" stoichiometry="0.000464" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_phe__L_c" stoichiometry="0.176" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pro__L_c" stoichiometry="0.21" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ps_EC_c" stoichiometry="5.2E-5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ptrc_c" stoichiometry="0.035" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_c" stoichiometry="0.205" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_spmd_c" stoichiometry="0.007" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succoa_c" stoichiometry="3E-6" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thr__L_c" stoichiometry="0.241" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trp__L_c" stoichiometry="0.054" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tyr__L_c" stoichiometry="0.131" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udpg_c" stoichiometry="0.003" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_utp_c" stoichiometry="0.136" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_val__L_c" stoichiometry="0.402" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="45.5608" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="45.56035" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="45.5628" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="0.7302" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_C120SN" metaid="R_C120SN" name="Fatty acid biosynthesis (n-C12:0)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_C120SN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/C120SN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96378" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
              <fbc:geneProductRef fbc:geneProduct="G_b2323" />
              <fbc:geneProductRef fbc:geneProduct="G_b0954" />
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
              <fbc:geneProductRef fbc:geneProduct="G_b0180" />
              <fbc:geneProductRef fbc:geneProduct="G_b1095" />
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
              <fbc:geneProductRef fbc:geneProduct="G_b2323" />
              <fbc:geneProductRef fbc:geneProduct="G_b0180" />
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
              <fbc:geneProductRef fbc:geneProduct="G_b0954" />
              <fbc:geneProductRef fbc:geneProduct="G_b1095" />
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_actACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="14" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="4" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="10" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="4" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="4" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ddcaACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="10" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_C140SN" metaid="R_C140SN" name="Fatty acid biosynthesis (n-C14:0)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_C140SN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/C140SN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96380" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
              <fbc:geneProductRef fbc:geneProduct="G_b0954" />
              <fbc:geneProductRef fbc:geneProduct="G_b1095" />
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
              <fbc:geneProductRef fbc:geneProduct="G_b0180" />
              <fbc:geneProductRef fbc:geneProduct="G_b2323" />
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
              <fbc:geneProductRef fbc:geneProduct="G_b2323" />
              <fbc:geneProductRef fbc:geneProduct="G_b0954" />
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
              <fbc:geneProductRef fbc:geneProduct="G_b0180" />
              <fbc:geneProductRef fbc:geneProduct="G_b1095" />
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_actACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="17" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="12" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_myrsACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="12" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_C141SN" metaid="R_C141SN" name="Fatty acid biosynthesis (n-C14:1)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_C141SN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/C141SN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96381" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1288" />
            <fbc:geneProductRef fbc:geneProduct="G_b0180" />
            <fbc:geneProductRef fbc:geneProduct="G_b2323" />
            <fbc:geneProductRef fbc:geneProduct="G_b0954" />
            <fbc:geneProductRef fbc:geneProduct="G_b1093" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_actACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="16" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="11" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="11" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tdeACP_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_C160SN" metaid="R_C160SN" name="Fatty acid biosynthesis (n-C16:0)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_C160SN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/C160SN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96386" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
              <fbc:geneProductRef fbc:geneProduct="G_b2323" />
              <fbc:geneProductRef fbc:geneProduct="G_b0954" />
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
              <fbc:geneProductRef fbc:geneProduct="G_b0180" />
              <fbc:geneProductRef fbc:geneProduct="G_b1095" />
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
              <fbc:geneProductRef fbc:geneProduct="G_b2323" />
              <fbc:geneProductRef fbc:geneProduct="G_b0180" />
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1288" />
              <fbc:geneProductRef fbc:geneProduct="G_b0954" />
              <fbc:geneProductRef fbc:geneProduct="G_b1095" />
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_actACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="20" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="14" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="14" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_palmACP_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_C161SN" metaid="R_C161SN" name="Fatty acid biosynthesis (n-C16:1)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_C161SN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/C161SN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96391" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            <fbc:geneProductRef fbc:geneProduct="G_b0180" />
            <fbc:geneProductRef fbc:geneProduct="G_b2323" />
            <fbc:geneProductRef fbc:geneProduct="G_b0954" />
            <fbc:geneProductRef fbc:geneProduct="G_b1288" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_actACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="19" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="13" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdeACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="13" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_C181SN" metaid="R_C181SN" name="Fatty acid biosynthesis (n-C18:1)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_C181SN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/C181SN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96398" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            <fbc:geneProductRef fbc:geneProduct="G_b0954" />
            <fbc:geneProductRef fbc:geneProduct="G_b2323" />
            <fbc:geneProductRef fbc:geneProduct="G_b0180" />
            <fbc:geneProductRef fbc:geneProduct="G_b1095" />
            <fbc:geneProductRef fbc:geneProduct="G_b1288" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_actACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="22" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="15" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="15" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_octeACP_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CADVt" metaid="R_CADVt" name="Lysine/Cadaverine antiporter" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CADVt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CADVt" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96445" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4132" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_15dap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lys__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_15dap_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lys__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CAT" metaid="R_CAT" name="Catalase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CATAL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.11.1.21" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.11.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00009" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96455" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20309" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20310" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20311" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20312" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1732" />
            <fbc:geneProductRef fbc:geneProduct="G_b3942" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o2_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_o2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CBIAT_DELETE" metaid="R_CBIAT_DELETE" name="Cobinamide adenyltransferase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CBIAT_DELETE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CBIAT_DELETE" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96466" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1270" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cbi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adocbi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CBL1abc" metaid="R_CBL1abc" name="Cob(1)alamin transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CBL1abc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CBL1abc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-5-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.33" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96471" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3966" />
            <fbc:geneProductRef fbc:geneProduct="G_b1709" />
            <fbc:geneProductRef fbc:geneProduct="G_b0158" />
            <fbc:geneProductRef fbc:geneProduct="G_b1711" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cbl1_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cbl1_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CBLAT_DELETE" metaid="R_CBLAT_DELETE" name="Cob(I)alamin adenosyltransferase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CBLAT_DELETE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CBLAT_DELETE" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.17" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96478" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1270" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cbl1_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adocbl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CBMKr" metaid="R_CBMKr" name="Carbamate kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CBMKr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CBMKr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBAMATE-KINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00150" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96482" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10152" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10153" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10154" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10155" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0323" />
            <fbc:geneProductRef fbc:geneProduct="G_b0521" />
            <fbc:geneProductRef fbc:geneProduct="G_b2874" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nh4_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cbp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CBPS" metaid="R_CBPS" name="Carbamoyl-phosphate synthase (glutamine-hydrolysing)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CBPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CBPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00575" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96485" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18633" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18634" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18635" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18636" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0032" />
            <fbc:geneProductRef fbc:geneProduct="G_b0033" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hco3_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cbp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CDAPPA_EC" metaid="R_CDAPPA_EC" name="CDP-Diacylglycerol pyrophostatase (Ecoli)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CDAPPA_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CDAPPA_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96509" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3918" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cdpdag_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pa_EC_c" stoichiometry="0.02" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CDPMEK" metaid="R_CDPMEK" name="4-(cytidine 5'-diphospho)-2-C-methyl-D-erythritol kinase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CDPMEK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CDPMEK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2.7.1.148-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.148" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05634" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96558" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18437" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18438" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18439" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18440" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1208" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4c2me_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2p4c2me_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CHLabc" metaid="R_CHLabc" name="Choline transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CHLabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CHLabc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96691" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/38715#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/38716#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/38717#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/38718#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2131" />
            <fbc:geneProductRef fbc:geneProduct="G_b2128" />
            <fbc:geneProductRef fbc:geneProduct="G_b2130" />
            <fbc:geneProductRef fbc:geneProduct="G_b2129" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_chol_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_chol_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CHLt2" metaid="R_CHLt2" name="Choline transport via proton symport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CHLt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CHLt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-99" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96692" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28843" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28844" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28845" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28846" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0314" />
            <fbc:geneProductRef fbc:geneProduct="G_b1801" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_chol_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_chol_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CHORM" metaid="R_CHORM" name="Chorismate mutase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CHORM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CHORM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CHORISMATEMUT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.99.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01715" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96708" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13897" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13898" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13899" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13900" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2600" />
            <fbc:geneProductRef fbc:geneProduct="G_b2599" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_chor_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pphn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CHORS" metaid="R_CHORS" name="Chorismate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CHORS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CHORS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CHORISMATE-SYNTHASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01714" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96709" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21020" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21021" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21022" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21023" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2329" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3psme_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_chor_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CHRPL" metaid="R_CHRPL" name="Chorismate pyruvate lyase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CHRPL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CHRPL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CHORPYRLY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.40" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01302" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96711" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16505" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16506" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16507" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16508" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4039" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_chor_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4hbz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CINNDO" metaid="R_CINNDO" name="Cinnamate dioxygenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CINNDO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CINNDO" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.14.12.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06783" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96722" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25058" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25059" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25060" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25061" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2538" />
            <fbc:geneProductRef fbc:geneProduct="G_b2542" />
            <fbc:geneProductRef fbc:geneProduct="G_b2539" />
            <fbc:geneProductRef fbc:geneProduct="G_b2540" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cinnm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cenchddd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CITL" metaid="R_CITL" name="Citrate lyase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CITL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CITL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CITLY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00362" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96731" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10760" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10761" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10762" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10763" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0616" />
            <fbc:geneProductRef fbc:geneProduct="G_b0615" />
            <fbc:geneProductRef fbc:geneProduct="G_b0617" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cit_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_oaa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CITt7" metaid="R_CITt7" name="Citrate transport via succinate antiport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CITt7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CITt7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-201" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96752" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28835" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28836" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28837" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28838" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0612" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cit_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cit_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CLPNS_EC" metaid="R_CLPNS_EC" name="Cardiolipin Synthase (E.coli)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CLPNS_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CLPNS_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96790" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1249" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pg_EC_c" stoichiometry="0.04" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_clpn_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CMPN" metaid="R_CMPN" name="CMP nucleosidase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CMPN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CMPN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.2.10" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96804" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_csn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CO2t" metaid="R_CO2t" name="CO2 transporter via diffusion" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CO2t">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CO2t" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-545" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96810" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_co2_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CPPPGO" metaid="R_CPPPGO" name="Coproporphyrinogen oxidase (O2 required)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CPPPGO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CPPPGO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-1461" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.3.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03220" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96880" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18257" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18258" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18259" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18260" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2436" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cpppg3_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pppg9_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CRNBTCT" metaid="R_CRNBTCT" name="Gamma-butyrobetainyl-CoA: carnitine CoA transferase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CRNBTCT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CRNBTCT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-3601" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.8.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10644" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96894" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28418" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28419" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28420" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28421" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0038" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_bbtcoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_crn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_crncoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gbbtn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CRNCBCT" metaid="R_CRNCBCT" name="Crotonobetainyl-CoA: carnitine CoA transferase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CRNCBCT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CRNCBCT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14936" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.8.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10643" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96899" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28526" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28527" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28528" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28529" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0038" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_crn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ctbtcoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_crncoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ctbt_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CRNCDH" metaid="R_CRNCDH" name="Carnityl-CoA dehydratse" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CRNCDH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CRNCDH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARNDETRU-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.149" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10675" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96900" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28338" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28339" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28340" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28341" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0036" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_crncoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ctbtcoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CRNt7" metaid="R_CRNt7" name="Carnitine/butyrobetaine antiporter" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CRNt7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CRNt7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-100" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96908" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29427" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29428" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29429" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29430" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0040" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_crn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gbbtn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_crn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gbbtn_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CS" metaid="R_CS" name="Citrate synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CITSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14905" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.3.16" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.3.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00351" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96920" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16845" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16846" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16847" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16848" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0720" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_oaa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cit_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CSND" metaid="R_CSND" name="Cytosine deaminase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CSND">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CSND" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYTDEAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00974" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96927" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20605" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20606" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20607" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20608" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0337" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_csn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ura_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CSNt2" metaid="R_CSNt2" name="Cytosine transport in via proton symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CSNt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CSNt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-116" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-199" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96929" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29935" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29936" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29937" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29938" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0336" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_csn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_csn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CTPS2" metaid="R_CTPS2" name="CTP synthase (glutamine)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CTPS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CTPS2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CTPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00573" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96945" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26426" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26427" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26428" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26429" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2780" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_utp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYANST" metaid="R_CYANST" name="Cyanide sulfurtransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYANST">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYANST" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THIOSULFATE-SULFURTRANSFERASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.8.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01931" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96956" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16881" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16882" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16883" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16884" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3425" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cyan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tsul_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_so3_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tcynt_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYNTAH" metaid="R_CYNTAH" name="Cyanate aminohydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYNTAH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYNTAH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-12893" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.104" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10079" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96961" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11120" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11121" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11122" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11123" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0340" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cynt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hco3_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYNTt2" metaid="R_CYNTt2" name="Cyanate transport via proton symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYNTt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYNTt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96962" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29231" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29232" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29233" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29234" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0341" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cynt_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cynt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYSDS" metaid="R_CYSDS" name="Cysteine Desulfhydrase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYSDS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYSDS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LCYSDESULF-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.4.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.4.1.28" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.4.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00782" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96994" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24931" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24932" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24933" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24934" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3708" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2s_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYSS" metaid="R_CYSS" name="Cysteine synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYSS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYSS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACSERLY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.47" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.65" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00897" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97007" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14829" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14830" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14831" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14832" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2414" />
            <fbc:geneProductRef fbc:geneProduct="G_b2421" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acser_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2s_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYSTL" metaid="R_CYSTL" name="Cystathionine b-lyase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYSTL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYSTL" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.4.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01286" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96469" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13965" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13966" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13967" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13968" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3008" />
            <fbc:geneProductRef fbc:geneProduct="G_b1622" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cyst__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hcys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYSabc" metaid="R_CYSabc" name="L-cysteine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYSabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYSabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-265" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97025" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29783#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29784#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29785#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29786#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32795#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32796#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32797#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32798#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYTBD" metaid="R_CYTBD" name="Cytochrome oxidase bd (ubiquinol-8: 2 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYTBD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYTBD" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97031" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0733" />
            <fbc:geneProductRef fbc:geneProduct="G_b0734" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="0.5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8h2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYTBO3" metaid="R_CYTBO3" name="Cytochrome oxidase bo3 (ubiquinol-8: 2.5 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYTBO3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYTBO3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97035" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0429" />
            <fbc:geneProductRef fbc:geneProduct="G_b0432" />
            <fbc:geneProductRef fbc:geneProduct="G_b0431" />
            <fbc:geneProductRef fbc:geneProduct="G_b0430" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2.5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="0.5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8h2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2.5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYTD" metaid="R_CYTD" name="Cytidine deaminase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYTD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYTD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYTIDEAM2-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01878" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97039" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16069" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16070" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16071" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16072" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2143" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cytd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uri_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYTDK2" metaid="R_CYTDK2" name="Cytidine kinase (GTP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYTDK2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYTDK2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYTIDINEKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.213" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.48" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00517" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97042" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28162" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28163" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28164" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28165" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2066" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cytd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_CYTDt2_copy1" metaid="R_CYTDt2_copy1" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYTDt2_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108B" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97044" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29983" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29984" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29985" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29986" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or>
            <fbc:geneProductRef fbc:geneProduct="G_b2964" />
            <fbc:geneProductRef fbc:geneProduct="G_b2393" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cytd_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cytd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CYTDt2_copy2" metaid="R_CYTDt2_copy2" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYTDt2_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108B" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97044" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29983" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29984" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29985" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29986" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2406" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cytd_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cytd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CYTK1" metaid="R_CYTK1" name="Cytidylate kinase (CMP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYTK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYTK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-11832" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.14" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.25" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00512" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97047" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11600" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11601" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11602" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11603" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0910" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_CYTK2" metaid="R_CYTK2" name="Cytidylate kinase (dCMP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_CYTK2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/CYTK2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-7913" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.13" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.14" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.25" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01665" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97053" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25094" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25095" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25096" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25097" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0910" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dcmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dcdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DAAD" metaid="R_DAAD" name="D-Amino acid dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DAAD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DAAD" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.99.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97073" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1189" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fadh2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DADA" metaid="R_DADA" name="Deoxyadenosine deaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DADA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DADA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADDALT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02556" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97076" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28190" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28191" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28192" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28193" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1623" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dad_2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_din_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DADK" metaid="R_DADK" name="Deoxyadenylate kinase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DADK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DADK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYADENYLATE-KINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.11" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.13" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01547" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96117" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23100" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23101" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23102" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23103" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0474" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_damp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DADNt2" metaid="R_DADNt2" name="Deoxyadenosine transport in via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DADNt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DADNt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108C" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97080" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29979" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29980" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29981" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29982" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2964" />
            <fbc:geneProductRef fbc:geneProduct="G_b2393" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dad_2_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dad_2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DAGK_EC" metaid="R_DAGK_EC" name="Diacylglycerol kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DAGK_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DAGK_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97109" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4042" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_12dgr_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pa_EC_c" stoichiometry="0.02" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DALAt2r" metaid="R_DALAt2r" name="D-alanine transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DALAt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DALAt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-62A" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97135" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28903" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28904" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28905" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28906" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4208" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__D_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ala__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DAPDC" metaid="R_DAPDC" name="Diaminopimelate decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DAPDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DAPDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIAMINOPIMDECARB-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00451" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97141" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15101" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15102" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15103" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15104" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2838" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_26dap__M_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lys__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DAPE" metaid="R_DAPE" name="Diaminopimelate epimerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DAPE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DAPE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIAMINOPIMEPIM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.1.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02735" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97143" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15393" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15394" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15395" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15396" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3809" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_26dap_LL_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_26dap__M_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DAPabc" metaid="R_DAPabc" name="M-diaminopimelic acid ABC transport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DAPabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DAPabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-291" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97146" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_26dap__M_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_26dap__M_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DASYN_EC" metaid="R_DASYN_EC" name="CDP-Diacylglycerol synthetase (Ecoli)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DASYN_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DASYN_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97168" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0175" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pa_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cdpdag_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DB4PS" metaid="R_DB4PS" name="3,4-Dihydroxy-2-butanone-4-phosphate synthase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DB4PS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DB4PS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.99.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R07281" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97178" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3041" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ru5p__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_db4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DBTS" metaid="R_DBTS" name="Dethiobiotin synthase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DBTS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DBTS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DETHIOBIOTIN-SYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.3.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03182" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97180" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15805" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15806" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15807" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15808" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0778" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dann_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtbt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DCTPD" metaid="R_DCTPD" name="DCTP deaminase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DCTPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DCTPD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DCTP-DEAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02325" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97201" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22680" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22681" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22682" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22683" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2065" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dutp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DCYTD" metaid="R_DCYTD" name="Deoxycytidine deaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DCYTD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DCYTD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CYTIDEAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02485" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97207" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13433" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13434" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13435" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13436" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2143" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dcyt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_duri_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DCYTt2" metaid="R_DCYTt2" name="Deoxycytidine transport in via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DCYTt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DCYTt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108D" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97209" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29975" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29976" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29977" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29978" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2964" />
            <fbc:geneProductRef fbc:geneProduct="G_b2393" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dcyt_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dcyt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DDGALK" metaid="R_DDGALK" name="2-dehydro-3-deoxygalactonokinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DDGALK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DDGALK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEHYDDEOXGALACTKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.178" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.58" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03387" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97213" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16525" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16526" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16527" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16528" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3693" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dh3dgal_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dh3dgal6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DDGLCNt2r" metaid="R_DDGLCNt2r"
      name="2-dehydro-3-deoxy-D-gluconate transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DDGLCNt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DDGLCNt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-113" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97214" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29943" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29944" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29945" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29946" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3909" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ddglcn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ddglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DDGLK" metaid="R_DDGLK" name="2-dehydro-3-deoxygluconokinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DDGLK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DDGLK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYGLUCONOKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.178" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.45" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01541" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97216" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14797" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14798" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14799" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14800" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3526" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ddglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ddg6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DDPA" metaid="R_DDPA" name="3-deoxy-D-arabino-heptulosonate 7-phosphate synthetase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DDPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DDPA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DAHPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.54" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01826" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97218" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14717" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14718" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14719" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14720" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1704" />
            <fbc:geneProductRef fbc:geneProduct="G_b2601" />
            <fbc:geneProductRef fbc:geneProduct="G_b0754" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_e4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dda7p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DDPGALA" metaid="R_DDPGALA" name="2-dehydro-3-deoxy-6-phosphogalactonate aldolase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DDPGALA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DDPGALA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEHYDDEOXPHOSGALACT-ALDOL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.21" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.55" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01064" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97220" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24464" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24465" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24466" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24467" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3692" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dh3dgal6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DGK1" metaid="R_DGK1" name="Deoxyguanylate kinase (dGMP:ATP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DGK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DGK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GMKALT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.12" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.13" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02090" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97320" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12697" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12698" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12699" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12700" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3648" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dgdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DGSNt2" metaid="R_DGSNt2" name="Deoxyguanosine transport in via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DGSNt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DGSNt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97325" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2964" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgsn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dgsn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHAD1" metaid="R_DHAD1" name="Dihydroxy-acid dehydratase (2,3-dihydroxy-3-methylbutanoate)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHAD1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHAD1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROXYISOVALDEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04441" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97358" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24809" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24810" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24811" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24812" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3771" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_23dhmb_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3mob_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHAD2" metaid="R_DHAD2"
      name="Dihydroxy-acid dehydratase (2,3-dihydroxy-3-methylpentanoate)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHAD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHAD2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROXYMETVALDEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05070" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97359" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27694" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27695" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27696" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27697" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3771" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_23dhmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3mop_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHAPT" metaid="R_DHAPT" name="Dihydroxyacetone phosphotransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHAPT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHAPT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2.7.1.121-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.121" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01012" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97364" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18381" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18382" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18383" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18384" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b1200" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b1199" />
            <fbc:geneProductRef fbc:geneProduct="G_b1198" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dha_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DHAt" metaid="R_DHAt" name="Dihydroxyacetone transport via facilitated diffusion"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHAt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHAt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-559" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97367" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dha_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dha_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DHBD" metaid="R_DHBD" name="2,3-dihydro-2,3-dihydroxybenzoate dehydrogenase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHBD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHBD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DHBDEHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.1.28" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01505" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97368" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23824" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23825" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23826" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23827" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0596" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_23ddhb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_23dhb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DHBSr" metaid="R_DHBSr" name="2,3-dihydroxybenzoate adenylate synthase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHBSr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHBSr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DHBAMPLIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.58" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01504" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97369" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20229" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20230" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20231" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20232" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0594" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_23dhb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_23dhba_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHCIND" metaid="R_DHCIND" name="2,3-dihydroxycinnamate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHCIND">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHCIND" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.1.87" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06785" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97378" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25066" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25067" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25068" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25069" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2541" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cenchddd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhcinnm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHCINDO_1" metaid="R_DHCINDO_1" name="2,3-dihydroxycinnamate 1,2-dioxygenase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHCINDO_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHCINDO_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.13.11.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06788" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97379" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0348" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhcinnm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hkntd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHDPRy" metaid="R_DHDPRy" name="Dihydrodipicolinate reductase (NADPH)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHDPRy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHDPRy" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.1.26" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97393" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0031" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_23dhdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHDPS" metaid="R_DHDPS" name="Dihydrodipicolinate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHDPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHDPS" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97394" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2478" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_aspsa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_23dhdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DHFR" metaid="R_DHFR" name="Dihydrofolate reductase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHFR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHFR" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.5.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00939" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97401" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15009" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15010" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15011" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15012" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0048" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHFS_1" metaid="R_DHFS_1" name="Dihydrofolate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHFS_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHFS_1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROFOLATESYNTH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.12" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.17" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02237" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97403" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23584" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23585" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23586" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23587" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2315" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhpt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHNAOT" metaid="R_DHNAOT" name="1,4-dihydroxy-2-naphthoate octaprenyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHNAOT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHNAOT" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97406" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3930" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhna_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_octdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHNPA2r" metaid="R_DHNPA2r" name="Dihydroneopterin aldolase reversible" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHNPA2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHNPA2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:H2NEOPTERINALDOL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-10857" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.25" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03504" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97415" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3058" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhnpt_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_6hmhpt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gcald_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHORD2" metaid="R_DHORD2" name="Dihydoorotic acid dehydrogenase (quinone8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHORD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHORD2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97418" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28687" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28688" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28689" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28690" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0945" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhor__S_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_orot_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHORD5" metaid="R_DHORD5" name="Dihydroorotic acid (menaquinone-8)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHORD5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHORD5" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97421" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29199" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29200" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29201" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29202" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0945" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhor__S_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_orot_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DHORTS" metaid="R_DHORTS" name="Dihydroorotase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHORTS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHORTS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDROOROT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01993" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97428" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24296" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24297" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24298" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24299" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1062" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhor__S_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cbasp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHPPD" metaid="R_DHPPD" name="2,3-dihydroxyphenylpropionate dehydrogenase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHPPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHPPD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHENPRODIOLDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.1.87" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97434" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25062" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25063" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25064" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25065" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2541" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cechddd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhpppn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHPPDA2" metaid="R_DHPPDA2"
      name="Diaminohydroxyphosphoribosylaminopryrimidine deaminase (25drapp)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHPPDA2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHPPDA2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBOFLAVINSYNDEAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.26" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03459" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97435" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21868" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21869" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21870" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21871" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0414" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_25drapp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5apru_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHPS2_1" metaid="R_DHPS2_1" name="Dihydropteroate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHPS2_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHPS2_1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:H2PTEROATESYNTH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.15" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03067" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97439" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19949" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19950" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19951" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19952" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3177" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4abz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_6hmhptpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhpt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHPTDCs" metaid="R_DHPTDCs" name="4,5-dihydroxy-2,3-pentanedione cyclization (spontaneous)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHPTDCs">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHPTDCs" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97441" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhptd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hmfurn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DHQTi" metaid="R_DHQTi" name="3-dehydroquinate dehydratase, irreversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHQTi">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHQTi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-DEHYDROQUINATE-DEHYDRATASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03084" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97449" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21096" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21097" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21098" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21099" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1693" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3dhq_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3dhsk_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DHQS" metaid="R_DHQS" name="3-dehydroquinate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DHQS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DHQS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-DEHYDROQUINATE-SYNTHASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.3.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03083" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97447" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21968" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21969" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21970" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21971" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3389" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dda7p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3dhq_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DINSt2" metaid="R_DINSt2" name="Deoxyinosine transport in via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DINSt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DINSt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108E" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97465" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29971" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29972" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29973" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29974" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2964" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_din_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_din_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DKGLCNR1" metaid="R_DKGLCNR1" name="2,5-diketo-D-gluconate reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DKGLCNR1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DKGLCNR1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97489" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0207" />
            <fbc:geneProductRef fbc:geneProduct="G_b3012" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_25dkglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dhguln_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DKGLCNR2x" metaid="R_DKGLCNR2x" name="2,5-diketo-D-gluconate reductase (NADH)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DKGLCNR2x">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DKGLCNR2x" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97490" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30103" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30104" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30105" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30106" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3553" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_25dkglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5dglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DKGLCNR2y" metaid="R_DKGLCNR2y" name="2,5-diketo-D-gluconate reductase (NADPH)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DKGLCNR2y">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DKGLCNR2y" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:YIAE1-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.215" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R08879" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97491" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27634" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27635" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27636" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27637" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3553" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_25dkglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5dglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DKMPPD" metaid="R_DKMPPD"
      name="2,3-diketo-5-methylthio-1-phosphopentane degradation reaction" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DKMPPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DKMPPD" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97497" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dkmpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2kmb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DKMPPD2" metaid="R_DKMPPD2"
      name="2 3 diketo 5 methylthio 1 phosphopentane degradation reaction" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DKMPPD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DKMPPD2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97498" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dkmpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="3" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2kmb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DMATT" metaid="R_DMATT" name="Dimethylallyltranstransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DMATT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DMATT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GPPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01658" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97512" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22408" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22409" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22410" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22411" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0421" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dmpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ipdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_grdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DMPPS" metaid="R_DMPPS"
      name="1-hydroxy-2-methyl-2-(E)-butenyl 4-diphosphate reductase (dmpp)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DMPPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DMPPS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97524" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0029" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2mb4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dmpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DMQMT" metaid="R_DMQMT" name="3-Dimethylubiquinonol 3-methyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DMQMT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DMQMT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DHHB-METHYLTRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.1.64" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97527" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28679" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28680" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28681" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28682" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2232" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2omhmbl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amet_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ahcys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DMSOR1" metaid="R_DMSOR1" name="Dimethyl sulfoxide reductase (Menaquinol 8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DMSOR1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DMSOR1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97529" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29203" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29204" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29205" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29206" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0896" />
            <fbc:geneProductRef fbc:geneProduct="G_b0894" />
            <fbc:geneProductRef fbc:geneProduct="G_b0895" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dmso_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mql8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dms_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mqn8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DMSOR1e" metaid="R_DMSOR1e" name="Dimethyl sulfoxide reductase (Menaquinol 8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DMSOR1e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DMSOR1e" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97530" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1873" />
            <fbc:geneProductRef fbc:geneProduct="G_b1872" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dmso_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mql8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dms_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mqn8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DMSOR2" metaid="R_DMSOR2" name="Dimethyl sulfoxide reductase (Demethylmenaquinol 8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DMSOR2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DMSOR2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97531" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30587" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30588" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30589" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30590" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0896" />
            <fbc:geneProductRef fbc:geneProduct="G_b0894" />
            <fbc:geneProductRef fbc:geneProduct="G_b0895" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dmso_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dms_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DMSOR2e" metaid="R_DMSOR2e" name="Dimethyl sulfoxide reductase (Demethylmenaquinol 8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DMSOR2e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DMSOR2e" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97532" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1873" />
            <fbc:geneProductRef fbc:geneProduct="G_b1872" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dmso_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dms_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DNMPPA" metaid="R_DNMPPA" name="Dihydroneopterin monophosphate dephosphorylase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DNMPPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DNMPPA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIHYDRONEOPTERIN-MONO-P-DEPHOS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04621" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97679" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhpmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhnpt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DNTPPA" metaid="R_DNTPPA" name="Dihydroneopterin triphosphate pyrophosphatase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DNTPPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DNTPPA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:H2NEOPTERINP3PYROPHOSPHOHYDRO-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.67" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04638" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97680" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25302" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25303" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25304" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25305" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0099" />
            <fbc:geneProductRef fbc:geneProduct="G_b1865" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ahdt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhpmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DOGULNR" metaid="R_DOGULNR" name="2,3 dioxo-L-gulonate reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DOGULNR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DOGULNR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-703" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.130" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02637" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97688" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21924" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21925" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21926" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21927" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3575" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_23doguln_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3dhguln_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DPCOAK" metaid="R_DPCOAK" name="Dephospho-CoA kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DPCOAK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DPCOAK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEPHOSPHOCOAKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.24" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00130" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97762" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18245" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18246" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18247" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18248" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dpcoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DPR" metaid="R_DPR" name="2-dehydropantoate 2-reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DPR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DPR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-DEHYDROPANTOATE-REDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.169" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02472" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97779" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16233" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16234" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16235" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16236" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0425" />
            <fbc:geneProductRef fbc:geneProduct="G_b3774" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dhp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pant__R_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DRPA" metaid="R_DRPA" name="Deoxyribose-phosphate aldolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DRPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DRPA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYRIBOSE-P-ALD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01066" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97787" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12821" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12822" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12823" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12824" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4381" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dr5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DSERt2" metaid="R_DSERt2" name="D-serine transport via proton symport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DSERt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DSERt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5130" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97798" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4208" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__D_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ser__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DTMPK" metaid="R_DTMPK" name="DTMP kinase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DTMPK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DTMPK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTMPKI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.12" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.13" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02094" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97804" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13517" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13518" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13519" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13520" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1098" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DURIK1" metaid="R_DURIK1" name="Deoxyuridine kinase (ATP:Deoxyuridine)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DURIK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DURIK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DURIDKI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.145" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.21" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02099" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97817" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28206" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28207" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28208" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28209" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1238" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_duri_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dump_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_DURIPP" metaid="R_DURIPP" name="Deoxyuridine phosphorylase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DURIPP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DURIPP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:URA-PHOSPH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02484" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97818" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22824" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22825" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22826" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22827" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4382" />
            <fbc:geneProductRef fbc:geneProduct="G_b4384" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_duri_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dr1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ura_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DURIt2" metaid="R_DURIt2" name="Deoxyuridine transport in via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DURIt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DURIt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108F" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97820" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29967" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29968" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29969" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29970" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2964" />
            <fbc:geneProductRef fbc:geneProduct="G_b2393" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_duri_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_duri_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DUTPDP" metaid="R_DUTPDP" name="DUTP diphosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DUTPDP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DUTPDP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DUTP-PYROP-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.23" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02100" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97822" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10248" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10249" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10250" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10251" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3640" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dutp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dump_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DXPRIi" metaid="R_DXPRIi" name="1-deoxy-D-xylulose reductoisomerase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DXPRIi">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DXPRIi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DXPREDISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.267" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05688" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97830" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13717" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13718" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13719" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13720" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0173" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dxyl5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2me4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DXPS" metaid="R_DXPS" name="1-deoxy-D-xylulose 5-phosphate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DXPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DXPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DXS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.2.1.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05636" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97831" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12605" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12606" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12607" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12608" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0420" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dxyl5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_DXYLK" metaid="R_DXYLK" name="1-Deoxy-D-xylulose kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_DXYLK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/DXYLK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-382" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97832" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27990" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27991" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27992" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27993" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3564" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dxyl_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dxyl5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_D_LACt2" metaid="R_D_LACt2" name="D lactate transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_D_LACt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/D_LACt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-515" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97838" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2975" />
            <fbc:geneProductRef fbc:geneProduct="G_b3603" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lac__D_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lac__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_E4PD" metaid="R_E4PD" name="Erythrose 4-phosphate dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_E4PD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/E4PD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ERYTH4PDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.72" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01825" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97842" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12056" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12057" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12058" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12059" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2927" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_e4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4per_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ECAP_EC" metaid="R_ECAP_EC" name="ECA polymerization (Ecoli)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ECAP_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ECAP_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97877" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_unagamuf_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_eca_EC_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udcpdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EDA" metaid="R_EDA" name="2-dehydro-3-deoxy-phosphogluconate aldolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EDA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/EDA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDPGALDOL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.14" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.55" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05605" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97897" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17089" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17090" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17091" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17092" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1850" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ddg6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EDD" metaid="R_EDD" name="6-phosphogluconate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EDD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/EDD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PGLUCONDEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02036" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97898" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17277" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17278" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17279" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17280" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1851" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_6pgc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ddg6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EDTXS1" metaid="R_EDTXS1" name="Endotoxin Synthesis (lauroyl transferase)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EDTXS1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/EDTXS1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.241" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97899" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1054" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ddcaACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_kdo2lipid4_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_kdo2lipid4L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EDTXS2" metaid="R_EDTXS2" name="Endotoxin Synthesis (myristoyl transferase)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EDTXS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/EDTXS2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.243" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97901" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1855" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_kdo2lipid4L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_myrsACP_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lipa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EDTXS3" metaid="R_EDTXS3" name="Endotoxin Synthesis (palmitoleoyl ACP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EDTXS3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/EDTXS3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97903" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2378" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hdeACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_kdo2lipid4_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_kdo2lipid4p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EDTXS4" metaid="R_EDTXS4" name="Endotoxin Synthesis (myristoyl transferase)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EDTXS4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/EDTXS4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97904" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1855" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_kdo2lipid4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_myrsACP_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lipa_cold_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ENO" metaid="R_ENO" name="Enolase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ENO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ENO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2PGADEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.11" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00658" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97932" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10164" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10165" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10166" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10167" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2779" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2pg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pep_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ENTCS_1" metaid="R_ENTCS_1" name="Enterochelin synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ENTCS_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ENTCS_1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97933" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28566" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28567" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28568" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28569" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0583" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_23dhba_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_seramp_c" stoichiometry="3" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_enter_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="6" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ETHAAL" metaid="R_ETHAAL" name="Ethanolamine ammonia-lyase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ETHAAL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ETHAAL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ETHAMLY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.1.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00749" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97968" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15313" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15314" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15315" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15316" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2441" />
            <fbc:geneProductRef fbc:geneProduct="G_b2440" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_etha_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ETOHt2r" metaid="R_ETOHt2r" name="Ethanol reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ETOHt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ETOHt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97981" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_etoh_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_etoh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_12ppd__S_e" metaid="R_EX_12ppd__S_e" name="(S)-Propane-1,2-diol exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_12ppd__S_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/12ppd__S" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_12ppd__S_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_15dap_e" metaid="R_EX_15dap_e" name="1,5-Diaminopentane exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_15dap_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/15dap" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_15dap_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_26dap__M_e" metaid="R_EX_26dap__M_e" name="Meso-2,6-Diaminoheptanedioate exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_26dap__M_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/26dap__M" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_26dap__M_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_2ddglcn_e" metaid="R_EX_2ddglcn_e" name="2-Dehydro-3-deoxy-D-gluconate exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_2ddglcn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/2ddglcn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ddglcn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_3hcinnm_e" metaid="R_EX_3hcinnm_e" name="3-hydroxycinnamic acid exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_3hcinnm_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/3hcinnm" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3hcinnm_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_3hpppn_e" metaid="R_EX_3hpppn_e" name="3-(3-hydroxy-phenyl)propionate exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_3hpppn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/3hpppn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3hpppn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_4abut_e" metaid="R_EX_4abut_e" name="4-Aminobutanoate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_4abut_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/4abut" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4abut_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ac_e" metaid="R_EX_ac_e" name="Acetate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ac_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ac" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ac_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_acac_e" metaid="R_EX_acac_e" name="Acetoacetate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_acac_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/acac" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acac_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_acald_e" metaid="R_EX_acald_e" name="Acetaldehyde exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_acald_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/acald" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acald_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_acgam_e" metaid="R_EX_acgam_e" name="N-Acetyl-D-glucosamine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_acgam_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/acgam" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acgam_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_acmana_e" metaid="R_EX_acmana_e" name="N-Acetyl-D-mannosamine exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_acmana_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/acmana" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acmana_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_acnam_e" metaid="R_EX_acnam_e" name="N-Acetylneuraminate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_acnam_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/acnam" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acnam_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ade_e" metaid="R_EX_ade_e" name="Adenine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ade_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ade" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ade_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_adn_e" metaid="R_EX_adn_e" name="Adenosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_adn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/adn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_akg_e" metaid="R_EX_akg_e" name="2-Oxoglutarate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_akg_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/akg" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ala__D_e" metaid="R_EX_ala__D_e" name="D-Alanine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ala__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ala__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ala__L_e" metaid="R_EX_ala__L_e" name="L-Alanine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ala__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ala__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_alltn_e" metaid="R_EX_alltn_e" name="Allantoin exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_alltn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/alltn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_alltn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_amp_e" metaid="R_EX_amp_e" name="AMP exchange" reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_amp_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/amp" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amp_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_arab__L_e" metaid="R_EX_arab__L_e" name="L-Arabinose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_arab__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/arab__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arab__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_arg__L_e" metaid="R_EX_arg__L_e" name="L-Arginine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_arg__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/arg__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_arg__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_asn__L_e" metaid="R_EX_asn__L_e" name="L-Asparagine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_asn__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/asn__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asn__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_asp__L_e" metaid="R_EX_asp__L_e" name="L-Aspartate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_asp__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/asp__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_but_e" metaid="R_EX_but_e" name="Butyrate (n-C4:0) exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_but_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/but" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_but_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_cbl1_e" metaid="R_EX_cbl1_e" name="Cob(I)alamin exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_cbl1_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/cbl1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cbl1_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_chol_e" metaid="R_EX_chol_e" name="Choline exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_chol_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/chol" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_chol_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_cit_e" metaid="R_EX_cit_e" name="Citrate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_cit_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/cit" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cit_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_co2_e" metaid="R_EX_co2_e" name="CO2 exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_co2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/co2" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_co2_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_crn_e" metaid="R_EX_crn_e" name="L-Carnitine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_crn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/crn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_crn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_csn_e" metaid="R_EX_csn_e" name="Cytosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_csn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/csn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_csn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_cynt_e" metaid="R_EX_cynt_e" name="Cyanate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_cynt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/cynt" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cynt_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_cys__L_e" metaid="R_EX_cys__L_e" name="L-Cysteine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_cys__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/cys__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_cytd_e" metaid="R_EX_cytd_e" name="Cytidine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_cytd_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/cytd" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cytd_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_dad_2_e" metaid="R_EX_dad_2_e" name="Deoxyadenosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_dad_2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/dad_2" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dad_2_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_dcyt_e" metaid="R_EX_dcyt_e" name="Deoxycytidine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_dcyt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/dcyt" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dcyt_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_dgsn_e" metaid="R_EX_dgsn_e" name="Deoxyguanosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_dgsn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/dgsn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgsn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_dha_e" metaid="R_EX_dha_e" name="Dihydroxyacetone exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_dha_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/dha" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dha_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_din_e" metaid="R_EX_din_e" name="Deoxyinosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_din_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/din" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_din_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_dms_e" metaid="R_EX_dms_e" name="Dimethyl sulfide exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_dms_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/dms" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dms_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_dmso_e" metaid="R_EX_dmso_e" name="Dimethyl sulfoxide exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_dmso_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/dmso" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dmso_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_duri_e" metaid="R_EX_duri_e" name="Deoxyuridine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_duri_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/duri" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_duri_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_etoh_e" metaid="R_EX_etoh_e" name="Ethanol exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_etoh_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/etoh" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_etoh_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_fe2_e" metaid="R_EX_fe2_e" name="Fe2+ exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_fe2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/fe2" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fe2_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_for_e" metaid="R_EX_for_e" name="Formate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_for_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/for" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_for_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_fru_e" metaid="R_EX_fru_e" name="D-Fructose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_fru_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/fru" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fru_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_fuc1p__L_e" metaid="R_EX_fuc1p__L_e" name="L-Fucose 1-phosphate exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_fuc1p__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/fuc1p__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fuc1p__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_fuc__L_e" metaid="R_EX_fuc__L_e" name="L-Fucose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_fuc__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/fuc__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fuc__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_fum_e" metaid="R_EX_fum_e" name="Fumarate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_fum_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/fum" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fum_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_g6p_e" metaid="R_EX_g6p_e" name="D-Glucose 6-phosphate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_g6p_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/g6p" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g6p_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_gal_e" metaid="R_EX_gal_e" name="D-Galactose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_gal_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/gal" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gal_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_galct__D_e" metaid="R_EX_galct__D_e" name="D-Galactarate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_galct__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/galct__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galct__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_galctn__D_e" metaid="R_EX_galctn__D_e" name="D-Galactonate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_galctn__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/galctn__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galctn__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_galt_e" metaid="R_EX_galt_e" name="Galactitol exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_galt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/galt" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galt_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_galur_e" metaid="R_EX_galur_e" name="D-Galacturonate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_galur_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/galur" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galur_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_gam_e" metaid="R_EX_gam_e" name="D-Glucosamine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_gam_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/gam" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gam_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_gbbtn_e" metaid="R_EX_gbbtn_e" name="Gamma-butyrobetaine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_gbbtn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/gbbtn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gbbtn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="R_EX_glc__D_e_lower_bound" fbc:upperFluxBound="cobra_0_bound" id="R_EX_glc__D_e" metaid="R_EX_glc__D_e" name="D-Glucose exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glc__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glc__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glc__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glcn_e" metaid="R_EX_glcn_e" name="D-Gluconate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glcn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glcn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glcr_e" metaid="R_EX_glcr_e" name="D-Glucarate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glcr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glcr" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcr_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glcur_e" metaid="R_EX_glcur_e" name="D-Glucuronate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glcur_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glcur" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcur_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_gln__L_e" metaid="R_EX_gln__L_e" name="L-Glutamine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_gln__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/gln__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glu__L_e" metaid="R_EX_glu__L_e" name="L-Glutamate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glu__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glu__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_gly_e" metaid="R_EX_gly_e" name="Glycine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_gly_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/gly" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gly_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glyald_e" metaid="R_EX_glyald_e" name="D-Glyceraldehyde exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glyald_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glyald" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyald_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glyb_e" metaid="R_EX_glyb_e" name="Glycine betaine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glyb_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glyb" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyb_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glyc3p_e" metaid="R_EX_glyc3p_e" name="Glycerol 3-phosphate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glyc3p_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glyc3p" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glyc_e" metaid="R_EX_glyc_e" name="Glycerol exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glyc_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glyc" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_glyclt_e" metaid="R_EX_glyclt_e" name="Glycolate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_glyclt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/glyclt" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyclt_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_gsn_e" metaid="R_EX_gsn_e" name="Guanosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_gsn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/gsn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gsn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_gua_e" metaid="R_EX_gua_e" name="Guanine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_gua_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/gua" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gua_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_h2o_e" metaid="R_EX_h2o_e" name="H2O exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_h2o_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/h2o" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_h_e" metaid="R_EX_h_e" name="H+ exchange" reversible="true" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_h_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/h" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_hdca_e" metaid="R_EX_hdca_e" name="Hexadecanoate (n-C16:0) exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_hdca_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/hdca" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hdca_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_his__L_e" metaid="R_EX_his__L_e" name="L-Histidine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_his__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/his__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_his__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_hxan_e" metaid="R_EX_hxan_e" name="Hypoxanthine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_hxan_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/hxan" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hxan_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_idon__L_e" metaid="R_EX_idon__L_e" name="L-Idonate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_idon__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/idon__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_idon__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ile__L_e" metaid="R_EX_ile__L_e" name="L-Isoleucine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ile__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ile__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ile__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_indole_e" metaid="R_EX_indole_e" name="Indole exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_indole_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/indole" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_indole_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ins_e" metaid="R_EX_ins_e" name="Inosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ins_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ins" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ins_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_k_e" metaid="R_EX_k_e" name="K+ exchange" reversible="true" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_k_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/k" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_k_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_lac__D_e" metaid="R_EX_lac__D_e" name="D-lactate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_lac__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/lac__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lac__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_lac__L_e" metaid="R_EX_lac__L_e" name="L-Lactate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_lac__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/lac__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lac__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_lcts_e" metaid="R_EX_lcts_e" name="Lactose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_lcts_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/lcts" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lcts_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_leu__L_e" metaid="R_EX_leu__L_e" name="L-Leucine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_leu__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/leu__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_leu__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_lys__L_e" metaid="R_EX_lys__L_e" name="L-Lysine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_lys__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/lys__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lys__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_mal__L_e" metaid="R_EX_mal__L_e" name="L-Malate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_mal__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/mal__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mal__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_malt_e" metaid="R_EX_malt_e" name="Maltose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_malt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/malt" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malt_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_malthx_e" metaid="R_EX_malthx_e" name="Maltohexaose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_malthx_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/malthx" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malthx_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_maltpt_e" metaid="R_EX_maltpt_e" name="Maltopentaose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_maltpt_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/maltpt" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltpt_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_malttr_e" metaid="R_EX_malttr_e" name="Maltotriose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_malttr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/malttr" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malttr_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_maltttr_e" metaid="R_EX_maltttr_e" name="Maltotetraose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_maltttr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/maltttr" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltttr_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_man6p_e" metaid="R_EX_man6p_e" name="D-Mannose 6-phosphate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_man6p_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/man6p" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_man6p_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_man_e" metaid="R_EX_man_e" name="D-Mannose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_man_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/man" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_man_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_melib_e" metaid="R_EX_melib_e" name="Melibiose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_melib_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/melib" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_melib_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_met__D_e" metaid="R_EX_met__D_e" name="D-Methionine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_met__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/met__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_met__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_met__L_e" metaid="R_EX_met__L_e" name="L-Methionine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_met__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/met__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_met__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_mnl_e" metaid="R_EX_mnl_e" name="D-Mannitol exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_mnl_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/mnl" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mnl_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_na1_e" metaid="R_EX_na1_e" name="Sodium exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_na1_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/na1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_nac_e" metaid="R_EX_nac_e" name="Nicotinate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_nac_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/nac" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nac_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_nad_e" metaid="R_EX_nad_e" name="Nicotinamide adenine dinucleotide exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_nad_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/nad" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_nh4_e" metaid="R_EX_nh4_e" name="Ammonia exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_nh4_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/nh4" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nh4_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_nmn_e" metaid="R_EX_nmn_e" name="NMN exchange" reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_nmn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/nmn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nmn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_no2_e" metaid="R_EX_no2_e" name="Nitrite exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_no2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/no2" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_no2_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_no3_e" metaid="R_EX_no3_e" name="Nitrate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_no3_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/no3" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_no3_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="R_EX_o2_e_lower_bound" fbc:upperFluxBound="cobra_0_bound" id="R_EX_o2_e" metaid="R_EX_o2_e" name="O2 exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_o2_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/o2" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ocdca_e" metaid="R_EX_ocdca_e" name="Octadecanoate (n-C18:0) exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ocdca_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ocdca" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ocdca_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_orn_e" metaid="R_EX_orn_e" name="Ornithine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_orn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/orn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_phe__L_e" metaid="R_EX_phe__L_e" name="L-Phenylalanine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_phe__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/phe__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_phe__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_pi_e" metaid="R_EX_pi_e" name="Phosphate exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_pi_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/pi" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_pnto__R_e" metaid="R_EX_pnto__R_e" name="(R)-Pantothenate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_pnto__R_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/pnto__R" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pnto__R_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_pppn_e" metaid="R_EX_pppn_e" name="Phenylpropanoate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_pppn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/pppn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pppn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_pro__L_e" metaid="R_EX_pro__L_e" name="L-Proline exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_pro__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/pro__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pro__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ptrc_e" metaid="R_EX_ptrc_e" name="Putrescine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ptrc_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ptrc" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ptrc_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_pyr_e" metaid="R_EX_pyr_e" name="Pyruvate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_pyr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/pyr" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_rib__D_e" metaid="R_EX_rib__D_e" name="D-Ribose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_rib__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/rib__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rib__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_rmn_e" metaid="R_EX_rmn_e" name="L-Rhamnose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_rmn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/rmn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rmn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_sbt__D_e" metaid="R_EX_sbt__D_e" name="D-Sorbitol exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_sbt__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/sbt__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sbt__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ser__D_e" metaid="R_EX_ser__D_e" name="D-Serine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ser__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ser__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ser__L_e" metaid="R_EX_ser__L_e" name="L-Serine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ser__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ser__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_EX_so4_e" metaid="R_EX_so4_e" name="Sulfate exchange" reversible="true"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_so4_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/so4" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_so4_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_spmd_e" metaid="R_EX_spmd_e" name="Spermidine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_spmd_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/spmd" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_spmd_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_succ_e" metaid="R_EX_succ_e" name="Succinate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_succ_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/succ" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_sucr_e" metaid="R_EX_sucr_e" name="Sucrose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_sucr_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/sucr" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucr_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_tartr__L_e" metaid="R_EX_tartr__L_e" name="L-tartrate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_tartr__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/tartr__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tartr__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_taur_e" metaid="R_EX_taur_e" name="Taurine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_taur_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/taur" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_taur_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_thm_e" metaid="R_EX_thm_e" name="Thiamin exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_thm_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/thm" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thm_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_thr__L_e" metaid="R_EX_thr__L_e" name="L-Threonine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_thr__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/thr__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thr__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_thymd_e" metaid="R_EX_thymd_e" name="Thymidine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_thymd_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/thymd" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thymd_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_tma_e" metaid="R_EX_tma_e" name="Trimethylamine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_tma_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/tma" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tma_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_tmao_e" metaid="R_EX_tmao_e" name="Trimethylamine N-oxide exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_tmao_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/tmao" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tmao_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_tre_e" metaid="R_EX_tre_e" name="Trehalose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_tre_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/tre" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tre_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_trp__L_e" metaid="R_EX_trp__L_e" name="L-Tryptophan exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_trp__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/trp__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trp__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_tsul_e" metaid="R_EX_tsul_e" name="Thiosulfate exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_tsul_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/tsul" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tsul_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ttdca_e" metaid="R_EX_ttdca_e" name="Tetradecanoate (n-C14:0) exchange"
      reversible="false" sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ttdca_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ttdca" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ttdca_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_tyr__L_e" metaid="R_EX_tyr__L_e" name="L-Tyrosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_tyr__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/tyr__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tyr__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_ura_e" metaid="R_EX_ura_e" name="Uracil exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_ura_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ura" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ura_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_urea_e" metaid="R_EX_urea_e" name="Urea exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_urea_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/urea" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_urea_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_uri_e" metaid="R_EX_uri_e" name="Uridine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_uri_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/uri" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uri_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_val__L_e" metaid="R_EX_val__L_e" name="L-Valine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_val__L_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/val__L" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_val__L_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_xan_e" metaid="R_EX_xan_e" name="Xanthine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_xan_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/xan" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xan_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_xtsn_e" metaid="R_EX_xtsn_e" name="Xanthosine exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_xtsn_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/xtsn" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xtsn_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_EX_xyl__D_e" metaid="R_EX_xyl__D_e" name="D-Xylose exchange" reversible="false"
      sboTerm="SBO:0000627">
        <annotation xmlns:sbml="http://www.sbml.org/sbml/level3/version1/core">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_EX_xyl__D_e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/xyl__D" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xyl__D_e" stoichiometry="1" />
        </listOfReactants>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_F6PA" metaid="R_F6PA" name="Fructose 6-phosphate aldolase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_F6PA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/F6PA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-313" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.-" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99075" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28002" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28003" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28004" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28005" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0825" />
            <fbc:geneProductRef fbc:geneProduct="G_b3946" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_f6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dha_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FAO1" metaid="R_FAO1" name="Fatty acid oxidation (tetradecanoate)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FAO1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FAO1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99245" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0221" />
            <fbc:geneProductRef fbc:geneProduct="G_b1805" />
            <fbc:geneProductRef fbc:geneProduct="G_b3845" />
            <fbc:geneProductRef fbc:geneProduct="G_b3846" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fad_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ttdca_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_accoa_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fadh2_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="6" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FAO2" metaid="R_FAO2" name="Fatty acid oxidation (n-C16:0)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FAO2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FAO2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99257" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0221" />
            <fbc:geneProductRef fbc:geneProduct="G_b1805" />
            <fbc:geneProductRef fbc:geneProduct="G_b3845" />
            <fbc:geneProductRef fbc:geneProduct="G_b3846" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fad_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hdca_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="7" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_accoa_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fadh2_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="7" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FAO3" metaid="R_FAO3" name="Fatty acid oxidation (octadecanoate)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FAO3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FAO3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99259" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0221" />
            <fbc:geneProductRef fbc:geneProduct="G_b1805" />
            <fbc:geneProductRef fbc:geneProduct="G_b3845" />
            <fbc:geneProductRef fbc:geneProduct="G_b3846" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="9" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fad_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ocdca_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_accoa_c" stoichiometry="9" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fadh2_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FAO4" metaid="R_FAO4" name="Fatty acid oxidation (Butanoyl-CoA )" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FAO4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FAO4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99260" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0221" />
            <fbc:geneProductRef fbc:geneProduct="G_b3846" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_btcoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aacoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fadh2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_FBA" metaid="R_FBA" name="Fructose-bisphosphate aldolase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FBA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FBA" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01068" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99459" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14729" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14730" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14731" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14732" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2925" />
            <fbc:geneProductRef fbc:geneProduct="G_b1773" />
            <fbc:geneProductRef fbc:geneProduct="G_b2097" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FBP" metaid="R_FBP" name="Fructose-bisphosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FBP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FBP" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.11" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99465" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11064" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11065" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11066" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11067" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4232" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_FCI" metaid="R_FCI" name="L-fucose isomerase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FCI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FCI" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99468" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2802" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fuc__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fcl__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FCLK" metaid="R_FCLK" name="L-fuculokinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FCLK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FCLK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FUCULOKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.51" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03241" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99469" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12376" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12377" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12378" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12379" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2803" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fcl__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fc1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_FCLPA" metaid="R_FCLPA" name="L-fuculose 1-phosphate aldolase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FCLPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FCLPA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FUCPALDOL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.17" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02262" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99470" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12933" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12934" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12935" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12936" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2800" />
            <fbc:geneProductRef fbc:geneProduct="G_b2738" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fc1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lald__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FCLT" metaid="R_FCLT" name="Ferrochelatase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FCLT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FCLT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PROTOHEMEFERROCHELAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.99.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00310" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99471" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22584" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22585" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22586" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22587" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0475" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fe2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppp9_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pheme_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FDH2" metaid="R_FDH2" name="Formate dehydrogenase (quinone-8: 2 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FDH2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FDH2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99476" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4079" />
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1474" />
              <fbc:geneProductRef fbc:geneProduct="G_b1476" />
              <fbc:geneProductRef fbc:geneProduct="G_b1475" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3894" />
              <fbc:geneProductRef fbc:geneProduct="G_b3892" />
              <fbc:geneProductRef fbc:geneProduct="G_b3893" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FDH3" metaid="R_FDH3" name="Formate Dehydrogenase (menaquinone-8: 2 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FDH3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FDH3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99477" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3894" />
              <fbc:geneProductRef fbc:geneProduct="G_b3892" />
              <fbc:geneProductRef fbc:geneProduct="G_b3893" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1474" />
              <fbc:geneProductRef fbc:geneProduct="G_b1476" />
              <fbc:geneProductRef fbc:geneProduct="G_b1475" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FE2abc" metaid="R_FE2abc" name="Iron (II) transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FE2abc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FE2abc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99504" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3409" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fe2_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fe2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FFSD" metaid="R_FFSD" name="Beta-fructofuranosidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FFSD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FFSD" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.26" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03921" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99590" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_suc6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fru_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FHL" metaid="R_FHL" name="Formate-hydrogen lyase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FHL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FHL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FHLMULTI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99594" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27610" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27611" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27612" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27613" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4079" />
            <fbc:geneProductRef fbc:geneProduct="G_b2719" />
            <fbc:geneProductRef fbc:geneProduct="G_b2722" />
            <fbc:geneProductRef fbc:geneProduct="G_b2723" />
            <fbc:geneProductRef fbc:geneProduct="G_b2721" />
            <fbc:geneProductRef fbc:geneProduct="G_b2724" />
            <fbc:geneProductRef fbc:geneProduct="G_b2720" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FMNAT" metaid="R_FMNAT" name="FMN adenylyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FMNAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FMNAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FADSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00161" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95501" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17237" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17238" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17239" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17240" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0025" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fmn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_FORt" metaid="R_FORt" name="Formate transport via diffusion" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FORt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FORt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99620" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29679" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29680" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29681" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29682" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2492" />
            <fbc:geneProductRef fbc:geneProduct="G_b0904" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_for_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FRD2" metaid="R_FRD2" name="Fumarate reductase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FRD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FRD2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99637" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29215" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29216" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29217" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29218" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4153" />
            <fbc:geneProductRef fbc:geneProduct="G_b4151" />
            <fbc:geneProductRef fbc:geneProduct="G_b4154" />
            <fbc:geneProductRef fbc:geneProduct="G_b4152" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fum_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mql8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mqn8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FRD3" metaid="R_FRD3" name="Fumarate reductase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FRD3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FRD3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.99.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99639" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4153" />
            <fbc:geneProductRef fbc:geneProduct="G_b4151" />
            <fbc:geneProductRef fbc:geneProduct="G_b4154" />
            <fbc:geneProductRef fbc:geneProduct="G_b4152" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fum_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FRUK" metaid="R_FRUK" name="Fructose-1-phosphate kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FRUK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FRUK" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.56" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99652" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14213" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14214" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14215" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14216" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2168" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_f1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FRUpts" metaid="R_FRUpts" name="D-fructose transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FRUpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FRUpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99661" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2167" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2169" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fru_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FRUpts2" metaid="R_FRUpts2" name="Fructose transport via PEP:Pyr PTS (f6p generating)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FRUpts2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FRUpts2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99662" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1819" />
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b1818" />
            <fbc:geneProductRef fbc:geneProduct="G_b1817" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fru_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FTHFD" metaid="R_FTHFD" name="Formyltetrahydrofolate deformylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FTHFD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FTHFD" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00944" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99669" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19833" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19834" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19835" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19836" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1232" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_10fthf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FUCPt6_2" metaid="R_FUCPt6_2" name="Fucose 1-phosphate transport via phosphate antiport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FUCPt6_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FUCPt6_2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99702" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3666" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fuc1p__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fuc1p__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_e" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_FUCt" metaid="R_FUCt" name="L-fucose transport via proton symport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FUCt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FUCt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-20" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99703" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2801" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fuc__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fuc__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_FUM" metaid="R_FUM" name="Fumarase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FUM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FUM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FUMHYDR-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01082" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99705" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12460" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12461" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12462" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12463" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1611" />
            <fbc:geneProductRef fbc:geneProduct="G_b1612" />
            <fbc:geneProductRef fbc:geneProduct="G_b4122" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fum_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mal__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FUMt2_2" metaid="R_FUMt2_2" name="Fumarate transport via proton symport (2 H)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FUMt2_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FUMt2_2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-121B" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99711" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29331" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29332" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29333" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29334" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3528" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fum_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_FUMt2_3" metaid="R_FUMt2_3" name="Fumarate transport via proton symport (3 H)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_FUMt2_3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/FUMt2_3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-299" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99712" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0621" />
            <fbc:geneProductRef fbc:geneProduct="G_b4123" />
            <fbc:geneProductRef fbc:geneProduct="G_b4138" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fum_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="3" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G1PACT" metaid="R_G1PACT" name="Glucosamine-1-phosphate N-acetyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G1PACT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G1PACT" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.157" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99841" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3730" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gam1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acgam1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G1PP" metaid="R_G1PP" name="Glucose-1-phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G1PP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G1PP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCOSE-1-PHOSPHAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-17330" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00304" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99844" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19933" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19934" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19935" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19936" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1002" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G1PTT" metaid="R_G1PTT" name="Glucose-1-phosphate thymidylyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G1PTT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G1PTT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDPGLUCOSEPP-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.24" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02328" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99846" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15225" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15226" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15227" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15228" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3789" />
            <fbc:geneProductRef fbc:geneProduct="G_b2039" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dttp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtdpglu_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G1SAT" metaid="R_G1SAT" name="Glutamate-1-semialdehyde aminotransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G1SAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G1SAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GSAAMINOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.3.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02272" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99850" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14265" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14266" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14267" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14268" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0154" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu1sa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5aop_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_G3PD2" metaid="R_G3PD2" name="Glycerol-3-phosphate dehydrogenase (NADP)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G3PD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G3PD2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.94" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00844" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99877" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11096" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11097" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11098" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11099" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3608" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G3PD5" metaid="R_G3PD5" name="Glycerol-3-phosphate dehydrogenase (ubiquinone-8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G3PD5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G3PD5" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99880" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28751" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28752" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28753" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28754" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2241" />
              <fbc:geneProductRef fbc:geneProduct="G_b2242" />
              <fbc:geneProductRef fbc:geneProduct="G_b2243" />
            </fbc:and>
            <fbc:geneProductRef fbc:geneProduct="G_b3426" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G3PD6" metaid="R_G3PD6" name="Glycerol-3-phosphate dehydrogenase (menaquinone-8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G3PD6">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G3PD6" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99881" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29223" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29224" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29225" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29226" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2241" />
            <fbc:geneProductRef fbc:geneProduct="G_b2242" />
            <fbc:geneProductRef fbc:geneProduct="G_b2243" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G3PD7" metaid="R_G3PD7" name="Glycerol-3-phosphate dehydrogenase (demethylmenaquinone-8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G3PD7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G3PD7" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.99.5" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99882" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2241" />
            <fbc:geneProductRef fbc:geneProduct="G_b2242" />
            <fbc:geneProductRef fbc:geneProduct="G_b2243" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G5SADs" metaid="R_G5SADs" name="L-glutamate 5-semialdehyde dehydratase (spontaneous)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G5SADs">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G5SADs" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SPONTPRO-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03314" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99897" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28234" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28235" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28236" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28237" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu5sa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_1pyr5c_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G5SD" metaid="R_G5SD" name="Glutamate-5-semialdehyde dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G5SD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G5SD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTSEMIALDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.41" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03313" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99896" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19541" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19542" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19543" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19544" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0243" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu5sa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G6PDA" metaid="R_G6PDA" name="Glucosamine-6-phosphate deaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G6PDA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G6PDA" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.99.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00765" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99905" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0678" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gam6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_G6PDH2r" metaid="R_G6PDH2r" name="Glucose 6-phosphate dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G6PDH2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G6PDH2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLU6PDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.363" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.49" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00835" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99907" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15841" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15842" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15843" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15844" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1852" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_6pgl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_G6Pt6_2" metaid="R_G6Pt6_2" name="Glucose-6-phosphate transport via phosphate antiport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_G6Pt6_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/G6Pt6_2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99913" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3666" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g6p_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_e" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GALCTD" metaid="R_GALCTD" name="Galactarate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALCTD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALCTD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GALACTARDEHYDRA-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.42" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05608" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99955" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16005" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16006" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16007" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16008" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3128" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galct__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5dh4dglc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GALCTND" metaid="R_GALCTND" name="Galactonate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALCTND">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALCTND" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GALACTONDEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.140" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03033" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99957" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18649" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18650" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18651" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18652" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3692" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galctn__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dh3dgal_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GALCTNt2r" metaid="R_GALCTNt2r"
      name="D-galactonate transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALCTNt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALCTNt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-16" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99960" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29835" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29836" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29837" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29838" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3691" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galctn__D_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_galctn__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GALCTt2r" metaid="R_GALCTt2r"
      name="D-galactarte transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALCTt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALCTt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99962" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3127" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galct__D_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_galct__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GALKr" metaid="R_GALKr" name="Galactokinase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALKr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALKr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GALACTOKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01092" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99985" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13553" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13554" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13555" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13556" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0757" />
            <fbc:geneProductRef fbc:geneProduct="G_b2045" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gal_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gal1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GALS3" metaid="R_GALS3" name="A-galactosidase (melibiose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALS3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALS3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALPHAGALACTOSID-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.22" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01101" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06091" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97317" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4119" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_melib_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gal_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GALTpts" metaid="R_GALTpts" name="Galactitol transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALTpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALTpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100018" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2092" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b2094" />
            <fbc:geneProductRef fbc:geneProduct="G_b2093" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galt_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_galt1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GALUi_copy1" metaid="R_GALUi_copy1" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALUi_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUC1PURIDYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.64" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00289" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100022" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19890" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19891" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19892" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1236" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_utp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udpg_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GALURt2r" metaid="R_GALURt2r"
      name="D-galacturonate transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALURt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALURt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100020" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3093" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galur_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_galur_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GALUi_copy2" metaid="R_GALUi_copy2" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALUi_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUC1PURIDYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.64" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00289" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100022" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19890" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19891" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19892" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or>
            <fbc:geneProductRef fbc:geneProduct="G_b1236" />
            <fbc:geneProductRef fbc:geneProduct="G_b2042" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_utp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udpg_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GALabc" metaid="R_GALabc" name="D-galactose transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-18-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.17" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100023" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2150" />
            <fbc:geneProductRef fbc:geneProduct="G_b2148" />
            <fbc:geneProductRef fbc:geneProduct="G_b2149" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gal_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gal_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GALt2" metaid="R_GALt2" name="D galactose transport in via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GALt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GALt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100026" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29019" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29020" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29021" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29022" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2943" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gal_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gal_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GAMpts" metaid="R_GAMpts" name="D-glucosamine transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GAMpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GAMpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100034" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1817" />
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b1818" />
            <fbc:geneProductRef fbc:geneProduct="G_b1819" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gam_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gam6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GAPD" metaid="R_GAPD" name="Glyceraldehyde-3-phosphate dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GAPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GAPD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GAPOXNPHOSPHN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.59" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01061" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100040" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10300" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10301" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10302" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10303" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1417" />
              <fbc:geneProductRef fbc:geneProduct="G_b1416" />
            </fbc:and>
            <fbc:geneProductRef fbc:geneProduct="G_b1779" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_13dpg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GARFT" metaid="R_GARFT" name="Phosphoribosylglycinamide formyltransferase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GARFT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GARFT" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04325" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99623" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15053" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15054" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15055" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15056" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2500" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_10fthf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gar_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fgam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GART" metaid="R_GART" name="GAR transformylase-T" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GART">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GART" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GARTRANSFORMYL2-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.2.-" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06974" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100042" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24829" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24830" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24831" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24832" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1849" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gar_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fgam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GCALDD" metaid="R_GCALDD" name="Glycolaldehyde dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GCALDD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GCALDD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCOLALD-DEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.21" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01333" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100060" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20001" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20002" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20003" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20004" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1415" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gcald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyclt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GDMANE" metaid="R_GDMANE" name="GDP-4-dehydro-6-deoxy-D-mannose epimerase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GDMANE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GDMANE" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100087" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2052" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gdpddman_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdpofuc_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GF6PTA" metaid="R_GF6PTA" name="Glutamine-fructose-6-phosphate transaminase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GF6PTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GF6PTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-GLN-FRUCT-6-P-AMINOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.16" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100107" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3729" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_f6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gam6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GHMT2r" metaid="R_GHMT2r" name="Glycine hydroxymethyltransferase, reversible"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GHMT2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GHMT2r" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00945" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100142" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15481" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15482" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15483" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15484" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2551" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thf_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gly_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mlthf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GK1" metaid="R_GK1" name="Guanylate kinase (GMP:ATP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANYL-KIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00332" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100144" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20780" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20781" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20782" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20783" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3648" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLCDe" metaid="R_GLCDe" name="Glucose dehydrogenase (ubiquinone-8 as acceptor)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCDe">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCDe" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100177" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0124" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glc__D_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glcn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLCNt2r" metaid="R_GLCNt2r" name="D-gluconate transport via proton symport, reversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCNt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCNt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-209" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100209" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28831" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28832" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28833" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28834" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3415" />
            <fbc:geneProductRef fbc:geneProduct="G_b4265" />
            <fbc:geneProductRef fbc:geneProduct="G_b4321" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLCP" metaid="R_GLCP" name="Glycogen phosphorylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100214" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3428" />
            <fbc:geneProductRef fbc:geneProduct="G_b3417" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glycogen_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g1p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLCRAL" metaid="R_GLCRAL" name="5-dehydro-4-deoxyglucarate aldolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCRAL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCRAL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDGALDOL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.20" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02754" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100219" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27726" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27727" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27728" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27729" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3126" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5dh4dglc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2h3oppan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLCRD" metaid="R_GLCRD" name="Glucarate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCRD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCRD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCARDEHYDRA-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.40" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02752" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100220" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14573" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14574" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14575" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14576" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2787" />
            <fbc:geneProductRef fbc:geneProduct="G_b2788" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5dh4dglc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLCRt2r" metaid="R_GLCRt2r" name="D-glucarate transport via proton symport, reversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCRt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCRt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100221" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3127" />
            <fbc:geneProductRef fbc:geneProduct="G_b2789" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcr_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glcr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLCS1" metaid="R_GLCS1" name="Glycogen synthase (ADPGlc)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCS1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCS1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100223" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3429" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adpglc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glycogen_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLCURt2r" metaid="R_GLCURt2r"
      name="D-glucuronate transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCURt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCURt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100234" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28955" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28956" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28957" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28958" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3093" />
            <fbc:geneProductRef fbc:geneProduct="G_b3909" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcur_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glcur_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLCpts" metaid="R_GLCpts" name="D-glucose transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100237" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1817" />
              <fbc:geneProductRef fbc:geneProduct="G_b2416" />
              <fbc:geneProductRef fbc:geneProduct="G_b2415" />
              <fbc:geneProductRef fbc:geneProduct="G_b1818" />
              <fbc:geneProductRef fbc:geneProduct="G_b1819" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2416" />
              <fbc:geneProductRef fbc:geneProduct="G_b2415" />
              <fbc:geneProductRef fbc:geneProduct="G_b2417" />
              <fbc:geneProductRef fbc:geneProduct="G_b1101" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2417" />
              <fbc:geneProductRef fbc:geneProduct="G_b1621" />
              <fbc:geneProductRef fbc:geneProduct="G_b2415" />
              <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glc__D_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLCt2" metaid="R_GLCt2" name="D-glucose transport in via proton symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLCt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLCt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-7077" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100238" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2943" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glc__D_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLGC" metaid="R_GLGC" name="Glucose-1-phosphate adenylyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLGC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLGC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUC1PADENYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.27" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00948" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100243" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12120" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12121" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12122" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12123" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3430" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adpglc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLNS" metaid="R_GLNS" name="Glutamine synthetase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLNS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLNS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTAMINESYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00253" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100024" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16169" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16170" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16171" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16172" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3870" />
            <fbc:geneProductRef fbc:geneProduct="G_b1297" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nh4_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLNabc" metaid="R_GLNabc" name="L-glutamine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLNabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLNabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-12-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100258" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29895#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29896#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29897#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29898#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0811" />
            <fbc:geneProductRef fbc:geneProduct="G_b0810" />
            <fbc:geneProductRef fbc:geneProduct="G_b0809" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLTPD" metaid="R_GLTPD" name="Galactitol-1-phosphate dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLTPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLTPD" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.251" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.M6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05571" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100275" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2091" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galt1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tag6p__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLU5K" metaid="R_GLU5K" name="Glutamate 5-kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLU5K">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLU5K" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.2.11" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00239" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100277" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14877" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14878" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14879" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14880" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0242" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLUABUTt7" metaid="R_GLUABUTt7" name="4-aminobutyrate/glutamate antiport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUABUTt7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUABUTt7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-261" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100276" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28919" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28920" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28921" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28922" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1492" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4abut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4abut_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUCYS" metaid="R_GLUCYS" name="Gamma-glutamylcysteine synthetase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUCYS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUCYS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTCYSLIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00894" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100074" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13285" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13286" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13287" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13288" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2688" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glucys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUDC" metaid="R_GLUDC" name="Glutamate Decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTDECARBOX-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.15" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00261" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95942" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17785" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17786" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17787" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17788" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3517" />
            <fbc:geneProductRef fbc:geneProduct="G_b1493" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4abut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLUDy" metaid="R_GLUDy" name="Glutamate dehydrogenase (NADP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUDy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUDy" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTDEHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.1.13" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.1.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00248" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100086" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11612" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11613" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11614" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11615" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1761" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUN" metaid="R_GLUN" name="Glutaminase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTAMIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.1.13" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.7.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.38" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.3.6" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00256" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100030" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15890" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15891" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15892" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1812" />
            <fbc:geneProductRef fbc:geneProduct="G_b2957" />
            <fbc:geneProductRef fbc:geneProduct="G_b0485" />
            <fbc:geneProductRef fbc:geneProduct="G_b1524" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUPRT" metaid="R_GLUPRT" name="Glutamine phosphoribosyldiphosphate amidotransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUPRT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUPRT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PRPPAMIDOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.14" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01072" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100287" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14905" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14906" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14907" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14908" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2312" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pram_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLUR" metaid="R_GLUR" name="Glutamate racemase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTRACE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00260" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100288" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12813" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12814" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12815" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12816" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3967" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUSy" metaid="R_GLUSy" name="Glutamate synthase (NADPH)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUSy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUSy" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTAMATESYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.1.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00114" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100291" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15501" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15502" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15503" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15504" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3212" />
            <fbc:geneProductRef fbc:geneProduct="G_b3213" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUTRR" metaid="R_GLUTRR" name="Glutamyl-tRNA reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUTRR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUTRR" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.70" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04109" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100294" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1210" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glutrna_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu1sa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trnaglu_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUTRS" metaid="R_GLUTRS" name="Glutamyl-tRNA synthetase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUTRS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUTRS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.1.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.1.1.17" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.1.1.24" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05578" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100295" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2400" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trnaglu_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glutrna_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUabc" metaid="R_GLUabc" name="L-glutamate transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-13-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100299" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29035#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29036#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29037#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29038#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0655" />
            <fbc:geneProductRef fbc:geneProduct="G_b0653" />
            <fbc:geneProductRef fbc:geneProduct="G_b0654" />
            <fbc:geneProductRef fbc:geneProduct="G_b0652" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLUt2r" metaid="R_GLUt2r" name="L glutamate transport via proton symport  reversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100300" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4077" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLUt4" metaid="R_GLUt4" name="Na+/glutamate symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLUt4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLUt4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100285" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29031" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29032" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29033" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29034" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3653" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_na1_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLXCL" metaid="R_GLXCL" name="Glyoxalate carboligase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLXCL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLXCL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYOCARBOLIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.47" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00013" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100303" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10136" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10137" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10138" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10139" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0507" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glx_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2h3oppan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLYALDt" metaid="R_GLYALDt" name="Glyceraldehyde facilitated diffusion"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYALDt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYALDt" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100311" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3927" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyald_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyald_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_0_bound" id="R_GLYAT" metaid="R_GLYAT" name="Glycine C-acetyltransferase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AKBLIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.29" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.37" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00371" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100313" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20736" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20737" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20738" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20739" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3617" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gly_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2aobut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYBabc" metaid="R_GLYBabc" name="Glycine betaine transport via ABC system"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYBabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYBabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-283" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100314" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32783#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32784#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32785#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32786#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2131" />
            <fbc:geneProductRef fbc:geneProduct="G_b2130" />
            <fbc:geneProductRef fbc:geneProduct="G_b2128" />
            <fbc:geneProductRef fbc:geneProduct="G_b2129" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyb_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLYBt2r" metaid="R_GLYBt2r"
      name="Glycine betaine transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYBt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYBt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-29A" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100315" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1801" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyb_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYC3Pabc" metaid="R_GLYC3Pabc" name="Sn-Glycerol 3-phosphate transport via ABC system"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYC3Pabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYC3Pabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.6.3.20-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-34-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.20" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100320" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21668#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21669#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21670#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21671#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3452" />
            <fbc:geneProductRef fbc:geneProduct="G_b3451" />
            <fbc:geneProductRef fbc:geneProduct="G_b3453" />
            <fbc:geneProductRef fbc:geneProduct="G_b3450" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYC3Pt6" metaid="R_GLYC3Pt6" name="Glycerol-3-phosphate : phosphate antiporter"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYC3Pt6">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYC3Pt6" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-22" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100321" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29015" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29016" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29017" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29018" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2240" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCDx" metaid="R_GLYCDx" name="Glycerol dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCDx">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCDx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCDEH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01034" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100326" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13769" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13770" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13771" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13772" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3945" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dha_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCK" metaid="R_GLYCK" name="Glycerate kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLY3KIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.31" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01514" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100328" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23516" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23517" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23518" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23519" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0514" />
            <fbc:geneProductRef fbc:geneProduct="G_b3124" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc__R_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3pg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCL" metaid="R_GLYCL" name="Glycine Cleavage System" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCL" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.8.1.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.2.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01221" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100330" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27758" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27759" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27760" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27761" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2905" />
            <fbc:geneProductRef fbc:geneProduct="G_b0116" />
            <fbc:geneProductRef fbc:geneProduct="G_b2903" />
            <fbc:geneProductRef fbc:geneProduct="G_b2904" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gly_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thf_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mlthf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCLTDx" metaid="R_GLYCLTDx" name="Glycolate dehydrogenase (NAD)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCLTDx">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCLTDx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCOLATE-REDUCTASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.26" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.29" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.79" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00717" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100331" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18229" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18230" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18231" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18232" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3553" />
            <fbc:geneProductRef fbc:geneProduct="G_b1033" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyclt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCLTDy" metaid="R_GLYCLTDy" name="Glycolate dehydrogenase (NADP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCLTDy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCLTDy" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.79" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00465" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100332" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10992" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10993" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10994" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10995" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3553" />
            <fbc:geneProductRef fbc:geneProduct="G_b1033" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyclt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCLTt2r" metaid="R_GLYCLTt2r"
      name="Glycolate transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCLTt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCLTt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-105" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100334" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29411" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29412" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29413" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29414" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3603" />
            <fbc:geneProductRef fbc:geneProduct="G_b2975" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyclt_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyclt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCTO2" metaid="R_GLYCTO2" name="Glycolate oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCTO2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCTO2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100339" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2979" />
            <fbc:geneProductRef fbc:geneProduct="G_b2978" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyclt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCTO3" metaid="R_GLYCTO3" name="Glycolate oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCTO3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCTO3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100340" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2979" />
            <fbc:geneProductRef fbc:geneProduct="G_b2978" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyclt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCTO4" metaid="R_GLYCTO4" name="Glycolate oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCTO4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCTO4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100341" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2979" />
            <fbc:geneProductRef fbc:geneProduct="G_b2978" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyclt_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glx_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLYCt" metaid="R_GLYCt" name="Glycerol transport via channel" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYCt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYCt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-131" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100343" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29675" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29676" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29677" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29678" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3927" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYK" metaid="R_GLYK" name="Glycerol kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCEROL-KIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.30" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00847" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100348" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21644" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21645" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21646" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21647" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3926" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GLYOX" metaid="R_GLYOX" name="Hydroxyacylglutathione hydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYOX">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYOX" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYOXII-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.2.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01736" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100353" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25245" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25246" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25247" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25248" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0212" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lgt__S_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gthrd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lac__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GLYt2r" metaid="R_GLYt2r" name="Glycine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GLYt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GLYt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-62B" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100368" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28899" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28900" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28901" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28902" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4208" />
            <fbc:geneProductRef fbc:geneProduct="G_b1801" />
            <fbc:geneProductRef fbc:geneProduct="G_b0007" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gly_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gly_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GMAND" metaid="R_GMAND" name="GDP-D-mannose dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GMAND">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GMAND" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GDPMANDEHYDRA-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.47" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00888" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100377" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23820" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23821" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23822" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23823" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2053" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gdpmann_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdpddman_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GMHEPAT" metaid="R_GMHEPAT" name="D-glycero-D-manno-hepose 1-phosphate adenyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GMHEPAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GMHEPAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-4342" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.70" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05644" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100378" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27465" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27466" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27467" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27468" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3052" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gmhep1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adphep_DD_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GMHEPK" metaid="R_GMHEPK" name="D-glycero-D-manno-heptose 7-phosphate kinase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GMHEPK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GMHEPK" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100379" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3052" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gmhep7p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gmhep17bp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GMHEPPA" metaid="R_GMHEPPA" name="D-glycero-D-manno-heptose 1,7-bisphosphate phosphatase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GMHEPPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GMHEPPA" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100380" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0200" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gmhep17bp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gmhep1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GMPR" metaid="R_GMPR" name="GMP reductase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GMPR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GMPR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GMP-REDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.7.1.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01134" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100382" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17185" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17186" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17187" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17188" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0104" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_imp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GMPS2" metaid="R_GMPS2" name="GMP synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GMPS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GMPS2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GMP-SYN-GLUT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01231" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100384" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11680" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11681" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11682" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11683" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2507" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GND" metaid="R_GND" name="Phosphogluconate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GND">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GND" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-9952" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.351" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.44" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01528" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100389" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10116" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10117" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10118" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10119" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2029" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_6pgc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ru5p__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GNK" metaid="R_GNK" name="Gluconokinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GNK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GNK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCONOKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01737" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100390" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19433" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19434" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19435" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19436" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4268" />
            <fbc:geneProductRef fbc:geneProduct="G_b3437" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_6pgc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GOFUCR" metaid="R_GOFUCR" name="GDP-4-oxo-L-fucose reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GOFUCR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GOFUCR" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.271" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100393" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2052" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gdpofuc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdpfuc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GP4GH" metaid="R_GP4GH" name="Gp4G hydrolase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GP4GH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GP4GH" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100398" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30203" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30204" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30205" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30206" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0049" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gp4g_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdp_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GPDDA1" metaid="R_GPDDA1"
      name="Glycerophosphodiester phosphodiesterase (Glycerophosphocholine)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GPDDA1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GPDDA1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.1.4.2-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.4.46" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01030" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100410" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16061" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16062" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16063" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16064" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2239" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g3pc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_chol_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GPDDA2" metaid="R_GPDDA2"
      name="Glycerophosphodiester phosphodiesterase (Glycerophosphoethanolamine)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GPDDA2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GPDDA2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14160" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.4.46" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01470" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100411" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29319" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29320" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29321" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29322" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2239" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g3pe_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_etha_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GPDDA3" metaid="R_GPDDA3"
      name="Glycerophosphodiester phosphodiesterase (Glycerophosphoserine)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GPDDA3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GPDDA3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14136" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.4.46" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100412" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29875" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29876" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29877" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29878" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2239" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g3ps_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ser__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GPDDA4" metaid="R_GPDDA4"
      name="Glycerophosphodiester phosphodiesterase (Glycerophosphoglycerol)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GPDDA4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GPDDA4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14073" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.4.46" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100413" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2239" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g3pg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GPDDA5" metaid="R_GPDDA5"
      name="Glycerophosphodiester phosphodiesterase (Glycerophosphoinositol)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GPDDA5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GPDDA5" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.1.4.44-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.4.44" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01193" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100414" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16501" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16502" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16503" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16504" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2239" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g3pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_inost_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GRTT" metaid="R_GRTT" name="Geranyltranstransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GRTT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GRTT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:FPPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02003" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99634" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19361" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19362" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19363" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19364" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0421" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_grdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ipdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_frdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GSNK" metaid="R_GSNK" name="Guanosine kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GSNK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GSNK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANOSINEKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.73" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01228" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100432" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27710" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27711" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27712" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27713" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0477" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gsn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GSNt2" metaid="R_GSNt2" name="Guanosine transport in via proton symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GSNt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GSNt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100434" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2964" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gsn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gsn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GSPMDA" metaid="R_GSPMDA" name="Glutathionylspermidine amidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GSPMDA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GSPMDA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GSPAMID-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.78" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01918" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100437" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17173" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17174" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17175" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17176" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2988" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtspmd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gthrd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_spmd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GSPMDS" metaid="R_GSPMDS" name="Glutathionylspermidine synthetase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GSPMDS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GSPMDS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GSPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01917" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100438" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21272" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21273" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21274" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21275" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2988" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gthrd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_spmd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gtspmd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GTHOr" metaid="R_GTHOr" name="Glutathione oxidoreductase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GTHOr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GTHOr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTATHIONE-REDUCT-NADPH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.8.1.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00115" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100098" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11740" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11741" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11742" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11743" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3500" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gthox_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gthrd_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GTHS" metaid="R_GTHS" name="Glutathione synthetase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GTHS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GTHS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTATHIONE-SYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00497" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100450" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13557" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13558" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13559" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13560" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2947" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glucys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gly_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gthrd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GTPCI_1" metaid="R_GTPCI_1" name="GTP cyclohydrolase I" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GTPCI_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GTPCI_1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GTP-CYCLOHYDRO-I-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00424" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100453" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17473" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17474" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17475" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17476" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2153" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ahdt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GTPCII2" metaid="R_GTPCII2" name="GTP cyclohydrolase II (25drapp)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GTPCII2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GTPCII2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GTP-CYCLOHYDRO-II-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.25" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00425" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100455" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23704" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23705" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23706" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23707" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1277" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="3" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_25drapp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GUAD" metaid="R_GUAD" name="Guanine deaminase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GUAD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GUAD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANINE-DEAMINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01676" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100464" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14665" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14666" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14667" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14668" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2883" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gua_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xan_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GUAPRT" metaid="R_GUAPRT" name="Guanine phosphoribosyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GUAPRT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GUAPRT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GUANPRIBOSYLTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.22" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.7" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01229" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100409" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25424" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25425" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25426" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25427" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0238" />
            <fbc:geneProductRef fbc:geneProduct="G_b0125" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gua_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GUAt" metaid="R_GUAt" name="Guanine transport" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GUAt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GUAt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-578" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100465" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gua_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gua_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_GUAt2r" metaid="R_GUAt2r" name="Guanine reversible transport via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GUAt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GUAt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100466" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3654" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gua_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gua_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GUI1" metaid="R_GUI1" name="Glucuronate isomerase (D-glucuronate)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GUI1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GUI1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01482" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100468" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13049" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13050" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13051" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13052" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3092" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glcur_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fruur_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_GUI2" metaid="R_GUI2" name="Glucuronate isomerase (D-galacturonate)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_GUI2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/GUI2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01983" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100469" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3092" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_galur_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tagur_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_H2Ot" metaid="R_H2Ot" name="H2O transport via diffusion" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_H2Ot">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/H2Ot" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-145" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-547" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR98641" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29667" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29668" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29669" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29670" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HBZOPT" metaid="R_HBZOPT" name="Hydroxybenzoate octaprenyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HBZOPT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HBZOPT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4OHBENZOATE-OCTAPRENYLTRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.39" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05615" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100566" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27782" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27783" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27784" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27785" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4040" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4hbz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_octdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3ophb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_HCINNMt2r" metaid="R_HCINNMt2r"
      name="3-hydroxycinnamic acid transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HCINNMt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HCINNMt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-457" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100570" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0353" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3hcinnm_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3hcinnm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_HCO3E" metaid="R_HCO3E" name="HCO3 equilibration reaction" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HCO3E">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HCO3E" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CARBODEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5224" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00132" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10092" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100482" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10748" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10749" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10750" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10751" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0339" />
            <fbc:geneProductRef fbc:geneProduct="G_b0126" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hco3_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HDCAt2" metaid="R_HDCAt2" name="Hexadecanoate transport via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HDCAt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HDCAt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100583" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2344" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hdca_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdca_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HEMEOS" metaid="R_HEMEOS" name="Heme O synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HEMEOS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HEMEOS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R07411" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100596" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28070" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28071" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28072" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28073" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0428" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_frdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pheme_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hemeO_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HETZK" metaid="R_HETZK" name="Hydroxyethylthiazole kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HETZK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HETZK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THIAZOLSYN3-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.50" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04448" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100611" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24212" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24213" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24214" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24215" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2104" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4mhetz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4mpetz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HEX1" metaid="R_HEX1" name="Hexokinase (D-glucose:ATP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HEX1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HEX1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCOKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00299" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100612" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17825" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17826" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17827" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17828" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2388" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glc__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HISTD" metaid="R_HISTD" name="Histidinol dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HISTD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HISTD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-8001" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.23" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01158" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100587" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20641" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20642" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20643" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20644" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2020" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_histd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_his__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HISTP" metaid="R_HISTP" name="Histidinol-phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HISTP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HISTP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HISTIDPHOS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.15" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03013" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100642" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14465" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14466" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14467" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14468" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2022" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hisp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_histd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HISabc" metaid="R_HISabc" name="L-histidine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HISabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HISabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-14-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100644" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29891#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29892#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29893#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29894#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2307" />
            <fbc:geneProductRef fbc:geneProduct="G_b2309" />
            <fbc:geneProductRef fbc:geneProduct="G_b2308" />
            <fbc:geneProductRef fbc:geneProduct="G_b2306" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_his__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_his__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_HISt2r" metaid="R_HISt2r" name="L histidine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HISt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HISt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100645" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/37047" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/37048" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/37049" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/37050" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0112" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_his__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_his__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HKNDDH_1" metaid="R_HKNDDH_1" name="2-hydroxy-6-ketonona-2,4-dienedioic acid hydrolase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HKNDDH_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HKNDDH_1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MHPCHYDROL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.7.1.14" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02603" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100654" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24789" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24790" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24791" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24792" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34187" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34188" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34189" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34190" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0349" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hkndd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_op4en_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HKNTDH_1" metaid="R_HKNTDH_1" name="2-hydroxy-6-ketononotrienedioate hydrolase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HKNTDH_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HKNTDH_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.7.1.14" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06789" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100655" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25046" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25047" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25048" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25049" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0349" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hkntd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_op4en_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HMBS" metaid="R_HMBS" name="Hydroxymethylbilane synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HMBS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HMBS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OHMETHYLBILANESYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.61" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00084" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100658" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13185" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13186" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13187" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13188" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3805" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppbng_c" stoichiometry="4" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hmbil_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="4" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HMPK1" metaid="R_HMPK1" name="Hydroxymethylpyrimidine kinase (ATP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HMPK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HMPK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OHMETPYRKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.49" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03471" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100664" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23096" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23097" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23098" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23099" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2103" />
            <fbc:geneProductRef fbc:geneProduct="G_b2418" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4ahmmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4ampm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HOPNTAL" metaid="R_HOPNTAL" name="4-hydroxy-2-oxopentanoate aldolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HOPNTAL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HOPNTAL" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.n4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.39" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00750" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100679" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0352" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4h2opntn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HPPK2" metaid="R_HPPK2" name="6-hydroxymethyl-dihydropterin pyrophosphokinase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HPPK2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HPPK2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:H2PTERIDINEPYROPHOSPHOKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.6.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03503" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100694" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11412" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11413" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11414" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11415" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0142" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_6hmhpt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_6hmhptpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HPPPNDO_1" metaid="R_HPPPNDO_1" name="2,3-dihydroxypheylpropionate 1,2-dioxygenase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HPPPNDO_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HPPPNDO_1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:1.13.11.16-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.13.11.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04376" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100695" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23840" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23841" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23842" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23843" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0348" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhpppn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hkndd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_HPPPNt2r" metaid="R_HPPPNt2r"
      name="3-(3-hydroxyphenyl)propionate transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HPPPNt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HPPPNt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-61" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100696" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28907" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28908" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28909" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28910" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0353" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3hpppn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3hpppn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_HPYRI" metaid="R_HPYRI" name="Hydroxypyruvate isomerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HPYRI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HPYRI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-305" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.22" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01394" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100692" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11952" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11953" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11954" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11955" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0508" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hpyr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2h3oppan_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HPYRRx" metaid="R_HPYRRx" name="Hydroxypyruvate reductase (NADH)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HPYRRx">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HPYRRx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYCERATE-DEHYDROGENASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.26" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.29" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.81" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01388" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100325" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17905" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17906" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17907" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17908" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1033" />
            <fbc:geneProductRef fbc:geneProduct="G_b3553" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hpyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc__R_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HPYRRy" metaid="R_HPYRRy" name="Hydroxypyruvate reductase (NADPH)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HPYRRy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HPYRRy" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-300" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.79" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.81" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01392" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100336" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18657" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18658" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18659" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18660" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1033" />
            <fbc:geneProductRef fbc:geneProduct="G_b3553" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hpyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc__R_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_HSDy" metaid="R_HSDy" name="Homoserine dehydrogenase (NADPH)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HSDy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HSDy" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01775" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100734" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15761" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15762" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15763" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15764" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3940" />
            <fbc:geneProductRef fbc:geneProduct="G_b0002" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hom__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aspsa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HSK" metaid="R_HSK" name="Homoserine kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HSK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HSK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HOMOSERKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.39" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01771" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100737" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13985" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13986" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13987" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13988" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0003" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hom__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_phom_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HSST" metaid="R_HSST" name="Homoserine O-succinyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HSST">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HSST" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HOMSUCTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.46" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01777" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100741" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22008" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22009" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22010" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22011" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4013" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hom__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_suchms_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HSTPT" metaid="R_HSTPT" name="Histidinol-phosphate transaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HSTPT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HSTPT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HISTAMINOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03243" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95852" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23744" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23745" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23746" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23747" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2021" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_imacp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hisp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HXPRT" metaid="R_HXPRT" name="Hypoxanthine phosphoribosyltransferase (Hypoxanthine)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HXPRT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HXPRT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HYPOXANPRIBOSYLTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01132" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100752" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17973" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17974" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17975" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17976" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0238" />
            <fbc:geneProductRef fbc:geneProduct="G_b0125" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hxan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_imp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HYD1" metaid="R_HYD1" name="Hydrogenase (ubiquinone-8: 2 protons)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HYD1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HYD1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100753" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29059#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29060#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29061#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29062#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2720" />
              <fbc:geneProductRef fbc:geneProduct="G_b2724" />
              <fbc:geneProductRef fbc:geneProduct="G_b2721" />
              <fbc:geneProductRef fbc:geneProduct="G_b2723" />
              <fbc:geneProductRef fbc:geneProduct="G_b2722" />
              <fbc:geneProductRef fbc:geneProduct="G_b2719" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0972" />
              <fbc:geneProductRef fbc:geneProduct="G_b0974" />
              <fbc:geneProductRef fbc:geneProduct="G_b0973" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2994" />
              <fbc:geneProductRef fbc:geneProduct="G_b2997" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HYD2" metaid="R_HYD2" name="Hydrogenase (menaquinone8: 2 protons)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HYD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HYD2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100754" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29051#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29052#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29053#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29054#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2994" />
              <fbc:geneProductRef fbc:geneProduct="G_b2997" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0972" />
              <fbc:geneProductRef fbc:geneProduct="G_b0974" />
              <fbc:geneProductRef fbc:geneProduct="G_b0973" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HYD3" metaid="R_HYD3" name="Hydrogenase (Demethylmenaquinone-8: 2 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HYD3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HYD3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.18.99.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100755" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0972" />
              <fbc:geneProductRef fbc:geneProduct="G_b0974" />
              <fbc:geneProductRef fbc:geneProduct="G_b0973" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2994" />
              <fbc:geneProductRef fbc:geneProduct="G_b2997" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_HYPOE" metaid="R_HYPOE" name="Hypothetical enyme" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HYPOE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HYPOE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14046" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.74" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02494" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100763" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25135" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25136" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25137" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25138" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyam5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pydam_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_HYXNt" metaid="R_HYXNt" name="Hypoxanthine transport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_HYXNt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/HYXNt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-562" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-579" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100749" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hxan_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hxan_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ICDHyr" metaid="R_ICDHyr" name="Isocitrate dehydrogenase (NADP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ICDHyr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ICDHyr" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.42" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00267" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100781" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19629" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19630" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19631" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19632" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1136" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_icit_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ICHORS" metaid="R_ICHORS" name="Isochorismate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ICHORS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ICHORS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ISOCHORSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01717" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100784" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18985" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18986" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18987" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18988" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2265" />
            <fbc:geneProductRef fbc:geneProduct="G_b0593" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_chor_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ichor_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ICHORT" metaid="R_ICHORT" name="Isochorismatase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ICHORT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ICHORT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ISOCHORMAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.3.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03037" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100785" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11112" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11113" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11114" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11115" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0595" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ichor_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_23ddhb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ICL" metaid="R_ICL" name="Isocitrate lyase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ICL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ICL" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00479" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100789" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13245" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13246" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13247" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13248" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4015" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_icit_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_IDOND" metaid="R_IDOND" name="L-idonate 5-dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IDOND">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IDOND" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-12107" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.264" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.366" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05683" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21172" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21173" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21174" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21175" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4267" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5dglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_idon__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IDOND2" metaid="R_IDOND2" name="L-indonate 5-dehydrogenase (NADP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IDOND2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IDOND2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:L-IDONATE-2-DEHYDROGENASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.264" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05684" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100802" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21176" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21177" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21178" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21179" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4267" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5dglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_idon__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_IDONt2r" metaid="R_IDONt2r" name="L-idonate transport via proton symport, reversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IDONt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IDONt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-181A" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.-" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100803" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29571" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29572" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29573" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29574" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4265" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_idon__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_idon__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IG3PS" metaid="R_IG3PS" name="Imidazole-glycerol-3-phosphate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IG3PS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IG3PS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUTAMIDOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.2.M2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04558" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100811" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24793" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24794" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24795" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24796" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2023" />
            <fbc:geneProductRef fbc:geneProduct="G_b2025" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prlp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aicar_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_eig3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IGPDH" metaid="R_IGPDH" name="Imidazoleglycerol-phosphate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IGPDH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IGPDH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:IMIDPHOSDEHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03457" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100813" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11040" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11041" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11042" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11043" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2022" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_eig3p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_imacp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IGPS" metaid="R_IGPS" name="Indole-3-glycerol-phosphate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IGPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IGPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:IGPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.48" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03508" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100814" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23476" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23477" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23478" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23479" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1262" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2cpr5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3ig3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ILETA" metaid="R_ILETA" name="Isoleucine transaminase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ILETA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ILETA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BRANCHED-CHAINAMINOTRANSFERILEU-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.42" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02199" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96228" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24802" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24803" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24804" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3770" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ile__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3mop_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ILEabc" metaid="R_ILEabc" name="L-isoleucine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ILEabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ILEabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-15-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100821" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29819#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29820#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29821#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29822#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3454" />
            <fbc:geneProductRef fbc:geneProduct="G_b3460" />
            <fbc:geneProductRef fbc:geneProduct="G_b3455" />
            <fbc:geneProductRef fbc:geneProduct="G_b3457" />
            <fbc:geneProductRef fbc:geneProduct="G_b3456" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ile__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ile__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ILEt2r" metaid="R_ILEt2r" name="L isoleucine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ILEt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ILEt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100822" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0401" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ile__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ile__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_IMPC" metaid="R_IMPC" name="IMP cyclohydrolase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IMPC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IMPC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:IMPCYCLOHYDROLASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01127" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100783" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18445" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18446" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18447" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18448" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4006" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_imp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fprica_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IMPD" metaid="R_IMPD" name="IMP dehydrogenase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IMPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IMPD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:IMP-DEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.205" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01130" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100830" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11708" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11709" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11710" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11711" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2508" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_imp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xmp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_INDOLEt2r" metaid="R_INDOLEt2r" name="Indole transport via proton symport, reversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_INDOLEt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/INDOLEt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-142" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100835" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29227" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29228" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29229" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29230" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3161" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_indole_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_indole_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_INSK" metaid="R_INSK" name="Insosine kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_INSK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/INSK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:INOSINEKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.73" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01131" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100845" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21140" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21141" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21142" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21143" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0477" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ins_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_imp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_INSt2_copy1" metaid="R_INSt2_copy1" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_INSt2_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108G" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100850" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29963" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29964" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29965" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29966" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2964" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ins_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ins_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_INSt2_copy2" metaid="R_INSt2_copy2" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_INSt2_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108G" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100850" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29963" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29964" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29965" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29966" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2406" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ins_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ins_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IPDDI" metaid="R_IPDDI" name="Isopentenyl-diphosphate D-isomerase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IPDDI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IPDDI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:IPPISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.3.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01123" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100796" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23284" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23285" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23286" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23287" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2889" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ipdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dmpp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IPDPS" metaid="R_IPDPS"
      name="1-hydroxy-2-methyl-2-(E)-butenyl 4-diphosphate reductase (ipdp)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IPDPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IPDPS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100874" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0029" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2mb4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ipdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IPMD" metaid="R_IPMD" name="3-isopropylmalate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IPMD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IPMD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-ISOPROPYLMALDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.85" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04426" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100878" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10892" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10893" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10894" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10895" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0073" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3c2hmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3c4mop_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_IPPMIa" metaid="R_IPPMIa" name="3-isopropylmalate dehydratase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IPPMIa">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IPPMIa" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-8991" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.33" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04001" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100827" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10676" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10677" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10678" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10679" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0072" />
            <fbc:geneProductRef fbc:geneProduct="G_b0071" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3c2hmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ippm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_IPPMIb" metaid="R_IPPMIb" name="2-isopropylmalate hydratase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IPPMIb">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IPPMIb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-ISOPROPYLMALISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.33" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03968" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100826" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16293" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16294" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16295" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16296" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0072" />
            <fbc:geneProductRef fbc:geneProduct="G_b0071" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ippm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3c3hmp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_IPPS" metaid="R_IPPS" name="2-isopropylmalate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_IPPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/IPPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-ISOPROPYLMALATESYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.3.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01213" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100879" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21524" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21525" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21526" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21527" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0074" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3mob_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3c3hmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_0_bound" id="R_KARA1" metaid="R_KARA1"
      name="Ketol-acid reductoisomerase (2,3-dihydroxy-3-methylbutanoate)" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KARA1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KARA1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETOLACTREDUCTOISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.383" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.86" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04439" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97523" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22068" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22069" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22070" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22071" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3774" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_23dhmb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_alac__S_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_KARA2" metaid="R_KARA2" name="Ketol-acid reductoisomerase (2-Acetolactate)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KARA2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KARA2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ACETOOHBUTREDUCTOISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.86" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100902" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13493" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13494" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13495" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13496" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3774" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ahbut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_23dhmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_KAS14" metaid="R_KAS14" name="Beta-ketoacyl-ACP synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KAS14">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KAS14" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-OXOACYL-ACP-SYNTH-BASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.179" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.180" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.41" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.85" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.86" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04355" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100912" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1095" />
            <fbc:geneProductRef fbc:geneProduct="G_b2323" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_actACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_KAS15" metaid="R_KAS15" name="Beta-ketoacyl-ACP synthase (2)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KAS15">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KAS15" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2.3.1.180-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.180" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10707" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100913" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1091" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_actACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_KAS16" metaid="R_KAS16" name="3-hydroxy-myristoyl-ACP synthesis" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KAS16">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KAS16" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100914" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2323" />
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1095" />
              <fbc:geneProductRef fbc:geneProduct="G_b1093" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ddcaACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3hmrsACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_KDOCT2" metaid="R_KDOCT2" name="3-deoxy-manno-octulosonate cytidylyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KDOCT2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KDOCT2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CPM-KDOSYNTH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.38" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100932" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23448" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23449" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23450" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23451" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0918" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_kdo_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ckdo_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_KDOPP" metaid="R_KDOPP" name="3-deoxy-manno-octulosonate-8-phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KDOPP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KDOPP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDO-8PPHOSPHAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.45" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100933" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11500" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11501" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11502" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11503" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_kdo8p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_kdo_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_KDOPS" metaid="R_KDOPS" name="3-deoxy -D-manno-octulosonic -acid 8-phosphate synthase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KDOPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KDOPS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.55" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03254" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100934" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14053" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14054" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14055" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14056" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1215" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ara5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_kdo8p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_KG6PDC" metaid="R_KG6PDC" name="3-keto-L-gulonate 6-phosphate decarboxylase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_KG6PDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/KG6PDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-705" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.85" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R07125" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100935" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14353" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14354" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14355" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14356" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4196" />
            <fbc:geneProductRef fbc:geneProduct="G_b3581" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3dhgulnp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xu5p__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_Kabc" metaid="R_Kabc" name="Potassium ABC transporter" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_Kabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/Kabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.6.3.10-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.6.3.12-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.10" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.12" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100657" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16777#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16778#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16779#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16780#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0697" />
            <fbc:geneProductRef fbc:geneProduct="G_b0698" />
            <fbc:geneProductRef fbc:geneProduct="G_b0696" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_k_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_k_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_Kt2r" metaid="R_Kt2r" name="Potassium reversible transport  via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_Kt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/Kt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100951" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28490" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28491" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28492" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28493" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1363" />
            <fbc:geneProductRef fbc:geneProduct="G_b3849" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_k_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_k_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LACZ" metaid="R_LACZ" name="B-galactosidase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LACZ">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LACZ" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.108" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.23" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01100" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06114" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101000" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10076" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10077" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10078" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10079" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0344" />
            <fbc:geneProductRef fbc:geneProduct="G_b2132" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lcts_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gal_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_LCADi_copy1" metaid="R_LCADi_copy1" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LCADi_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LACTALDDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.22" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01446" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101018" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14277" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14278" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14279" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14280" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or>
            <fbc:geneProductRef fbc:geneProduct="G_b1300" />
            <fbc:geneProductRef fbc:geneProduct="G_b1415" />
            <fbc:geneProductRef fbc:geneProduct="G_b0356" />
            <fbc:geneProductRef fbc:geneProduct="G_b3588" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lald__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lac__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LCADi_copy2" metaid="R_LCADi_copy2" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LCADi_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LACTALDDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.22" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01446" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101018" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14277" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14278" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14279" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14280" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1241" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lald__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lac__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_LCARS" metaid="R_LCARS" name="Lacaldehyde reductase (S-propane-1,2-diol forming)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LCARS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LCARS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LACTALDREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.77" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02258" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95714" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15933" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15934" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15935" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15936" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2799" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lald__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_12ppd__S_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_LCTSt" metaid="R_LCTSt" name="Lactose transport via proton symport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LCTSt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LCTSt" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101025" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28867" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28868" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28869" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28870" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0343" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lcts_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lcts_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_LDH_D" metaid="R_LDH_D" name="D-lactate dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LDH_D">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LDH_D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DLACTDEHYDROGNAD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.28" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00704" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101037" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16369" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16370" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16371" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16372" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2133" />
            <fbc:geneProductRef fbc:geneProduct="G_b1380" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lac__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LDH_D2" metaid="R_LDH_D2" name="D-lactate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LDH_D2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LDH_D2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101039" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29079" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29080" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29081" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29082" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2133" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lac__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LEUTAi" metaid="R_LEUTAi" name="Leucine transaminase (irreversible)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LEUTAi">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LEUTAi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BRANCHED-CHAINAMINOTRANSFERLEU-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.42" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.67" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01090" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96229" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18321" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18322" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18323" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18324" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4054" />
            <fbc:geneProductRef fbc:geneProduct="G_b3770" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4mop_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_leu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LEUabc" metaid="R_LEUabc" name="L-leucine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LEUabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LEUabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-35-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101054" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29807#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29808#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29809#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29810#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3454" />
              <fbc:geneProductRef fbc:geneProduct="G_b3460" />
              <fbc:geneProductRef fbc:geneProduct="G_b3455" />
              <fbc:geneProductRef fbc:geneProduct="G_b3457" />
              <fbc:geneProductRef fbc:geneProduct="G_b3456" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3458" />
              <fbc:geneProductRef fbc:geneProduct="G_b3454" />
              <fbc:geneProductRef fbc:geneProduct="G_b3455" />
              <fbc:geneProductRef fbc:geneProduct="G_b3457" />
              <fbc:geneProductRef fbc:geneProduct="G_b3456" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_leu__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_leu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_LEUt2r" metaid="R_LEUt2r" name="L leucine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LEUt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LEUt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101055" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0401" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_leu__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_leu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LGTHL" metaid="R_LGTHL" name="Lactoylglutathione lyase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LGTHL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LGTHL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYOXI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.4.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02530" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100355" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19069" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19070" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19071" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19072" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1651" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gthrd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mthgxl_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lgt__S_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LPADSS" metaid="R_LPADSS" name="Lipid A disaccaride synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LPADSS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LPADSS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LIPIDADISACCHARIDESYNTH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.182" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04606" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101119" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22668" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22669" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22670" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22671" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0182" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lipidX_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_u23ga_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lipidAds_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LPLIPA1" metaid="R_LPLIPA1" name="Lysophospholipase L (acyl-glycerophosphoglycerol)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LPLIPA1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LPLIPA1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101135" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3825" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_agpg_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3pg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdca_c" stoichiometry="0.36" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdcea_c" stoichiometry="0.07" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ocdcea_c" stoichiometry="0.5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdca_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdcea_c" stoichiometry="0.05" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LPLIPA2" metaid="R_LPLIPA2" name="Lysophospholipase L (acyl-glycerophosphoethanolamine)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LPLIPA2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LPLIPA2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101136" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3825" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_agpe_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3pe_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdca_c" stoichiometry="0.36" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdcea_c" stoichiometry="0.07" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ocdcea_c" stoichiometry="0.5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdca_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdcea_c" stoichiometry="0.05" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LPLIPA3" metaid="R_LPLIPA3" name="Lysophospholipase L (acyl-glycerophosphocholine)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LPLIPA3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LPLIPA3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101137" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3825" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_agpc_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3pc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdca_c" stoichiometry="0.36" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdcea_c" stoichiometry="0.07" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ocdcea_c" stoichiometry="0.5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdca_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdcea_c" stoichiometry="0.05" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LPLIPA4" metaid="R_LPLIPA4"
      name="Lysophospholipase L (acyl transfer to phosphatidylglycerol)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LPLIPA4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LPLIPA4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101138" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3825" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_agpe_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pg_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_apg_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3pe_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LPLIPA5" metaid="R_LPLIPA5"
      name="Lysophospholipase L (acyl transfer to phosphatidylglycerol)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LPLIPA5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LPLIPA5" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101139" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3825" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_agpc_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pg_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_apg_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3pc_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LPLIPA6" metaid="R_LPLIPA6"
      name="Lysophospholipase L (acyl transfer to phosphatidylglycerol)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LPLIPA6">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LPLIPA6" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101140" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3825" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_agpg_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pg_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_apg_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3pg_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LPSSYN_EC" metaid="R_LPSSYN_EC" name="Lipopolysaccharide synthesis (Ecoli)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LPSSYN_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LPSSYN_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101229" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3621" />
            <fbc:geneProductRef fbc:geneProduct="G_b3631" />
            <fbc:geneProductRef fbc:geneProduct="G_b3622" />
            <fbc:geneProductRef fbc:geneProduct="G_b3626" />
            <fbc:geneProductRef fbc:geneProduct="G_b3627" />
            <fbc:geneProductRef fbc:geneProduct="G_b3620" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adphep_LD_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cdpea_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ckdo_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lipa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udpg_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cdp_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="10" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lps_EC_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udp_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LYSDC" metaid="R_LYSDC" name="Lysine decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LYSDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LYSDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LYSDECARBOX-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.18" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00462" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101256" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22352" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22353" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22354" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22355" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0186" />
            <fbc:geneProductRef fbc:geneProduct="G_b4131" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lys__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_15dap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_LYSabc" metaid="R_LYSabc" name="L-lysine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LYSabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LYSabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-3-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101265" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29887#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29888#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29889#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29890#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2308" />
            <fbc:geneProductRef fbc:geneProduct="G_b2310" />
            <fbc:geneProductRef fbc:geneProduct="G_b2307" />
            <fbc:geneProductRef fbc:geneProduct="G_b2306" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lys__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_LYSt2r" metaid="R_LYSt2r" name="L lysine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_LYSt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/LYSt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-58" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101266" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28911" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28912" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28913" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28914" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2156" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lys__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lys__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_L_LACD2" metaid="R_L_LACD2" name="L-Lactate dehydrogenase (ubiquinone)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_L_LACD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/L_LACD2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101273" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3605" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lac__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_L_LACD3" metaid="R_L_LACD3" name="L-Lactate dehydrogenase (menaquinone)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_L_LACD3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/L_LACD3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101275" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3605" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lac__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_L_LACt2r" metaid="R_L_LACt2r" name="L lactate reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_L_LACt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/L_LACt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-104" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101277" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29415" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29416" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29417" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29418" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3603" />
            <fbc:geneProductRef fbc:geneProduct="G_b2975" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lac__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lac__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_M1PD" metaid="R_M1PD" name="Mannitol-1-phosphate 5-dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_M1PD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/M1PD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MANNPDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.17" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02703" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101311" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19661" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19662" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19663" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19664" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3600" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mnl1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MACPD" metaid="R_MACPD" name="Malonyl-ACP decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MACPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MACPD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALONYL-ACPDECARBOX-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.87" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R09482" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101327" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2323" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malACP_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALS" metaid="R_MALS" name="Malate synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.3.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00472" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101347" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18181" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18182" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18183" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18184" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4014" />
            <fbc:geneProductRef fbc:geneProduct="G_b2976" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mal__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALTHXabc" metaid="R_MALTHXabc" name="Maltohexaose transport via ABC system"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALTHXabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALTHXabc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101353" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4036" />
            <fbc:geneProductRef fbc:geneProduct="G_b4035" />
            <fbc:geneProductRef fbc:geneProduct="G_b4032" />
            <fbc:geneProductRef fbc:geneProduct="G_b4033" />
            <fbc:geneProductRef fbc:geneProduct="G_b4034" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malthx_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malthx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALTPTabc" metaid="R_MALTPTabc" name="Maltopentaose transport via ABC system"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALTPTabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALTPTabc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101355" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4036" />
            <fbc:geneProductRef fbc:geneProduct="G_b4032" />
            <fbc:geneProductRef fbc:geneProduct="G_b4035" />
            <fbc:geneProductRef fbc:geneProduct="G_b4033" />
            <fbc:geneProductRef fbc:geneProduct="G_b4034" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltpt_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_maltpt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALTTRabc" metaid="R_MALTTRabc" name="Maltotriose transport via ABC system"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALTTRabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALTTRabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-503" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.-" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101358" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4036" />
            <fbc:geneProductRef fbc:geneProduct="G_b4035" />
            <fbc:geneProductRef fbc:geneProduct="G_b4032" />
            <fbc:geneProductRef fbc:geneProduct="G_b4033" />
            <fbc:geneProductRef fbc:geneProduct="G_b4034" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malttr_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malttr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALTTTRabc" metaid="R_MALTTTRabc" name="Maltotetraose transport via ABC system"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALTTTRabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALTTTRabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-504" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.-" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101360" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4036" />
            <fbc:geneProductRef fbc:geneProduct="G_b4032" />
            <fbc:geneProductRef fbc:geneProduct="G_b4035" />
            <fbc:geneProductRef fbc:geneProduct="G_b4033" />
            <fbc:geneProductRef fbc:geneProduct="G_b4034" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltttr_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_maltttr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALTabc" metaid="R_MALTabc" name="Maltose transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALTabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALTabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.6.3.19-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-16-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.19" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101362" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22132#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22133#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22134#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22135#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b4036" />
            <fbc:geneProductRef fbc:geneProduct="G_b4032" />
            <fbc:geneProductRef fbc:geneProduct="G_b4035" />
            <fbc:geneProductRef fbc:geneProduct="G_b4033" />
            <fbc:geneProductRef fbc:geneProduct="G_b4034" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malt_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALTpts" metaid="R_MALTpts" name="Maltose transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALTpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALTpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101363" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b1621" />
            <fbc:geneProductRef fbc:geneProduct="G_b2417" />
            <fbc:geneProductRef fbc:geneProduct="G_b4036" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malt_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malt6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALt2_2" metaid="R_MALt2_2" name="Malate transport via proton symport (2 H)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALt2_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALt2_2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-121A" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101370" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29339" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29340" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29341" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29342" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3528" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mal__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mal__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MALt2_3" metaid="R_MALt2_3" name="Malate transport via proton symport (3 H)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MALt2_3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MALt2_3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101371" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4123" />
            <fbc:geneProductRef fbc:geneProduct="G_b4138" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mal__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mal__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MAN1PT2" metaid="R_MAN1PT2" name="Mannose-1-phosphate guanylyltransferase (GDP)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MAN1PT2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MAN1PT2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MANNPGUANYLTRANGDP-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.22" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00883" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101376" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12905" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12906" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12907" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12908" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2049" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_man1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdpmann_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MAN6PI" metaid="R_MAN6PI" name="Mannose-6-phosphate isomerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MAN6PI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MAN6PI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MANNPISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01819" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101382" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1613" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_man6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MAN6Pt6_2" metaid="R_MAN6Pt6_2" name="Mannose-6-phosphate transport via phosphate antiport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MAN6Pt6_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MAN6Pt6_2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101384" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3666" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_man6p_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_man6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_e" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MANAO" metaid="R_MANAO" name="Mannonate oxidoreductase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MANAO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MANAO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MANNONOXIDOREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.57" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02454" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101389" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15729" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15730" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15731" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15732" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4323" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mana_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fruur_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MANpts" metaid="R_MANpts" name="D-mannose transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MANpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MANpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101401" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1819" />
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b1818" />
            <fbc:geneProductRef fbc:geneProduct="G_b1817" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_man_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_man6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MCITD" metaid="R_MCITD" name="2-methylcitrate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MCITD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MCITD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-METHYLCITRATE-DEHYDRATASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.79" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04424" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101413" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17725" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17726" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17727" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17728" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0334" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2mcit_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2mcacn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MCITL2" metaid="R_MCITL2" name="Methylisocitrate lyase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MCITL2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MCITL2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHYLISOCITRATE-LYASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.30" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00409" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101415" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16809" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16810" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16811" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16812" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0331" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_micit_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MCITS" metaid="R_MCITS" name="2-methylcitrate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MCITS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MCITS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-METHYLCITRATE-SYNTHASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00931" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101417" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23780" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23781" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23782" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23783" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0333" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_oaa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppcoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2mcit_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MCOATA" metaid="R_MCOATA" name="Malonyl-CoA-ACP transacylase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MCOATA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MCOATA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALONYL-COA-ACP-TRANSACYL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.0" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.39" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.85" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.86" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01626" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101421" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1092" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malcoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malACP_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MDH" metaid="R_MDH" name="Malate dehydrogenase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MDH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MDH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALATE-DEH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.299" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.37" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00342" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101439" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21432" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21433" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21434" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21435" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3236" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mal__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_oaa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MDH2" metaid="R_MDH2" name="Malate dehydrogenase (ubiquinone 8 as acceptor)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MDH2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MDH2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101440" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29083" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29084" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29085" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29086" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2210" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mal__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_oaa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MDH3" metaid="R_MDH3" name="Malate dehydrogenase (menaquinone 8 as acceptor)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MDH3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MDH3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101441" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29087" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29088" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29089" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29090" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2210" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mal__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_oaa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MDRPD" metaid="R_MDRPD" name="5-Methylthio-5-deoxy-D-ribulose 1-phosphate dehydratase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MDRPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MDRPD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:R145-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.109" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R07392" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101445" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5mdru1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dkmpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ME1" metaid="R_ME1" name="Malic enzyme (NAD)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ME1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ME1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:1.1.1.39-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.38" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.39" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00214" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101446" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12653" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12654" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12655" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12656" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1479" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mal__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ME2" metaid="R_ME2" name="Malic enzyme (NADP)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ME2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ME2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALIC-NADP-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.40" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00216" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101443" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18253" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18254" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18255" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18256" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2463" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mal__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MECDPDH" metaid="R_MECDPDH" name="2C-methyl-D-erythritol 2,4 cyclodiphosphate dehydratase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MECDPDH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MECDPDH" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101447" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2515" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2mecdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2mb4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MECDPS" metaid="R_MECDPS" name="2-C-methyl-D-erythritol 2,4-cyclodiphosphate synthase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MECDPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MECDPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-302" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.6.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05637" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101424" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23864" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23865" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23866" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23867" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2746" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2p4c2me_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2mecdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MELIBt2" metaid="R_MELIBt2" name="Melibiose transport in via symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MELIBt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MELIBt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-94" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101460" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28855" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28856" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28857" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28858" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4120" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_melib_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_melib_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MEPCT" metaid="R_MEPCT" name="2-C-methyl-D-erythritol 4-phosphate cytidylyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MEPCT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MEPCT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2.7.7.60-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.60" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05633" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96807" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13429" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13430" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13431" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13432" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2747" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2me4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4c2me_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_METAT" metaid="R_METAT" name="Methionine adenosyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_METAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/METAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:S-ADENMETSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00177" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101407" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21080" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21081" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21082" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21083" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2942" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_met__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amet_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_METDabc" metaid="R_METDabc" name="D-methionine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_METDabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/METDabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-202" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101472" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29767#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29768#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29769#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29770#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0199" />
            <fbc:geneProductRef fbc:geneProduct="G_b0197" />
            <fbc:geneProductRef fbc:geneProduct="G_b0198" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_met__D_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_met__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_METS_1" metaid="R_METS_1" name="Methionine synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_METS_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/METS_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.1.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00946" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101481" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11172" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11173" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11174" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11175" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4019" />
            <fbc:geneProductRef fbc:geneProduct="G_b3829" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5mthf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hcys__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_met__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_METabc" metaid="R_METabc" name="L-methionine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_METabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/METabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-4522" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101489" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29779#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29780#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29781#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29782#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32827#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32828#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32829#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32830#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0199" />
            <fbc:geneProductRef fbc:geneProduct="G_b0197" />
            <fbc:geneProductRef fbc:geneProduct="G_b0198" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_met__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_met__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MGSA" metaid="R_MGSA" name="Methylglyoxal synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MGSA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MGSA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHGLYSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.3.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01016" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101551" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17937" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17938" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17939" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17940" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0963" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhap_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mthgxl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MI1PP" metaid="R_MI1PP" name="Myo-inositol 1-phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MI1PP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MI1PP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5408" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.25" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01185" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101570" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27670" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27671" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27672" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27673" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2533" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mi1p__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_inost_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MICITDr" metaid="R_MICITDr" name="2-methylisocitrate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MICITDr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MICITDr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4.2.1.99-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.99" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04425" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101580" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17941" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17942" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17943" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17944" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1276" />
            <fbc:geneProductRef fbc:geneProduct="G_b0118" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2mcacn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_micit_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MLTG1" metaid="R_MLTG1" name="Maltodextrin glucosidase (maltotriose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MLTG1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MLTG1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5183" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101619" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0403" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malttr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malt_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MLTG2" metaid="R_MLTG2" name="Maltodextrin glucosidase (maltotetraose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MLTG2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MLTG2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MALTET-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101620" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0403" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltttr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malttr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MLTG3" metaid="R_MLTG3" name="Maltodextrin glucosidase (maltopentaose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MLTG3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MLTG3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14281" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.74" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101621" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0403" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltpt_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_maltttr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MLTG4" metaid="R_MLTG4" name="Maltodextrin glucosidase (maltohexaose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MLTG4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MLTG4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14282" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101622" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0403" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malthx_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_maltpt_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MLTG5" metaid="R_MLTG5" name="Maltodextrin glucosidase (maltoheptaose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MLTG5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MLTG5" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14283" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101623" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0403" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malthp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malthx_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MLTP1" metaid="R_MLTP1" name="Maltodextrin phosphorylase (maltopentaose)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MLTP1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MLTP1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14284" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101632" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29647" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29648" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29649" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29650" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3417" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_maltpt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_maltttr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MLTP2" metaid="R_MLTP2" name="Maltodextrin phosphorylase (maltohexaose)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MLTP2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MLTP2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14285" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101633" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29683" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29684" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29685" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29686" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3417" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malthx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_maltpt_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MLTP3" metaid="R_MLTP3" name="Maltodextrin phosphorylase (maltoheptaose)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MLTP3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MLTP3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14286" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101634" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3417" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_malthp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_malthx_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MMCD" metaid="R_MMCD" name="Methylmalonyl-CoA decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MMCD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MMCD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHYLMALONYL-COA-DECARBOXYLASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.41" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00923" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101656" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21396" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21397" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21398" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21399" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2919" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mmcoa__S_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppcoa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MME" metaid="R_MME" name="Methylmalonyl-CoA epimerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MME">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MME" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHYLMALONYL-COA-EPIM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.99.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02765" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101657" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20553" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20554" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20555" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20556" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mmcoa__R_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mmcoa__S_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MMM2" metaid="R_MMM2" name="Methylmalonyl-CoA mutase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MMM2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MMM2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHYLMALONYL-COA-MUT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.99.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00833" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101661" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22888" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22890" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22891" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2917" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mmcoa__R_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MNLpts" metaid="R_MNLpts" name="Mannitol transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MNLpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MNLpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101677" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b3599" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mnl_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mnl1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MNNH" metaid="R_MNNH" name="D-mannonate hydrolyase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MNNH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MNNH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:MANNONDEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05606" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101679" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20097" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20098" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20099" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20100" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4322" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mana_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ddglcn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MOAT" metaid="R_MOAT" name="3-deoxy-D-manno-octulosonic acid transferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MOAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MOAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.99.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04658" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101683" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28066" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28067" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28068" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28069" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3633" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ckdo_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lipidA_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_kdolipid4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MOAT2" metaid="R_MOAT2" name="3-deoxy-D-manno-octulosonic acid transferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MOAT2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MOAT2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KDOTRANS2-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.99.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05074" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06285" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101684" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28062" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28063" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28064" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28065" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3633" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ckdo_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_kdolipid4_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_kdo2lipid4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MOHMT" metaid="R_MOHMT" name="3-methyl-2-oxobutanoate hydroxymethyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MOHMT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MOHMT" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.2.11" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01226" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101702" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11824" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11825" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11826" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11827" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0134" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3mob_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mlthf_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dhp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MTAN" metaid="R_MTAN" name="Methylthioadenosine nucleosidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MTAN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MTAN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:METHYLTHIOADENOSINE-NUCLEOSIDASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.2.16" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.2.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01401" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101744" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13617" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13618" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13619" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13620" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0159" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5mta_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5mtr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ade_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MTHFC_1" metaid="R_MTHFC_1" name="Methenyltetrahydrofolate cyclohydrolase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MTHFC_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MTHFC_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01655" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101748" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23700" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23701" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23702" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23703" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0529" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_methf_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_10fthf_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MTHFD_1" metaid="R_MTHFD_1" name="Methylenetetrahydrofolate dehydrogenase (NADP)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MTHFD_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MTHFD_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.5.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01220" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101749" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22812" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22813" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22814" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22815" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0529" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mlthf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_methf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MTHFR2_1" metaid="R_MTHFR2_1" name="5,10-methylenetetrahydrofolate reductase (NADH)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MTHFR2_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MTHFR2_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.5.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R07168" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101751" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19821" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19822" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19823" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19824" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3941" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mlthf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5mthf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_MTRI" metaid="R_MTRI" name="5-methylthioribose-1-phosphate isomerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MTRI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MTRI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5.3.1.23-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.23" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04420" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101759" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19989" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19990" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19991" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19992" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5mdr1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5mdru1p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_MTRK" metaid="R_MTRK" name="5-methylthioribose kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_MTRK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/MTRK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5-METHYLTHIORIBOSE-KINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.100" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04143" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101760" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22312" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22313" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22314" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22315" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5mtr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5mdr1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NACODA" metaid="R_NACODA" name="N-acetylornithine deacetylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NACODA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NACODA" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101860" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30503" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30504" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30505" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30506" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3957" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acg5sa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu5sa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NACt" metaid="R_NACt" name="Nicotinic acid transport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NACt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NACt" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101808" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nac_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nac_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADDP" metaid="R_NADDP" name="NAD diphosphatase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADDP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADDP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NADPYROPHOSPHAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.22" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00103" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101861" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11800" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11802" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11803" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2411" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nmn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADDPe" metaid="R_NADDPe" name="NAD diphosphatase (extracellular)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADDPe">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADDPe" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NADPYROPHOSPHAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.22" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00103" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101861" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11800" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11802" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11803" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nmn_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADH10" metaid="R_NADH10" name="NADH dehydrogenase (menaquinone-8 &amp; 0 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADH10">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADH10" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101863" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29243" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29244" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29245" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29246" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1109" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADH5" metaid="R_NADH5" name="NADH dehydrogenase (ubiquinone-8 )" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADH5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADH5" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101872" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29107" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29108" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29109" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29110" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1109" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADH6" metaid="R_NADH6" name="NADH dehydrogenase (ubiquinone-8 &amp; 3.5 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADH6">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADH6" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.6.5.3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101873" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2280" />
            <fbc:geneProductRef fbc:geneProduct="G_b2277" />
            <fbc:geneProductRef fbc:geneProduct="G_b2282" />
            <fbc:geneProductRef fbc:geneProduct="G_b2283" />
            <fbc:geneProductRef fbc:geneProduct="G_b2287" />
            <fbc:geneProductRef fbc:geneProduct="G_b2284" />
            <fbc:geneProductRef fbc:geneProduct="G_b2285" />
            <fbc:geneProductRef fbc:geneProduct="G_b2281" />
            <fbc:geneProductRef fbc:geneProduct="G_b2276" />
            <fbc:geneProductRef fbc:geneProduct="G_b2288" />
            <fbc:geneProductRef fbc:geneProduct="G_b2286" />
            <fbc:geneProductRef fbc:geneProduct="G_b2278" />
            <fbc:geneProductRef fbc:geneProduct="G_b2279" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="4.5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="3.5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADH7" metaid="R_NADH7" name="NADH dehydrogenase (menaquinone-8 &amp; 2 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADH7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADH7" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.6.5.3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101874" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2280" />
            <fbc:geneProductRef fbc:geneProduct="G_b2277" />
            <fbc:geneProductRef fbc:geneProduct="G_b2282" />
            <fbc:geneProductRef fbc:geneProduct="G_b2283" />
            <fbc:geneProductRef fbc:geneProduct="G_b2287" />
            <fbc:geneProductRef fbc:geneProduct="G_b2284" />
            <fbc:geneProductRef fbc:geneProduct="G_b2285" />
            <fbc:geneProductRef fbc:geneProduct="G_b2281" />
            <fbc:geneProductRef fbc:geneProduct="G_b2276" />
            <fbc:geneProductRef fbc:geneProduct="G_b2288" />
            <fbc:geneProductRef fbc:geneProduct="G_b2286" />
            <fbc:geneProductRef fbc:geneProduct="G_b2278" />
            <fbc:geneProductRef fbc:geneProduct="G_b2279" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mqn8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADH8" metaid="R_NADH8" name="NADH dehydrogenase (demethylmenaquinone-8 &amp; 2.8 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADH8">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADH8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.6.5.3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101875" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2280" />
            <fbc:geneProductRef fbc:geneProduct="G_b2277" />
            <fbc:geneProductRef fbc:geneProduct="G_b2282" />
            <fbc:geneProductRef fbc:geneProduct="G_b2283" />
            <fbc:geneProductRef fbc:geneProduct="G_b2287" />
            <fbc:geneProductRef fbc:geneProduct="G_b2284" />
            <fbc:geneProductRef fbc:geneProduct="G_b2285" />
            <fbc:geneProductRef fbc:geneProduct="G_b2281" />
            <fbc:geneProductRef fbc:geneProduct="G_b2276" />
            <fbc:geneProductRef fbc:geneProduct="G_b2288" />
            <fbc:geneProductRef fbc:geneProduct="G_b2286" />
            <fbc:geneProductRef fbc:geneProduct="G_b2278" />
            <fbc:geneProductRef fbc:geneProduct="G_b2279" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="3.8" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2.8" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADH9" metaid="R_NADH9" name="NADH dehydrogenase (demethylmenaquinone-8 &amp; 0 protons)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADH9">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADH9" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101876" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30507" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30508" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30509" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30510" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1109" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADK" metaid="R_NADK" name="NAD kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NAD-KIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.23" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00104" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101882" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18629" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18630" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18631" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18632" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2615" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADPPPS" metaid="R_NADPPPS" name="NADP phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADPPPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADPPPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-5822" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00118" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101798" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28050" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28051" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28052" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28053" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADS1" metaid="R_NADS1" name="NAD synthase (nh3)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADS1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADS1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NAD-SYNTH-NH3-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00189" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101897" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21188" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21189" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21190" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21191" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1740" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dnad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nh4_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NADTRHD" metaid="R_NADTRHD" name="NAD transhydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NADTRHD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NADTRHD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRNUTRANSHYDROGEN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.6.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.6.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.6.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00112" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101898" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11692" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11693" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11694" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11695" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1602" />
              <fbc:geneProductRef fbc:geneProduct="G_b1603" />
            </fbc:and>
            <fbc:geneProductRef fbc:geneProduct="G_b3962" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NAMNPP" metaid="R_NAMNPP" name="Nicotinic acid mononucleotide pyrophosphorylase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NAMNPP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NAMNPP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NICOTINATEPRIBOSYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.4.21" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01724" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101909" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/36163" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/36164" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/36165" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/36166" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0931" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nicrnt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NAt3_1" metaid="R_NAt3_1" name="Sodium proton antiporter  HNA is 11 " reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NAt3_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NAt3_1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-101" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101803" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29419" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29420" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29421" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29422" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1216" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_na1_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NAt3_15" metaid="R_NAt3_15" name="Sodium proton antiporter (H:NA is 1.5)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NAt3_15">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NAt3_15" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-130" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101914" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29247" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29248" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29249" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29250" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1186" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_na1_e" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NAt3_2" metaid="R_NAt3_2" name="Sodium proton antiporter (H:NA is 2)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NAt3_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NAt3_2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-129" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101915" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29251" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29252" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29253" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29254" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0019" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_na1_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NDPK1" metaid="R_NDPK1" name="Nucleoside-diphosphate kinase (ATP:GDP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NDPK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NDPK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GDPKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00330" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96119" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27686" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27687" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27688" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27689" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2518" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gtp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NDPK2" metaid="R_NDPK2" name="Nucleoside-diphosphate kinase (ATP:UDP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NDPK2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NDPK2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00156" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101935" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25098" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25099" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25100" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25101" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2518" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_utp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NDPK3" metaid="R_NDPK3" name="Nucleoside-diphosphate kinase (ATP:CDP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NDPK3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NDPK3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CDPKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00570" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101936" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25237" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25238" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25239" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25240" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2518" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ctp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NDPK4" metaid="R_NDPK4" name="Nucleoside-diphosphate kinase (ATP:dTDP)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NDPK4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NDPK4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDPKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02093" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101937" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27682" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27683" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27684" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27685" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2518" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dttp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NDPK5" metaid="R_NDPK5" name="Nucleoside-diphosphate kinase (ATP:dGDP)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NDPK5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NDPK5" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DGDPKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01857" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96118" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27690" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27691" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27692" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27693" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2518" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dgtp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NDPK6" metaid="R_NDPK6" name="Nucleoside-diphosphate kinase (ATP:dUDP)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NDPK6">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NDPK6" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DUDPKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02331" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101938" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28582" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28583" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28584" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28585" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2518" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dudp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dutp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NDPK7" metaid="R_NDPK7" name="Nucleoside-diphosphate kinase (ATP:dCDP)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NDPK7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NDPK7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DCDPKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02326" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101939" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27678" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27679" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27680" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27681" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2518" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dcdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dctp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NDPK8" metaid="R_NDPK8" name="Nucleoside-diphosphate kinase (ATP:dADP)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NDPK8">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NDPK8" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DADPKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01137" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101940" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27674" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27675" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27676" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27677" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2518" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_datp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NH4t" metaid="R_NH4t" name="Ammonia reversible transport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NH4t">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NH4t" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-9615" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-206" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-544" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101950" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28747" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28748" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28749" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28750" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0451" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nh4_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NMNAT" metaid="R_NMNAT" name="Nicotinamide-nucleotide adenylyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NMNAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NMNAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2.7.7.1-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.18" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00137" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95841" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21360" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21361" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21362" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21363" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0639" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nmn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NMNDA" metaid="R_NMNDA" name="Nicotinamide-nucleotide amidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NMNDA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NMNDA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NMNAMIDOHYDRO-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.42" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02322" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101969" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12400" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12401" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12402" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12403" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nmn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nicrnt_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NMNN" metaid="R_NMNN" name="NMN nucleosidase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NMNN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NMNN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.2.14" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101971" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nmn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ncam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NMNP" metaid="R_NMNP" name="NMN permease" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NMNP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NMNP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101972" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28482" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28483" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28484" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28485" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0751" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nmn_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nmn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NMNt7" metaid="R_NMNt7" name="NMN transport via NMN glycohydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NMNt7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NMNt7" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.2.14" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101974" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nmn_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ncam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NNAM" metaid="R_NNAM" name="Nicotinamidase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NNAM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NNAM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NICOTINAMID-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01268" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101977" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14545" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14546" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14547" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14548" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1768" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ncam_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NNATr" metaid="R_NNATr" name="Nicotinate-nucleotide adenylyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NNATr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NNATr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:NICONUCADENYLYLTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.18" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03005" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101978" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22860" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22861" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22862" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22863" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0639" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nicrnt_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dnad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NNDMBRT" metaid="R_NNDMBRT"
      name="Nicotinate-nucleotide dimethylbenzimidazole phosphoribosyltransferase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NNDMBRT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NNDMBRT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DMBPPRIBOSYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.21" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04148" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101979" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11196" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11197" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11198" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11199" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1991" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dmbzid_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nicrnt_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5prdmbz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nac_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NNDPR" metaid="R_NNDPR" name="Nicotinate-nucleotide diphosphorylase (carboxylating)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NNDPR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NNDPR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:QUINOPRIBOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03348" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101980" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12733" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12734" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12735" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12736" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0109" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_quln_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nicrnt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_NO2t2r" metaid="R_NO2t2r" name="Nitrite transport in via proton symport, reversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NO2t2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NO2t2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101984" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1223" />
            <fbc:geneProductRef fbc:geneProduct="G_b3367" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_no2_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_no2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NO3R1" metaid="R_NO3R1" name="Nitrate reductase (Ubiquinol-8)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NO3R1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NO3R1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101986" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29147#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29148#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29149#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29150#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1227" />
            <fbc:geneProductRef fbc:geneProduct="G_b1225" />
            <fbc:geneProductRef fbc:geneProduct="G_b1226" />
            <fbc:geneProductRef fbc:geneProduct="G_b1224" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_no3_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8h2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_no2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NO3R2" metaid="R_NO3R2" name="Nitrate reductase (Menaquinol-8)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NO3R2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NO3R2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101988" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29139#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29140#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29141#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29142#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1227" />
            <fbc:geneProductRef fbc:geneProduct="G_b1225" />
            <fbc:geneProductRef fbc:geneProduct="G_b1226" />
            <fbc:geneProductRef fbc:geneProduct="G_b1224" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_no3_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mqn8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_no2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NO3t7" metaid="R_NO3t7" name="Nitrate transport in via nitrite antiport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NO3t7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NO3t7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-239" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101995" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28743" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28744" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28745" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28746" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1223" />
            <fbc:geneProductRef fbc:geneProduct="G_b1469" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_no2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_no3_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_no2_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_no3_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NPHS" metaid="R_NPHS" name="Naphthoate synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NPHS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NPHS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.3.36" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04150" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102012" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15525" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15526" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15527" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15528" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2262" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sbzcoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhna_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD1" metaid="R_NTD1" name="5'-nucleotidase (dUMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14143" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.89" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02102" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102028" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29355" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29356" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29357" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29358" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dump_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_duri_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD10" metaid="R_NTD10" name="5'-nucleotidase (XMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD10">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD10" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XMPXAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02719" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102029" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28530" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28531" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28532" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28533" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xtsn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD11" metaid="R_NTD11" name="5'-nucleotidase (IMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD11">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD11" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-7607" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.99" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01126" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102030" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27718" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27719" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27720" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27721" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_imp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ins_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD2" metaid="R_NTD2" name="5'-nucleotidase (UMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14025" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00963" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102032" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29359" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29360" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29361" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29362" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ump_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uri_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD3" metaid="R_NTD3" name="5'-nucleotidase (dCMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5292" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.89" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01664" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102033" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29363" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29364" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29365" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29366" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dcmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dcyt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD4" metaid="R_NTD4" name="5'-nucleotidase (CMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14026" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.91" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00511" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102034" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29367" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29368" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29369" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29370" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cytd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD5" metaid="R_NTD5" name="5'-nucleotidase (dTMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD5" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THYMIDYLATE-5-PHOSPHATASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.35" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.89" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01569" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102035" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11080" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11081" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11082" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11083" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thymd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD6" metaid="R_NTD6" name="5'-nucleotidase (dAMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD6">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD6" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14161" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.89" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02088" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102036" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29371" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29372" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29373" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29374" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_damp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dad_2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD7" metaid="R_NTD7" name="5'-nucleotidase (AMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMP-DEPHOSPHORYLATION-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00183" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102037" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29375" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29376" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29377" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29378" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD8" metaid="R_NTD8" name="5'-nucleotidase (dGMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD8">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD8" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14142" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.89" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01968" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102038" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29379" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29380" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29381" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29382" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dgsn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTD9" metaid="R_NTD9" name="5'-nucleotidase (GMP)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTD9">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTD9" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-7609" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01227" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100381" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27714" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27715" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27716" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27717" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gsn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPP1" metaid="R_NTPP1" name="Nucleoside triphosphate pyrophosphorylase (dgtp)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPP1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPP1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-385" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01855" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102049" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28362" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28363" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28364" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28365" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0099" />
            <fbc:geneProductRef fbc:geneProduct="G_b2781" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dgmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPP2" metaid="R_NTPP2" name="Nucleoside triphosphate pyrophosphorylase (gtp)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPP2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPP2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14140" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00426" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102052" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29391" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29392" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29393" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29394" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2781" />
            <fbc:geneProductRef fbc:geneProduct="G_b0099" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPP3" metaid="R_NTPP3" name="Nucleoside triphosphate pyrophosphorylase (dctp)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPP3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPP3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DCTP-PYROPHOSPHATASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.65" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01668" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97203" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22636" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22637" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22638" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22639" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2781" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dcmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPP4" metaid="R_NTPP4" name="Nucleoside triphosphate pyrophosphorylase (ctp)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPP4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPP4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-383" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.65" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00515" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102053" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27762" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27763" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27764" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27765" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2781" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPP5" metaid="R_NTPP5" name="Nucleoside triphosphate pyrophosphorylase (datp)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPP5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPP5" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-384" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97173" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28334" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28335" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28336" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28337" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2781" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_datp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_damp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPP6" metaid="R_NTPP6" name="Nucleoside triphosphate pyrophosphorylase (atp)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPP6">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPP6" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ATP-PYROPHOSPHATASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00087" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102054" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14245" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14246" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14247" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14248" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2781" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPP7" metaid="R_NTPP7" name="Nucleoside triphosphate pyrophosphorylase (dttp)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPP7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPP7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5107" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R11323" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97809" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28534" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28535" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28536" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28537" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2781" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dttp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPP8" metaid="R_NTPP8" name="Nucleoside triphosphate pyrophosphorylase (utp)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPP8">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPP8" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14139" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00662" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102055" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29395" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29396" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29397" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29398" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2781" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_utp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ump_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPTP1" metaid="R_NTPTP1" name="Nucleoside triphosphate tripolyhydrolase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPTP1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPTP1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DGTPTRIPHYDRO-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.5.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01856" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102057" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15193" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15194" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15195" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15196" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0160" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dgsn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTPTP2" metaid="R_NTPTP2" name="Nucleoside triphosphate tripolyhydrolase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTPTP2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTPTP2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102058" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30559" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30560" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30561" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30562" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0160" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gsn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_NTRIR2x" metaid="R_NTRIR2x" name="Nitrite Reductase (NADH)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_NTRIR2x">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/NTRIR2x" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-13854" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.7.1.15" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.7.1.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00787" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102062" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24628" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24629" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24630" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24631" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3365" />
            <fbc:geneProductRef fbc:geneProduct="G_b3366" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_no2_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_O2t" metaid="R_O2t" name="O2 transport  diffusion " reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_O2t">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/O2t" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-474" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102090" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_o2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OBTFL" metaid="R_OBTFL" name="2-Oxobutanoate formate lyase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OBTFL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OBTFL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:KETOBUTFORMLY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.54" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06987" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102135" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28054" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28055" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28056" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28057" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3114" />
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0902" />
              <fbc:geneProductRef fbc:geneProduct="G_b0903" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2obut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppcoa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_OCBT" metaid="R_OCBT" name="Ornithine carbamoyltransferase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OCBT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OCBT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ORNCARBAMTRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-13482" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.3.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01398" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102137" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19513" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19514" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19515" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19516" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4254" />
            <fbc:geneProductRef fbc:geneProduct="G_b0273" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cbp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_citr__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OCDCAt2" metaid="R_OCDCAt2" name="Octadecanoate transport via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OCDCAt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OCDCAt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102142" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2344" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ocdca_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ocdca_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OCTDPS" metaid="R_OCTDPS" name="Octaprenyl pyrophosphate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OCTDPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OCTDPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-8992" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.90" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R09248" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102153" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27798" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27799" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27800" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27801" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3187" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_frdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ipdp_c" stoichiometry="5" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_octdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="5" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_OHPBAT" metaid="R_OHPBAT"
      name="O-Phospho-4-hydroxy-L-threonine:2-oxoglutarate aminotransferase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OHPBAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OHPBAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PSERTRANSAMPYR-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.52" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05085" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102169" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16573" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16574" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16575" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16576" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0907" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ohpb_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_phthr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OHPHM" metaid="R_OHPHM" name="2-octaprenyl-6-hydroxyphenol methylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OHPHM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OHPHM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2-OCTAPRENYL-6-OHPHENOL-METHY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.1.222" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04988" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102170" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27770" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27771" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27772" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27773" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2232" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ohph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amet_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2omph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ahcys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OMBZLM" metaid="R_OMBZLM" name="2-Octaprenyl-6-methoxy-benzoquinol methylase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OMBZLM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OMBZLM" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102178" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3833" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ombzl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amet_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ommbl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ahcys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OMCDC" metaid="R_OMCDC" name="2-Oxo-4-methyl-3-carboxypentanoate decarboxylation"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OMCDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OMCDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-7800" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.85" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01652" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102180" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25078" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25079" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25080" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25081" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0073" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3c4mop_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4mop_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OMMBLHX" metaid="R_OMMBLHX"
      name="2-Octaprenyl-3-methyl-6-methoxy-1,4-benzoquinol hydroxylase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OMMBLHX">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OMMBLHX" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102187" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0662" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2ommbl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="0.5" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2omhmbl_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OMPDC" metaid="R_OMPDC" name="Orotidine-5'-phosphate decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OMPDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OMPDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OROTPDECARB-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.23" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00965" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102190" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11596" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11597" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11598" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11599" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1281" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orot5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ump_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OMPHHX" metaid="R_OMPHHX" name="2-octaprenyl-6-methoxyphenol hydroxylase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OMPHHX">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OMPHHX" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102191" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2907" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2omph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="0.5" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ombzl_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OP4ENH" metaid="R_OP4ENH" name="2-oxopent-4-enoate hydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OP4ENH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OP4ENH" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.80" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02601" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102198" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0350" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_op4en_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4h2opntn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OPHBDC" metaid="R_OPHBDC" name="Octaprenyl-hydroxybenzoate decarboxylase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OPHBDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OPHBDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3-OCTAPRENYL-4-OHBENZOATE-DECARBOX-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.98" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04986" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102201" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27778" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27779" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27780" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27781" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3843" />
            <fbc:geneProductRef fbc:geneProduct="G_b2311" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3ophb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2oph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OPHHX" metaid="R_OPHHX" name="2-Octaprenylphenol hydroxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OPHHX">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OPHHX" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102203" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3835" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2oph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="0.5" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2ohph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ORNDC" metaid="R_ORNDC" name="Ornithine Decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ORNDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ORNDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ORNDECARBOX-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.17" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00670" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102209" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22964" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22965" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22966" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22967" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2965" />
            <fbc:geneProductRef fbc:geneProduct="G_b0693" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ptrc_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ORNTA" metaid="R_ORNTA" name="Ornithine transaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ORNTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ORNTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ORNITHINE-GLU-AMINOTRANSFERASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00667" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102220" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25160" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25161" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25162" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25163" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3073" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu5sa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_ORNabc" metaid="R_ORNabc" name="Ornithine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ORNabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ORNabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-37-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102222" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29883#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29884#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29885#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29886#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2307" />
            <fbc:geneProductRef fbc:geneProduct="G_b2310" />
            <fbc:geneProductRef fbc:geneProduct="G_b2308" />
            <fbc:geneProductRef fbc:geneProduct="G_b2306" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orn_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_orn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_ORPT" metaid="R_ORPT" name="Orotate phosphoribosyltransferase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_ORPT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/ORPT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:OROPRIBTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01870" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102231" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10380" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10381" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10382" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10383" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3642" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orot5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_orot_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_prpp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_OXGDC2" metaid="R_OXGDC2" name="2-oxoglutarate decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_OXGDC2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/OXGDC2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.71" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102244" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2264" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thmpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ssaltpp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_P5CD" metaid="R_P5CD" name="1-pyrroline-5-carboxylate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_P5CD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/P5CD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRROLINECARBDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.88" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.5.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00707" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102301" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16417" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16418" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16419" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16420" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1014" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_1pyr5c_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_P5CR" metaid="R_P5CR" name="Pyrroline-5-carboxylate reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_P5CR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/P5CR" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.5.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01251" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102302" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14109" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14110" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14111" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14112" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0386" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_1pyr5c_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pro__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PACCOAL" metaid="R_PACCOAL" name="Phenylacetate-CoA ligase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PACCOAL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PACCOAL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-10819" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.2.1.30" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02539" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102315" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20956" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20957" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20958" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20959" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1398" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pac_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_phaccoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PANTS" metaid="R_PANTS" name="Pantothenate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PANTS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PANTS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PANTOATE-BETA-ALANINE-LIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02473" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102345" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10912" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10913" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10914" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10915" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0133" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala_B_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pant__R_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pnto__R_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PAPA_EC" metaid="R_PAPA_EC" name="Phosphatidate phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PAPA_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PAPA_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102369" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pa_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_12dgr_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PAPPT3" metaid="R_PAPPT3"
      name="Phospho-N-acetylmuramoyl-pentapeptide-transferase (meso-2,6-diaminopimelate)" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PAPPT3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PAPPT3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.8.13" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102373" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0087" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udcpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ugmda_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uagmda_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ump_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PAPSR" metaid="R_PAPSR" name="Phosphoadenylyl-sulfate reductase (thioredoxin)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PAPSR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PAPSR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:1.8.4.8-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.8.4.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02021" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102376" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2762" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_paps_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_so3_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PASYN_EC" metaid="R_PASYN_EC" name="Phosphatidic acid synthase (Ecoli)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PASYN_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PASYN_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102383" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3018" />
            <fbc:geneProductRef fbc:geneProduct="G_b4041" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hdeACP_c" stoichiometry="0.14" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_myrsACP_c" stoichiometry="0.04" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_octeACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_palmACP_c" stoichiometry="0.72" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tdeACP_c" stoichiometry="0.1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pa_EC_c" stoichiometry="0.02" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PDH" metaid="R_PDH" name="Pyruvate dehydrogenase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PDH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PDH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRUVDEH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.51" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.4.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.8.1.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00209" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102425" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28042" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28043" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28044" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28045" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0116" />
            <fbc:geneProductRef fbc:geneProduct="G_b0114" />
            <fbc:geneProductRef fbc:geneProduct="G_b0115" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PDX5POi" metaid="R_PDX5POi" name="Pyridoxine 5'-phosphate oxidase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PDX5POi">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PDX5POi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PNPOXI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00278" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102438" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15149" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15150" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15151" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15152" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1638" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pdx5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pydx5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PDX5PS" metaid="R_PDX5PS" name="Pyridoxine 5'-phosphate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PDX5PS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PDX5PS" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102439" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2564" />
            <fbc:geneProductRef fbc:geneProduct="G_b0052" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dxyl5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_phthr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pdx5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PDXPP" metaid="R_PDXPP" name="Pyridoxine 5-phosphate phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PDXPP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PDXPP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14181" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.74" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01911" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102440" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25112" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25113" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25114" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25115" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pdx5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pydxn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PEAMNO" metaid="R_PEAMNO" name="Phenethylamine oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PEAMNO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PEAMNO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:AMINEPHEN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-10817" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.3.21" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.3.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.3.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02613" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102461" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25265" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25266" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25267" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25268" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1386" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_peamn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pacald_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PEPT_EC" metaid="R_PEPT_EC" name="Ethanolamine phosphotransferase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PEPT_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PEPT_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.8.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102491" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pe_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_12dgr_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cdpea_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PERD" metaid="R_PERD" name="Erythronate 4-phosphate (4per) dehydrogenase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PERD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PERD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ERYTHRON4PDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.290" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04210" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102495" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18829" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18830" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18831" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18832" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2320" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4per_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ohpb_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PFK" metaid="R_PFK" name="Phosphofructokinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PFK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PFK" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.11" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102507" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16109" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16110" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16111" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16112" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3916" />
            <fbc:geneProductRef fbc:geneProduct="G_b1723" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_f6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PFK_2" metaid="R_PFK_2" name="Phosphofructokinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PFK_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PFK_2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TAGAKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.11" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.144" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.56" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03236" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102509" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12420" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12421" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12422" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12423" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1723" />
            <fbc:geneProductRef fbc:geneProduct="G_b2168" />
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3137" />
              <fbc:geneProductRef fbc:geneProduct="G_b3132" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tag6p__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tagdp__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PFL" metaid="R_PFL" name="Pyruvate formate lyase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PFL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PFL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRUVFORMLY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.54" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00212" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102514" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11844" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11845" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11846" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11847" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0902" />
              <fbc:geneProductRef fbc:geneProduct="G_b0903" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3951" />
              <fbc:geneProductRef fbc:geneProduct="G_b3952" />
            </fbc:and>
            <fbc:geneProductRef fbc:geneProduct="G_b3114" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_for_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PGAMT" metaid="R_PGAMT" name="Phosphoglucosamine mutase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGAMT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGAMT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:5.4.2.10-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.10" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02060" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102526" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23424" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23425" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23426" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23427" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3176" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gam1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gam6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PGCD" metaid="R_PGCD" name="Phosphoglycerate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGCD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGCD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PGLYCDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.95" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01513" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102527" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12641" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12642" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12643" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12644" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2913" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3pg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3php_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PGI" metaid="R_PGI" name="Glucose-6-phosphate isomerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PGLUCISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102535" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4025" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PGK" metaid="R_PGK" name="Phosphoglycerate kinase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSGLYPHOS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01512" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102538" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14802" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14803" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14804" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2926" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3pg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_13dpg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PGL" metaid="R_PGL" name="6-phosphogluconolactonase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:6PGLUCONOLACT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.1.31" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02035" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102539" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12556" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12557" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12558" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12559" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_6pgl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_6pgc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PGLYCP" metaid="R_PGLYCP" name="Phosphoglycolate phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGLYCP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGLYCP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GPH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.18" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01334" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102543" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14369" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14370" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14371" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14372" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3385" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2pglyc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyclt_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PGM" metaid="R_PGM" name="Phosphoglycerate mutase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3PGAREARR-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-15513" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.11" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01518" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102547" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15901" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15902" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15903" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15904" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4395" />
            <fbc:geneProductRef fbc:geneProduct="G_b3612" />
            <fbc:geneProductRef fbc:geneProduct="G_b0755" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2pg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3pg_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PGMT" metaid="R_PGMT" name="Phosphoglucomutase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGMT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGMT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSPHOGLUCMUT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-16999" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R08639" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102548" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0688" />
            <fbc:geneProductRef fbc:geneProduct="G_b2690" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PGPP_EC" metaid="R_PGPP_EC" name="Phosphatidylglycerol phosphate phosphatase (Ecoli)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGPP_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGPP_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.27" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102581" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0418" />
            <fbc:geneProductRef fbc:geneProduct="G_b1278" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pgp_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pg_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PGSA_EC" metaid="R_PGSA_EC" name="Phosphatidylglycerol synthase (Ecoli)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PGSA_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PGSA_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102603" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1912" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cdpdag_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glyc3p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pgp_EC_c" stoichiometry="0.02" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PHETA1" metaid="R_PHETA1" name="Phenylalanine transaminase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PHETA1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PHETA1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHEAMINOTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-10814" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.42" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.57" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.58" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00694" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102631" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4054" />
            <fbc:geneProductRef fbc:geneProduct="G_b3770" />
            <fbc:geneProductRef fbc:geneProduct="G_b0928" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_phe__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_phpyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PHEt2r" metaid="R_PHEt2r" name="L phenylalanine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PHEt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PHEt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-56" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102635" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28923" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28924" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28925" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28926" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0112" />
            <fbc:geneProductRef fbc:geneProduct="G_b0576" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_phe__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_phe__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PIabc" metaid="R_PIabc" name="Phosphate transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PIabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PIabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-27-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.27" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102870" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24440#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24441#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24442#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24443#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3726" />
            <fbc:geneProductRef fbc:geneProduct="G_b3725" />
            <fbc:geneProductRef fbc:geneProduct="G_b3727" />
            <fbc:geneProductRef fbc:geneProduct="G_b3728" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PIt2r" metaid="R_PIt2r" name="Phosphate reversible transport via symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PIt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PIt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-114" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102872" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29939" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29940" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29941" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29942" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3493" />
            <fbc:geneProductRef fbc:geneProduct="G_b2987" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PLIPA1" metaid="R_PLIPA1" name="Phospholipase A (phosphatidylglycerol)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PLIPA1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PLIPA1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102966" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3821" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pg_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_agpg_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdca_c" stoichiometry="0.36" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdcea_c" stoichiometry="0.07" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ocdcea_c" stoichiometry="0.5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdca_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdcea_c" stoichiometry="0.05" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PLIPA2" metaid="R_PLIPA2" name="Phosphlipase A (phosphatidylethanolamine)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PLIPA2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PLIPA2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102988" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3821" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pe_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_agpe_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdca_c" stoichiometry="0.36" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdcea_c" stoichiometry="0.07" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ocdcea_c" stoichiometry="0.5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdca_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdcea_c" stoichiometry="0.05" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PLIPA3" metaid="R_PLIPA3" name="Phospholipase A (phosphatidylcholine)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PLIPA3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PLIPA3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103010" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3821" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pc_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_agpc_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdca_c" stoichiometry="0.36" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hdcea_c" stoichiometry="0.07" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ocdcea_c" stoichiometry="0.5" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdca_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdcea_c" stoichiometry="0.05" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PMANM" metaid="R_PMANM" name="Phosphomannomutase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PMANM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PMANM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSMANMUT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01818" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR101729" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11140" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11141" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11142" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11143" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2048" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_man1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_man6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PMDPHT" metaid="R_PMDPHT" name="Pyrimidine phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PMDPHT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PMDPHT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBOPHOSPHAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.104" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R07280" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96145" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25197" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25198" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25199" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25200" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5aprbu_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4r5au_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PMPK" metaid="R_PMPK" name="Phosphomethylpyrimidine kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PMPK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PMPK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRIMSYN3-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04509" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103045" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19893" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19894" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19895" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19896" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2103" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4ampm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2mahmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PNTK" metaid="R_PNTK" name="Pantothenate kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PNTK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PNTK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PANTOTHENATE-KIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.33" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03018" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103050" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16373" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16374" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16375" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16376" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3974" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pnto__R_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4ppan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PNTOt4" metaid="R_PNTOt4" name="Pantothenate sodium symporter" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PNTOt4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PNTOt4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-117" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103054" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29927" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29928" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29929" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29930" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3258" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pnto__R_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_na1_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pnto__R_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_POX" metaid="R_POX" name="Pyruvate oxidase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_POX">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/POX" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103067" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29167" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29168" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29169" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29170" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0871" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPA" metaid="R_PPA" name="Inorganic diphosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:INORGPYROPHOSPHAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00004" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100808" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24576" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24577" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24578" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24579" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4226" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PPAKr" metaid="R_PPAKr" name="Propionate kinase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPAKr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPAKr" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.2.15" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95897" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3115" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppap_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPBNGS" metaid="R_PPBNGS" name="Porphobilinogen synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPBNGS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPBNGS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PORPHOBILSYNTH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.24" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00036" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103095" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24064" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24065" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24066" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24067" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0369" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5aop_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppbng_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPC" metaid="R_PPC" name="Phosphoenolpyruvate carboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPC" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.31" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00345" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103096" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23072" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23073" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23074" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23075" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3956" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_oaa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPCDC" metaid="R_PPCDC" name="Phosphopantothenoylcysteine decarboxylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPCDC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPCDC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:P-PANTOCYSDECARB-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.36" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03269" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103098" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16793" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16794" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16795" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16796" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4ppcys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pan4p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPCK" metaid="R_PPCK" name="Phosphoenolpyruvate carboxykinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPCK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPCK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PEPCARBOXYKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.49" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00341" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103099" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18617" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18618" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18619" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18620" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3403" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_oaa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pep_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPCSCT" metaid="R_PPCSCT" name="Propanoyl-CoA: succinate CoA-transferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPCSCT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPCSCT" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.8.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103103" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2920" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppcoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succoa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PPM" metaid="R_PPM" name="Phosphopentomutase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14456" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.7" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103115" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4383" />
            <fbc:geneProductRef fbc:geneProduct="G_b3380" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_r1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PPM2" metaid="R_PPM2" name="Phosphopentomutase 2 (deoxyribose)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPM2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPM2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:D-PPENTOMUT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.2.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02749" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103116" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25205" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25206" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25207" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25208" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27658" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27659" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27660" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27661" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4383" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dr1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dr5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPNCL2" metaid="R_PPNCL2" name="Phosphopantothenate-cysteine ligase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPNCL2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPNCL2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:P-PANTOCYSLIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04231" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103119" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19397" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19398" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19399" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19400" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4ppan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4ppcys_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPND" metaid="R_PPND" name="Prephenate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPND">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPND" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PREPHENATEDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.1.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01728" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103121" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13869" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13870" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13871" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13872" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2600" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pphn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_34hpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPNDH" metaid="R_PPNDH" name="Prephenate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPNDH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPNDH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PREPHENATEDEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.51" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.91" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01373" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103123" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2599" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pphn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_phpyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPPGO" metaid="R_PPPGO" name="Protoporphyrinogen oxidase (aerobic)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPPGO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPPGO" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.3.4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103128" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3850" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1.5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pppg9_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppp9_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPPNDO" metaid="R_PPPNDO" name="Phenylpropanoate Dioxygenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPPNDO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPPNDO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HCAMULTI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.14.12.19" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103134" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20357" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20358" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20359" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20360" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2538" />
            <fbc:geneProductRef fbc:geneProduct="G_b2542" />
            <fbc:geneProductRef fbc:geneProduct="G_b2539" />
            <fbc:geneProductRef fbc:geneProduct="G_b2540" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pppn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cechddd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PPPNt2r" metaid="R_PPPNt2r"
      name="3-phenylpropionate transport via proton symport, reversible" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPPNt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPPNt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103135" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2536" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pppn_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pppn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPS" metaid="R_PPS" name="Phosphoenolpyruvate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PEPSYNTH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.9.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00199" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103140" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11364" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11365" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11366" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11367" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1702" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pep_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PPTGS" metaid="R_PPTGS" name="Peptidoglycan subunit synthesis" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PPTGS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PPTGS" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103142" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uaagmda_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_peptido_EC_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udcpdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PRAGSr" metaid="R_PRAGSr" name="Phosphoribosylglycinamide synthase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRAGSr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRAGSr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLYRIBONUCSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.4.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04144" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103139" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4005" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gly_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pram_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gar_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PRAIS" metaid="R_PRAIS" name="Phosphoribosylaminoimidazole synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRAIS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRAIS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103157" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23032" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23033" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23034" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23035" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2499" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fpram_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_air_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PRAIi" metaid="R_PRAIi" name="Phosphoribosylanthranilate isomerase (irreversible)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRAIi">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRAIi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PRAISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.24" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03509" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103156" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21540" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21541" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21542" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21543" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1262" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pran_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2cpr5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PRAMPC" metaid="R_PRAMPC" name="Phosphoribosyl-AMP cyclohydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRAMPC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRAMPC" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HISTCYCLOHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.4.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04037" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103154" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20049" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20050" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20051" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20052" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2026" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prbamp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_prfp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PRASCSi" metaid="R_PRASCSi"
      name="Phosphoribosylaminoimidazolesuccinocarboxamide synthase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRASCSi">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRASCSi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SAICARSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.6" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103158" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2476" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5aizc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_asp__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_25aics_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PRATPP" metaid="R_PRATPP" name="Phosphoribosyl-ATP pyrophosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRATPP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRATPP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:HISTPRATPHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.31" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04035" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103155" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22828" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22829" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22830" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22831" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2026" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prbatp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_prbamp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PRFGS" metaid="R_PRFGS" name="Phosphoribosylformylglycinamidine synthase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRFGS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRFGS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.5.3" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103165" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17129" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17130" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17131" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17132" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2557" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fgam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gln__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fpram_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PRMICI" metaid="R_PRMICI"
      name="1-(5-phosphoribosyl)-5-[(5-phosphoribosylamino)methylideneamino)imidazole-4-carboxamide isomerase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRMICI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRMICI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PRIBFAICARPISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04640" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103169" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15469" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15470" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15471" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15472" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2024" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prfp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_prlp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PROD2" metaid="R_PROD2" name="Proline dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PROD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PROD2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.5.-.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.5.99.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10507" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103187" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26124" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26125" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26126" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26127" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1014" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pro__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_1pyr5c_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fadh2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PROabc" metaid="R_PROabc" name="L-proline transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PROabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PROabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-26-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103210" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29815#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29816#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29817#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29818#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2677" />
            <fbc:geneProductRef fbc:geneProduct="G_b2679" />
            <fbc:geneProductRef fbc:geneProduct="G_b2678" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pro__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pro__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PROt2r" metaid="R_PROt2r" name="L proline reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PROt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PROt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-29" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103211" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28963" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28964" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28965" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28966" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4111" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pro__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pro__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PROt4" metaid="R_PROt4" name="Na+/Proline-L symporter" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PROt4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PROt4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-118" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103197" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28967" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28968" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28969" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28970" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1015" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pro__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_na1_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pro__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PRPPS" metaid="R_PRPPS" name="Phosphoribosylpyrophosphate synthetase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PRPPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PRPPS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.6.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103215" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1207" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_r5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_prpp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PSCVT" metaid="R_PSCVT" name="3-phosphoshikimate 1-carboxyvinyltransferase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PSCVT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PSCVT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2.5.1.19-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03460" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103226" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21256" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21257" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21258" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21259" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0908" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_skm5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3psme_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PSD_EC" metaid="R_PSD_EC" name="Phosphatidylserine decarboxylase (Ecoli)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PSD_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PSD_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.65" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103237" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4160" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ps_EC_c" stoichiometry="0.02" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pe_EC_c" stoichiometry="0.02" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PSERT" metaid="R_PSERT" name="Phosphoserine transaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PSERT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PSERT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PSERTRANSAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.52" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04173" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103225" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14329" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14330" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14331" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14332" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0907" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3php_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pser__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PSP_L" metaid="R_PSP_L" name="Phosphoserine phosphatase (L-serine)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PSP_L">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PSP_L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5114" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00582" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103260" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21208" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21209" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21210" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21211" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4388" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pser__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ser__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PSSA_EC" metaid="R_PSSA_EC" name="Phosphatidylserine syntase (Ecoli)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PSSA_EC">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PSSA_EC" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103307" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2585" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cdpdag_EC_c" stoichiometry="0.02" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ps_EC_c" stoichiometry="0.02" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PTA2" metaid="R_PTA2" name="Phosphate acetyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PTA2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PTA2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PTAALT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.222" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00921" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR102416" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28046" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28047" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28048" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28049" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2297" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ppcoa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppap_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PTAr" metaid="R_PTAr" name="Phosphotransacetylase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PTAr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PTAr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PHOSACETYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00230" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103319" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19521" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19522" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19523" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19524" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2297" />
            <fbc:geneProductRef fbc:geneProduct="G_b2458" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_actp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PTPATi" metaid="R_PTPATi" name="Pantetheine-phosphate adenylyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PTPATi">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PTPATi" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PANTEPADENYLYLTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03035" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95892" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19802" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19803" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19804" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3634" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pan4p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dpcoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PTRCORNt7" metaid="R_PTRCORNt7" name="Putrescine/ornithine antiporter" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PTRCORNt7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PTRCORNt7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-211" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103336" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28827" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28828" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28829" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28830" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0692" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_orn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ptrc_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_orn_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ptrc_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PTRCTA" metaid="R_PTRCTA" name="Putrescine Transaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PTRCTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PTRCTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PUTTRANSAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.29" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.82" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01155" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103337" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23816" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23817" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23818" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23819" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ptrc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4abutn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PTRCabc" metaid="R_PTRCabc" name="Putrescine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PTRCabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PTRCabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-25-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.31" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103338" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29995#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29996#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29997#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29998#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1440" />
              <fbc:geneProductRef fbc:geneProduct="G_b1443" />
              <fbc:geneProductRef fbc:geneProduct="G_b1442" />
              <fbc:geneProductRef fbc:geneProduct="G_b1441" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0854" />
              <fbc:geneProductRef fbc:geneProduct="G_b0855" />
              <fbc:geneProductRef fbc:geneProduct="G_b0857" />
              <fbc:geneProductRef fbc:geneProduct="G_b0856" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1124" />
              <fbc:geneProductRef fbc:geneProduct="G_b1123" />
              <fbc:geneProductRef fbc:geneProduct="G_b1125" />
              <fbc:geneProductRef fbc:geneProduct="G_b1126" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ptrc_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ptrc_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PTRCt2" metaid="R_PTRCt2" name="Putrescine transport in via proton symport, reversible"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PTRCt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PTRCt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-69" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103340" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28891" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28892" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28893" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28894" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0692" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ptrc_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ptrc_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PUNP1" metaid="R_PUNP1" name="Purine-nucleoside phosphorylase (Adenosine)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PUNP1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PUNP1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADENPHOSPHOR-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01561" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103343" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27642" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27643" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27644" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27645" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2407" />
            <fbc:geneProductRef fbc:geneProduct="G_b4384" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ade_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r1p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PUNP2" metaid="R_PUNP2" name="Purine-nucleoside phosphorylase (Deoxyadenosine)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PUNP2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PUNP2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYADENPHOSPHOR-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02557" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103344" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27742" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27743" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27744" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27745" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2407" />
            <fbc:geneProductRef fbc:geneProduct="G_b4384" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dad_2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dr1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ade_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PUNP3" metaid="R_PUNP3" name="Purine-nucleoside phosphorylase (Guanosine)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PUNP3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PUNP3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5199" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.15" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02147" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103345" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13233" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13234" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13235" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13236" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2407" />
            <fbc:geneProductRef fbc:geneProduct="G_b4384" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gsn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gua_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r1p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PUNP4" metaid="R_PUNP4" name="Purine-nucleoside phosphorylase (Deoxyguanosine)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PUNP4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PUNP4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYGUANPHOSPHOR-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01969" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103346" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27738" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27739" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27740" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27741" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2407" />
            <fbc:geneProductRef fbc:geneProduct="G_b4384" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dgsn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dr1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gua_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PUNP5" metaid="R_PUNP5" name="Purine-nucleoside phosphorylase (Inosine)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PUNP5">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PUNP5" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:INOPHOSPHOR-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.15" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01863" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103347" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27646" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27647" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27648" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27649" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2407" />
            <fbc:geneProductRef fbc:geneProduct="G_b4384" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ins_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hxan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r1p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PUNP6" metaid="R_PUNP6" name="Purine-nucleoside phosphorylase (Deoxyinosine)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PUNP6">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PUNP6" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DEOXYINOPHOSPHOR-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02748" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103348" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27750" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27751" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27752" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27753" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2407" />
            <fbc:geneProductRef fbc:geneProduct="G_b4384" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_din_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dr1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hxan_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PUNP7" metaid="R_PUNP7" name="Purine-nucleoside phosphorylase (Xanthosine)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PUNP7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PUNP7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XANTHOSINEPHOSPHORY-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02297" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103349" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27638" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27639" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27640" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27641" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2407" />
            <fbc:geneProductRef fbc:geneProduct="G_b4384" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xtsn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xan_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PYAM5PO" metaid="R_PYAM5PO" name="Pyridoxamine 5'-phosphate oxidase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PYAM5PO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PYAM5PO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PMPOXI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.4.3.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00277" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103355" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15817" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15818" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15819" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15820" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1638" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyam5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pydx5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PYDAMK" metaid="R_PYDAMK" name="Pyridoxamine kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PYDAMK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PYDAMK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRAMKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.35" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02493" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103357" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25104" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25105" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25106" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25107" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2418" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pydam_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyam5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PYDXK" metaid="R_PYDXK" name="Pyridoxal kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PYDXK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PYDXK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PYRIDOXKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.35" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00174" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103361" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10224" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10225" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10226" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10227" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1636" />
            <fbc:geneProductRef fbc:geneProduct="G_b2418" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pydx_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pydx5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PYDXNK" metaid="R_PYDXNK" name="Pyridoxine kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PYDXNK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PYDXNK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PNKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.35" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01909" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103362" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25108" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25109" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25110" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25111" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2418" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pydxn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pdx5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PYDXPP" metaid="R_PYDXPP" name="Pyridoxal 5-phosphate phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PYDXPP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PYDXPP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.1.3.74-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.74" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00173" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103368" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20533" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20534" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20535" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20536" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pydx5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pydx_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_PYK" metaid="R_PYK" name="Pyruvate kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PYK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PYK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:PEPDEPHOS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.40" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00200" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103371" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18157" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18158" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18159" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18160" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1854" />
            <fbc:geneProductRef fbc:geneProduct="G_b1676" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PYNP2r" metaid="R_PYNP2r" name="Pyrimidine-nucleoside phosphorylase (uracil)"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PYNP2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PYNP2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:URPHOS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01876" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103377" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24388" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24389" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24390" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24391" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3831" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uri_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ura_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_PYRt2" metaid="R_PYRt2" name="Pyruvate transport in via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_PYRt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/PYRt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103385" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pyr_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_QULNS_1" metaid="R_QULNS_1" name="Quinolinate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_QULNS_1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/QULNS_1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.72" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04292" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103400" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25888" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25890" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25891" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0750" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_iasp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_quln_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RBFK" metaid="R_RBFK" name="Riboflavin kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RBFK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RBFK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBOFLAVINKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.26" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00549" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103429" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14357" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14358" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14359" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14360" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0025" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ribflv_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fmn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RBFSa" metaid="R_RBFSa" name="Riboflavin synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RBFSa">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RBFSa" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.78" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04457" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR97501" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1662" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4r5au_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_db4p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dmlz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RBFSb" metaid="R_RBFSb" name="Riboflavin synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RBFSb">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RBFSb" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBOFLAVIN-SYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00066" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103430" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20772" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20773" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20774" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20775" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0415" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dmlz_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4r5au_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ribflv_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RBK" metaid="R_RBK" name="Ribokinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RBK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RBK" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.15" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103431" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3752" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rib__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_r5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RBK_L1" metaid="R_RBK_L1" name="L-ribulokinase (L-ribulose)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RBK_L1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RBK_L1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5116" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02439" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103433" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22072" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22073" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22074" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22075" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0063" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rbl__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ru5p__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_RBP4E" metaid="R_RBP4E" name="L-ribulose-phosphate 4-epimerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RBP4E">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RBP4E" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBULPEPIM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.3.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05850" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103434" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22368" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22369" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22370" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22371" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3583" />
            <fbc:geneProductRef fbc:geneProduct="G_b0061" />
            <fbc:geneProductRef fbc:geneProduct="G_b4198" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ru5p__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xu5p__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RHCCE" metaid="R_RHCCE" name="S-ribosylhomocysteine cleavage enzyme" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RHCCE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RHCCE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBOSYLHOMOCYSTEINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.4.1.21" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01291" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104026" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17753" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17754" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17755" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17756" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2687" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rhcys_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhptd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_hcys__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RIBabc" metaid="R_RIBabc" name="D-ribose transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RIBabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RIBabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-28-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.17" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104034" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29903#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29904#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29905#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29906#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b4227" />
              <fbc:geneProductRef fbc:geneProduct="G_b4231" />
              <fbc:geneProductRef fbc:geneProduct="G_b4229" />
              <fbc:geneProductRef fbc:geneProduct="G_b4230" />
              <fbc:geneProductRef fbc:geneProduct="G_b4228" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3751" />
              <fbc:geneProductRef fbc:geneProduct="G_b3750" />
              <fbc:geneProductRef fbc:geneProduct="G_b3749" />
              <fbc:geneProductRef fbc:geneProduct="G_b3748" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rib__D_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_rib__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_RMI" metaid="R_RMI" name="L-rhamnose isomerase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RMI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RMI" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.14" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02437" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104038" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23160" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23161" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23162" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23163" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3903" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rmn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_rml_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RMK" metaid="R_RMK" name="Rhamnulokinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RMK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RMK" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03014" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104039" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20117" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20118" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20119" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20120" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3904" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rml_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_rml1p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RMNt" metaid="R_RMNt" name="L-rhamnose transport via proton symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RMNt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RMNt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-112" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104041" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29947" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29948" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29949" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29950" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3907" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rmn_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_rmn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_RMPA" metaid="R_RMPA" name="Rhamnulose-1-phosphate aldolase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RMPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RMPA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RHAMNULPALDOL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02263" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104043" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19689" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19690" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19691" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19692" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3902" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_rml1p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lald__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RNDR1" metaid="R_RNDR1" name="Ribonucleoside-diphosphate reductase (ADP)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RNDR1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RNDR1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ADPREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.4.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02017" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104060" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2234" />
            <fbc:geneProductRef fbc:geneProduct="G_b2235" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RNDR2" metaid="R_RNDR2" name="Ribonucleoside-diphosphate reductase (GDP)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RNDR2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RNDR2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GDPREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.4.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02019" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104062" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2234" />
            <fbc:geneProductRef fbc:geneProduct="G_b2235" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dgdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RNDR3" metaid="R_RNDR3" name="Ribonucleoside-diphosphate reductase (CDP)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RNDR3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RNDR3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:CDPREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.4.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02024" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104064" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2234" />
              <fbc:geneProductRef fbc:geneProduct="G_b2235" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2675" />
              <fbc:geneProductRef fbc:geneProduct="G_b2676" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dcdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RNDR4" metaid="R_RNDR4" name="Ribonucleoside-diphosphate reductase (UDP)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RNDR4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RNDR4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.4.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02018" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104066" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2234" />
            <fbc:geneProductRef fbc:geneProduct="G_b2235" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dudp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RNTR1" metaid="R_RNTR1" name="Ribonucleoside triphosphate reductase  ATP "
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RNTR1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RNTR1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02014" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104070" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4238" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_datp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RNTR2" metaid="R_RNTR2" name="Ribonucleoside triphosphate reductase  GTP "
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RNTR2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RNTR2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02020" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104073" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4238" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dgtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RNTR3" metaid="R_RNTR3" name="Ribonucleoside triphosphate reductase  CTP "
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RNTR3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RNTR3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02022" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104076" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4238" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dctp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RNTR4" metaid="R_RNTR4" name="Ribonucleoside triphosphate reductase  UTP "
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RNTR4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RNTR4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.17.4.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02023" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104079" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4238" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdrd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_utp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dutp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdox_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_RPE" metaid="R_RPE" name="Ribulose 5-phosphate 3-epimerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RPE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RPE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBULP3EPIM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01529" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104083" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13677" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13678" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13679" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13680" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3386" />
            <fbc:geneProductRef fbc:geneProduct="G_b4301" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ru5p__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xu5p__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_RPI" metaid="R_RPI" name="Ribose-5-phosphate isomerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RPI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RPI" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104084" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2914" />
            <fbc:geneProductRef fbc:geneProduct="G_b4090" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_r5p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ru5p__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_RZ5PP" metaid="R_RZ5PP" name="Alpha-ribazole 5-phosphate phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_RZ5PP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/RZ5PP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RIBAZOLEPHOSPHAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-16788" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.73" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04594" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104116" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24456" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24457" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24458" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24459" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0638" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_5prdmbz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_rdmbzi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_S7PI" metaid="R_S7PI" name="Sedoheptulose 7-phosphate isomerase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_S7PI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/S7PI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-4301" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.28" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05645" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R09768" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104235" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0222" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_s7p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gmhep7p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SADH" metaid="R_SADH" name="Succinylarginine dihydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SADH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SADH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCARGDIHYDRO-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.3.23" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04189" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104239" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19533" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19534" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19535" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19536" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1748" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucarg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sucorn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SADT2" metaid="R_SADT2" name="Sulfate adenyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SADT2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SADT2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104241" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2751" />
            <fbc:geneProductRef fbc:geneProduct="G_b2752" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_so4_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aps_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SBTPD" metaid="R_SBTPD" name="Sorbitol-6-phosphate dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SBTPD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SBTPD" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.140" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05607" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104285" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19837" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19838" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19839" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19840" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2705" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sbt6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SBTpts" metaid="R_SBTpts" name="D-sorbitol transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SBTpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SBTpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104290" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b2703" />
            <fbc:geneProductRef fbc:geneProduct="G_b2702" />
            <fbc:geneProductRef fbc:geneProduct="G_b2704" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sbt__D_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sbt6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SDPDS" metaid="R_SDPDS" name="Succinyl-diaminopimelate desuccinylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SDPDS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SDPDS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCDIAMINOPIMDESUCC-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.18" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02734" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104300" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22608" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22609" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22610" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22611" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2472" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sl26da_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_26dap_LL_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SDPTA" metaid="R_SDPTA" name="Succinyldiaminopimelate transaminase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SDPTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SDPTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCINYLDIAMINOPIMTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.17" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04475" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104301" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11960" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11961" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11962" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/11963" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sl26da_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sl2a6o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SELNPS" metaid="R_SELNPS" name="Selenophosphate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SELNPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SELNPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2.7.9.3-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.9.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03595" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104322" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18737" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18738" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18739" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18740" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1764" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_seln_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_selnp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SERASr" metaid="R_SERASr" name="(L-seryl)adenylate synthase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SERASr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SERASr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ENTF-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104332" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28570" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28571" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28572" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28573" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0586" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_seramp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SERAT" metaid="R_SERAT" name="Serine O-acetyltransferase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SERAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SERAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SERINE-O-ACETTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.30" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00586" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104275" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24560" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24561" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24562" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24563" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3607" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acser_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SERD_D" metaid="R_SERD_D" name="D-serine deaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SERD_D">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SERD_D" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DSERDEAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.1.18" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00221" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104338" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13977" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13978" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13979" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13980" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2366" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SERD_L" metaid="R_SERD_L" name="L-serine deaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SERD_L">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SERD_L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:4.3.1.17-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.1.15" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.1.17" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00220" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00223" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104339" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19169" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19170" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19171" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19172" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3708" />
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3112" />
              <fbc:geneProductRef fbc:geneProduct="G_b3111" />
            </fbc:and>
            <fbc:geneProductRef fbc:geneProduct="G_b1814" />
            <fbc:geneProductRef fbc:geneProduct="G_b2797" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SERt2r" metaid="R_SERt2r" name="L serine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SERt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SERt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-71" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104352" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28887" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28888" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28890" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3116" />
            <fbc:geneProductRef fbc:geneProduct="G_b2796" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ser__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SERt4" metaid="R_SERt4" name="L-serine via sodium symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SERt4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SERt4" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-4083" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104346" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29575" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29576" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29577" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29578" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3089" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_na1_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ser__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SGDS" metaid="R_SGDS" name="Succinylglutamate desuccinylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SGDS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SGDS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCGLUDESUCC-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.96" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00411" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104362" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15169" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15170" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15171" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15172" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1744" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucglu_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SGSAD" metaid="R_SGSAD" name="Succinylglutamic semialdehyde dehydrogenase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SGSAD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SGSAD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCGLUALDDEHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.71" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05049" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104368" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10812" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10813" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10814" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10815" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1746" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucgsa_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sucglu_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SHCHCS2" metaid="R_SHCHCS2"
      name="2-succinyl-6-hydroxy-2,4-cyclohexadiene 1-carboxylate synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SHCHCS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SHCHCS2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104371" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2264" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ichor_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ssaltpp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2shchc_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thmpp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SHCHD2" metaid="R_SHCHD2" name="Sirohydrochlorin dehydrogenase (NAD)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SHCHD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SHCHD2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DIMETHUROPORDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.1.76" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03947" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104373" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15613" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15614" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15615" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15616" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3368" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dscl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_scl_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SHCHF" metaid="R_SHCHF" name="Sirohydrochlorin ferrochelatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SHCHF">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SHCHF" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SIROHEME-FERROCHELAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.99.1.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02864" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104375" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24360" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24361" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24362" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24363" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3368" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fe2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_scl_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sheme_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SHK3Dr" metaid="R_SHK3Dr" name="Shikimate dehydrogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SHK3Dr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SHK3Dr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SHIKIMATE-5-DEHYDROGENASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.25" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.282" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02413" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104378" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17737" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17738" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17739" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17740" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3281" />
            <fbc:geneProductRef fbc:geneProduct="G_b1692" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3dhsk_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_skm_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SHKK" metaid="R_SHKK" name="Shikimate kinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SHKK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SHKK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SHIKIMATE-KINASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.71" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02412" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104379" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13121" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13122" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13123" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13124" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3390" />
            <fbc:geneProductRef fbc:geneProduct="G_b0388" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_skm_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_skm5p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SHSL1" metaid="R_SHSL1" name="O-succinylhomoserine lyase (L-cysteine)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SHSL1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SHSL1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:O-SUCCHOMOSERLYASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.48" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03260" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104380" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20397" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20398" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20399" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20400" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3939" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_suchms_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cyst__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SOTA" metaid="R_SOTA" name="Succinylornithine transaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SOTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SOTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCORNTRANSAM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.81" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04217" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104477" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16953" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16954" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16955" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16956" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1745" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucorn_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sucgsa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SPMDAT1" metaid="R_SPMDAT1" name="Spermidine acetyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SPMDAT1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SPMDAT1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SPERMACTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.57" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104491" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28150" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28151" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28152" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28153" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1584" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_spmd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_N1aspmd_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SPMDAT2" metaid="R_SPMDAT2" name="Spermidine acetyltransferase (N8)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SPMDAT2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SPMDAT2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-7165" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.57" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.29" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104492" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28270" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28271" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28272" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28273" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1584" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_spmd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_n8aspmd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SPMDabc" metaid="R_SPMDabc" name="Spermidine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SPMDabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SPMDabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-24-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.31" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104494" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29999#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30000#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30001#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30002#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1440" />
              <fbc:geneProductRef fbc:geneProduct="G_b1443" />
              <fbc:geneProductRef fbc:geneProduct="G_b1442" />
              <fbc:geneProductRef fbc:geneProduct="G_b1441" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1125" />
              <fbc:geneProductRef fbc:geneProduct="G_b1123" />
              <fbc:geneProductRef fbc:geneProduct="G_b1124" />
              <fbc:geneProductRef fbc:geneProduct="G_b1126" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_spmd_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_spmd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SPMS" metaid="R_SPMS" name="Spermidine synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SPMS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SPMS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SPERMIDINESYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01920" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95860" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12721" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12722" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12723" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12724" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0121" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ametam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ptrc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_5mta_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_spmd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SPODM" metaid="R_SPODM" name="Superoxide dismutase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SPODM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SPODM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUPEROX-DISMUT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.15.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00275" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104498" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20696" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20697" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20698" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20699" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b1646" />
            <fbc:geneProductRef fbc:geneProduct="G_b3908" />
            <fbc:geneProductRef fbc:geneProduct="G_b1656" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2s_c" stoichiometry="2" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_o2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SSALx" metaid="R_SSALx" name="Succinate-semialdehyde dehydrogenase (NAD)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SSALx">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SSALx" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCINATE-SEMIALDEHYDE-DEHYDROGENASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.16" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.24" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00713" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104540" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13217" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13218" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13219" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13220" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucsal_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SSALy" metaid="R_SSALy" name="Succinate-semialdehyde dehydrogenase (NADP)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SSALy">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SSALy" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCSEMIALDDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.16" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.2.1.79" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00714" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104541" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13213" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13214" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13215" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13216" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2661" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucsal_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SUCBZL" metaid="R_SUCBZL" name="O-succinylbenzoate-CoA ligase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCBZL">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCBZL" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:O-SUCCINYLBENZOATE-COA-LIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.2.1.26" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04030" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104612" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17009" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17010" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17011" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17012" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2260" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucbz_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sbzcoa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SUCBZS" metaid="R_SUCBZS" name="O-succinylbenzoate-CoA synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCBZS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCBZS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.113" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104613" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2261" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2shchc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sucbz_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SUCCabc" metaid="R_SUCCabc" name="Succinate transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCCabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCCabc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104618" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3579" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SUCCt2_2" metaid="R_SUCCt2_2" name="Succinate transport via proton symport (2 H)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCCt2_2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCCt2_2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-121" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104620" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29303" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29304" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29305" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29306" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3528" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SUCCt2_3" metaid="R_SUCCt2_3" name="Succintate transport via proton symport (3 H)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCCt2_3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCCt2_3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-300" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104621" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4123" />
            <fbc:geneProductRef fbc:geneProduct="G_b4138" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_0_bound" id="R_SUCCt2r" metaid="R_SUCCt2r" name="Succinate transport via proton symport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCCt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCCt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104622" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0621" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SUCD1" metaid="R_SUCD1" name="Succinate dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCD1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCD1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.5.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00408" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99636" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30343" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30344" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30345" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30346" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0723" />
            <fbc:geneProductRef fbc:geneProduct="G_b0722" />
            <fbc:geneProductRef fbc:geneProduct="G_b0721" />
            <fbc:geneProductRef fbc:geneProduct="G_b0724" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fadh2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SUCD4" metaid="R_SUCD4" name="Succinate dehyrdogenase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCD4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCD4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104630" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0723" />
            <fbc:geneProductRef fbc:geneProduct="G_b0722" />
            <fbc:geneProductRef fbc:geneProduct="G_b0721" />
            <fbc:geneProductRef fbc:geneProduct="G_b0724" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fadh2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_q8_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_q8h2_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SUCFUMt" metaid="R_SUCFUMt" name="Succinate:fumarate antiporter" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCFUMt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCFUMt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-106" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104633" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29323" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29324" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29325" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29326" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0621" />
            <fbc:geneProductRef fbc:geneProduct="G_b4138" />
            <fbc:geneProductRef fbc:geneProduct="G_b4123" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_fum_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fum_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SUCOAS" metaid="R_SUCOAS" name="Succinyl-CoA synthetase (ADP-forming)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCOAS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCOAS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SUCCCOASYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.2.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00405" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104635" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17661" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17662" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17663" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17664" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0728" />
            <fbc:geneProductRef fbc:geneProduct="G_b0729" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succoa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SUCpts" metaid="R_SUCpts" name="Sucrose transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SUCpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SUCpts" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104643" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2429" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b2417" />
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_sucr_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_suc6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_SULR" metaid="R_SULR" name="Sulfite reductase (NADPH2)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SULR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SULR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:SULFITE-REDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.8.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00858" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104650" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13802" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13803" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13804" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2764" />
            <fbc:geneProductRef fbc:geneProduct="G_b2763" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="5" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_so3_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2s_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="3" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_SULabc" metaid="R_SULabc" name="Sulfate transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_SULabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/SULabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.6.3.25-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-70-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.25" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104655" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10192#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10193#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10194#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10195#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2423" />
              <fbc:geneProductRef fbc:geneProduct="G_b2422" />
              <fbc:geneProductRef fbc:geneProduct="G_b2425" />
              <fbc:geneProductRef fbc:geneProduct="G_b2424" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2413" />
              <fbc:geneProductRef fbc:geneProduct="G_b2423" />
              <fbc:geneProductRef fbc:geneProduct="G_b2422" />
              <fbc:geneProductRef fbc:geneProduct="G_b2424" />
              <fbc:geneProductRef fbc:geneProduct="G_b3917" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_so4_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_so4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TAGURr" metaid="R_TAGURr" name="Tagaturonate reductase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TAGURr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TAGURr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ALTRO-OXIDOREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.58" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02555" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104710" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17813" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17814" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17815" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17816" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1521" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_altrn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tagur_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TALA" metaid="R_TALA" name="Transaldolase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TALA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TALA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANSALDOL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.2.1.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01827" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104715" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17053" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17054" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17055" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17056" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2464" />
            <fbc:geneProductRef fbc:geneProduct="G_b0008" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_s7p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_e4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TARTD" metaid="R_TARTD" name="L(+)-tartrate dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TARTD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TARTD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LTARTDEHYDRA-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.32" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00339" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104718" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15413" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15414" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15415" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15416" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3062" />
            <fbc:geneProductRef fbc:geneProduct="G_b3061" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tartr__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_oaa_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TARTRt7" metaid="R_TARTRt7" name="Tartrate/succinate antiporter" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TARTRt7">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TARTRt7" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-127" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104723" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29259" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29260" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29261" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29262" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3063" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succ_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tartr__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tartr__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TAUDO" metaid="R_TAUDO" name="Taurine dioxygenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TAUDO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TAUDO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-299" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.14.11.17" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R05320" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104730" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15909" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15910" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15911" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15912" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0368" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_o2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_taur_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_aacald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_so3_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_succ_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TAURabc" metaid="R_TAURabc" name="Taurine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TAURabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TAURabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:3.6.3.36-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-64-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.36" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104732" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14613#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14614#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14615#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14616#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0367" />
            <fbc:geneProductRef fbc:geneProduct="G_b0365" />
            <fbc:geneProductRef fbc:geneProduct="G_b0366" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_taur_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_taur_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TDPADGAT" metaid="R_TDPADGAT" name="DTDP-4-amino-4,6-dideoxy-D-glucose acetyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TDPADGAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TDPADGAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TDPFUCACTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.210" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R10142" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104757" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28763" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28764" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28765" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28766" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34443" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34444" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34445" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/34446" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3790" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_accoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtdp4addg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtdp4aaddg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TDPAGTA" metaid="R_TDPAGTA" name="DTDP-4-amino-4,6-dideoxy-D-glucose transaminase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TDPAGTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TDPAGTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RFFTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.59" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04438" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104758" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28631" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28632" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28633" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28634" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3791" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtdp4d6dg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtdp4addg_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TDPDRE" metaid="R_TDPDRE" name="DTDP-4-dehydrorhamnose 3,5-epimerase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TDPDRE">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TDPDRE" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDPDEHYDRHAMEPIM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.3.13" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06514" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104759" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16969" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16970" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16971" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16972" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2038" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtdp4d6dg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtdp4d6dm_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TDPDRR" metaid="R_TDPDRR" name="DTDP-4-dehydrorhamnose reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TDPDRR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TDPDRR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDPDEHYRHAMREDUCT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.133" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02777" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104760" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21796" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21797" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21798" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/21799" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2040" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtdp4d6dm_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtdprmn_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TDPGDH" metaid="R_TDPGDH" name="DTDPglucose 4,6-dehydratase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TDPGDH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TDPGDH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:DTDPGLUCDEHYDRAT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.46" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06513" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104761" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17221" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17222" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17223" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17224" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2041" />
            <fbc:geneProductRef fbc:geneProduct="G_b3788" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dtdpglu_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtdp4d6dg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TDSK" metaid="R_TDSK" name="Tetraacyldisaccharide 4'kinase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TDSK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TDSK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TETRAACYLDISACC4KIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.130" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04657" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104762" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20700" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20701" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20702" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20703" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0915" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_lipidAds_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lipidA_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TGBPA" metaid="R_TGBPA" name="Tagatose-bisphosphate aldolase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TGBPA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TGBPA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TAGAALDOL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.40" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01069" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104798" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22948" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22949" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22950" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22951" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3137" />
              <fbc:geneProductRef fbc:geneProduct="G_b3132" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2096" />
              <fbc:geneProductRef fbc:geneProduct="G_b2095" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tagdp__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhap_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THD2" metaid="R_THD2" name="NAD(P) transhydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THD2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THD2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.6.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104805" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1602" />
            <fbc:geneProductRef fbc:geneProduct="G_b1603" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadph_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THDPS" metaid="R_THDPS" name="Tetrahydrodipicolinate succinylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THDPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THDPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TETHYDPICSUCC-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.117" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04365" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104806" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17325" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17326" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17327" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17328" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0166" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_succoa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_coa_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_sl2a6o_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THMDt2_copy1" metaid="R_THMDt2_copy1" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THMDt2_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108H" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104818" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29955" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29956" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29957" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29958" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or>
            <fbc:geneProductRef fbc:geneProduct="G_b2964" />
            <fbc:geneProductRef fbc:geneProduct="G_b2393" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thymd_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thymd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_THMDt2_copy2" metaid="R_THMDt2_copy2" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THMDt2_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108H" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104818" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29955" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29956" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29957" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29958" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2406" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thymd_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thymd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THMabc" metaid="R_THMabc" name="Thiamine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THMabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THMabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-32-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104827" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29811#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29812#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29813#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29814#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0067" />
            <fbc:geneProductRef fbc:geneProduct="G_b0066" />
            <fbc:geneProductRef fbc:geneProduct="G_b0068" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thm_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thm_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_THRA" metaid="R_THRA" name="Threonine aldolase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THRA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THRA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THREONINE-ALDOLASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.48" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00751" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104833" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19625" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19626" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19627" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19628" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0870" />
            <fbc:geneProductRef fbc:geneProduct="G_b2551" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thr__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_acald_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gly_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THRD" metaid="R_THRD" name="L-threonine dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THRD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THRD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THREODEHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.103" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01465" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR103402" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13161" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13162" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13163" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13164" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3616" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thr__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2aobut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THRD_L" metaid="R_THRD_L" name="L-threonine deaminase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THRD_L">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THRD_L" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THREDEHYD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.3.1.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00996" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104714" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22108" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22109" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22110" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22111" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2797" />
            <fbc:geneProductRef fbc:geneProduct="G_b3772" />
            <fbc:geneProductRef fbc:geneProduct="G_b3117" />
            <fbc:geneProductRef fbc:geneProduct="G_b1814" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thr__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2obut_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THRS" metaid="R_THRS" name="Threonine synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THRS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THRS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THRESYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.3.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01466" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104846" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10840" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10841" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10842" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10843" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0004" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_phom_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thr__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THRabc" metaid="R_THRabc" name="L-threonine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THRabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THRabc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104849" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3454" />
            <fbc:geneProductRef fbc:geneProduct="G_b3457" />
            <fbc:geneProductRef fbc:geneProduct="G_b3455" />
            <fbc:geneProductRef fbc:geneProduct="G_b3460" />
            <fbc:geneProductRef fbc:geneProduct="G_b3456" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thr__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thr__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_THRt2r" metaid="R_THRt2r" name="L threonine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THRt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THRt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-72" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104850" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28883" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28884" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28885" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28886" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3116" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thr__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thr__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THRt4" metaid="R_THRt4" name="L-threonine  via sodium symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THRt4">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THRt4" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104842" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3089" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_na1_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thr__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_na1_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thr__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_THZPSN" metaid="R_THZPSN" name="Thiazole phosphate synthesis" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_THZPSN">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/THZPSN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104858" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3992" />
            <fbc:geneProductRef fbc:geneProduct="G_b4407" />
            <fbc:geneProductRef fbc:geneProduct="G_b3991" />
            <fbc:geneProductRef fbc:geneProduct="G_b3990" />
            <fbc:geneProductRef fbc:geneProduct="G_b0423" />
            <fbc:geneProductRef fbc:geneProduct="G_b2530" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_cys__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dxyl5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tyr__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4hba_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_4mpetz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ala__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_amp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TKT1" metaid="R_TKT1" name="Transketolase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TKT1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TKT1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.2.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104868" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2935" />
            <fbc:geneProductRef fbc:geneProduct="G_b2465" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_r5p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xu5p__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_s7p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TKT2" metaid="R_TKT2" name="Transketolase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TKT2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TKT2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:2TRANSKETO-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.2.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01830" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104869" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27626" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27627" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27628" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27629" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b2465" />
            <fbc:geneProductRef fbc:geneProduct="G_b2935" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_e4p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xu5p__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_f6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TMAOR1" metaid="R_TMAOR1" name="Trimethylamine N-oxide reductase (menaquinol 8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMAOR1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMAOR1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104878" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29279" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29280" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29281" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29282" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0896" />
            <fbc:geneProductRef fbc:geneProduct="G_b0894" />
            <fbc:geneProductRef fbc:geneProduct="G_b0895" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tmao_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mqn8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tma_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TMAOR1e" metaid="R_TMAOR1e" name="Trimethylamine N-oxide reductase (menaquinol 8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMAOR1e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMAOR1e" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104879" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0996" />
              <fbc:geneProductRef fbc:geneProduct="G_b0997" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1873" />
              <fbc:geneProductRef fbc:geneProduct="G_b1872" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tmao_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_mqn8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tma_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TMAOR2" metaid="R_TMAOR2" name="Trimethylamine N-oxide reductase (demethylmenaquinol 8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMAOR2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMAOR2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104880" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30583" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30584" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30585" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30586" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b0896" />
            <fbc:geneProductRef fbc:geneProduct="G_b0894" />
            <fbc:geneProductRef fbc:geneProduct="G_b0895" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tmao_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tma_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TMAOR2e" metaid="R_TMAOR2e" name="Trimethylamine N-oxide reductase (demethylmenaquinol 8)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMAOR2e">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMAOR2e" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104881" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b0996" />
              <fbc:geneProductRef fbc:geneProduct="G_b0997" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b1873" />
              <fbc:geneProductRef fbc:geneProduct="G_b1872" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2dmmql8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tmao_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dmmq8_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tma_e" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TMDK1" metaid="R_TMDK1" name="Thymidine kinase (ATP:thymidine)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMDK1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMDK1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THYKI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.145" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.21" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01567" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104885" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19129" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19130" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19131" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19132" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1238" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thymd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TMDPP" metaid="R_TMDPP" name="Thymidine phosphorylase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMDPP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMDPP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THYM-PHOSPH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.2" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01570" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104887" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16037" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16038" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16039" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16040" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4382" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thymd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_2dr1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thym_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TMDS" metaid="R_TMDS" name="Thymidylate synthase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMDS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMDS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.1.45" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02101" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12104" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12105" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12106" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12107" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2827" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dump_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_mlthf_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dhf_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dtmp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TMK" metaid="R_TMK" name="Thiamine kinase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THIKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.89" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02134" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104898" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12012" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12013" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12014" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12015" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thm_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thmmp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TMPK" metaid="R_TMPK" name="Thiamine-phosphate kinase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMPK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMPK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THI-P-KIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.16" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00617" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104907" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15913" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15914" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15915" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15916" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0417" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_thmmp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thmpp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TMPPP" metaid="R_TMPPP" name="Thiamine-phosphate diphosphorylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TMPPP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TMPPP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THI-P-SYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.3" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03223" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104908" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22328" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22329" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22330" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22331" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3993" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2mahmp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_4mpetz_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_thmmp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TPI" metaid="R_TPI" name="Triose-phosphate isomerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TPI">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TPI" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRIOSEPISOMERIZATION-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01015" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104918" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18585" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18586" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18587" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18588" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3919" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dhap_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TRDR" metaid="R_TRDR" name="Thioredoxin reductase (NADPH)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRDR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRDR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:THIOREDOXIN-REDUCT-NADPH-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.6.4.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.8.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02016" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104766" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0888" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trdox_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trdrd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TRE6PH" metaid="R_TRE6PH" name="Trehalose-6-phosphate hydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRE6PH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRE6PH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRE6PHYDRO-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.122" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.93" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00837" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06113" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104925" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23008" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23009" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23010" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23011" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4239" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tre6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TRE6PP" metaid="R_TRE6PP" name="Trehalose-phosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRE6PP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRE6PP" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TREHALOSEPHOSPHA-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.1.3.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02778" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06228" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104926" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23420" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23421" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23422" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23423" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1897" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tre6p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tre_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TRE6PS" metaid="R_TRE6PS" name="Alpha,alpha-trehalose-phosphate synthase (UDP-forming)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRE6PS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRE6PS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TREHALOSE6PSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.15" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00836" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06043" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104927" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18890" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18891" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18892" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1896" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_g6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udpg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tre6p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TREH" metaid="R_TREH" name="Alpha,alpha-trehalase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TREH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TREH" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.28" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00010" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06103" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95163" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20868" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20869" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20870" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20871" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3519" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tre_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TREHe" metaid="R_TREHe" name="Alpha,alpha-trehalase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TREHe">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TREHe" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.2.1.28" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00010" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R06103" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95163" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20868" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20869" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20870" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/20871" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1197" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tre_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glc__D_e" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TREpts" metaid="R_TREpts" name="Trehalose transport via PEP:Pyr PTS" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TREpts">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TREpts" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.69" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104931" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b2416" />
            <fbc:geneProductRef fbc:geneProduct="G_b2415" />
            <fbc:geneProductRef fbc:geneProduct="G_b2417" />
            <fbc:geneProductRef fbc:geneProduct="G_b4240" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tre_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tre6p_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TRPAS2" metaid="R_TRPAS2" name="Tryptophanase (L-tryptophan)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRPAS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRPAS2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRYPTOPHAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.99.1" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00673" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104941" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19553" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19554" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19555" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19556" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3708" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trp__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_indole_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TRPS1" metaid="R_TRPS1" name="Tryptophan synthase (indoleglycerol phosphate)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRPS1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRPS1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRYPSYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02722" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104343" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10532" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10533" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10534" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10535" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1260" />
            <fbc:geneProductRef fbc:geneProduct="G_b1261" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3ig3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trp__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TRPS2" metaid="R_TRPS2" name="Tryptophan synthase (indole)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRPS2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRPS2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-2382" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.122" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00674" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104946" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26434" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26435" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26436" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/26437" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1260" />
            <fbc:geneProductRef fbc:geneProduct="G_b1261" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_indole_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ser__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trp__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TRPS3" metaid="R_TRPS3" name="Tryptophan synthase (indoleglycerol phosphate)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRPS3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRPS3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-2381" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.20" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02340" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR99889" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14081" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14082" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14083" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/14084" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b1260" />
            <fbc:geneProductRef fbc:geneProduct="G_b1261" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3ig3p_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g3p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_indole_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TRPt2r" metaid="R_TRPt2r" name="L tryptophan reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRPt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRPt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-76" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104950" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28879" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28880" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28881" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28882" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3709" />
            <fbc:geneProductRef fbc:geneProduct="G_b0112" />
            <fbc:geneProductRef fbc:geneProduct="G_b3161" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_trp__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_trp__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TRSARr" metaid="R_TRSARr" name="Tartronate semialdehyde reductase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TRSARr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TRSARr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN0-5289" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.60" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01745" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR100682" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18845" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18846" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18847" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18848" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3125" />
            <fbc:geneProductRef fbc:geneProduct="G_b0509" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2h3oppan_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadh_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glyc__R_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nad_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TSULabc" metaid="R_TSULabc" name="Thiosulfate transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TSULabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TSULabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-7-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.3.25" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104964" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29871#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29872#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29873#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29874#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b3917" />
              <fbc:geneProductRef fbc:geneProduct="G_b2422" />
              <fbc:geneProductRef fbc:geneProduct="G_b2423" />
              <fbc:geneProductRef fbc:geneProduct="G_b2424" />
            </fbc:and>
            <fbc:and sboTerm="SBO:0000173">
              <fbc:geneProductRef fbc:geneProduct="G_b2423" />
              <fbc:geneProductRef fbc:geneProduct="G_b2422" />
              <fbc:geneProductRef fbc:geneProduct="G_b2425" />
              <fbc:geneProductRef fbc:geneProduct="G_b2424" />
            </fbc:and>
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tsul_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tsul_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_TTDCAt2" metaid="R_TTDCAt2" name="Tetradecanoate transport via proton symport"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TTDCAt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TTDCAt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR104971" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2344" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ttdca_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ttdca_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TYRTA" metaid="R_TYRTA" name="Tyrosine transaminase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TYRTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TYRTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TYROSINE-AMINOTRANSFERASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.1" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.57" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00734" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105000" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15093" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15094" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15095" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/15096" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b4054" />
            <fbc:geneProductRef fbc:geneProduct="G_b0928" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tyr__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_34hpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_TYRt2r" metaid="R_TYRt2r" name="L tyrosine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_TYRt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/TYRt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-77" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105003" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28875" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28876" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28877" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28878" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0112" />
            <fbc:geneProductRef fbc:geneProduct="G_b0576" />
            <fbc:geneProductRef fbc:geneProduct="G_b1907" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_tyr__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_tyr__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_U23GAAT" metaid="R_U23GAAT" name="UDP-3-O-(3-hydroxymyristoyl)glucosamine acyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_U23GAAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/U23GAAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPHYDROXYMYRGLUCOSAMNACETYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.191" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04550" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105007" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0179" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3hmrsACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_u3hga_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_u23ga_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UAAGDS" metaid="R_UAAGDS"
      name="UDP-N-acetylmuramoyl-L-alanyl-D-glutamyl-meso-2,6-diaminopimelate synthetase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAAGDS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAAGDS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.13" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105013" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0085" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_26dap__M_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uamag_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ugmd_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UACMAMO" metaid="R_UACMAMO" name="UDP-N-acetyl-D-mannosamine oxidoreductase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UACMAMO">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UACMAMO" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPMANNACADEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.336" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03317" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105022" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25780" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25781" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25782" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25783" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3787" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uacmam_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uacmamu_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UAG2E" metaid="R_UAG2E" name="UDP-N-acetylglucosamine 2-epimerase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAG2E">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAG2E" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPGLCNACEPIM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.3.14" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00420" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105024" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17213" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17214" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17215" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17216" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3786" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uacgam_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uacmam_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_UAGAAT" metaid="R_UAGAAT" name="UDP-N-acetylglucosamine acyltransferase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAGAAT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAGAAT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPNACETYLGLUCOSAMACYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.3.1.129" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04567" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105028" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0181" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3hmrsACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uacgam_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ACP_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_u3aga_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UAGCVT" metaid="R_UAGCVT" name="UDP-N-acetylglucosamine 1-carboxyvinyltransferase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAGCVT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAGCVT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPNACETYLGLUCOSAMENOLPYRTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.7" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00660" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105032" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18681" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18682" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18683" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18684" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3189" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_pep_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uacgam_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uaccg_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UAGDP" metaid="R_UAGDP" name="UDP-N-acetylglucosamine diphosphorylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAGDP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAGDP" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.23" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105033" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3730" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_acgam1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_utp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uacgam_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UAGPT3" metaid="R_UAGPT3"
      name="UDP-N-acetylglucosamine-N-acetylmuramyl-(pentapeptide)pyrophosphoryl-undecaprenol N-acetylglucosamine transferase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAGPT3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAGPT3" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.1.227" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105037" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0090" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uacgam_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uagmda_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uaagmda_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UAMAGS" metaid="R_UAMAGS" name="UDP-N-acetylmuramoyl-L-alanyl-D-glutamate synthetase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAMAGS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAMAGS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-NACMURALA-GLU-LIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02783" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105038" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16429" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16430" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16431" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/16432" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0088" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__D_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uama_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uamag_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UAMAS" metaid="R_UAMAS" name="UDP-N-acetylmuramoyl-L-alanine synthetase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAMAS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAMAS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDP-NACMUR-ALA-LIG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03193" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105039" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23372" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23373" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23374" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23375" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0091" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uamr_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uama_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UAPGR" metaid="R_UAPGR" name="UDP-N-acetylenolpyruvoylglucosamine reductase"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UAPGR">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UAPGR" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPNACETYLMURAMATEDEHYDROG-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.3.1.98" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03192" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105041" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12248" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12249" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12250" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/12251" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3972" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nadph_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uaccg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uamr_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UDCPDP" metaid="R_UDCPDP" name="Undecaprenyl-diphosphatase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UDCPDP">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UDCPDP" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105042" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17069" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17070" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17071" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/17072" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udcpdp_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udcpp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UDCPDPS" metaid="R_UDCPDPS" name="Undecaprenyl diphosphate synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UDCPDPS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UDCPDPS" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-9138" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.31" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.5.1.M1" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105043" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0174" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_frdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ipdp_c" stoichiometry="8" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="8" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udcpdp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_UDPG4E" metaid="R_UDPG4E" name="UDPglucose 4-epimerase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UDPG4E">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UDPG4E" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPGLUCEPIM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.3.2" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00291" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105057" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22168" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22169" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22170" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22171" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0759" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udpg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udpgal_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UDPGALM" metaid="R_UDPGALM" name="UDPgalactopyranose mutase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UDPGALM">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UDPGALM" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GALPMUT-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.4.99.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00505" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105058" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24132" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24133" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24134" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24135" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2036" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udpgal_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udpgalfur_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UDPGD" metaid="R_UDPGD" name="UDPglucose 6-dehydrogenase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UDPGD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UDPGD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UGD-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/1.1.1.22" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00286" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105063" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23596" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23597" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23598" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/23599" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2028" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_nad_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udpg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="3" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nadh_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udpglcur_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_UGLT" metaid="R_UGLT" name="UDPglucose--hexose-1-phosphate uridylyltransferase"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UGLT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UGLT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GALACTURIDYLYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.7.12" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00955" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105090" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13989" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13990" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13991" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13992" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0758" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gal1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_udpg_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_g1p_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udpgal_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UGLYCH" metaid="R_UGLYCH" name="Ureidoglycolate hydrolase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UGLYCH">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UGLYCH" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UREIDOGLYCOLATE-HYDROLASE-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.116" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.3.19" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00469" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105091" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19809" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19810" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19811" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19812" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0505" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_urdglyc_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glx_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_nh4_c" stoichiometry="2" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UGMDDS" metaid="R_UGMDDS"
      name="UDP-N-acetylmuramoyl-L-alanyl-D-glutamyl-meso-2,6-diaminopimeloyl-D-alanyl-D-alanine synthetase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UGMDDS">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UGMDDS" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/6.3.2.10" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105094" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0086" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_alaala_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ugmd_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ugmda_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UHGADA" metaid="R_UHGADA" name="UDP-3-O-acetylglucosamine deacetylase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UHGADA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UHGADA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UDPACYLGLCNACDEACETYL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.5.1.108" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04587" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105110" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25209" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25210" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25211" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25212" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0096" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_u3aga_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ac_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_u3hga_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_UMPK" metaid="R_UMPK" name="UMP kinase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UMPK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UMPK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-12002" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.14" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.22" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.4" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00158" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105118" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24400" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24401" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24402" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24403" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b0171" />
            <fbc:geneProductRef fbc:geneProduct="G_b0910" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ump_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_udp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UNK3" metaid="R_UNK3" name="2-keto-4-methylthiobutyrate transamination" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UNK3">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UNK3" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14147" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.57" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R07396" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR95923" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25796" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25797" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25798" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25799" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_2kmb_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glu__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_met__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UPP3MT" metaid="R_UPP3MT" name="Uroporphyrinogen methyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UPP3MT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UPP3MT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-13403" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.1.1.107" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03194" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105137" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32459" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32460" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32461" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32462" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or sboTerm="SBO:0000174">
            <fbc:geneProductRef fbc:geneProduct="G_b3803" />
            <fbc:geneProductRef fbc:geneProduct="G_b3368" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_amet_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uppg3_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ahcys_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dscl_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UPP3S" metaid="R_UPP3S" name="Uroporphyrinogen-III synthase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UPP3S">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UPP3S" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UROGENIIISYN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.2.1.75" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03165" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105139" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18965" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18966" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18967" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18968" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3804" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_hmbil_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uppg3_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UPPDC1" metaid="R_UPPDC1" name="Uroporphyrinogen decarboxylase (uroporphyrinogen III)"
      reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UPPDC1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UPPDC1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:UROGENDECARBOX-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/4.1.1.37" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03197" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105140" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19865" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19866" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19867" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/19868" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3997" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_c" stoichiometry="4" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uppg3_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_co2_c" stoichiometry="4" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_cpppg3_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_UPPRT" metaid="R_UPPRT" name="Uracil phosphoribosyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UPPRT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UPPRT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:URACIL-PRIBOSYLTRANS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00966" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105142" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13017" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13018" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13019" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/13020" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2498" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ura_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ump_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_URAt2_copy1" metaid="R_URAt2_copy1" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_URAt2_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-132" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105148" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29239" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29240" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29241" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29242" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2497" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ura_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ura_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_URAt2_copy2" metaid="R_URAt2_copy2" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_URAt2_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-132" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105148" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29239" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29240" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29241" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29242" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b1006" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ura_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ura_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_UREAt" metaid="R_UREAt" name="Urea transport via facilitate diffusion" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_UREAt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/UREAt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-460" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105156" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32799" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32800" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/32802" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3927" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_urea_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_urea_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_URIDK2r" metaid="R_URIDK2r" name="Uridylate kinase (dUMP)" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_URIDK2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/URIDK2r" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-14122" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.13" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.4" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.4.9" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02098" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105160" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30655" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30656" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30657" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/30658" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0171" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_dump_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_dudp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_URIK2" metaid="R_URIK2" name="Uridine kinase (GTP:Uridine)" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_URIK2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/URIK2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:URKI-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.48" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00968" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105163" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27650" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27651" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27652" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/27653" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2066" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_gtp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uri_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_gdp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ump_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_URIt2_copy1" metaid="R_URIt2_copy1" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_URIt2_copy1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108I" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105166" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29951" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29952" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29953" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29954" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:or>
            <fbc:geneProductRef fbc:geneProduct="G_b2964" />
            <fbc:geneProductRef fbc:geneProduct="G_b2393" />
          </fbc:or>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uri_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uri_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_URIt2_copy2" metaid="R_URIt2_copy2" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_URIt2_copy2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-108I" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105166" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29951" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29952" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29953" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29954" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2406" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_uri_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_uri_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_USHD" metaid="R_USHD" name="UDP-sugar hydrolase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_USHD">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/USHD" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LIPIDXSYNTHESIS-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/3.6.1.54" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R04549" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105170" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25213" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25214" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25215" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/25216" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0480" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_u23ga_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="2" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_lipidX_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ump_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_VALTA" metaid="R_VALTA" name="Valine transaminase" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_VALTA">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/VALTA" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:BRANCHED-CHAINAMINOTRANSFERVAL-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.42" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.6" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01214" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR96230" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24813" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24814" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24815" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/24816" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3770" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_akg_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_val__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_3mob_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_glu__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_VALabc" metaid="R_VALabc" name="L-valine transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_VALabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/VALabc" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:ABC-36-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105187" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29803#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29804#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29805#2" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29806#2" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3457" />
            <fbc:geneProductRef fbc:geneProduct="G_b3454" />
            <fbc:geneProductRef fbc:geneProduct="G_b3455" />
            <fbc:geneProductRef fbc:geneProduct="G_b3460" />
            <fbc:geneProductRef fbc:geneProduct="G_b3456" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_val__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_val__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_VALt2r" metaid="R_VALt2r" name="L valine reversible transport via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_VALt2r">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/VALt2r" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105188" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0401" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_val__L_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_val__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_VPAMTr" metaid="R_VPAMTr" name="Valine-pyruvate aminotransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_VPAMTr">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/VPAMTr" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:VALINE-PYRUVATE-AMINOTRANSFER-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.6.1.66" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01215" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105217" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22912" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22913" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22914" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22915" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3572" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_3mob_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_ala__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pyr_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_val__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_X5PL3E" metaid="R_X5PL3E" name="L-xylulose 5-phosphate 3-epimerase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_X5PL3E">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/X5PL3E" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:LXULRU5P-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.-.-.-" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.1.3.22" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R03244" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105224" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18497" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18498" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18499" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/18500" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4197" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xu5p__L_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ru5p__L_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_XANt" metaid="R_XANt" name="Xanthine reversible transport" reversible="true"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XANt">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XANt" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN0-561" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105226" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xan_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xan_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_XANt2" metaid="R_XANt2" name="Xanthine transport in via proton symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XANt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XANt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:RXN-5076" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-206" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105227" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29663" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29664" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29665" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29666" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3654" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xan_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xan_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_XPPT" metaid="R_XPPT" name="Xanthine phosphoribosyltransferase" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XPPT">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XPPT" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XANPRIBOSYLTRAN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.22" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.4.2.8" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R02142" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105243" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10800" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10801" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10802" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10803" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b0238" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_prpp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xan_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_ppi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xmp_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_XTSNt2" metaid="R_XTSNt2" name="Xanthosine transport in via proton symport"
      reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XTSNt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XTSNt2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:TRANS-RXN-31" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105246" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28939" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28940" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28941" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28942" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b2406" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xtsn_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xtsn_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_default_ub" id="R_XYLI1" metaid="R_XYLI1" name="Xylose isomerase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XYLI1">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XYLI1" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XYLISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01432" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105253" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22816" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22817" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22818" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/22819" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3565" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xyl__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xylu__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_default_lb" fbc:upperFluxBound="cobra_0_bound" id="R_XYLI2" metaid="R_XYLI2" name="Xylose isomerase" reversible="true" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XYLI2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XYLI2" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:GLUCISOM-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/5.3.1.5" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R00307" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105254" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28546" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28547" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28548" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28549" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3565" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_glc__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_fru_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_XYLK" metaid="R_XYLK" name="Xylulokinase" reversible="false" sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XYLK">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XYLK" />
                  <rdf:li rdf:resource="http://identifiers.org/biocyc/META:XYLULOKIN-RXN" />
                  <rdf:li rdf:resource="http://identifiers.org/ec-code/2.7.1.17" />
                  <rdf:li rdf:resource="http://identifiers.org/kegg.reaction/R01639" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105255" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10964" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10965" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10966" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/10967" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b3564" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xylu__D_c" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xu5p__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_XYLabc" metaid="R_XYLabc" name="D-xylose transport via ABC system" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XYLabc">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XYLabc" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105268" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29899#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29900#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29901#1" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/29902#1" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:and sboTerm="SBO:0000173">
            <fbc:geneProductRef fbc:geneProduct="G_b3566" />
            <fbc:geneProductRef fbc:geneProduct="G_b3568" />
            <fbc:geneProductRef fbc:geneProduct="G_b3567" />
          </fbc:and>
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_atp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h2o_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xyl__D_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_adp_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_pi_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xyl__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
      <reaction fast="false" fbc:lowerFluxBound="cobra_0_bound" fbc:upperFluxBound="cobra_default_ub" id="R_XYLt2" metaid="R_XYLt2" name="D-xylose transport in via proton symport" reversible="false"
      sboTerm="SBO:0000375">
        <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/">
            <rdf:Description rdf:about="#R_XYLt2">
              <bqbiol:is>
                <rdf:Bag>
                  <rdf:li rdf:resource="http://identifiers.org/bigg.reaction/XYLt2" />
                  <rdf:li rdf:resource="http://identifiers.org/metanetx.reaction/MNXR105271" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28959" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28960" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28961" />
                  <rdf:li rdf:resource="http://identifiers.org/rhea/28962" />
                </rdf:Bag>
              </bqbiol:is>
            </rdf:Description>
          </rdf:RDF>
        </annotation>
        <fbc:geneProductAssociation xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2">
          <fbc:geneProductRef fbc:geneProduct="G_b4031" />
        </fbc:geneProductAssociation>
        <listOfReactants>
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_h_e" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000010" species="M_xyl__D_e" stoichiometry="1" />
        </listOfReactants>
        <listOfProducts>
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_h_c" stoichiometry="1" />
          <speciesReference constant="true" sboTerm="SBO:0000011" species="M_xyl__D_c" stoichiometry="1" />
        </listOfProducts>
      </reaction>
    </listOfReactions>
  </model>
</sbml>

"""
